#include "Surface3D.h"
#include "Basic3DMath.h"
#include "../logger.h"


Surface3D::Surface3D()
{
    setSize( 0, 0 );
}


Surface3D::Surface3D( size_t rows, size_t cols )
{
    setSize( rows, cols );
}


Surface3D::~Surface3D()
{
    WX_CLEAR_ARRAY( m_transformedPoints );
}


void Surface3D::setSize( size_t rows, size_t cols )
{
    m_rows = rows;
    m_cols = cols;

    m_indicies.Clear();
    WX_CLEAR_ARRAY( m_points );
    WX_CLEAR_ARRAY( m_transformedPoints );


    for ( size_t i = 0; i <  m_rows * m_cols; ++i )
    {
        m_points.Add( new Point4D() );
        m_transformedPoints.Add( new Point4D() );
    }

    ///add corners
    if ( m_rows > 1 && m_cols > 1 )
    {
        for ( size_t i = 0; i < 4; ++i )
        {
            m_points.Add( new Point4D() );
            m_transformedPoints.Add( new Point4D() );
        }
    }

    createIndicies();

    m_minY = 0.0;
}


void Surface3D::setCornerPoint( size_t row, size_t col, const Point4D& p )
{
    if ( m_rows <= 1 || m_cols <= 1 ) { return; }

    Point4D point(p);
    point.y = m_minY;

    if ( row == 0 && col == 0 )
    {
        *m_points[ m_rows * m_cols ] = point;
    }

    if ( row == 0 && col == m_cols - 1 )
    {
        *m_points[ m_rows * m_cols  + 1 ] = point;
    }

    if ( row == m_rows - 1 && col == 0 )
    {
        *m_points[ m_rows * m_cols + 2 ] = point;
    }

    if ( row == m_rows - 1 && col == m_cols - 1 )
    {
        *m_points[ m_rows * m_cols + 3 ] = point;
    }
}


///   3 - 0
///   |   |
///   2 - 1
void Surface3D::createIndicies()
{
    for ( size_t r = 1; r < m_rows; ++r )
    {
        for ( size_t c = 1; c < m_cols; ++c )
        {
            int idx0 = m_cols * r + c;
            int idx1 = m_cols * ( r - 1 ) + c;
            int idx2 = m_cols * ( r - 1 ) + c - 1;
            int idx3 = m_cols * r + c - 1;

            m_indicies.Add( idx0 );
            m_indicies.Add( idx1 );
            m_indicies.Add( idx1 );
            m_indicies.Add( idx2 );
            m_indicies.Add( idx2 );
            m_indicies.Add( idx3 );
            m_indicies.Add( idx3 );
            m_indicies.Add( idx0 );
        }
    }

    /// add corners edges
    if ( m_rows <= 1 || m_cols <= 1 ) { return; }

    /// horizontal edges
    m_indicies.Add( m_rows * m_cols );
    m_indicies.Add( m_rows * m_cols + 1 );
    m_indicies.Add( m_rows * m_cols + 1);
    m_indicies.Add( m_rows * m_cols + 3 );
    m_indicies.Add( m_rows * m_cols + 3 );
    m_indicies.Add( m_rows * m_cols + 2 );
    m_indicies.Add( m_rows * m_cols + 2 );
    m_indicies.Add( m_rows * m_cols );

    /// vertical edges
    m_indicies.Add( m_rows * m_cols );
    m_indicies.Add( 0 );
    m_indicies.Add( m_rows * m_cols + 1 );
    m_indicies.Add( m_cols - 1 );
    m_indicies.Add( m_rows * m_cols + 2 );
    m_indicies.Add( m_rows * m_cols - m_cols );
    m_indicies.Add( m_rows * m_cols + 3 );
    m_indicies.Add( m_rows * m_cols - 1 );
}



Surface3D::Surface3D( const Point4DArray& points, int cols, int rows )
{
    if ( points.GetCount() != (size_t)( cols * rows + 4 ) || !points.GetCount() ) {  return; }

    for ( size_t i = 0; i < points.GetCount(); ++i )
    {
        m_points.Add( new Point4D( *points[i] ) );
    }

    m_cols = cols;
    m_rows = rows;
}


void Surface3D::draw( wxDC& dc, int w, int h )
{
    findInvisibleEdges();

    dc.SetPen( m_color );

    int x0 = 0;
    int y0 = 0;
    int x1 = 0;
    int y1 = 0;

    for ( size_t i = 1; i < m_indicies.GetCount(); i +=2 )
    {
        if ( m_invisibleEdges.Index( i ) != wxNOT_FOUND ) { continue; }

        x0 = (int)(m_transformedPoints[ m_indicies[i] ]->x);
        y0 = h - (int)(m_transformedPoints[ m_indicies[i] ]->y);
        x1 = (int)(m_transformedPoints[ m_indicies[i-1] ]->x);
        y1 = h - (int)(m_transformedPoints[ m_indicies[i-1] ]->y);

        dc.DrawLine( x0, y0, x1, y1 );
    }

}


void Surface3D::findInvisibleEdges()
{
    if ( m_rows <= 1 || m_cols <= 1 ) { return; }

    /// show visible edges
    IntArray cornerPoints;

    cornerPoints.Add( m_rows * m_cols );
    cornerPoints.Add( m_rows * m_cols + 1 );
    cornerPoints.Add( m_rows * m_cols + 2 );
    cornerPoints.Add( m_rows * m_cols + 3 );
    cornerPoints.Add( 0 );
    cornerPoints.Add( m_cols - 1 );
    cornerPoints.Add( m_rows * m_cols - m_cols );
    cornerPoints.Add( m_rows * m_cols - 1 );

    /// find invisible point
    size_t idxMinZ = 0;
    double minZ = m_transformedPoints[ cornerPoints[idxMinZ] ]->z;

    for( size_t i = 1; i < 4; ++i )
    {
        double z = m_transformedPoints[ cornerPoints[i] ]->z;

        if ( z < minZ )
        {
            minZ = z;
            idxMinZ = i;
        }
    }

    cornerPoints.RemoveAt( idxMinZ );

    m_invisibleEdges.Clear();

    size_t firstEdgeIndex = ( m_rows - 1 ) * ( m_cols - 1 ) * 8;

    for ( size_t i = firstEdgeIndex + 1; i < m_indicies.GetCount(); i += 2 )
    {
        int idx0 = m_indicies[i-1];
        int idx1 = m_indicies[i];

        if ( cornerPoints.Index( idx0 ) == wxNOT_FOUND || cornerPoints.Index( idx1 ) == wxNOT_FOUND )
        {
            m_invisibleEdges.Add( i-1 );
            m_invisibleEdges.Add( i );
        }
    }
}

const Point4D& Surface3D::getPoint( size_t row, size_t col )
{
    return *m_points[ m_cols * row + col ];
}


const Point4D& Surface3D::getPoint( size_t index )
{
    return *m_points[ index ];
}


const Point4D& Surface3D::getTransformedPoint( size_t row, size_t col )
{
    return *m_transformedPoints[ m_cols * row + col ];
}


void Surface3D::setPoint( size_t row, size_t col, const Point4D& p )
{
    *m_points[ m_cols * row + col ] = p;
    setCornerPoint( row, col, p );
}


void Surface3D::transform( const Matrix* m )
{
    for ( size_t i = 0; i < m_points.GetCount(); ++i )
    {
        *m_transformedPoints[i] = Basic3DMath::transform( *m_points[i], *m );
        //Logger::logf( LOG_MSG, _T("Surface3D::transform() - m_transformedPoints[%d] = ") + m_transformedPoints[i]->toString(), i );
    }


}

void Surface3D::homDiv()
{
    for ( size_t i = 0; i < m_points.GetCount(); ++i )
    {
        m_transformedPoints[i]->homDiv();
    }
}

void Surface3D::setMinY( double min )
{
    m_minY = min;

    if ( m_rows <= 1 || m_cols <= 1 ) { return; }

    size_t firstCornerIndex = m_rows * m_cols;

    for ( size_t i = firstCornerIndex; i < m_points.GetCount(); ++i )
    {
        m_points[i]->y = m_minY;
    }
}


///---------------------------------------------------------------------------------------------Marker3D
Marker3D::Marker3D( const Point3D& p, const Point3D& p0, int size, MarkerType m, wxColor c, bool drawLines )
{
    init( p, p0, size, m, c, drawLines );
}


void Marker3D::init( const Point3D& p, const Point3D& p0, int size, MarkerType m, wxColor c, bool drawLines )
{
    m_size = size;
    m_color = c;
    m_markerType = m;
    m_isDrawLines = drawLines;
    m_points.Add( new Point4D( p.x, p0.y, p.z ) );
    m_points.Add( new Point4D( p ) );               // marked point
    m_points.Add( new Point4D( p.x, p0.y, p0.z ) );
    m_points.Add( new Point4D( p0.x, p0.y, p.z ) );
}


void Marker3D::draw( wxDC& dc, int w, int h )
{
    dc.SetPen( m_color );
    dc.SetBrush( m_color );

    if ( m_isDrawLines )
    {
        for ( size_t i = 1; i < 4; ++i )
        {
            dc.DrawLine( (int)( m_points[ 0 ]->x ),
                             (int)( h - m_points[ 0 ]->y ),
                             (int)( m_points[ i ]->x),
                             (int)( h - m_points[ i ]->y ) );
        }
    }

    int x = (int)(m_points[ 1 ]->x);
    int y = h - (int)(m_points[ 1 ]->y);
    int r = m_size;

    switch ( m_markerType )
    {
        case MARKER_CIRCLE :
            dc.DrawCircle( x, y, r );
            break;
        case MARKER_CROSS :
            wxPen p = dc.GetPen();
            p.SetWidth( 2 );
            dc.SetPen( p );

            dc.DrawLine( x - r, y - r, x - 1, y - 1 );
            dc.DrawLine( x + r, y - r, x + 1, y - 1 );
            dc.DrawLine( x - r, y + r, x - 1, y + 1 );
            dc.DrawLine( x + r, y + r, x + 1, y + 1 );
            break;
    }
}


Axis3D::Axis3D()
{
    createPoints( Point4D() );
    createIndicies();
    createColors();

    m_isVisible = true;
}


Axis3D::~Axis3D()
{
    //TODO
}


void Axis3D::draw( wxDC& dc, int w, int h )
{
    if ( !m_isVisible ) { return; }

    size_t i = 1;
    size_t c = 0;

    for ( ; i < m_indicies.GetCount(); i += 2, ++c )
    {
        int idx0 = m_indicies[i-1];
        int idx1 = m_indicies[i];

        const Point4D& p0 = *m_points[idx0];
        const Point4D& p1 = *m_points[idx1];

        dc.SetPen( *m_colors[c] );

        dc.DrawLine( (int)p0.x, (int)( h - p0.y ), (int)( p1.x), (int)( h - p1.y ) );
    }
}


void Axis3D::setAxisColour( Axis axis, wxColour color )
{
    *m_colors[ axis ] = color;
}


void Axis3D::setAxisLength( Axis axis, double length )
{
    switch ( axis )
    {
        case X_AXIS:
            m_points[1]->x = length;
            break;
        case Y_AXIS:
            m_points[2]->y = length;
            break;
        case Z_AXIS:
            m_points[3]->z = length;
            break;
    }
}


void Axis3D::setCenter( const Point4D& center )
{
    createPoints( center );
}


wxColour Axis3D::getAxisColour( Axis axis ) const
{
    return *m_colors[ axis ];
}


void Axis3D::createPoints( const Point4D& pc )
{
    m_points.Clear();

    m_points.Add( new Point4D( pc ) );
    m_points.Add( new Point4D( pc.x + 1.0, pc.y, pc.z ) );
    m_points.Add( new Point4D( pc.x, pc.y + 1.0, pc.z ) );
    m_points.Add( new Point4D( pc.x, pc.y, pc.z + 1.0 ) );
}


void Axis3D::createIndicies()
{
    m_indicies.Add( 0 );
    m_indicies.Add( 1 );
    m_indicies.Add( 0 );
    m_indicies.Add( 2 );
    m_indicies.Add( 0 );
    m_indicies.Add( 3 );
}


void Axis3D::createColors()
{
    m_colors.Add( new wxColor( *wxWHITE ) );
    m_colors.Add( new wxColor( *wxWHITE ) );
    m_colors.Add( new wxColor( *wxWHITE ) );
}
