#include "LogDirManager.h"
#include "DirHelper.h"
#include "wxSqliteHelper.h"

#include <wx/wx.h>
#include <wx/regex.h>
#include <wx/stdpaths.h>
#include <wx/filename.h>

#include "gui/DT.h"

#include "logger.h"

const wxString VEMSFiles = _T("VEMS_Files");

wxString LogDirManager::m_dbpath = wxEmptyString;
wxString LogDirManager::m_defworkdir = wxEmptyString;
wxSQLite3Database LogDirManager::db;
RecentECUs LogDirManager::rECU[MAX_RECENT_ECU_NUM];

void LogDirManager::openDB()
{
    if ( db.IsOpen() ) return;
    wxString dbfile = m_dbpath + _T("workingdirs.db");
    db.Open ( dbfile );
}

void LogDirManager::dropTables()
{
    openDB();
    try
    {
        wxString qs = _T("DROP TABLE [workingdirs]"); db.ExecuteQuery(qs);
        qs = _T("DROP TABLE [serialnumber]"); db.ExecuteQuery(qs);
    }
    catch (wxSQLite3Exception e)
    {
        Logger::logf(LOG_MSG, wxT("Database error. Check/delete workingdirs.db"));
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
    }
    checkDb();
}

void LogDirManager::SetDbPath(const wxString & dbpath)
{
    m_dbpath = dbpath;
}

void LogDirManager::checkDb()
{
    wxString dbName = m_dbpath + DirHelper::Separ() + _T("workingdirs.db");

    //working dir table
    // CREATE TABLE [workingdirs] ([id] integer  PRIMARY KEY AUTOINCREMENT NOT NULL,[path] varchar(255)  NULL)
    ColumnDescArray wd_cols;

    ColumnDesc wd_id( _T("id"), _T("INTEGER"), wxNullVariant, true, true );
    ColumnDesc wd_path( _T("path"), _T("VARCHAR(255)"), wxNullVariant );
    wd_cols.Add( wd_id ); wd_cols.Add( wd_path );

    // serialnumber table
    // CREATE TABLE [serialnumber] ([serialnum] varchar(255)  UNIQUE NULL,[name] varchar(255)  NULL, [tstamp] DATETIME default CURRENT_TIMESTAMP NULL)
    ColumnDescArray sn_cols;
    ColumnDesc sn_sn( _T("serialnum"), _T("VARCHAR(255)"), wxNullVariant, false, false, false );
    ColumnDesc sn_name( _T("name"), _T("VARCHAR(255)"), wxNullVariant );
    wxVariant dtvar;
	dtvar = _T("CURRENT_TIMESTAMP");
	wxString type = dtvar.GetType();
    ColumnDesc sn_tstamp( _T("tstamp"), _T("DATETIME"), dtvar );

    sn_cols.Add( sn_tstamp ); sn_cols.Add( sn_sn ); sn_cols.Add( sn_name );

    //
    ColumnDescArrayHash tables;
    tables[_T("workingdirs")] = wd_cols;
    tables[_T("serialnumber")] = sn_cols;

    wxSqliteHelper::DatabaseCheck( dbName, tables );
}

void LogDirManager::SetDefaultWorkDir(const wxString & workdir)
{
    openDB();
    checkDb(); // create tables if they don't exist

    wxString qs = _T("INSERT INTO [workingdirs] (path) VALUES ('") + workdir + _T("')");
    try {

        wxString qs1 = _T("SELECT path FROM workingdirs ORDER BY id DESC LIMIT 0,1");
        wxSQLite3ResultSet res = db.ExecuteQuery(qs1);
        if (res.Eof())
        {
            db.ExecuteQuery(qs);
        }
        else
        {
            res.NextRow();
            wxString currDir = res.GetString(_T("path"));
            if ( currDir != workdir )
                db.ExecuteQuery(qs);
        }

        m_defworkdir = workdir;
        DirHelper::SetPath(dWORKINGDIR, workdir);

        wxString logdir = workdir;

        if ( selectedECU() != wxEmptyString ) logdir += DirHelper::Separ() + selectedECU();

        if ( selectedECU() != wxEmptyString )
        {
            if (wxDirExists( logdir ) == false)
            {
                DirHelper::MkDir(logdir);
            }
        }
        DirHelper::SetPath(dLOG, logdir);

    }
    catch (wxSQLite3Exception e)
    {
        Logger::logf(LOG_MSG, wxT("Database error. Check/delete workingdirs.db"));
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
    }
}

wxString LogDirManager::GetDefaultWorkDir()
{
    openDB();

    checkDb(); // create tables if they don't exist
    wxString qs = _T("SELECT path FROM workingdirs ORDER BY id DESC LIMIT 0,1");
    try
    {
        wxSQLite3ResultSet res = db.ExecuteQuery( qs );

        // default is MyDocuments/VEMS_Files // ~/VEMS_Files
        if ( res.Eof() )
        {
            wxStandardPaths * paths = new wxStandardPaths();
            wxString logDir = paths->GetDocumentsDir() + wxFileName::GetPathSeparator() + VEMSFiles;
            if (wxDirExists( logDir ) == false)
            {
                DirHelper::MkDir(logDir);
            }
            SetDefaultWorkDir( logDir );
            return logDir;
        }
        else{
            DirHelper::SetPath(dWORKINGDIR, res.GetString(_T("path") ));

            wxString logdir = res.GetString(_T("path"));
            if ( selectedECU() != wxEmptyString ) logdir += DirHelper::Separ() + selectedECU();

            if ( selectedECU() != wxEmptyString )
            {
                if (wxDirExists( logdir ) == false)
                {
                    DirHelper::MkDir(logdir);
                }
            }
            //DirHelper::SetPath(dLOG, logdir);

            return res.GetString(_T("path"));
        }
    }
    catch (wxSQLite3Exception e)
    {
        Logger::logf(LOG_MSG, wxT("Database error. Check/delete workingdirs.db"));
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
        return wxEmptyString;
    }
}

wxArrayString LogDirManager::GetRecentECUs(int limit)
{
    openDB();

    checkDb(); // create tables if they don't exist
    wxString qs;
    /*if (limit != -1)
        qs = _T("SELECT serialnum, name FROM serialnumber ORDER BY tstamp DESC LIMIT 0,")+wxString::Format(_T("%d"), limit);
    else qs = _T("SELECT serialnum, name FROM serialnumber ORDER BY tstamp DESC name, serialnum ASC");*/

    qs = wxString::Format( _T("SELECT serialnum, name FROM serialnumber ORDER by tstamp DESC LIMIT 0,%d"), limit ) ;

    wxArrayString result;

    try {
        wxSQLite3ResultSet res = db.ExecuteQuery( qs );

        if ( res.Eof() ) return result;
        else
        {
            while ( res.NextRow() )
            {
                wxString id = res.GetString(_T("serialnum"));

                // round
//                if ( id.StartsWith(_T("R") )
//                    id = id.Mid( 1 );

                wxString name = res.GetString(_T("name"));
                wxString dir = name + wxT("-") + id;
                result.Add(dir);
            }
        }
        return result;
    }
    catch (wxSQLite3Exception e)
    {
        Logger::logf(LOG_MSG, wxT("Database error. Check/delete workingdirs.db"));
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
        return result;
    }
}

wxArrayString LogDirManager::GetECUList()
{
    // select all
    return GetRecentECUs(-1);
}

wxString LogDirManager::selectedECU()
{
    wxArrayString s = GetRecentECUs();
    if (s.GetCount() == 0) return wxEmptyString;
    else return s.Item(0);
}

RecentECUs & LogDirManager::GetRecentECU(int i)
{
    return rECU[i];
}

void LogDirManager::SetRecentECU(int i, RecentECUs& r)
{
    rECU[i] = r;
}

/// ECU Serial Number handlers (get,add,update,delete)

wxArrayString LogDirManager::searchForDirectories( unsigned int SerialNumber, wxString ECUType )
{
    wxString ECUSerial = SNToString( SerialNumber, ECUType );

    wxArrayString tmp;
    DT traverser( tmp );
    wxDir dir( DirHelper::GetPath( dWORKINGDIR ) );
    dir.Traverse( traverser );

    // subdir name format (name)-(number)
    wxRegEx regN(wxT("^([[:alnum:]_.-]+)-") + ECUSerial + _T("$"));

    wxArrayString foundDirs;

    for (unsigned int i = 0; i < tmp.GetCount(); ++i)
    {
        // tmp[i] is a fullpath
        wxString subdirName = tmp[i].AfterLast( wxFileName::GetPathSeparator() );

        if ( regN.Matches(subdirName) )
        {
            subdirName = subdirName.BeforeLast('-');
            foundDirs.Add( subdirName );
        }
    }

    // sort them and create gui
    foundDirs.Sort();

    return foundDirs;
}

wxString LogDirManager::getNameBySerialNumber(unsigned int SerialNumber, wxString ECUType )
{
    wxString ECUSerial = SNToString(SerialNumber, ECUType);
    return getNameBySerialNumber( ECUSerial );
}

wxString LogDirManager::getNameBySerialNumber(const wxString& SerialNumber)
{
    openDB();

    checkDb(); // create tables if they don't exist
    wxString qs = _T("SELECT name FROM serialnumber WHERE serialnum = '") + SerialNumber + _T("'");

    wxString name = wxEmptyString;

    try
    {
        wxSQLite3ResultSet res = db.ExecuteQuery( qs );
        if ( res.Eof() ) return name;
        else return res.GetString(_T("name"));
    }
    catch (wxSQLite3Exception e)
    {
        Logger::logf(LOG_MSG, wxT("Database error. Check/delete workingdirs.db"));
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
        return name;
    }

}

void LogDirManager::updateSerialNumber(unsigned int SerialNumber, wxString ECUType )
{
    wxString ECUSerial = SNToString( SerialNumber, ECUType );
    updateSerialNumber( ECUSerial );
}

void LogDirManager::updateSerialNumber(const wxString& SerialNumber)
{
    openDB();
    checkDb();
    wxString qs = wxT("UPDATE serialnumber SET tstamp = CURRENT_TIMESTAMP WHERE serialnum='") + SerialNumber + wxT("'");

    try {
        wxSQLite3ResultSet res = db.ExecuteQuery( qs );
    }
    catch (wxSQLite3Exception e)
    {
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
    }
}

bool LogDirManager::insertSerialNumber(unsigned int SerialNumber, const wxString & name, wxString ECUType, int timestamp)
{
    wxString ECUSerial = SNToString( SerialNumber, ECUType );
    return insertSerialNumber( ECUSerial, name, timestamp);
}

bool LogDirManager::insertSerialNumber(const wxString & SerialNumber, const wxString & name, int timestamp)
{
    openDB();
    checkDb(); // create tables if they don't exist
    wxString qs;
    if ( timestamp == -1 )
        qs = _T("INSERT INTO serialnumber (serialnum, name, tstamp) VALUES ('") + SerialNumber + _T("','") + name + _T("', CURRENT_TIMESTAMP )");
    else
        qs = _T("INSERT INTO serialnumber (serialnum, name, tstamp) VALUES ('") + SerialNumber + _T("','") + name + _T("', ") +
            wxString::Format(_T("%d"), timestamp) + _T(")");

    try {
        wxSQLite3ResultSet res = db.ExecuteQuery( qs );
        wxString logDir = m_defworkdir + wxFileName::GetPathSeparator() + name + _T("-") + SerialNumber;
        if (wxDirExists( logDir ) == false)
        {
            DirHelper::MkDir(logDir);
        }
        // SetDefaultWorkDir( logDir );
        return true;
    }
    catch (wxSQLite3Exception e)
    {
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
        return false;
    }
}

bool LogDirManager::deleteSerialNumber(unsigned int SerialNumber, wxString ECUType)
{
    wxString ECUName = SNToString( SerialNumber, ECUType );
    return deleteSerialNumber(ECUName);
}

bool LogDirManager::deleteSerialNumber(const wxString& SerialNumber)
{
    openDB();

    checkDb(); // create tables if they don't exist
    wxString qs = _T("DELETE FROM serialnumber WHERE serialnum = '") + SerialNumber + _T("'");

    wxString name = wxEmptyString;

    try
    {
        db.ExecuteUpdate( qs );
        return true;
    }
    catch (wxSQLite3Exception e)
    {
        Logger::logf(LOG_MSG, wxT("Database error. Check/delete workingdirs.db"));
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
        return false;
    }

}

int LogDirManager::checkForUsedECU()
{
    openDB();

    checkDb(); // create tables if they don't exist
    wxString qs = _T("SELECT tstamp FROM serialnumber ORDER BY tstamp desc");

    try
    {
        wxSQLite3ResultSet res = db.ExecuteQuery( qs );
        wxString tstamp = res.GetString(0);
        Logger::logf(LOG_MSG, _T("checkedForUsedECU() num = ")+ tstamp);
        return (tstamp != wxEmptyString && tstamp != wxT("0"));
    }
    catch (wxSQLite3Exception e)
    {
        Logger::logf(LOG_MSG, wxT("Database error. Check/delete workingdirs.db"));
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
        return 0;
    }
}

wxString LogDirManager::SNToString(unsigned num, wxString ECUType )
{
    wxString ret;
    if ( ECUType == _T("") )
        ret = wxString::Format(_T("%d"), num );
    else
    {
		wxString et = (ECUType.MakeUpper()).Left(1);
        ret = et + wxString::Format(_T("%d"), num);
    }

    return ret;
}

/// deprecated
/*
void LogDirManager::updateByName(const wxString& ECUName)
{
    openDB();
    checkDb();
    wxString qs = wxT("UPDATE serialnumber SET tstamp = CURRENT_TIMESTAMP WHERE name='") + ECUName + wxT("'");

    try {
        wxSQLite3ResultSet res = db.ExecuteQuery( qs );
    }
    catch (wxSQLite3Exception e)
    {
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
    }
}

bool LogDirManager::ecuNameExists(const wxString& ECUName)
{
    openDB();
    checkDb();
    wxString qs = _T("SELECT * FROM serialnumber WHERE name = '") + ECUName + _T("'");
    try {
        wxSQLite3ResultSet res = db.ExecuteQuery( qs );
        if ( res.Eof() ) return false;
        else return true;
    }
    catch (wxSQLite3Exception e)
    {
        wxString error = e.GetMessage();
        wxMessageBox(error, _T("Database error. Check workingdirs.db"));
        return false;
    }
}

*/
