#include "BitmapCanvasWidgetBase.h"

#include "../logger.h"

#include "../wxVariantHelper.h"
#include "../agg/wxAgg.h"
#include "../agg/wxAgg2.h"
#include "../vemsdialog/WindowTipperAgg.h"

#include "BitmapCanvasParentBase.h"
#include "bcWrapper.h"
#include "bcToolTip.h"

#include "BitmapCanvasSystemSettings.h"

BEGIN_EVENT_TABLE(BitmapCanvasWidgetBase, wxEvtHandler)

    EVT_TIMER(idSlideTimer , BitmapCanvasWidgetBase::OnSlideTimer)

    EVT_ENTER_WINDOW(BitmapCanvasWidgetBase::OnEnterWin)
    EVT_LEAVE_WINDOW(BitmapCanvasWidgetBase::OnLeaveWin)

END_EVENT_TABLE()

IMPLEMENT_ABSTRACT_CLASS(BitmapCanvasWidgetBase, wxEvtHandler);

/******************
    Constructors
*/

BitmapCanvasWidgetBase::BitmapCanvasWidgetBase()
{

}

BitmapCanvasWidgetBase::BitmapCanvasWidgetBase(BitmapCanvasParentBase* parent, int id, const wxPoint& pos, const wxSize& size, int z_index, long style)
    :m_mainTimer(this, idSlideTimer)
{
    //ctor
    CreateBcWidget(parent,id,pos,size,z_index,style);
}

BitmapCanvasWidgetBase::BitmapCanvasWidgetBase(BitmapCanvasParentBase* parent, wxVariantIntHash& properties)
    :m_mainTimer(this, idSlideTimer)
{
    CreateBcWidget(parent,properties);
}

BitmapCanvasWidgetBase::~BitmapCanvasWidgetBase()
{
    //dtor
    m_mainTimer.Stop();
    ptrdel(m_canvas);
    ptrdel(m_bitmapCanvas);
    ptrdel(m_disabledCanvas);

    if(m_canvasAlphaBackup) delete[] m_canvasAlphaBackup;

    m_parent->RemoveWidget(this);

    ptrdel(m_wrapper);

    ptrdel(m_toolTipWindow);
    //ptrdel(m_toolTipWidget);
}

/***********
    Init
*/

void BitmapCanvasWidgetBase::CreateBcWidget(BitmapCanvasParentBase* parent, wxVariantIntHash& properties)
{
    wxRealSize s = wxVariantHelper::GetRealSize(properties[BCWB_SIZE]);
    wxSize size((int)s.w, (int)s.h);
    wxRealPoint p = wxVariantHelper::GetRealPointEx(properties[BCWB_POSITION]);

    int id = properties.find(BCWB_ID) == properties.end() ? -1 : properties[BCWB_ID].GetLong();

    CreateBcWidget(parent,id,wxPoint((int)p.x,(int)p.y),size);

    m_properties = properties;
}

void BitmapCanvasWidgetBase::CreateBcWidget(BitmapCanvasParentBase* parent, int id, const wxPoint& pos, const wxSize& size, int z_index, long style)
{
    m_toolTipWindow = NULL;
    m_toolTipWidget = NULL;
    m_canvas = NULL;
    m_bitmapCanvas = NULL;
    m_canvasAlphaBackup = NULL;
    m_disabledCanvas = NULL;
    m_wrapper = NULL;
    m_style = style;

    if(id == wxID_ANY) id = wxNewId();

    m_properties[BCWB_ID] = (long)id;
    m_properties[BCWB_ISSHOWN] = true;
    m_properties[BCWB_ISENABLED] = true;

    m_parent = parent;
    parent->AddWidget(this);

    m_properties[BCWB_POSITION] << wxRealPointEx(pos);
    m_properties[BCWB_SIZE] << wxRealSize(size);

    m_properties[BCWB_Z_INDEX] = (long)(z_index == bcDefaultZindex ? parent->GetNewZindex() : z_index);


    m_timeBase = 20; // 50 fps

    m_targetPos = wxDefaultPosition;
    m_timerShoots = 0;

    m_currentAlpha = ALPHA_MAX;
    m_targetAlpha = ALPHA_MAX;

    m_slideType = 0;
}


/*************
    Setters
*/

void BitmapCanvasWidgetBase::SetProperties(wxVariantIntHash hash, bool refresh)
{
    for(wxVariantIntHash::iterator it = hash.begin(); it != hash.end(); it++) {
        m_properties[ it->first ] = it->second;
    }

    if(refresh) Refresh();
}

void BitmapCanvasWidgetBase::SetProperty(int key, const wxVariant& value, bool refresh)
{
    m_properties[ key ] = value;

    if(refresh) Refresh();
}

void BitmapCanvasWidgetBase::SetProperty(int key, bool val, bool refresh)
{
    m_properties[ key ] = val;
    if(refresh) Refresh();
}

void BitmapCanvasWidgetBase::SetProperty(int key, double val, bool refresh)
{
    m_properties[ key ] = val;
    if(refresh) Refresh();
}

void BitmapCanvasWidgetBase::SetProperty(int key, long val, bool refresh)
{
    m_properties[ key ] = val;
    if(refresh) Refresh();
}

void BitmapCanvasWidgetBase::SetProperty(int key, const wxColor& val, bool refresh)
{
    m_properties[ key ] = wxVariantHelper::ToVariant(val);
    if(refresh) Refresh();
}

void BitmapCanvasWidgetBase::SetProperty(int key, const wxRealSize& val, bool refresh)
{
    m_properties[ key ] = wxVariantHelper::ToVariant(val);
    if(refresh) Refresh();
}

void BitmapCanvasWidgetBase::SetProperty(int key, const wxRealPointEx& val, bool refresh)
{
    m_properties[ key ] = wxVariantHelper::ToVariant(val);
    if(refresh) Refresh();
}

void BitmapCanvasWidgetBase::SetProperty(int key, const wxString& val, bool refresh)
{
    m_properties[ key ] = val;
    if(refresh) Refresh();
}

void BitmapCanvasWidgetBase::SetStyle(long style)
{
    m_style = style;
    Refresh();
}

void BitmapCanvasWidgetBase::SetStyleFlag(long flag)
{
    if( m_style & flag ) return;
    m_style |= flag;
    Refresh();
}

void BitmapCanvasWidgetBase::DelStyleFlag(long flag)
{
    if( !(m_style & flag) ) return;
    m_style &= ~flag;
    Refresh();
}

void BitmapCanvasWidgetBase::SwapZindex(BitmapCanvasWidgetBase* widget)
{
    int oz = widget->GetZindex();
    widget->SetZindex( GetZindex(), false );
    SetZindex( oz );
}

void BitmapCanvasWidgetBase::SetZindex(int idx, bool refresparent)
{
    m_properties[BCWB_Z_INDEX] = (long)idx;
    if(refresparent) RefreshParent();
}

void BitmapCanvasWidgetBase::SetPosition(const wxPoint& pos)
{
    SetPosition(wxRealPoint(pos.x,pos.y));
}

void BitmapCanvasWidgetBase::SetPosition(const wxRealPoint& pos)
{
    wxRect refreshrect = GetRect();
    m_properties[BCWB_POSITION] << wxRealPointEx(pos);
    refreshrect.Union( GetRect() );
    m_parent->RefreshWindow(refreshrect);
}

void BitmapCanvasWidgetBase::SetSize(const wxRealSize& size)
{
    wxRect refreshrect = GetRect();
    if( size.w > GetSize().w) refreshrect.width = (int)ceil(size.w);
    if( size.h > GetSize().h) refreshrect.height = (int)ceil(size.h);

    m_properties[BCWB_SIZE] << size;
    DrawWidget();

    refreshrect.Union( GetRect() );

    m_parent->RefreshWindow(refreshrect);
}

void BitmapCanvasWidgetBase::SetPosition(double x, double y)
{
    SetPosition(wxRealPoint(x,y));
}

void BitmapCanvasWidgetBase::SetSize(double w, double h)
{
    SetSize(wxRealSize(w,h));
}

void BitmapCanvasWidgetBase::Centre()
{
    wxSize size = GetSize().Ceil();
    wxSize parent_size = m_parent->GetCanvasSize();

    SetPosition( (parent_size.x - size.x)/2, (parent_size.y - size.y)/2 );
}

/*************
    Getters
*/

wxBitmap* BitmapCanvasWidgetBase::GetBitmap(bool enabled)
{
    return m_bitmapCanvas;//enabled ? m_bitmapCanvas : m_disabledCanvas;
}

long BitmapCanvasWidgetBase::GetStyle()
{
    return m_style;
}

wxRealSize BitmapCanvasWidgetBase::GetSize()
{
    return wxVariantHelper::GetRealSize( m_properties[BCWB_SIZE] );
}

wxRealPointEx BitmapCanvasWidgetBase::GetPosition()
{
    return wxVariantHelper::GetRealPointEx( m_properties[BCWB_POSITION] );
}

#define PROPCHECK if( m_properties.find(key) == m_properties.end() ) return defval;

bool BitmapCanvasWidgetBase::GetPropBool(int key, bool defval)
{
    PROPCHECK
    return m_properties[key].GetBool();
}

wxRealPoint BitmapCanvasWidgetBase::GetPropRealPoint(int key, const wxRealPoint& defval)
{
    PROPCHECK
    return wxVariantHelper::GetRealPointEx(m_properties[key]);
}

wxRealSize BitmapCanvasWidgetBase::GetPropRealSize(int key, const wxRealSize& defval)
{
    PROPCHECK
    return wxVariantHelper::GetRealSize(m_properties[key]);
}

long BitmapCanvasWidgetBase::GetPropLong(int key, long defval)
{
    PROPCHECK
    return m_properties[key].GetLong();
}

double BitmapCanvasWidgetBase::GetPropDouble(int key, double defval)
{
    PROPCHECK
    return m_properties[key].GetDouble();
}

wxColor BitmapCanvasWidgetBase::GetPropColor(int key, const wxColor& defval)
{
    PROPCHECK
    return wxVariantHelper::GetColor(m_properties[key]);
}

wxSize BitmapCanvasWidgetBase::GetPropSize(int key, const wxSize& defval)
{
    PROPCHECK
    return wxVariantHelper::GetSizeEx(m_properties[key]);
}

wxPoint BitmapCanvasWidgetBase::GetPropPoint(int key, const wxPoint& defval)
{
    PROPCHECK
    return wxVariantHelper::GetPointEx(m_properties[key]);
}

wxString BitmapCanvasWidgetBase::GetPropString(int key, const wxString& defval)
{
    PROPCHECK
    return m_properties[key].GetString();
}

bool BitmapCanvasWidgetBase::GetSysBool(int key)
{
    if( m_properties.find(key) == m_properties.end() ) return BitmapCanvasSystemSettings::GetBool(key);

    return m_properties[key].GetBool();
}

long BitmapCanvasWidgetBase::GetSysLong(int key)
{
    if( m_properties.find(key) == m_properties.end() ) return BitmapCanvasSystemSettings::GetLong(key);

    return m_properties[key].GetLong();
}

double BitmapCanvasWidgetBase::GetSysDouble(int key)
{
    if( m_properties.find(key) == m_properties.end() ) return BitmapCanvasSystemSettings::GetDouble(key);

    return m_properties[key].GetDouble();
}

wxColor BitmapCanvasWidgetBase::GetSysColor(int key)
{
    if( m_properties.find(key) == m_properties.end() ) return BitmapCanvasSystemSettings::GetColor(key);

    return wxVariantHelper::GetColor(m_properties[key]);
}

wxSize BitmapCanvasWidgetBase::GetSysSize(int key)
{
    if( m_properties.find(key) == m_properties.end() ) return BitmapCanvasSystemSettings::GetSize(key);

    return wxVariantHelper::GetSizeEx(m_properties[key]);
}

wxRealSize BitmapCanvasWidgetBase::GetSysRealSize(int key)
{
    if( m_properties.find(key) == m_properties.end() ) return BitmapCanvasSystemSettings::GetRealSize(key);

    return wxVariantHelper::GetRealSize(m_properties[key]);
}

wxRealPoint BitmapCanvasWidgetBase::GetSysRealPoint(int key)
{
    if( m_properties.find(key) == m_properties.end() ) return BitmapCanvasSystemSettings::GetRealPoint(key);

    return wxVariantHelper::GetRealPointEx(m_properties[key]);
}

wxPoint BitmapCanvasWidgetBase::GetSysPoint(int key)
{
    if( m_properties.find(key) == m_properties.end() ) return BitmapCanvasSystemSettings::GetPoint(key);

    return wxVariantHelper::GetPointEx(m_properties[key]);
}

wxString BitmapCanvasWidgetBase::GetSysString(int key)
{
    if( m_properties.find(key) == m_properties.end() ) return BitmapCanvasSystemSettings::GetString(key);

    return m_properties[key].GetString();
}

wxWindow* BitmapCanvasWidgetBase::GetWindowParent()
{
    wxWindow* parent = dynamic_cast<wxWindow*>( m_parent );
    return parent;
}

wxRect BitmapCanvasWidgetBase::GetRect()
{
    wxRealSize s = GetSize();
    wxRealPoint p = GetPosition();
    return wxRect( (int)p.x, (int)p.y, (int)s.w, (int)s.h );
}

wxRealRect BitmapCanvasWidgetBase::GetRRect()
{
    return wxRealRect(GetPosition(), GetSize());
}

/*******************
    widget render
*/


void BitmapCanvasWidgetBase::render(wxAgg2& agg)
{
    ptrdel(m_canvas);
    if(!agg.isOk()) return;
    m_canvas = agg.renderCopyToImage();
    CopyCanvasToBitmap();
    AlphaBackup();
    //CreateDisabledImage();

    if(m_currentAlpha != ALPHA_MAX) SetAlpha(m_currentAlpha,false);
}

void BitmapCanvasWidgetBase::render(wxAgg& agg)
{
    ptrdel(m_canvas);
    m_canvas = agg.renderCopyToImage();
    CopyCanvasToBitmap();
    AlphaBackup();
    //CreateDisabledImage();

    if(m_currentAlpha != ALPHA_MAX) SetAlpha(m_currentAlpha,false);
}

/**************************
    widget base features
*/

void BitmapCanvasWidgetBase::CreateDisabledImage()
{
    ptrdel(m_disabledCanvas);
    m_disabledCanvas = new wxBitmap(m_canvas->ConvertToGreyscale( 0.299, 0.587, 0.114 ));
}

void BitmapCanvasWidgetBase::RefreshParent()
{
    m_parent->RefreshWindow();
}

void BitmapCanvasWidgetBase::AlphaBackup()
{
    if(!m_canvas) return;

    unsigned char* tmp = m_canvas->GetAlpha();

    if(tmp) {
        size_t size = m_canvas->GetWidth() * m_canvas->GetHeight();

        if(m_canvasAlphaBackup) delete[] m_canvasAlphaBackup;

        m_canvasAlphaBackup = new unsigned char[size];
        memcpy(m_canvasAlphaBackup,tmp,size);
    }
}

void BitmapCanvasWidgetBase::CopyCanvasToBitmap()
{
    ptrdel(m_bitmapCanvas);

    if(!m_canvas) return;

    m_bitmapCanvas = new wxBitmap(*m_canvas);
}

void BitmapCanvasWidgetBase::Update(bool redraw)
{
    if(redraw) {
        DrawWidget();
    }

    m_parent->UpdateWindow();
}

void BitmapCanvasWidgetBase::Refresh(bool redraw)
{
    if(redraw) {
        DrawWidget();
    }

    m_parent->RefreshWindow(GetRect());
}

void BitmapCanvasWidgetBase::SetCursor(const wxCursor& cursor)
{
    wxWindow* win = GetWindowParent();
    if(win)
        win->SetCursor(cursor);
}

void BitmapCanvasWidgetBase::SetCursor(int cursor)
{
    SetCursor(wxCursor(cursor));
}

void BitmapCanvasWidgetBase::DropEvent(wxEvent& evt)
{
    wxWindow* win = GetWindowParent();
    if(win)
        //win->AddPendingEvent(evt);
        win->ProcessEvent(evt);
}

void BitmapCanvasWidgetBase::Enable(bool enable)
{
    bool ise = GetPropBool(BCWB_ISENABLED);

    m_properties[BCWB_ISENABLED] = enable;

    if(ise != enable) Refresh();
}

void BitmapCanvasWidgetBase::Show(bool isshow)
{
    m_properties[BCWB_ISSHOWN] = isshow;

    m_parent->RefreshWindow();
}

void BitmapCanvasWidgetBase::StartSlide(long type)
{
    m_timerShoots = 0;
    m_slideType |= type;
    m_mainTimer.Start(m_timeBase);
}

void BitmapCanvasWidgetBase::SlidePosition(const wxPoint& pos)
{
    m_targetPos = pos;
    StartSlide(BCST_POSITION);
}

void BitmapCanvasWidgetBase::SetAlpha(int alpha, bool isRefresh)
{

    if( !m_canvas || !m_canvas->IsOk() || !m_canvasAlphaBackup ) return;

    unsigned char * alphaBuffer = m_canvas->GetAlpha();

    if( alphaBuffer == NULL ) return;

    chkminmax(alpha,ALPHA_MIN,ALPHA_MAX);

    //if(m_currentAlpha == alpha) return;

    size_t imgw = m_canvas->GetWidth();
    size_t imgh = m_canvas->GetHeight();
    size_t x,y,idx;
    double a,ratio;

    ratio = 1.0 / (ALPHA_MAX - ALPHA_MIN);

    if(alpha == ALPHA_MAX) {

        memcpy(alphaBuffer,m_canvasAlphaBackup,imgw*imgh);

    } else {

        for(y = 0; y < imgh; y++) {
            for(x = 0; x < imgw; x++) {
                idx = x + imgw * y;
                a = m_canvasAlphaBackup[idx];
                if(a < 1) continue;
                a = a * alpha * ratio;
                chkminmax(a,0,255);
                alphaBuffer[idx] = (int)a;
            }
        }
    }

    m_currentAlpha = alpha;

    CopyCanvasToBitmap();

    if(isRefresh) RefreshParent();
}

void BitmapCanvasWidgetBase::SlideAlpha(bool isMaxAlpha)
{
    if(!IsShown()) return;

    m_targetAlpha = isMaxAlpha ? ALPHA_MAX : ALPHA_MIN;

    StartSlide(BCST_ALPHA);
}

void BitmapCanvasWidgetBase::StopSlide(long type)
{
    m_slideType &= ~type;

    if(m_slideType == BCST_NOSLIDE) {
        m_mainTimer.Stop();
        m_timerShoots = 0;
    }
}

void BitmapCanvasWidgetBase::OnSlideTimer(wxTimerEvent& evt)
{
    if( m_timerShoots >= 100 || m_slideType == BCST_NOSLIDE ) StopSlide();

    if(m_slideType & BCST_POSITION) {

        wxRealPoint cpos = GetPosition();

        if(wxRealPointEx( m_targetPos ) == cpos) {
            StopSlide(BCST_POSITION);
            return;
        }

        wxRealSize distance = wxSizeEx(m_targetPos - GetRect().GetTopLeft());

        distance *= GetPropDouble(BCWB_SLIDE_SPEED,0.4);

        if(fabs(distance.w) < 1.0 && fabs(distance.h) < 1.0) {
            StopSlide(BCST_POSITION);
            SetPosition(m_targetPos);
            return;
        }

    //    Logger::logf(LOG_MSG,_T("distance: w=%f, h=%f"),distance.w,distance.h);

        wxRealPointEx npos = wxRealPoint(distance.w,distance.h) + cpos;

        wxVariant v;
        v << npos;

        SetProperty(BCWB_POSITION, v, false);
    }

    if(m_slideType & BCST_ALPHA) {

        int delta = (int)ceil(GetPropDouble(BCWB_SLIDE_SPEED,0.4) * (m_targetAlpha-m_currentAlpha));

        SetAlpha( abs(delta) < 1 ? m_targetAlpha : (m_currentAlpha + delta), false );

        if( m_currentAlpha == m_targetAlpha ) StopSlide(BCST_ALPHA);
    }

    RefreshParent();

    m_timerShoots++;
}

void BitmapCanvasWidgetBase::OnEnterWin(wxMouseEvent& event)
{
    if(m_toolTipWindow)
        m_toolTipWindow->ShowWindow( wxGetMousePosition() + wxPoint(0,16) );
    else if(m_toolTipWidget) {
        wxPoint pos = wxGetMousePosition() - GetWindowParent()->GetScreenPosition() + wxPoint(0,16);
        m_toolTipWidget->ShowTip( pos );
    }

    event.Skip(1);
}

void BitmapCanvasWidgetBase::OnLeaveWin(wxMouseEvent& event)
{
    if(m_toolTipWindow)
        m_toolTipWindow->HideWin();
    else if(m_toolTipWidget)
        m_toolTipWidget->HideTip();


    event.Skip(1);
}

void BitmapCanvasWidgetBase::SetToolTip(const wxString& msg, bool useBcToolTip)
{
    if(useBcToolTip) {
        if(!m_toolTipWidget) m_toolTipWidget = new bcToolTip(m_parent);
        m_toolTipWidget->SetTipMessage(msg);
    } else {
    if(!m_toolTipWindow) m_toolTipWindow = new WindowTipperAgg(GetWindowParent());
    m_toolTipWindow->SetMessage(msg);
}
}

void BitmapCanvasWidgetBase::SetBcToolTip(bcToolTip* tooltip)
{
    ptrdel(m_toolTipWidget);
    m_toolTipWidget = tooltip;
}

void BitmapCanvasWidgetBase::SetToolTip(WindowTipperAgg* tooltip)
{
    ptrdel(m_toolTipWindow);
    m_toolTipWindow = tooltip;
}

wxString BitmapCanvasWidgetBase::GetToolTipMsg()
{
    return m_toolTipWindow ? m_toolTipWindow->GetMessage() : _T("");
}

WindowTipperAgg* BitmapCanvasWidgetBase::GetToolTip()
{
    return m_toolTipWindow;
}

void BitmapCanvasWidgetBase::SetFocus()
{
    m_parent->SetWidgetFocus(this);
}

bool BitmapCanvasWidgetBase::IsFocused()
{
    return m_parent->m_focusedWidget == this ? true : false;
}

int BitmapCanvasWidgetBase::zindex_compare(BitmapCanvasWidgetBase** widget_A, BitmapCanvasWidgetBase** widget_B)
{
    int widA = (*widget_A)->GetId();
    int widB = (*widget_B)->GetId();

    int zidA = (*widget_A)->GetZindex();
    int zidB = (*widget_B)->GetZindex();

    return zidA != zidB ? ( zidA > zidB ? 1 : -1 ) : ( widA != widB ? ( widA > widB ? 1 : -1 ) : 0 );
}

