#ifndef BITMAPCANVASWIDGETBASE_H
#define BITMAPCANVASWIDGETBASE_H

#include <wx/wx.h>
#include "../wxRealSize.h"
#include "../wxRealRect.h"
#include "../wxRealPointEx.h"

#include "../common/types/wxVariantIntHash.h"

#define ptrdel(ptr) \
    if( (ptr) != NULL ) { delete (ptr); (ptr) = NULL; }

#define chkminmax(val,min,max) \
    if( (val) > (max) ) (val) = (max); \
    if( (val) < (min) ) (val) = (min);

#define rastu(val) ceil(val)+0.5
#define rastd(val) ceil(val)-0.5

#define ALPHA_MIN 0
#define ALPHA_MAX 100

#define bcDefaultZindex -1

class wxAgg;
class wxAgg2;
class bcWrapper;
class bcToolTip;
class WindowTipperAgg;
class BitmapCanvasParentBase;

enum BitmapCanvasWidgetBasePropTypes {
    BCWB_ID = 1000,
    BCWB_POSITION,
    BCWB_SIZE,
    BCWB_ISSHOWN,
    BCWB_ISENABLED,
    BCWB_SLIDE_SPEED,
    BCWB_Z_INDEX
};



class BitmapCanvasWidgetBase: public wxEvtHandler
{
    DECLARE_ABSTRACT_CLASS(BitmapCanvasWidgetBase);

    public:
        BitmapCanvasWidgetBase();
        BitmapCanvasWidgetBase(BitmapCanvasParentBase* parent, int id = wxID_ANY, const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxDefaultSize, int z_index = bcDefaultZindex, long style = 0);
        BitmapCanvasWidgetBase(BitmapCanvasParentBase* parent, wxVariantIntHash& properties);
        void CreateBcWidget(BitmapCanvasParentBase* parent, int id = wxID_ANY, const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxDefaultSize, int z_index = bcDefaultZindex, long style = 0);
        void CreateBcWidget(BitmapCanvasParentBase* parent, wxVariantIntHash& properties);

        virtual ~BitmapCanvasWidgetBase();

        virtual void DrawWidget() = 0;

        virtual void Create(BitmapCanvasParentBase* parent, wxVariantIntHash& properties) = 0;

        virtual bool Focusable() { return true; };

        void SetPosition(const wxPoint& pos);
        void SetPosition(const wxRealPoint& pos);
        void SetPosition(double x, double y);
        void SetSize(const wxRealSize& size);
        void SetSize(double w, double h);

        void Centre();

        int  GetZindex() { return GetPropLong(BCWB_Z_INDEX); }
        void SetZindex(int idx, bool refresparent = true);
        void SwapZindex(BitmapCanvasWidgetBase* widget);

        wxRect GetRect();
        wxRealRect GetRRect();

        BitmapCanvasParentBase* GetParent() { return m_parent; }

        int GetId() { return GetPropLong(BCWB_ID); };

        wxBitmap* GetBitmap(bool enabled = true);

        void Refresh(bool redraw = true);
        void Update(bool redraw = true);

        long GetStyle();
        void SetStyle(long style);
        void SetStyleFlag(long flag);
        void DelStyleFlag(long flag);

        bool        GetPropBool(int key, bool defval = false);
        long        GetPropLong(int key, long defval = 0L);
        double      GetPropDouble(int key, double defval = 0.0);
        wxColor     GetPropColor(int key, const wxColor& defval = wxNullColour);
        wxSize      GetPropSize(int key, const wxSize& defval = wxDefaultSize);
        wxRealSize  GetPropRealSize(int key, const wxRealSize& defval = wxDefaultRealSize);
        wxRealPoint GetPropRealPoint(int key, const wxRealPoint& defval = wxRealPoint(0,0));
        wxPoint     GetPropPoint(int key, const wxPoint& defval = wxDefaultPosition);
        wxString    GetPropString(int key, const wxString& defval = wxEmptyString);



        void SetProperties(wxVariantIntHash hash, bool refresh = false);
        void SetProperty(int key, const wxVariant& value, bool refresh = false);

        void SetProperty(int key, bool val, bool refresh = false);
        void SetProperty(int key, double val, bool refresh = false);
        void SetProperty(int key, long val, bool refresh = false);
        void SetProperty(int key, const wxColor& val, bool refresh = false);
        void SetProperty(int key, const wxRealSize& val, bool refresh = false);
        void SetProperty(int key, const wxRealPointEx& val, bool refresh = false);
        void SetProperty(int key, const wxString& val, bool refresh = false);


        wxRealSize      GetSize();
        wxRealPointEx   GetPosition();

        void Show(bool isshow = true);
        void Hide() { Show(false); }

        void Enable(bool enable=true);
        void Disable() { Enable(false); }

        bool IsShown() { return GetPropBool(BCWB_ISSHOWN); }
        bool IsEnabled() { return GetPropBool(BCWB_ISENABLED); }


        void SlidePosition(const wxPoint& pos);
        void SlideAlpha(bool isMaxAlpha = true);

        void SetSlideSpeed(double speed) { SetProperty(BCWB_SLIDE_SPEED, speed, false); };
        void StopSlide(long type = BCST_NOSLIDE);

        void SetAlpha(int alpha, bool isRefresh = true); /// value: 0...100

        short GetAlpha() { return m_currentAlpha; }

        static int zindex_compare(BitmapCanvasWidgetBase** widget_A, BitmapCanvasWidgetBase** widget_B);

        void SetFocus();
        bool IsFocused();


        void SetWindowWrapper(bcWrapper* w) { m_wrapper = w; }
        bcWrapper* GetWindowWrapper() { return m_wrapper; }

        void SetToolTip(const wxString& msg, bool useBcToolTip = false);
        void SetToolTip(WindowTipperAgg* tooltip);
        void SetBcToolTip(bcToolTip* tooltip);

        wxString GetToolTipMsg();
        WindowTipperAgg* GetToolTip();

    protected:

        bool        GetSysBool(int key);
        long        GetSysLong(int key);
        double      GetSysDouble(int key);
        wxColor     GetSysColor(int key);
        wxSize      GetSysSize(int key);
        wxRealSize  GetSysRealSize(int key);
        wxRealPoint GetSysRealPoint(int key);
        wxPoint     GetSysPoint(int key);
        wxString    GetSysString(int key);

        void DropEvent(wxEvent& evt);

        BitmapCanvasParentBase* m_parent;

        wxVariantIntHash        m_properties;

        wxWindow* GetWindowParent();
        void render(wxAgg& agg);
        void render(wxAgg2& agg);
        void SetCursor(const wxCursor& cursor);
        void SetCursor(int cursor);

        void RefreshParent();

        void StartSlide(long type);

        long m_style;

        void OnEnterWin(wxMouseEvent& event);
        void OnLeaveWin(wxMouseEvent& event);

        virtual void CreateDisabledImage();



    private:

        void CopyCanvasToBitmap();
        void AlphaBackup();

        wxBitmap*   m_bitmapCanvas; // wxDC need wxBitmap, that's why need variable. this way is faster
        wxImage*    m_canvas;       // must be private, because access and modify the buffers of the canvas
        wxBitmap*   m_disabledCanvas;

    //protected:

        enum {
            idSlideTimer = 500
        };

        enum SlideType {
            BCST_NOSLIDE = 0,
            BCST_POSITION = 1,
            BCST_ALPHA = 2,
            BCST_SIZE = 4 // not yet implemented
        };

        void OnSlideTimer(wxTimerEvent& evt);

        wxTimer     m_mainTimer;
        int         m_timeBase;
        int         m_timerShoots;

        wxPoint     m_targetPos;
        short       m_targetAlpha;
        short       m_currentAlpha;
        wxSize      m_targetSize;
        long        m_slideType;

        bcWrapper*  m_wrapper;

        WindowTipperAgg* m_toolTipWindow;
        bcToolTip*      m_toolTipWidget;

        unsigned char* m_canvasAlphaBackup; // to restore the alpha buffer of the m_canvas, after the fade in or out

        DECLARE_EVENT_TABLE();
};

#endif // BITMAPCANVASWIDGETBASE_H



