#include "Log2CsvConverter.h"

#include <wx/filename.h>

#include "ParserHelper.h"

#include "../common/types/DoubleHash.h"

#include "../loghistory/LogHistoryMgr.h"

#include "../rtdatamanager/RTDataMgrFacade.h"
#include "../rtdatamanager/RTDataInputSIPR.h"
#include "../stdatamanager/STDataMgrFacade.h"

#include "../vemsprotocol/SIPRPacket.h"

#include "../DirHelper.h"
#include "../MTConfigData.h"
#include "../RawLogHelper.h"
#include "../FirmwareInfo.h"
#include "../Symbol.h"
#include "../MTConfigHelper.h"
#include "../logger.h"
#include "../math.h"


Log2CsvConverter::Log2CsvConverter( const wxString& filename )
{
    m_fileName = filename;
}

void Log2CsvConverter::convert()
{
    m_mtConfigData = MTConfigHelper::getMTConfigData();
    m_symbolMgr = &(m_mtConfigData->getSymbolMgr());

    m_rtViewerModel = new DoubleHash();

    m_rtMgr = new RTDataMgrFacade();
    m_inputSIPR = new RTDataInputSIPR( m_symbolMgr, m_rtViewerModel );
    m_rtMgr->registerInput( m_inputSIPR );
    m_rtMgr->selectInput( INPUT_TYPE_SIPR );

    m_stMgr = new STDataMgrFacade();

    m_myModelSerial = new STDataModelSerial( m_mtConfigData->getSymbolTable(), m_mtConfigData->getSIPRDescriptor() );
    m_stMgr->setModel( m_myModelSerial );

    m_deviceModel = new STDataModelSerial( m_mtConfigData->getSymbolTable(), m_mtConfigData->getSIPRDescriptor() );
    m_stMgr->setDeviceModel( m_deviceModel );

    m_logHistoryMgr = new LogHistoryMgr( m_rtMgr, m_stMgr, m_symbolMgr );

    m_logHistoryMgr->initST(*m_mtConfigData->getSIPRDescriptor() );

    m_logHistoryMgr->loadHistoryFromLog( m_fileName );

    StringHash sh = m_logHistoryMgr->getHeaderHash();

    if ( sh[_T("LogVersion")] < _T("3") )
    {
        observerNotify( -1, this );
        return;
    }

    observerNotify( 1, this );

    wxString fwStr; fwStr = _T("firmware_") + sh[_T("Version")];
    wxString dirname = DirHelper::GetPath( dPROGRAMROOT ) + _T("commmodel\\") +fwStr;

    FirmwareInfo fwinfo;

    RawLogHelper::fillFirmwareInfo(fwinfo, m_logHistoryMgr->getHeaderHash());

    MTConfigHelper::parseMTConfigDataV2( fwinfo, m_mtConfigData, false );

    ParserHelper helper( dirname, m_mtConfigData );

    // parse mymake, map, global.h,eventqueue.h
    helper.doParsing();
    const wxArrayString& requiredAddrs = helper.getRequiredAddresses();

    // create header
    createCSVFile( requiredAddrs );

    // parse sipr packets

    unsigned int mdxmarkers = m_logHistoryMgr->getMdxFrameMarkersCount();

    SIPRPacket** p;
    unsigned int noofpackets = 0;

    for ( unsigned int j = 0; j < mdxmarkers; ++j )
    {
        m_logHistoryMgr->getMdxSiprPackets( j, &p, noofpackets );
        writePackets( p, noofpackets );
        delete[] p;
    }
    closeCsvFile();
}

void Log2CsvConverter::createCSVFile( const wxArrayString& addresses )
{
    wxFileName fn( m_fileName );
    fn.SetExt( _T("csv") );

    m_fileName = fn.GetFullPath();

    if ( fn.FileExists() )
    {
        m_csvfile.Open( m_fileName );
        m_csvfile.Clear();
    }
    else
    {
        m_csvfile.Create( m_fileName );
    }

    wxString header = _T("");
    wxString header2 = _T("");

    m_variables.clear();
    m_addr2index.clear();

    const SymbolTable& st = m_symbolMgr->getSymbolTable();

    SymbolTable::const_iterator it = st.begin();

    for ( ; it != st.end(); ++it )
    {
        Symbol* sym = it->second;
        m_variables[sym->getOffset()] = sym->getName();
    }

    for ( size_t i = 0; i < addresses.GetCount(); ++i )
    {
        int addr = Math::hex2int( addresses[i] ) * 16;
        for ( unsigned char j = 0; j < 32; ++j )
        {
            header += wxString::Format( _T("0x%04x;"), addr+j );

            Int2StringHash::const_iterator it2 = m_variables.find( addr+j );
            if ( it2 != m_variables.end() )
            {
                header2 += it2->second + _T(";");
            }
            else header2 += _T("--;");
            m_addr2index[addr+j] = i * 32 + j;
        }
    }
    header = header.RemoveLast();
    header2 = header2.RemoveLast();
    m_csvfile.AddLine( header );
    m_csvfile.AddLine( header2 );
}

void Log2CsvConverter::writePackets( SIPRPacket** packets, unsigned int packetNo )
{
    if ( packetNo < 1 )
        return;


    unsigned char* values = new unsigned char[m_addr2index.size()];
    memset( values, 0, m_addr2index.size() );

    for ( unsigned int i = 0; i < packetNo; ++i )
    {
        SIPRPacket* packet = packets[i];

        unsigned char* data = packet->data();

        unsigned int addr = 0;
        addr += data[0] << 24;
        addr += data[1] << 16;
        addr += data[2] << 8;
        addr += data[3];

        for ( unsigned j = 0; j < 32; ++j )
        {
            int idx = m_addr2index[addr+j];
            values[idx] = data[4+j];
        }
    }

    wxString line = _T("");

    for ( unsigned int i = 0; i < m_addr2index.size(); ++i )
    {
        line += wxString::Format( _T("%d;"), values[i] );
    }
    line = line.RemoveLast();

    m_csvfile.AddLine( line );

    delete[] values;
}

void Log2CsvConverter::closeCsvFile()
{
    m_csvfile.Write();
    m_csvfile.Close();
    wxString s = _T("csv|") + m_fileName;
    observerNotify( s );
}
