#ifndef VEMSMATRIX_H
#define VEMSMATRIX_H
#include "../common/typedef.h"
#include <wx/wx.h>

/// TODO
/// handle xOrder
/// ...

const size_t VEMSMATRIX_MAX_COLS = 32;
const size_t VEMSMATRIX_MAX_ROWS = 32;

template <class T>
class VemsMatrix
{
    public:
        enum { VERTICAL, HORIZONTAL };
    public:
        /// constructors always alloc MAX size buffer
        VemsMatrix( T value = T() );                                // set size to max values
        //VemsMatrix( T value );                                    // set size to max values, fill with value
        VemsMatrix( size_t rows, size_t cols, T value = T() );      // set rows, cols, fill with value
        VemsMatrix( const VemsMatrix<T>& m );

        virtual ~VemsMatrix()
        {
            delete[] m_pData;
            delete[] m_pDirtys;
        }

        /// setters
        /// set only dimensions, buffer and data are not changed, use for parsing
        void    setRowCount( size_t rows );
        void    setColCount( size_t cols );
        void    setSize( size_t rows, size_t cols );

        void    resize( size_t rows, size_t cols );

        /// set values
        void    set( size_t row, size_t col, T value );         // set dirty 1
        void    set( size_t index, T value );                   // set dirty 1
        void    fill( T value );                                // set dirtys 0

        /// getters
        size_t  rows() const { return m_iRows; }
        size_t  cols() const { return m_iCols; }
        size_t  size() const { return m_iRows * m_iCols; }
        size_t  maxSize() const { return VEMSMATRIX_MAX_ROWS * VEMSMATRIX_MAX_COLS; }
        T*      getData(){ return m_pData; }
        T       get( size_t row, size_t col ) const;
        T       get( size_t index ) const;
        bool    dirty( size_t row, size_t col ) const;
        bool    dirty( size_t index ) const;
        T&      operator[]( size_t index ) { return m_pData[index]; }
        const T& operator[]( size_t index ) const { return m_pData[index]; }

        /// utilities
        void    flip( int direction = VERTICAL );
        bool    insert_row( size_t row );
        bool    insert_row_before( size_t row );
        bool    insert_col( size_t col );
        bool    insert_col_before( size_t col );
        void    shift_left();
        void    shift_up();

        // use extarnal data, doesn't take ownership, wrap linear data
        /// !!!should  use / make other class?
        /// !!!should use VemsMatrix instead of external data buffers
        //VemsMatrix( int rows, int cols, T* pData );
        //VemsMatrix( const VemsMatrix<T>& m );
        /// develop, test functions
        //void print();
        //void    fill( int rows, int cols, T value );
        // bool    error() { return m_bError; }
    private:
        //wxString 	m_name;                 // table name
        size_t          m_iCols;	            // number of cols
        size_t		    m_iRows;                // nomber of rows
        T*              m_pData;                // matrix data buffer
        uint8_t*        m_pDirtys;              // matrix dirty buffer

        // bool        m_bError;               // last operation result
        //bool        m_bUseInternalData;    //

    private:
        //void        init( int rows, int cols );
        void        init( size_t rows, size_t cols, T value );
        bool        isOutOfBounds( size_t row, size_t cols ) const;
        bool        isOutOfBounds( size_t index ) const;
};

///-------------------------------------------------------------------------- constructors

/**
*/
template <class T>
VemsMatrix<T>::VemsMatrix( T value )
{
    init( VEMSMATRIX_MAX_ROWS, VEMSMATRIX_MAX_COLS, value );
}

/**
*/
template <class T>
VemsMatrix<T>::VemsMatrix( size_t rows, size_t cols, T value )
{
    init( rows, cols, value );
}

/**
*/
template <class T>
VemsMatrix<T>::VemsMatrix( const VemsMatrix<T>& m )
{
    init( m.rows(), m.cols(), T() );
    for ( size_t i = 0; i < maxSize(); ++i )
    {
        set( i, m.get( i ) );
    }
}

/**
*/
template <class T>
void VemsMatrix<T>::init( size_t rows, size_t cols, T value )
{
    setSize( rows, cols );
    m_pData     = new T[ maxSize() ];
    m_pDirtys   = new uint8_t[ maxSize() ];
    fill( value );
    //memset( m_pDirtys, 0x00, maxSize() );
}

///-------------------------------------------------------------------------- setters
template <class T>
void VemsMatrix<T>::setRowCount( size_t rows )
{
    if ( rows < 1 ) { rows = 1; }
    if ( rows > VEMSMATRIX_MAX_ROWS ) { rows = VEMSMATRIX_MAX_ROWS; }

    m_iRows = rows;
}

template <class T>
void VemsMatrix<T>::setColCount( size_t cols )
{
    if ( cols < 1 ) { cols = 1; }
    if ( cols > VEMSMATRIX_MAX_COLS ) { cols = VEMSMATRIX_MAX_COLS; }

    m_iCols = cols;
}

template <class T>
void VemsMatrix<T>::setSize( size_t rows, size_t cols )
{
    setRowCount( rows );
    setColCount( cols );
}

template <class T>
void VemsMatrix<T>::resize( size_t newRow, size_t newCol )
{

    if( cols() == newCol ){
        setRowCount( newRow );
        return;
    }
    VemsMatrix<T> old(*this);
    setSize(newRow, newCol);
    fill( 0 );
    size_t sizeT = sizeof(T);

    size_t minRow = ( old.rows() < rows() ) ? old.rows() : rows() ;
    size_t minCol = ( old.cols() < cols() ) ? old.cols() : cols() ;

    for ( size_t r = 0; r < minRow; ++r )
    {
        size_t oldIdx = r * old.cols();
        size_t newIdx = r * cols();
        memcpy( &m_pData[newIdx], &old.m_pData[oldIdx],  sizeT * minCol );
    }
}

template <class T>
void VemsMatrix<T>::set( size_t row, size_t col, T value )
{
    if ( isOutOfBounds( row, col ) ) { return; }
    m_pData[ m_iCols * row + col ] = value;
    m_pDirtys[ m_iCols * row + col ] = 0x01;
}

template <class T>
void VemsMatrix<T>::set( size_t index, T value )
{
    if ( isOutOfBounds( index ) ) { return; }
    m_pData[ index ] = value;
    m_pDirtys[ index ] = 0x01;
}

template <class T>
void VemsMatrix<T>::fill( T value )
{
    for ( size_t i = 0; i < size(); ++i )
    {
        m_pData[i]      = value;
        m_pDirtys[i]    = 0x00;
    }
}

///-------------------------------------------------------------------------- getters
template <class T>
T VemsMatrix<T>::get( size_t row, size_t col ) const
{
    if ( isOutOfBounds( row, col ) ) { return T(); }

    return m_pData[ m_iCols * row + col ];
}

template <class T>
T VemsMatrix<T>::get( size_t index ) const
{
    if ( isOutOfBounds( index ) ) { return T(); }

    return m_pData[ index ];
}

template <class T>
bool VemsMatrix<T>::dirty( size_t row, size_t col ) const
{
    if ( isOutOfBounds( row, col ) ) { return false; }

    return m_pDirtys[ m_iCols * row + col ];
}

template <class T>
bool VemsMatrix<T>::dirty( size_t index ) const
{
    if ( isOutOfBounds( index ) ) { return false; }

    return m_pDirtys[ index ]!=0?true:false;
}

///-------------------------------------------------------------------------- utilities
template <class T>
void VemsMatrix<T>::flip( int direction )
{
    switch ( direction )
    {
        case VERTICAL:
        {
            size_t size        = sizeof(T) * m_iCols;
			T *tmp= (T*)malloc(m_iCols*sizeof(T));
            //T tmp[ m_iCols ];

            for ( size_t i = 0; i < m_iRows / 2; ++i )
            {
                // copy row[i] to tmp
                memcpy( tmp ,&m_pData[ i * m_iCols ], size );
                // copy row[ rows - i] to row[i]
                memcpy( &m_pData[ i * m_iCols ], &m_pData[ ( m_iRows - i - 1 ) * m_iCols  ], size );
                // copy tmp( row[i] ) to [ rows - i]
                memcpy( &m_pData[ ( m_iRows - i - 1 ) * m_iCols ], tmp, size );
            }
			free(tmp);
        } break;
        case HORIZONTAL:
        {
            T tmp;

            for ( size_t i = 0; i < m_iCols; ++i )
            {
                for ( size_t j = 0; j < m_iCols / 2; ++j )
                {
                    tmp = get( i , j );
                    set ( i, j, get( i, m_iCols - 1 - j ) );
                    set( i, m_iCols - 1 - j, tmp );
                }
            }
        } break;
    }
}


/**
    insert an empty row after row, remove last row
*/
template <class T>
bool VemsMatrix<T>::insert_row( size_t row )
{
    if ( rows() >= VEMSMATRIX_MAX_ROWS ) { return false; }

    resize( rows() + 1, cols() );

    for ( size_t c = 0; c < cols(); ++c )
    {
        for ( size_t r = rows() - 1; r > row; --r )
        {
            if ( r == row + 1 )
            {
                set ( r, c , T() );
                continue;
            }

            set( r, c, get( r - 1, c ) );
        }
    }

    return true;
}


/**
    insert an empty row after row, remove last row
*/
template <class T>
bool VemsMatrix<T>::insert_row_before( size_t row )
{
    if ( rows() >= VEMSMATRIX_MAX_ROWS ) { return false; }

    resize( rows() + 1, cols() );

    for ( size_t c = 0; c < cols(); ++c )
    {
        for ( size_t r = 0; r < row; ++r )
        {
            if ( r == row - 1 )
            {
                set ( r, c , T() );
                continue;
            }

            set( r, c, get( r + 1, c ) );
        }
    }

    return true;
}


/**
    insert an empty column after col if it is possible, resize matrix
*/
template <class T>
bool VemsMatrix<T>::insert_col( size_t col )
{
    if ( cols() >= VEMSMATRIX_MAX_COLS ) { return false; }

    resize( rows(), cols() + 1 );

    for ( size_t r = 0; r < rows(); ++r )
    {
        for ( size_t c = cols() - 1; c > col; --c )
        {
            set( r, c, get( r, c - 1 ) );

            if ( c == col + 1 )
            {
                set ( r, c , T() );
                continue;
            }


        }
    }

    return true;
}



/**
    insert an empty column before col,
*/
template <class T>
bool VemsMatrix<T>::insert_col_before( size_t col )
{
    if ( cols() >= VEMSMATRIX_MAX_COLS ) { return false; }

    resize( rows(), cols() + 1 );

    for ( size_t r = 0; r < rows(); ++r )
    {
        for ( size_t c = 0; c < col; ++c )
        {
            if ( c == col - 1 )
            {
                set ( r, c , T() );
                continue;
            }

            set( r, c, get( r, c + 1 ) );
        }
    }

    return true;
}


/**
    shift columns left T() enter in the end
*/
template <class T>
void VemsMatrix<T>::shift_left()
{
    for ( size_t r = 0; r < rows(); ++r )
    {
        for ( size_t c = 0; c < cols(); ++c )
        {
            if ( c == cols() - 1 )
            {
                set ( r, c , T() );
                continue;
            }

            set( r, c, get( r, c + 1 ) );
        }
    }
}


/**
    shift columns left T() enter in the end
*/
template <class T>
void VemsMatrix<T>::shift_up()
{
    for ( size_t r = 1; r < rows(); ++r )
    {
        memcpy( &m_pData[ ( r - 1 ) * cols() ], &m_pData[ r * cols()], cols() * sizeof(T) );
    }
}


///-------------------------------------------------------------------------- private
template <class T>
bool VemsMatrix<T>::isOutOfBounds( size_t row, size_t col ) const
{
    return ( row < 0 || col < 0 || row >= m_iRows || col >= m_iCols );
}

template <class T>
bool VemsMatrix<T>::isOutOfBounds( size_t index ) const
{
    return ( index < 0 || index  >= size() );
}

///  VemsMatrix<uint8_t> wx containers
WX_DECLARE_STRING_HASH_MAP( VemsMatrix<uint8_t>*, VemsMatrixByteHash );
WX_DEFINE_ARRAY( VemsMatrix<uint8_t>* ,VemsMatrixByteArray );
///  VemsMatrix<double> wx containers
WX_DECLARE_STRING_HASH_MAP( VemsMatrix<double>*, VemsMatrixDoubleHash );
WX_DEFINE_ARRAY( VemsMatrix<double>* ,VemsMatrixDoubleArray );

///typedefs
typedef VemsMatrix<double>  VemsMatrixDouble;
typedef VemsMatrix<uint8_t> VemsMatrixByte;

/*
template <class T>
VemsMatrix<T>::VemsMatrix( int rows, int cols, T* pData  )
{
    m_bUseInternalData  = false;
    //m_bError            = false;
    // check size
    if ( rows < 1 ) { rows = 1; }
    if ( cols < 1 ) { cols = 1; }
    if ( rows > VEMSMATRIX_MAX_ROWS ) { rows = VEMSMATRIX_MAX_ROWS; }
    if ( cols > VEMSMATRIX_MAX_COLS ) { cols = VEMSMATRIX_MAX_COLS; }

    m_iRows = rows;
    m_iCols = cols;
    //m_pData = new T[ rows * cols ];
    m_pData = pData;
}
*/


/*
// create new internal buffer and copy data
template <class T>
VemsMatrix<T>::VemsMatrix( const VemsMatrix<T>& m  )
{
    m_bUseInternalData  = true;
    //m_bError            = false;

    //m_name      = m.m_name;
    m_iCols     = m.m_iCols;
    m_iRows     = m.m_iRows;

    m_pData = new T[ size() ];

    //memset( m_pData, initialValue, sizeof( T ) * rows * cols );
    for ( int i = 0; i < size(); ++i )
    {
        m_pData[i] = m.get( i );
    }
}
*/


#endif // VEMSMATRIX_H
