#include "BaudRateSelector.h"
#include "DeviceModeSelector.h"
#include "BaudRateSelectorListener.h"
#include "PortScanner.h"
#include "PortScanner.h"
#include "PortScannerEvent.h"
#include "../../CommunicationBase.h"
#include "../../common/TxData.h"
#include "../../logger.h"
#include "../../vemsprotocol/recognizers/StringRecognizer.h"
#include "../../common/types/IntArray.h"



BaudRateSelector::BaudRateSelector( BaudRateSelectorListener* listener,
                                    TimeoutListener* timeoutListener,
                                    CommunicationBase* commBase,
                                    const IntArray& baudrates,
                                    ECUInfo::EcuMode deviceMode,
                                    size_t currentBaudrate,
                                    size_t targetBaudrate )
{
    m_listener          = listener;
    m_timeoutListener   = timeoutListener;
    m_commBase          = commBase;
    m_deviceMode        = deviceMode;
    m_currentBaudrate   = currentBaudrate;
    m_targetBaudrate    = targetBaudrate;
    m_baudrates         = baudrates;

    m_baudrateChangedRecognizer = new StringRecognizer( _T("OK\r\n") );

    m_client            = NULL;

    if ( m_deviceMode == ECUInfo::MAN ) {
        m_isSelectManMode = false;
        m_state = SEND_CHANGE_BAUD_RATE;
    } else {
        m_isSelectManMode = true;
        m_client    = new DeviceModeSelector( this, this,commBase, baudrates, m_deviceMode, ECUInfo::MAN );
        m_state     = CLIENT_MODE;
    }
}

BaudRateSelector::~BaudRateSelector()
{
}

bool BaudRateSelector::startProcess()
{
    if ( m_state == CLIENT_MODE ) {
        m_client->setTimeout( getTimeout() );
        m_client->setInterWriteDelay( getInterWriteDelay() );
        m_client->startProcess();
    }

    return true;
}

void BaudRateSelector::write( uint8_t* buffer, int& size )
{
    switch ( m_state )
    {
        case CLIENT_MODE:
            sendClientMode( buffer, size );
            break;

        case SEND_CHANGE_BAUD_RATE:
            sendChangeBaudRate( buffer, size );
            break;

        default:
            break;
    }
}

void BaudRateSelector::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case CLIENT_MODE:
            waitClientMode( buffer, size );
            break;

        case WAIT_BAUD_RATE_CHANGED:
            waitBaudRateChanged( buffer, size );
            break;

        default:
            break;
    }
}

void BaudRateSelector::finish()
{
    switch ( m_state )
    {
        case FINISHED:
            m_listener->OnBaudrateChanged( m_targetBaudrate );
            break;

        case TIMEDOUT:
           m_timeoutListener->OnTimeout();
            break;

        default:
            break;
    }
}

bool BaudRateSelector::isFinished()
{
    return ( m_state == FINISHED || m_state == TIMEDOUT );
}

//void BaudRateSelector::handleEvent( CommunicationClientEvent::Type eventType )
//{
//    switch ( eventType )
//    {
//        case CommunicationClientEvent::DEVICE_MODE_SELECTED:
//        {
//            switch ( m_state )
//            {
//                case WAIT_MAN_MODE_SELECTED:
//                LOG
//                    m_state = SEND_CHANGE_BAUD_RATE;
//                    break;
//
//                case WAIT_PREVIOUS_MODE_SELECTED:
//                    LOG
//                    m_state = FINISHED;
//                    break;
//
//                default:
//                    break;
//            }
//        } break;
//
//        case CommunicationClientEvent::TIMEOUT:
//            m_state = TIMEDOUT;
//            break;
//
//        default:
//            break;
//    }
//}

void BaudRateSelector::sendClientMode( uint8_t* buffer, int& size )
{
    m_client->write( buffer, size );
}

void BaudRateSelector::waitClientMode( uint8_t* buffer, int size )
{
    m_client->read( buffer, size );

    if ( m_client->isFinished () ) {
        m_client->finish();
    }
}

void BaudRateSelector::sendChangeBaudRate( uint8_t* buffer, int& size )
{
    int baudrateCommand;

    switch ( m_targetBaudrate )
    {
        case 9600:
            baudrateCommand = 104; //68
            break;

        case 19200:
            baudrateCommand = 52; //34
            break;

        case 38400:
            baudrateCommand = 26;
            break;

        case 57600:
            baudrateCommand = 16;
            break;

        case 115200:
            baudrateCommand = 8;
            break;

        default:
            baudrateCommand = 1000000 / m_targetBaudrate - 1;
            //baudrateCommand = 52;
            break;
    }

    wxString commandStr = wxString::Format( _T("mcB%02X"), baudrateCommand );

    TxData command = TxData( commandStr );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_timeoutTimer.start(100);

    m_state = WAIT_BAUD_RATE_CHANGED;
}

void BaudRateSelector::waitBaudRateChanged( uint8_t* buffer, int size )
{
    if ( m_baudrateChangedRecognizer->feed( buffer[0] ) ) {
        m_commBase->setBaudRate( m_targetBaudrate );
        if ( m_deviceMode == ECUInfo::MAN ) {
            m_state = FINISHED;
        } else {
            delete m_client;
            IntArray baudrates;
            baudrates.Add(m_targetBaudrate );
            m_client = new DeviceModeSelector( this, this, m_commBase, baudrates, ECUInfo::MAN, m_deviceMode );
            m_state = CLIENT_MODE;
        }
    } else {
        if ( m_timeoutTimer.isTimedout() ) {
            delete m_client;
            m_client = new PortScanner( this, m_commBase, m_baudrates );
            m_state = CLIENT_MODE;
        }
    }

}

void BaudRateSelector::OnDeviceModeSelected( ECUInfo::EcuMode deviceMode )
{
    if ( m_isSelectManMode ) {
        m_state = SEND_CHANGE_BAUD_RATE;
    } else {
        m_state = FINISHED;
    }
}

void BaudRateSelector::OnDeviceFound( PortScannerEvent& event )
{
    LOGP( event.toString() );

    if ( event.mode == m_deviceMode ) {
        m_state = FINISHED;
    } else {
        m_isSelectManMode = false;
        delete m_client;
        m_client = new DeviceModeSelector( this, this, m_commBase, m_baudrates, event.mode, m_deviceMode );
        m_state = CLIENT_MODE;
    }
}

void BaudRateSelector::OnDeviceNotFound()
{
    m_state = TIMEDOUT;
}

void BaudRateSelector::OnTimeout()
{
    m_state = TIMEDOUT;
}
