#include "ConfigDownloader.h"
#include "ConfigDownloadListener.h"
#include "TimeoutListener.h"
#include "DeviceDetector.h"
#include "../CommunicationClientEventHandler.h"
#include "../../vemsprotocol/SiprPacketReceiver.h"
#include "../../vemsprotocol/SIPRPacket.h"
#include "../../vemsprotocol/SIPRTypes.h"
#include "../../common/TxData.h"
#include "../../logger.h"

ConfigDownloader::ConfigDownloader( ConfigDownloadListener* downloadListener,
                                    TimeoutListener* timeoutListener,
                                    SiprPacketReceiver* receiver )
    : DOWNLOAD_MIN_LINE_COUNT ( 2 )
    , DOWNLOAD_SKIP_EMPTY_LINES ( 2 )
    , TIMEOUT ( 100 )
{
    m_downloadListener = downloadListener;
    m_timeoutListener = timeoutListener;
    m_receiver = receiver;

    m_state = SEND;
    m_rxDataLength = 0;

    m_lineCount         = 0;
    m_emptyLineCount    = 0;

    m_isDownloadingConfig= true;

    m_client = NULL;

}

ConfigDownloader::~ConfigDownloader()
{

}

void ConfigDownloader::write( uint8_t* buffer, int& size )
{
    switch ( m_state )
    {
        case SEND:
            send( buffer, size );
            break;

        case CLIENT_MODE:
            sendClientMode( buffer, size );
            break;

        default:
            break;
    }
}

void ConfigDownloader::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case WAIT:
            wait( buffer, size );
            break;

        case CLIENT_MODE:
            waitClientMode( buffer, size );
            break;

        default:
            break;
    }
}

void ConfigDownloader::finish()
{
    if ( m_state == TIMEDOUT ) {
        m_timeoutListener->OnTimeout();
    } else {
        m_downloadListener->OnConfigDownloaded();
    }
}

bool ConfigDownloader::isFinished()
{
    return ( ( m_state == FINISHED ) || ( m_state == TIMEDOUT ) );
}

void ConfigDownloader::send( uint8_t* buffer, int& size )
{
    TxData command;

    if ( m_isDownloadingConfig ) {
        command = TxData( _T("mcd") );
    } else {
        command = TxData( _T("mct") );
    }

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_timeoutTimer.start( m_timeout + TIMEOUT );

    m_state = WAIT;
}

void ConfigDownloader::wait( uint8_t* buffer, int size )
{
    if ( size > 0 )
    {
        m_rxBuffer[ m_rxDataLength++ ] = buffer[0];

        if ( isEOL() ) {
            ++m_lineCount;
            if ( m_rxDataLength == 2 ) {
                ///TooMuchEmptyLines
                if ( ++m_emptyLineCount >= DOWNLOAD_SKIP_EMPTY_LINES ) {
                    m_state = TIMEDOUT;
                    return;
                }
            }
            sendLinePacket();
            m_timeoutTimer.start(m_timeout + TIMEOUT);
            m_rxDataLength = 0;
        }
    }
    else
    {
        wxThread::Sleep( m_sleep );

        if ( !m_timeoutTimer.isTimedout() ) {
            return;
        }

        if ( m_lineCount < DOWNLOAD_MIN_LINE_COUNT ) {
            m_state = TIMEDOUT;
            return;
        }

        if ( m_isDownloadingConfig ) {
            m_isDownloadingConfig = false;
            m_timeoutTimer.start(m_timeout + TIMEOUT);
            m_rxDataLength = 0;
            m_state = SEND;
            return;
        }

        delete m_client;
        m_client = new DeviceDetector( this );
        m_client->setTimeout( m_timeout );
        m_client->setInterWriteDelay( m_interWriteDelay );

        m_state = CLIENT_MODE;
    }
}

void ConfigDownloader::sendClientMode( uint8_t* buffer, int& size )
{
    m_client->write( buffer, size );
}

void ConfigDownloader::waitClientMode( uint8_t* buffer, int size )
{
    m_client->read( buffer, size );

    if ( m_client->isFinished () ) {
        m_client->finish();
    }
}

void ConfigDownloader::sendLinePacket()
{
    m_rxBuffer[ m_rxDataLength - 2 ] = '\0';
    SIPRPacket packet( 0, SIPR_CONFIG_LINE, m_rxDataLength - 1, m_rxBuffer );
    m_receiver->receive( packet );
}


bool ConfigDownloader::isEOL() const
{
    if ( m_rxDataLength > 1 )
    {
        if (    m_rxBuffer[ m_rxDataLength - 2 ] == '\r' &&
                m_rxBuffer[ m_rxDataLength - 1 ] == '\n' ) { return true; }
    }

    return false;
}

void ConfigDownloader::OnDeviceFound( ECUInfo::Device device, ECUInfo::EcuMode mode )
{
    if ( mode == ECUInfo::MAN ) {
        m_state = FINISHED;
    } else {
        m_state = TIMEDOUT;
    }
}

void ConfigDownloader::OnDeviceNotFound()
{
    m_state = TIMEDOUT;
}
