#include "DeviceModeSelector.h"
#include "DeviceModeSelectorListener.h"
#include "PortScanner.h"
#include "PortScannerEvent.h"
#include "DeviceDetectorListener.h"
#include "BaudRateSelector.h"
#include "DeviceDetectorListener.h"
#include "../../CommunicationBase.h"
#include "../../common/TxData.h"
#include "../../logger.h"


#include <wx/arrimpl.cpp>
WX_DEFINE_OBJARRAY(ModeSelectCmdArray);


DeviceModeSelector::DeviceModeSelector( DeviceModeSelectorListener* listener,
                                        TimeoutListener* timeoutListener,
                                        CommunicationBase* commBase,
                                        const IntArray& baudrates,
                                        ECUInfo::EcuMode sourceMode,
                                        ECUInfo::EcuMode targetMode,
                                        ECUInfo::EcuMode* selectedMode )
{
    m_listener              = listener;
    m_timeoutListener       = timeoutListener;
    m_sourceMode            = sourceMode;
    m_targetMode            = targetMode;
    m_selectedMode          = selectedMode;

    m_detectedDevice        = ECUInfo::DEVICE_UNKNOWN;
    m_detectedMode          = ECUInfo::UNKNOWN;

    m_state                 = SEND_CHANGE_MODE;

    m_commBase = commBase;
    m_baudrates = baudrates;
    m_sourceBaudrate = commBase->getBaudRate();

    m_client = new PortScanner( this, m_commBase, m_baudrates, m_sourceBaudrate );

    m_timeoutCount  = 0;
    m_clearCount    = 0;
    m_clearCountMax = 0;
    m_detectTryCount    = 0;
    m_detectTryCountMax = 3;

    m_modeSelectCommands.Add( ModeSelectCmd( ECUInfo::BOOT, ECUInfo::MEGATUNE, _T("E.") ) );
    m_modeSelectCommands.Add( ModeSelectCmd( ECUInfo::MEGATUNE, ECUInfo::BOOT, _T("Mancb33cb.") ) );
    m_modeSelectCommands.Add( ModeSelectCmd( ECUInfo::MEGATUNE, ECUInfo::MAN, _T("Man.") ) );
    m_modeSelectCommands.Add( ModeSelectCmd( ECUInfo::MAN, ECUInfo::MEGATUNE, _T("byebye.") ) );
    m_modeSelectCommands.Add( ModeSelectCmd( ECUInfo::MAN, ECUInfo::BOOT, _T("cb33cb.") ) );

    LOGP( _T("m_sourceMode: ") + ECUInfo::getStateAsString(m_sourceMode) +
        _T(", targetMode: ") + ECUInfo::getStateAsString(m_targetMode) );
}

DeviceModeSelector::~DeviceModeSelector()
{
    delete m_client;
}

bool DeviceModeSelector::startProcess()
{
    if ( m_state == CLIENT_MODE ) {
        m_client->startProcess();
    }

    return true;
}

void DeviceModeSelector::write( uint8_t* buffer, int& size )
{
    switch ( m_state )
    {
        case SEND_CHANGE_MODE:
            sendChangeMode( buffer, size );
            break;

        case CLIENT_MODE:
            sendClient( buffer, size );
            break;

        default:
            break;
    }
}

void DeviceModeSelector::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case CLIENT_MODE:
            waitClient( buffer, size );
            break;

        case CLEAR:
            clear( buffer, size );
            break;

        default:
            break;
    }
}

void DeviceModeSelector::finish()
{

    switch ( m_state )
    {
        case FINISHED:
            m_listener->OnDeviceModeSelected( m_detectedMode );
            break;

        case TIMEDOUT:
            m_timeoutListener->OnTimeout();
            break;

        default:
            break;
    }
}

bool DeviceModeSelector::isFinished()
{
    return ( m_state == FINISHED || m_state == TIMEDOUT );
}

void DeviceModeSelector::OnDeviceFound( PortScannerEvent& event )
{
    LOGP( ECUInfo::getStateAsString( event.mode ) );

    m_detectedMode  = event.mode;
    if ( event.mode == m_targetMode ) {
        //restore baudrate after leave BOOT mode
        if ( ( m_sourceMode == ECUInfo::BOOT ) && ( m_sourceBaudrate != event.baudrate ) ) {

            delete m_client;
            m_client = new BaudRateSelector( this, this, m_commBase, m_baudrates, m_detectedMode, event.baudrate, m_sourceBaudrate );
            m_client->setTimeout( getTimeout() );
            m_client->setInterWriteDelay( getInterWriteDelay() );
            m_state = CLIENT_MODE;
        } else {
            m_clearingTimer.start(100);
            m_state = CLEAR;
        }
    } else {
        if ( ++m_detectTryCount < m_detectTryCountMax ) {
            m_state = SEND_CHANGE_MODE;
            m_sourceMode = m_detectedMode;
            m_client->reset();
        } else {
            m_state = TIMEDOUT;
        }
    }
}

void DeviceModeSelector::OnDeviceNotFound()
{
    LOG
    m_state = TIMEDOUT;
}

void DeviceModeSelector::OnBaudrateChanged( size_t baudrate )
{
    m_clearingTimer.start(100);
    m_state = CLEAR;
}

void DeviceModeSelector::OnTimeout()
{
    //m_timeoutListener->OnTimeout();
    m_state = TIMEDOUT;
}

void DeviceModeSelector::sendChangeMode( uint8_t* buffer, int& size )
{
    if ( m_sourceMode == m_targetMode ) {
        m_detectedMode = m_targetMode;
        m_state = FINISHED;
        return;
    }

    for ( size_t i = 0; i < m_modeSelectCommands.GetCount(); ++i ){
        if ( m_modeSelectCommands[i].sourceMode == m_sourceMode &&
                m_modeSelectCommands[i].targetMode == m_targetMode )
        {
            TxData command = TxData( m_modeSelectCommands[i].command );

            size = command.size;
            memcpy( buffer, command.data, command.size );

            m_client->setTimeout( getTimeout() );
            m_client->setInterWriteDelay( getInterWriteDelay() );

            m_state = CLIENT_MODE;

            return;
        }
    }

    LOGP( _T("INVALID MODE SELECT - source: ") + ECUInfo::getStateAsString( m_sourceMode ) +
        _T(", target: ") + ECUInfo::getStateAsString( m_targetMode ) );

}

void DeviceModeSelector::sendClient( uint8_t* buffer, int& size )
{
    m_client->write( buffer, size );
}


void DeviceModeSelector::waitClient( uint8_t* buffer, int size )
{
    m_client->read( buffer, size );

    if ( m_client->isFinished () ) {
        m_client->finish();
    }
}

void DeviceModeSelector::clear( uint8_t* buffer, int size )
{
    if ( m_clearingTimer.isTimedout() ) {
        m_state = FINISHED;
    } else {
        wxThread::Sleep(1);
    }
}
