#include "Megatune.h"
#include "TimeoutListener.h"
#include "../../vemsprotocol/SIPRTypes.h"
#include "../../vemsprotocol/SIPRPacket.h"
#include "../../vemsprotocol/SiprPacketReceiver.h"
#include "../../vemsprotocol/SIPRDescriptor.h"
#include "../../vemsprotocol/SiprDescHelper.h"
#include "../../vemsprotocol/MegatuneHelper.h"
#include "../CommunicationClientEventHandler.h"
#include "../../common/TxData.h"
#include "../../logger.h"
#include "../../Math.h"

#include "../../vemsprotocol/SIPRPacketHelper.h"
#include "../../events/InputEvent.h"
#include "../../triggerlog/TriggerFrame.h"

#include <wx/listimpl.cpp>

WX_DEFINE_LIST(PageRequestList);

/**TODO
-read page timeout
-handle no sync after timeout

*/

Megatune::Megatune( TimeoutListener* timeoutListener,
                    wxEvtHandler* eventHandler,
                    const MegatuneConfig& config,
                    SIPRDescriptor* descriptor )
{
    m_timeoutListener       = timeoutListener;
    m_config    = config;
    m_eventHandler = eventHandler;

    m_descriptor = descriptor; ///TODO - refactor SIPRDescriptor

    m_isReadRealtimeEnabled = false;

    m_state                 = REQUEST;

    m_rxDataLength          = 0;
    m_expectedRxSize        = 0;
    m_sentReadRTCmdCount    = 0;
    m_realtimeErrorCount    = 0;

    m_crc   = 0xFFFF;
    m_rxCrc = 0;

    m_rxBuffer = &m_buffer[3];

    m_startTime = wxDateTime::UNow();

    m_clearCount = 0;
    m_interWriteDelay = 0;
    m_sendAck = false;
}

Megatune::~Megatune()
{
    WX_CLEAR_LIST( PageRequestList, m_readPageRequests );
}

void Megatune::write( uint8_t* buffer, int& size )
{
    switch ( m_state )
    {
        case REQUEST:
            sendNextCommand( buffer, size );
            break;

        default:
            break;
    }
}

void Megatune::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case REQUEST:
            wxThread::Sleep( m_sleep );
            break;

        case WAIT_RT:
            waitRt( buffer, size );
            break;

        case WAIT_PAGE:
            waitPage( buffer, size );
            break;

        case HANDLE_CRC_ERROR:
            handleCrcError(  buffer, size );
            break;

        case CLEAR:
            clear();
            break;

        default:
            break;
    }
}

void Megatune::finish()
{
    m_timeoutListener->OnTimeout();
}


bool Megatune::isFinished()
{
    return m_state == END;
}

size_t Megatune::getInterWriteDelay()
{
    switch ( m_state )
    {
        case WAIT_RT:
        case WAIT_PAGE:
            return m_config.interWriteDelay;

        default:
            return m_interWriteDelay;
    }
}

void Megatune::request( VemsCommand* command )
{
    if ( !command ) {
        LOGP( _T("ERROR - command is NULL!") );
        return;
    }

    ReadAllPageCommand* readAllCmd = dynamic_cast<ReadAllPageCommand*>( command );

    if ( readAllCmd ) {
        PageRequest* pageReq = new PageRequest();

        for ( size_t i = 0; i < m_descriptor->getPageCount(); ++i ) {
            int pageId = m_descriptor->getPageNumber( i );

            if ( pageId != -1 ) {
                ReadPageCommand* command = new ReadPageCommand( pageId );
                wxCriticalSectionLocker locker(m_criticialSection);
                m_queue.push_back( command );
                pageReq->requestedPages.insert( pageId );
            }
        }

        if ( pageReq->requestedPages.size() > 0 ) {
            m_readPageRequests.push_back( pageReq );
        } else {
            delete pageReq;
        }

        return;
    }

    ReadPageCommand* readPageCmd = dynamic_cast<ReadPageCommand*>( command );

    if ( readPageCmd ) {
        int pageId = readPageCmd->pageId;

        if ( !m_descriptor->isPageExist( pageId ) ) {
            LOGP( wxString::Format( _T("Page doesn't exist: %d"), pageId ) );
            return;
        }

        PageRequest* pageReq = new PageRequest();
        pageReq->requestedPages.insert( pageId );
        m_readPageRequests.push_back( pageReq );
    }

    wxCriticalSectionLocker locker(m_criticialSection);
    m_queue.push_back( command );
}

bool Megatune::readPage( int pageId )
{
    m_page  = pageId;
    wxString readCommand = m_descriptor->getPageReadCommand( pageId );

    TxData* cmd = SiprDescHelper::getPageReadCommand( readCommand, m_config.device );

    if ( cmd ) {
        m_command = *cmd;
        delete cmd;
    } else {
        LOGP( _T("INVALID COMMAND - ") + readCommand );
        return false;
    }

    activatePageRequest();

    m_expectedRxSize = m_config.pageSize[pageId];
    m_rxDataLength          = 0;
    m_timeoutTimer.start( m_config.timeout + 200 );

    m_state = WAIT_PAGE;

    LOGP( wxString::Format( _T("m_page: %d, m_expectedRxSize: %d, readCommand: "), m_page, m_expectedRxSize ) + readCommand );

    return true;
}

bool Megatune::writePage( int pageId, int pageOffset, uint8_t value )
{
    wxString pageWriteCommand = m_descriptor->getPageWriteCommand( pageId );

    int CMD_SIZE = 3;

    if ( m_config.device == ECUInfo::DEVICE_ROUND ) {
        CMD_SIZE = 4;
    }

    uint8_t* writeCommand = new uint8_t[ CMD_SIZE ];

    unsigned long val = 0 ;

    if ( pageWriteCommand.StartsWith( _T("0x") ) || pageWriteCommand.StartsWith( _T("0X") ) ) {
        if ( pageWriteCommand.ToULong(&val,0 ) ) {
            writeCommand[0] = val;
        } else {
            delete[] writeCommand;
            return false;
        }
    } else if ( pageWriteCommand.Len() >= 1 ) {
        writeCommand[0]  = pageWriteCommand.GetChar( 0 );
    } else {
        delete[] writeCommand;
        return false;
    }

    writeCommand[1]  = (uint8_t)pageOffset;
    writeCommand[2]  = value;

    if ( m_config.device == ECUInfo::DEVICE_ROUND ) {
        writeCommand[3]  = '.';
    }

//    m_rxDataLength          = 0;
//    resetTimeoutTime();

    m_command =  TxData( writeCommand, CMD_SIZE );
    m_state = REQUEST;

    sendAckPacket();

	delete[] writeCommand;

	return true;
}

void Megatune::burn()
{
    wxString burnCommand = m_descriptor->getBurnCommand();

    if ( m_config.device == ECUInfo::DEVICE_ROUND ) {
        burnCommand += _T('.');
    }

    m_command =  TxData( burnCommand );
    m_state = REQUEST;
}

void Megatune::readRealtime()
{
}

void Megatune::executeStringCommand( const wxString& command, size_t sleepAfterWrite, bool sendAck )
{
    m_command =  TxData( command );
    m_clearCount = 50;
    m_interWriteDelay = sleepAfterWrite;
    m_sendAck = sendAck;
    m_state = CLEAR;
}

void Megatune::executeRawCommand( const TxData& data )
{
    m_command =  data;
    m_clearCount = 50;
    m_state = CLEAR;
}


void Megatune::enableRealtime()
{
    m_isReadRealtimeEnabled = true;
}

void Megatune::sendNextCommand( uint8_t* buffer, int& size )
{
    wxCriticalSectionLocker locker(m_criticialSection);

    m_interWriteDelay = 0;
    m_sendAck = false;

    if ( m_queue.IsEmpty() ) {
        if ( !m_isReadRealtimeEnabled ) {
            return;
        }
        m_expectedRxSize = m_config.realtimeLength;
        m_command = TxData( m_config.readRealtimeCommand.To8BitData() );
        start = ::wxGetLocalTimeMillis();
        ++m_sentReadRTCmdCount;

        size = m_command.size;
        memcpy( buffer, m_command.data, m_command.size );

        m_timeoutTimer.start( m_config.timeout );

        m_state = WAIT_RT;
    } else {
//        LOGP( wxString::Format( _T("m_sentReadRTCmdCount: %d"),m_sentReadRTCmdCount ) );
        if ( m_sentReadRTCmdCount > 0 ) {
            m_state = WAIT_RT;
        } else {
            VemsCommand* command = m_queue.front();
            //command->execute( this );
            bool isExecuted = execute( command );
            m_queue.pop_front();
            delete command;

            if ( isExecuted ) {
                size = m_command.size;
                memcpy( buffer, m_command.data, m_command.size );
            } else {
                m_state = REQUEST;
            }

        }
    }
}

bool Megatune::execute( VemsCommand* command )
{
   ReadPageCommand* readCmd = dynamic_cast<ReadPageCommand*>( command );

    if ( readCmd ) {
//        LOGP( wxString::Format(_T("readCmd->pageId: %d"), readCmd->pageId ) );
        return readPage( readCmd->pageId );
//        return true;
    }

    WritePageCommand* writeCmd = dynamic_cast<WritePageCommand*>( command );

    if ( writeCmd ) {
//        LOGP( wxString::Format(_T("writeCmd->pageId: %d"), writeCmd->pageId ) );
        return writePage( writeCmd->pageId, writeCmd->pageOffset, writeCmd->value );
//        return true;
    }

    BurnCommand* burnCmd = dynamic_cast<BurnCommand*>( command );

    if ( burnCmd ) {
//        LOGP( _T("burnCmd->pageId: %d") );
        burn();
        return true;
    }

    StringCommand* stringCmd = dynamic_cast<StringCommand*>( command );

    if ( stringCmd ) {
//        LOGP( _T("stringCmd: ") + stringCmd->m_str );
        LOGP( wxString::Format( _T("stringCmd: ") + stringCmd->m_str + _T(", sleepAfterWrite: %d, sendAck: %d")
            , stringCmd->sleepAfterWrite, stringCmd->sendAck ) );
        executeStringCommand( stringCmd->m_str, stringCmd->sleepAfterWrite, stringCmd->sendAck );
        return true;
    }

    RawCommand* rawCmd = dynamic_cast<RawCommand*>( command );

    if ( rawCmd ) {
        executeRawCommand( rawCmd->rawData );
        return true;
    }


    return false;
}

void Megatune::waitRt( uint8_t* buffer, int size )
{
    if ( size > 0 ) {
        waitRtRead( buffer, size );
    } else {
        waitRtEmptyRead();
    }
}

void Megatune::waitRtRead( uint8_t* buffer, int size )
{
    switch ( m_config.rtCheckType )
    {
        case MegatuneConfig::TRIGGER:
            waitRtReadTriggerFrame( buffer, size );
            break;

        case MegatuneConfig::CRC:
            m_rxBuffer[ m_rxDataLength++ ] = buffer[0];
            waitRtReadCheckCrc();
            break;

        case MegatuneConfig::LENGTH:
            m_rxBuffer[ m_rxDataLength++ ] = buffer[0];
            waitRtReadNoCrc();
            break;

    }
}

void Megatune::waitRtReadTriggerFrame( uint8_t* buffer, int size )
{
    if( m_triggerFrameFSM.parse( buffer[0] ) ){

        if ( m_triggerFrameFSM.getType() != TriggerFrame::EVENT_REALTIME ) return;

        --m_sentReadRTCmdCount;

        sendRtTimePacket();
        sendRtPacket(m_triggerFrameFSM.getFrame());
        updateRealtimeFrequency();

        m_timeoutTimer.start(m_config.timeout);
        m_state = REQUEST;
        return;
    }

    const IntArray& trFrame = m_triggerFrameFSM.getFrame();

    if( trFrame.GetCount() == 33 && m_sentReadRTCmdCount < 2 ){
        m_state = REQUEST;
    }
}

void Megatune::waitRtReadCheckCrc()
{
    if ( m_rxDataLength <= m_config.realtimeLength - 2 )
    {
        m_crc = Math::crc16_sipr( m_crc, m_rxBuffer[m_rxDataLength-1] );
    }

    ///TODO  use constants
    if ( m_rxDataLength == 33 && m_sentReadRTCmdCount < 2 ) {
        m_state = REQUEST;
    }

    if ( m_rxDataLength < m_config.realtimeLength ) return;

    updateRealtimeFrequency();

    --m_sentReadRTCmdCount;

    m_rxCrc = m_rxBuffer[m_rxDataLength-2];
    m_rxCrc |= ((uint16_t)m_rxBuffer[m_rxDataLength-1]<<8);

///TEST ONLY, handle crc error user message
    if ( m_rxCrc != m_crc ) {
        sendRTCRCErrorPacket();
        LOGP( wxString::Format( _T("CRC ERROR m_crc: %4X, m_rxCrc: %04X"), m_crc, m_rxCrc ) );
        int crcErrorsPerSec = m_crcErrorMeter.update();
        LOGP( wxString::Format( _T("crcErrorsPerSec: %d"), crcErrorsPerSec ) );
        if ( crcErrorsPerSec >= MAX_CRC_ERROR_COUNT_PER_SEC ) {
            m_state = END;
            return;
        } else {
            LOG
            m_rxDataLength = 0;
            m_timeoutTimer.start(m_config.timeout);
            m_state = HANDLE_CRC_ERROR;
        }

    } else {
        sendRtTimePacket();
        sendRtPacket();

        m_crc   = 0xFFFF;
        m_rxCrc = 0;
        m_rxDataLength          = 0;
        m_realtimeErrorCount    = 0;
        m_timeoutTimer.start(m_config.timeout);
        m_state = REQUEST;
    }
}

void Megatune::waitRtReadNoCrc()
{
    if ( m_rxDataLength == 33 && m_sentReadRTCmdCount < 2 ) {///), remove magic constants 33
        m_state = REQUEST;
    }

    if ( m_rxDataLength == m_config.realtimeLength )
    {
        updateRealtimeFrequency();

        --m_sentReadRTCmdCount;

        sendRtTimePacket();
        sendRtPacket();

        m_rxDataLength          = 0;

        m_timeoutTimer.start(m_config.timeout);
        m_realtimeErrorCount    = 0;

        m_state = REQUEST;
    }
}

void Megatune::updateRealtimeFrequency()
{
    int updateResult = m_updateFrequencyMeter.update();

    if ( updateResult != -1 ) {
        sendUpdateFrequencyPacket( updateResult );
    }
}

void Megatune::waitRtEmptyRead()
{
    wxThread::Sleep( m_sleep );

    if ( !m_timeoutTimer.isTimedout() ) return;

    if ( ++m_realtimeErrorCount > m_config.maxRealtimeError ) {
        m_state = END;
    } else {
        sendRTTimeoutPacket();
        m_timeoutTimer.start( m_config.timeout );
    }
}

void Megatune::waitPage( uint8_t* buffer, int size )
{
    if ( size > 0 ) {
        waitPageRead( buffer, size );
    } else {
        waitPageEmptyRead();
    }
}

void Megatune::waitPageRead( uint8_t* buffer, int size )
{
    m_rxBuffer[ m_rxDataLength++ ] = buffer[0];

    if ( m_rxDataLength < m_expectedRxSize ) return;

    if ( m_config.pageCheckType == MegatuneConfig::CRC ) {

        if ( !MegatuneHelper::checkCrc( m_rxBuffer, m_rxDataLength ) ) {
            m_state = END;
            return;
        }

        m_rxDataLength -= 2;
    }

    sendPagePacket();
    handlePageRequest();

    m_rxDataLength  = 0;

    m_state = REQUEST;
}

void Megatune::waitPageEmptyRead()
{
    wxThread::Sleep( m_sleep );

    if ( m_timeoutTimer.isTimedout() ) {
        LOGP( wxString::Format( _T("m_expectedRxSize: %d"), m_expectedRxSize ) );
        LOGP( wxString::Format( _T("m_rxDataLength: %d"), m_rxDataLength ) );
        m_state = END;
    }
}

void Megatune::handleCrcError( uint8_t* buffer, int size )
{
    if ( size > 0 ) {

        ++m_rxDataLength;

        if ( m_rxDataLength == m_expectedRxSize ) {

            if ( m_sentReadRTCmdCount > 0 )  {
                --m_sentReadRTCmdCount;
            }

            if ( m_sentReadRTCmdCount == 0 ) {
                LOG
                ///handle crc finished
                m_crc   = 0xFFFF;
                m_rxCrc = 0;
                m_rxDataLength          = 0;
                m_realtimeErrorCount    = 0;

                m_timeoutTimer.start(m_config.timeout);

                m_state = REQUEST;
            }
        }
    } else {
        wxThread::Sleep(m_sleep);
        if ( m_timeoutTimer.isTimedout() ) {
                LOG
                ///handle crc finished
                m_sentReadRTCmdCount = 0;
                m_crc   = 0xFFFF;
                m_rxCrc = 0;
                m_rxDataLength          = 0;
                m_realtimeErrorCount    = 0;

                m_timeoutTimer.start(m_config.timeout);

                m_state = REQUEST;
        }
    }
}

void Megatune::clear()
{
    if ( m_clearCount > 0 ) {
        wxThread::Sleep( m_sleep );
        --m_clearCount;
    } else {
        m_state = REQUEST;
    }
    if ( m_sendAck ) {
        sendAckPacket();
        m_sendAck = false;
    }
}


void Megatune::sendInputEvent( const SIPRPacket& siprPacket )
{
    InputEvent event;
    event.setPacket( siprPacket );
    m_eventHandler->AddPendingEvent( event );
}

void Megatune::sendRtPacket()
{
    m_buffer[0] = 0;
    m_buffer[1] = 0;
    m_buffer[2] = m_config.siprSensorType;

    SIPRPacket packet( 0, SIPR_READ_SENSORS, m_config.realtimeLength + 3, m_buffer );
    InputEvent event;
    event.setPacket( packet );

    m_eventHandler->AddPendingEvent( event );
}

void Megatune::sendRtPacket(const IntArray& frame)
{
    SIPRPacket packet( 0, SIPR_READ_SENSORS, frame.GetCount()-3, frame );
    InputEvent event;
    event.setPacket( packet );

    m_eventHandler->AddPendingEvent( event );
}

void Megatune::sendRtTimePacket()
{
    wxDateTime now = wxDateTime::UNow();
    wxTimeSpan timeSpan = now.Subtract( m_startTime );
    long elapsedTime = timeSpan.GetMilliseconds().ToLong();

    Math::getBytes( elapsedTime, m_timePacketBuffer, 4 );
    SIPRPacket packet( 0, SIPR_RT_TIME, 4, m_timePacketBuffer );
    //m_receiver->receive( packet );
    InputEvent event;
    event.setPacket( packet );
    m_eventHandler->AddPendingEvent( event );
}

void Megatune::sendUpdateFrequencyPacket( int updateFrequency )
{
    uint8_t buffer[4];
    Math::getBytes( updateFrequency, buffer );

    SIPRPacket packet( 0, SIPR_UPDATE_FREQUENCY, 4, buffer );
    //m_receiver->receive( packet );
    InputEvent event;
    event.setPacket( packet );
    m_eventHandler->AddPendingEvent( event );
}

void Megatune::sendPagePacket()
{
    SIPRPacketList packets = SIPRPacketHelper::createSIPRPackets( *m_descriptor, m_rxBuffer, m_page );

    for ( SIPRPacketList::iterator it = packets.begin(); it != packets.end(); ++it ){
        SIPRPacket* packet = *it;
        //m_receiver->receive( *packet );
        InputEvent event;
        event.setPacket( *packet );
        m_eventHandler->AddPendingEvent( event );
        delete packet;
    }
}

void Megatune::sendAckPacket()
{
    SIPRPacket packet( 0, SIPR_ACK );
    InputEvent event;
    event.setPacket( packet );
    m_eventHandler->AddPendingEvent( event );
}

void Megatune::sendRTCRCErrorPacket()
{
    SIPRPacket packet( 0, SIPR_RT_CRC_ERROR );
    InputEvent event;
    event.setPacket( packet );
    m_eventHandler->AddPendingEvent( event );
}

void Megatune::sendRTTimeoutPacket()
{
    SIPRPacket packet( 0, SIPR_TIMEOUT );
    InputEvent event;
    event.setPacket( packet );
    m_eventHandler->AddPendingEvent( event );
}

void Megatune::sendRefreshStartEvent()
{
    uint8_t packetData = SIPR_REFRESH_START;
    SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &packetData );
    sendInputEvent( packet );
}

void Megatune::sendRefreshStopEvent()
{
    uint8_t packetData = SIPR_REFRESH_STOP;
    SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &packetData );
    sendInputEvent( packet );
}

void Megatune::handlePageRequest()
{
    if ( m_readPageRequests.empty() ) {
        LOGP( _T("TODO!") );
        return;
    }

    PageRequest* request = m_readPageRequests.front();

    if ( !request ) {
        m_readPageRequests.pop_front();
        LOGP( _T("TODO!") );
        return;
    }

    if ( !request->requestedPages.erase( m_page ) ) {
        LOGP( _T("TODO!") );
        return;
    }

//    LOGP( wxString::Format( _T("page id: %d removed!"), m_page ) );

    if ( request->requestedPages.empty() ) {
//        LOGP( wxString::Format( _T("request removed!"), m_page ) );

        m_readPageRequests.pop_front();
        delete request;

        sendRefreshStopEvent();
    }
}

void Megatune::activatePageRequest()
{
    if ( m_readPageRequests.empty() )
        return;

    PageRequest* request = m_readPageRequests.front();

    if ( !request )
        return;

    if ( !request->active ) {
            request->active = true;
            sendRefreshStartEvent();
    }
}

bool Megatune::isTriggerFrame()
{
    FirmwareVersion fw(1,1,86);
    return (m_config.fwVersion >= fw);
}
