#include "PortScanner.h"
#include "DeviceDetector.h"
#include "PortScannerListener.h"
#include "PortScannerEvent.h"
#include "../../CommunicationBase.h"
#include "../../logger.h"


PortScanner::PortScanner(   PortScannerListener* listener,
                            CommunicationBase* commBase,
                            const IntArray& baudrates,
                            size_t preferredBaudrate )
{
    m_listener  = listener;
    m_commBase  = commBase;
    m_baudrates = baudrates;

    if ( preferredBaudrate > 0 ) {
        int idx = m_baudrates.Index( preferredBaudrate );
        if (  idx != wxNOT_FOUND ) {
            m_baudrates.RemoveAt(idx);
        }
        m_baudrates.Insert( preferredBaudrate, 0u );
    }

    m_commClient    = NULL;
    m_currBaudrateIdx = 0;

    m_device    = ECUInfo::DEVICE_UNKNOWN;
    m_mode      = ECUInfo::UNKNOWN;

    m_state = START_PROCESS;
}

PortScanner::~PortScanner()
{
    delete m_commClient;
    m_commClient = NULL;
}

///CommunicationControllerClient methods
bool PortScanner::startProcess()
{
    if ( m_state == START_PROCESS ) {
        m_commBase->setBaudRate( m_baudrates[m_currBaudrateIdx] );
        m_state = WRITE;
    }

    return true;
}

void PortScanner::write( uint8_t* buffer, int& size )
{
    if ( m_state == WRITE ) {
        registerDeviceDetector();
        m_commClient->write( buffer, size );
        m_state = READ;
    }
}

void PortScanner::read( uint8_t* buffer, int size )
{
    if ( m_state == READ ) {
        m_commClient->read( buffer, size );

        if ( m_commClient->isFinished () ) {
            m_commClient->finish();
        }
    }
}

void PortScanner::finish()
{
    if ( m_state == DEVICE_FOUND  ) {
        PortScannerEvent event;
        event.device    = m_device;
        event.mode      = m_mode;
        event.baudrate  = m_baudrates[m_currBaudrateIdx];
        m_listener->OnDeviceFound(event);
    } else {
        m_listener->OnDeviceNotFound();
    }
}

bool PortScanner::isFinished()
{
    return ( ( m_state == DEVICE_FOUND ) || ( m_state == DEVICE_NOT_FOUND ) );
}

void PortScanner::reset()
{
    m_currBaudrateIdx = 0;
    m_state = START_PROCESS;

    m_device    = ECUInfo::DEVICE_UNKNOWN;
    m_mode      = ECUInfo::UNKNOWN;
}

/// DeviceDetectorListener methods
void PortScanner::OnDeviceFound( ECUInfo::Device device, ECUInfo::EcuMode deviceMode )
{
    m_device    = device;
    m_mode      = deviceMode;
    m_state = DEVICE_FOUND;
}

void PortScanner::OnDeviceNotFound()
{
    if ( ++m_currBaudrateIdx < m_baudrates.GetCount() ) {
        m_state = START_PROCESS;
        LOG
    } else {
        LOG
        m_state = DEVICE_NOT_FOUND;
    }
}

void PortScanner::registerDeviceDetector()
{
    delete m_commClient;
    m_commClient = new DeviceDetector( this );
    m_commClient->setTimeout( m_timeout );
    m_commClient->setInterWriteDelay( m_interWriteDelay );
}
