#include "SerialPortScanner.h"
#include "DeviceDetector.h"
#include "../CommunicationController2.h"
#include "../CommunicationControllerEvent.h"
#include "../../CommunicationBase.h"
#include "../SerialPortHandlerFactory.h"
#include "../../SerialException.h"
#include "../../serialconfig.h"
#include "../../logger.h"
#include "../../events/SerialPortScannerEvent.h"
#include "../../events/CommunicationEvent.h"


SerialPortScanner::SerialPortScanner(   wxEvtHandler* eventHandler,
                                        const wxArrayString& ports, const IntArray& baudrates,
                                        size_t timeout, size_t interWriteDelay,
                                        const wxString& firstPort, int firstBaudrate )
{
    m_commBase          = NULL;
    m_commClient        = NULL;

    m_eventHandler      = eventHandler;

    m_commController    = new CommunicationController2();
    m_commController->setEventHandler( this );

    m_ports             = ports;
    m_baudrates         = baudrates;
    m_timeout           = timeout;
    m_interWriteDelay   = interWriteDelay;
    m_firstPort         = firstPort;
    m_firstBaudrate     = firstBaudrate;

    m_currPortIdx       = 0;
    m_currBaudrateIdx   = 0;

    m_maxBaudrateTryCount = 3;
    m_baudrateTryCount = 0;


    Connect( COMMUNICATION_CONTROLLER_START, wxCommandEventHandler( SerialPortScanner::onStartEvent ), NULL, this );
    Connect( wxID_ANY, wxEVT_COMMUNICATION, CommunicationEventHandler( SerialPortScanner::OnCommunicationEvent ), NULL, this );

    scanPort();
}

SerialPortScanner::~SerialPortScanner()
{
    Disconnect( COMMUNICATION_CONTROLLER_START, wxCommandEventHandler( SerialPortScanner::onStartEvent ), NULL, this );

    m_commController->stop();
	delete m_commController;

    delete m_commClient;
    delete m_commBase;
}

///// CommunicationClientEventHandler methods
//void SerialPortScanner::handleEvent( CommunicationClientEvent::Type eventType )
//{
//    switch ( eventType )
//    {
//        case DEVICE_FOUND:
//            deviceFound();
//            break;
//
//        case DEVICE_NOT_FOUND:
//            deviceNotFound();
//            break;
//
//        default:
//            break;
//    }
//}

void SerialPortScanner::OnDeviceFound( ECUInfo::Device device, ECUInfo::EcuMode deviceMode )
{
    m_deviceInfo.device = device;
    m_deviceInfo.deviceMode = deviceMode;

    deviceFound();
}

void SerialPortScanner::OnDeviceNotFound()
{
    deviceNotFound();
}

void SerialPortScanner::scanPort()
{
    setCommunicationBase();
    registerDeviceDetector();
    startCommunicationController();
}

void SerialPortScanner::setCommunicationBase()
{
    LOGP( wxString::Format( _T("PORT: ") + m_ports[m_currPortIdx] + _T(", BAUDRATE: %6d"), m_baudrates[m_currBaudrateIdx] ) );

    m_deviceInfo.serialConfig.setPortName( m_ports[m_currPortIdx] );
    m_deviceInfo.serialConfig.setBaudRate( m_baudrates[m_currBaudrateIdx] );

    delete m_commBase;
    m_commBase = SerialPortHandlerFactory::create( &m_deviceInfo.serialConfig );

    m_commController->setCommunicationBase( m_commBase );
}

void SerialPortScanner::registerDeviceDetector()
{
    delete m_commClient;
    m_commClient = new DeviceDetector( this );
    m_commClient->setTimeout( m_timeout );
    m_commClient->setInterWriteDelay( m_interWriteDelay );

    m_commController->registerClient( m_commClient );
}

void SerialPortScanner::startCommunicationController()
{
    wxCommandEvent startEvent( COMMUNICATION_CONTROLLER_START, wxID_ANY );
    AddPendingEvent( startEvent );
}

void SerialPortScanner::deviceFound()
{
    sendSerialPortScannerEvent( SerialPortScannerEvent::SPS_EVT_DEVICE_FOUND, true );
}

void SerialPortScanner::deviceNotFound()
{
    bool isScanningFinished = false;

    if ( ++m_baudrateTryCount < m_maxBaudrateTryCount ) {
        scanPort();
        return;
    } else {
        m_baudrateTryCount = 0;
    }

    if ( ( m_currBaudrateIdx + 1 ) < m_baudrates.GetCount() ) {
        sendSerialPortScannerEvent( SerialPortScannerEvent::SPS_EVT_DEVICE_NOT_FOUND, isScanningFinished );
        ++m_currBaudrateIdx;
    } else if ( ( m_currPortIdx + 1 ) < m_ports.GetCount() ) {
        sendSerialPortScannerEvent( SerialPortScannerEvent::SPS_EVT_DEVICE_NOT_FOUND, isScanningFinished );
        m_currBaudrateIdx = 0;
        ++m_currPortIdx;
    } else {
        isScanningFinished = true;
        sendSerialPortScannerEvent( SerialPortScannerEvent::SPS_EVT_DEVICE_NOT_FOUND, isScanningFinished );
    }

    if ( !isScanningFinished ) {
        scanPort();
    }

}

void SerialPortScanner::onStartEvent( wxCommandEvent& event )
{
    m_commController->start();
}

void SerialPortScanner::OnCommunicationEvent( CommunicationEvent& commEvent )
{
    LOG

    switch ( commEvent.GetId() )
    {
        case CommunicationEvent::CANNOT_OPEN_PORT:
        case CommunicationEvent::PORT_CLOSED:
        {
            bool isScanningFinished = false;

            if ( ( m_currPortIdx + 1 ) < m_ports.GetCount() ) {
                sendSerialPortScannerEvent( SerialPortScannerEvent::SPS_EVT_CANNOT_OPEN_PORT, isScanningFinished );
                ++m_currPortIdx;
                m_currBaudrateIdx = 0;
            } else {
                isScanningFinished = true;
                sendSerialPortScannerEvent( SerialPortScannerEvent::SPS_EVT_CANNOT_OPEN_PORT, isScanningFinished );
            }

            if ( !isScanningFinished ) {
                scanPort();
            }
        } break;

        case CommunicationEvent::REMOVE_USB:
            LOGP( wxString::Format( _T("m_eventHandler: %p"), m_eventHandler ) );
            if ( m_eventHandler ) {
                LOG
                m_eventHandler->AddPendingEvent( commEvent );
            }
            break;
    }
}

void SerialPortScanner::sendSerialPortScannerEvent( int eventId, bool isScanningFinished )
{
    SerialPortScannerEvent event;

    event.SetId( eventId );
    event.setDeviceInfo( m_deviceInfo );
    event.setScanningFinished( isScanningFinished );
    event.setNumPorts( m_ports.GetCount() );
    event.setNumBauds( m_baudrates.GetCount() );
    event.setCurrPortIdx( m_currPortIdx );
    event.setCurrBaudIdx( m_currBaudrateIdx );

    if ( !isScanningFinished ) {
        event.setNextPortName( m_ports[m_currPortIdx] );
    }

    if ( m_eventHandler ) {
        m_eventHandler->AddPendingEvent( event );
    } else {
        LOGP( _T("ERROR - m_eventHandler is NULL!") );
    }
}
