#include "TriggerLogReader.h"
#include "TimeoutListener.h"
#include "../../common/TxData.h"
#include "../../logger.h"
#include "../../events/RawTriggerLogEvent.h"
#include "../../events/CommunicationEvent.h"
#include "../../events/GeneralEvent.h"

TriggerLogReader::TriggerLogReader( TimeoutListener* timeoutListener,
                                    wxEvtHandler* eventHandler )

{
    m_timeoutListener       = timeoutListener;
    m_eventHandler          = eventHandler;

    m_state                 = SEND_START_CMD;

    m_rxDataLength          = 0;
    m_timeoutCount          = 0;
    m_mde40counter          = 0;
}

TriggerLogReader::~TriggerLogReader()
{
}

void TriggerLogReader::stop()
{
    m_state = SEND_STOP_CMD;
}

void TriggerLogReader::write( uint8_t* buffer, int& size )
{
    switch ( m_state )
    {
        case SEND_START_CMD:
            sendStartCmd( buffer, size );
            break;

        case SEND_STOP_CMD:
            LOG
            sendStopCmd( buffer, size );
            break;

        case TIMEDOUT:
            sendRestart( buffer, size );
            break;

        case SEND_MXC:
            sendMxc( buffer, size );
            break;

        //case INCOMING_AIM:

        default:
            break;
    }
}

void TriggerLogReader::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case WAIT_LOG:
            waitLog( buffer, size );
            break;

        case WAIT_MXC:
            waitMxc( buffer, size );
            break;

        default:
            break;
    }
}

void TriggerLogReader::finish()
{
    if ( m_state == STOPPED ) {
        GeneralEvent event;
        event.SetString( _T("TRIGGER_LOG_STOPPED") );
        m_eventHandler->AddPendingEvent( event );
    } else {
        m_timeoutListener->OnTimeout();
    }
}


bool TriggerLogReader::isFinished()
{
    //LOG
    return ( ( m_state == FINISHED ) || ( m_state == STOPPED ) );
}

void TriggerLogReader::sendMxc( uint8_t* buffer, int& size )
{
    m_mxcsize = 0;
    m_mde40counter = 0;
    m_mxctimeout = 0;

    TxData command( _T("Manmxc") );
    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_state = WAIT_MXC;
}

void TriggerLogReader::waitMxc( uint8_t* buffer, int size )
{
    if ( size > 0 )
    {
        m_mxcbuff[m_mxcsize++] = buffer[0];
        m_mxctimeout = 0;
        if ( m_mxcsize == 25 )
        {
            if ( m_mxcbuff[0] == 'Y' && m_mxcbuff[1] == 'e' && m_mxcbuff[2] == 's' && m_mxcbuff[3] == ',' )
            {
                LOGP( _T("COMM WORX, NO TRIGGER INPUT") )
                if ( m_eventHandler )
                {
                    CommunicationEvent event( CommunicationEvent::COMM_WORX );
                    m_eventHandler->AddPendingEvent( event );
                }
            }
            else
            {
                LOGP(_T("OOPS! SOMETHING STRANGE REPLY CAME FROM THE ECU AFTER MXC COMMAND"))
            }
            m_state = TIMEDOUT;
        }
    }
    else
    {
        wxThread::Sleep( 1 );
        ++m_mxctimeout;
        if ( m_mxctimeout > 100 )
        {
            LOGP( _T("COMM DOES NOT WORK"))
            if ( m_eventHandler )
            {
                CommunicationEvent event( CommunicationEvent::COMMUNICATION_ERROR );
                m_eventHandler->AddPendingEvent( event );
            }
            m_mxctimeout = 0;

            //m_state = FINISHED;
            m_state = TIMEDOUT;
        }
    }
}
void TriggerLogReader::sendStartCmd( uint8_t* buffer, int& size )
{
    TxData command( _T("mde40") );
    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_rxDataLength          = 0;
    m_timeoutCount          = 0;

    m_state = WAIT_LOG;
}

void TriggerLogReader::sendRestart( uint8_t* buffer, int& size )
{
    //LOGP( _T("Sent Manmde40"))

    m_mde40counter++;

    if ( m_mde40counter > 5 )
    {
        m_state = SEND_MXC;
        size = 0;
        return;
    }

    TxData command( _T("Manmde40") );
    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_rxDataLength = 0;
    m_timeoutCount = 0;

    m_state = WAIT_LOG;
}

void TriggerLogReader::waitLog( uint8_t* buffer, int size )
{
    if ( size > 0 ) {
        //LOGP( _T("waitLog size>0"))
        m_rxBuffer[ m_rxDataLength++ ] = buffer[0];
        if ( m_rxDataLength < BUFFER_SIZE ) return;
        sendEvent(m_rxDataLength);

    } else {
        wxThread::Sleep( m_sleep );
        m_timeoutCount++;
        if(m_timeoutCount < 100){
            //LOG
            return;
        }
       //LOGP( _T("TIMEOUT") )
        m_timeoutCount = 0;
        if(m_rxDataLength) {
            //LOG
            sendEvent(m_rxDataLength);
        }
        else {
            //LOG
            m_state = TIMEDOUT;
        }
    }
}

void TriggerLogReader::sendEvent(size_t size)
{
    if ( checkAIM( size ) )
    {
        m_rxDataLength = 0;
        m_timeoutCount = 0;
        m_mde40counter = 0;
        LOGP( _T("Incoming AIM data -> Manmde40") )
        m_state = TIMEDOUT;
    }
    else
    {
        wxMemoryBuffer eventBuffer( size );
        eventBuffer.AppendData( m_rxBuffer, size );
        RawTriggerLogEvent event;
        event.SetData( eventBuffer );
        m_eventHandler->AddPendingEvent( event );

        m_rxDataLength  = 0;
        m_timeoutCount  = 0;
    }
}

void TriggerLogReader::sendStopCmd( uint8_t* buffer, int& size )
{
    TxData command( _T("mde00") );
    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_rxDataLength          = 0;
    m_timeoutCount          = 0;

    m_state = STOPPED;

    //m_state = WAIT_LOG;
}

bool TriggerLogReader::checkAIM( size_t size )
{
    unsigned char aimcounter = 0;

    for ( size_t i = 0; i < size; ++i )
    {
        if ( m_rxBuffer[i] == 0xA3 && (i > 0 ) && (i + 4 < size ) )
        {
            if ( isValidAIM( m_rxBuffer + i - 1 ) )
                ++aimcounter;
        }
        if ( aimcounter > 2 )
            return true;
    }
    return false;
}

bool TriggerLogReader::isValidAIM( uint8_t* buffer )
{
    uint8_t sum = 0;
    for ( unsigned char i = 0; i < 4; ++i )
    {
        sum += buffer[i];
    }
    return ( sum == buffer[4] );
}
