#include "EepromCfgSaverBase.h"

#include <wx/textfile.h>

#include "../bcWidgets/bcProgressDialog.h"
#include "../common/VemsContext.h"
#include "../events/GeneralEvent.h"
#include "../globalhparser/HeaderParser.h"
#include "../globalhparser/SimpleMapParser.h"
#include "../globalhparser/SymbolTableFiller.h"
#include "../DirHelper.h"
#include "../logger.h"
#include "../SymbolMgr.h"
#include "../Symbol.h"
#include "../FirmwareInfo.h"
#include "../UserLog.h"

BEGIN_EVENT_TABLE( EepromCfgSaverBase, wxEvtHandler )
    EVT_GENERAL( wxID_ANY, EepromCfgSaverBase::onGeneralEvent )
END_EVENT_TABLE()

EepromCfgSaverBase::EepromCfgSaverBase()
{
    m_symbolMgr = NULL;
}

EepromCfgSaverBase::EepromCfgSaverBase( SymbolMgr* symbolMgr )
{
    m_symbolMgr = symbolMgr;
}

EepromCfgSaverBase::~EepromCfgSaverBase()
{
    delete m_headerparser;
}

void EepromCfgSaverBase::save( bool createProgressDialog, bcProgressDialog* dlg )
{
    m_progDialog = dlg;
    m_selfCreatedPrgDlg = createProgressDialog;

    if ( m_selfCreatedPrgDlg )
    {
        m_progDialog = new bcProgressDialog( _T("Config download"), _T("Config download in progress"), 100 );
    }

    if ( m_progDialog )
    {
        wxSafeYield( m_progDialog );

        m_progDialog->Show();
    }

    getEepromData();
}

void EepromCfgSaverBase::parseGlobalH( const wxString& headerfile, const wxString& mymakefile )
{
    m_headerparser = new HeaderParser( mymakefile );
    m_headerparser->parse( headerfile );
    m_parsedstructures = m_headerparser->getParsedStructs();
}

void EepromCfgSaverBase::parseMapFile( const wxString& mapfile )
{
    SimpleMapParser parser( &m_structAddresses );
    parser.parse( mapfile );
}

void EepromCfgSaverBase::onGeneralEvent( GeneralEvent& event )
{
    if ( event.GetString() == _T("EEPROM_DL_PROGRESS") )
    {
        double percent = event.GetDouble();

        if ( m_progDialog != NULL )
        {
            m_progDialog->Update( (int) percent );
        }

        return;
    }

    int status = event.GetLong();

    if ( status == EEPROM_SAVE_OK )
    {
        _saveintern();
    }
    else if ( status == EEPROM_SAVE_ERROR )
    {
        //LOGP( _T("Error when saving eeprom"))
        UserLog::uLog( ULOG_ERROR, _T("Could not save config") );
        wxMessageBox( _T("Config has NOT been saved."), _T("Error"), wxOK | wxICON_EXCLAMATION );
    }

    if ( m_selfCreatedPrgDlg )
    {
        m_progDialog->Destroy();
        m_progDialog = NULL;
}
}

void EepromCfgSaverBase::_saveintern()
{
    bool destroySymMgr = false;

    if ( m_symbolMgr == NULL )
    {
        m_symbolMgr = new SymbolMgr;
        destroySymMgr = true;
    }

    SymbolTableFiller stfiller( m_symbolMgr );

    String2IntHash::const_iterator it = m_structAddresses.begin();

    for ( ; it != m_structAddresses.end(); ++it )
    {
        stfiller.fillSymbolTable( m_parsedstructures, it->first, it->second ); // name, offset
    }
    saveToCfgFile( &m_symbolMgr->getSymbolTable() );

    if ( destroySymMgr )
    {
        delete m_symbolMgr;
        m_symbolMgr = NULL;
    }
}


void EepromCfgSaverBase::saveToCfgFile( SymbolTable* symTable )
{
    wxDateTime dt = wxDateTime::Now();

    wxString ds = dt.FormatISODate() + _T("_") + dt.FormatISOTime();

    ds.Replace(_T(":"),_T(".")); ds.Replace(_T("-"),_T("."));

    wxString fname = DirHelper::GetPath( dLOG ) + _T("eepromcfg_") + ds + _T(".vemscfg2");

    wxTextFile f( fname );

    if ( f.Exists() )
    {
        f.Open();
        f.Clear();
    }
    else
        f.Create();

    addHeader( f );

    SymbolTable::iterator it = symTable->begin();

    for ( ; it != symTable->end(); ++it )
    {
        Symbol* sym = it->second;
        unsigned long offset = sym->getOffset();
        unsigned char size = sym->getSize();

        wxString line;

        if ( sym->isArray() == false )
        {
            if ( offset + size >= m_size )
                continue;

            unsigned long value = 0;

            for ( unsigned int j = 0; j < size; ++j )
            {
                value = 256 * value + m_eepromdata[offset+j];
            }

            line = sym->getName() + wxString::Format( _T("=%02x"), value );
        }
        else
        {
            size_t arrsize = sym->getArraySize();
            if ( offset + size * arrsize >= m_size )
                continue;

            line = sym->getName() + _T("=");
            for ( unsigned int i = 0; i < arrsize; ++i )
            {
                unsigned long  value = 0;
                for ( unsigned int j = 0; j < size; ++j )
                {
                    value = 256 * value + m_eepromdata[offset + i*size + j];
                }
                line += wxString::Format( _T("%02x,"), value );
            }
            line.RemoveLast(1);
        }
        f.AddLine( line );
    }
    f.Write();
    f.Close();

    UserLog::uLog( ULOG_OK, _T("Config has been saved to: ") + fname );
    wxMessageBox( _T("Config has been saved. See log for details"), _T("Information"), wxOK | wxICON_INFORMATION );
}

void EepromCfgSaverBase::addHeader( wxTextFile& f )
{
    wxDateTime dt = wxDateTime::Now();

    wxString ds = dt.FormatISODate() + _T(" ") + dt.FormatISOTime();

    wxString hdr = _T("# CurrentTime: ") + ds + _T("\n");

    hdr += _T("# ConfigVersion: 2.0\n");

    if ( toString() == _T("ECU") )
    {
        wxObject* fwobj = VemsContext::getObject( _T("FirmwareInfo") );
        FirmwareInfo* fwInfo = dynamic_cast<FirmwareInfo*>( fwobj );

        hdr += _T("# Version: ") + fwInfo->getFirmwareVersionAsString() + _T("\n");
        hdr += _T("# SerialNumber: ") + fwInfo->getSerialNumber() + _T("\n");
        hdr += _T("# CompilationTime: ") + fwInfo->getCompilationTime() + _T("\n");
        hdr += _T("# Signature: ") + fwInfo->getSignature() + _T("\n");
        hdr += _T("# CompileOptions: ") + fwInfo->getCompileOptions() + _T("\n");
    }

    f.AddLine( hdr );
}

