#include "RPNParser.h"

#include <wx/arrstr.h>

#include "../Utils.h"
#include "ExpressionFactory.h"
#include "../logger.h"
#include "StringUtil.h"
#include <wx/regex.h>

/**
Reverse Polish Notation parser
*/
RPNParser::RPNParser()
{
    m_rpnTokenA = NULL;
    //m_variableH = NULL;
    m_data = NULL;

    m_shExceptions = true;
}

RPNParser::~RPNParser()
{
    clearCache();
}

void RPNParser::reset()
{
    m_infixTokenA.clear();

    m_stack.clear();

    m_cache.clear();

    if ( m_rpnTokenA )
        m_rpnTokenA->clear();
}


//  String expression = "42,2,1,-,+";
// String expression = ""fileName",42,table"; table("fileName", 42)
// String expression = "a,42,table("; lookup: a=123 table(123, 42)
//VARIABLE -> lokup and put number;
//fv        -> remove some item and put number;
// string   -> put string;
// number   -> put number;

// a12b : variable;
// "str": string
// fv(  : fv
/*
void RPNParser::setVariableHash( DoubleHash* varH )
{
    m_variableH = varH;
}
*/

void RPNParser::setVariableData( VariableData* data )
{
    m_data = data;
}

void RPNParser::parseInfix( const wxString& str )
{
    m_infixTokenA.clear();

    int pNumber = 0;
    bool firstToken = true;

    ExpressionBase::TokenType prevToken = ExpressionBase::UNKNOWN;

    size_t length = str.length();
    uint32_t idx = 0;
    wxString token = wxEmptyString;

    bool isPrevDelim = false;

    while( idx < length )
    {
        token = wxEmptyString;

        switch( (wxChar)str[idx] )
        {
        case _T(' '):
        {
            idx++;
        } break;
        case _T('*'):
        case _T('/'):
        case _T('+'):
        case _T('='):
        case _T(','):
        case _T('?'):
        case _T(':'):
        case _T('&'):
        case _T('|'):
        case _T('~'):
        case _T('!'):
        {
            token += str[idx];
            idx++;

            if( idx < length )
            {
                if( ( str[idx] == _T('=') ) || ( str[idx] == _T('&') ) || ( str[idx] == _T('|') ) )
                {
                    token += str[idx];
                    idx++;
                }
            }
        } break;
        case _T('('):
        {
            pNumber++;
            token += str[idx];
            idx++;
        } break;
        case _T(')'):
        {
            pNumber--;
            token += str[idx];
            idx++;
        } break;
        case _T('>'):
        {
            token += str[idx];
            idx++;

            if( idx < length )
            {
                if( str[idx] == _T('>') || str[idx] == _T('=') )
                {
                    token += str[idx];
                    idx++;
                }
            }

        } break;
        case _T('<'):
        {
            token += str[idx];
            idx++;

            if( idx < length )
            {
                if( str[idx] == _T('<') || str[idx] == _T('=') )
                {
                    token += str[idx];
                    idx++;
                }
            }
        } break;
        case _T('"'):
        {
            do {
                token += str[idx];
                idx++;
            }while( idx < length && ( str[idx] != _T('"') ) );

            if( idx < length && ( str[idx] == _T('"') ) )
            {
                token += str[idx];
                idx++;
            }
            else
            {
                throw ExpError::SyntaxError( _T("Parsing error : missing \"") );
            }

        } break;
        case _T('-'):
        {
            if ( isPrevDelim == false )
            {
            token += str[idx];
            idx++;

            if( !( ( prevToken == ExpressionBase::OPERATOR ) || ( prevToken == ExpressionBase::LP ) || firstToken ||
                ( prevToken == ExpressionBase::COMMA )  || ( prevToken == ExpressionBase::COLON ) ) )
            {
                break;
            }
            }

        }
        case _T('0'):
        case _T('1'):
        case _T('2'):
        case _T('3'):
        case _T('4'):
        case _T('5'):
        case _T('6'):
        case _T('7'):
        case _T('8'):
        case _T('9'):
        {
            if ( isPrevDelim == false )
            {
            while( idx < length )
            {
                if( StringUtil::isNumber( str[idx] ) ||  str[idx] == _T('.') )
                {
                    token += str[idx];
                    idx++;
                }
                else
                {
                    break;
                }
            }
            }
        } break;
        default:
        {
            if( StringUtil::isAlpha( str[idx] ) )//|| str[idx] == '[' || str[idx] == ']')
            {
                do {
                    if ( str[idx] == '[' )
                        isPrevDelim = true;

                    token += str[idx];

                    if (  str[idx] == ']' )
                    {
                        isPrevDelim = false;
                        idx++;
                        break;
                    }

                    idx++;

                }while( idx < length  && ( StringUtil::isANum( str[idx] ) || ( str[idx] == _T('_') ) ||
                        ( str[idx] == _T('[') ) || ( str[idx] == _T(']') ) || (str[idx] == '-') ) );
            }
            else
            {
                idx++;
            }
        }
        }

        if( !token.IsEmpty() )
        {
            ExpressionBase* b = ExpressionFactory::create( token, m_data );

            if( b )
            {
                if( firstToken )
                {
                    firstToken = !firstToken;
                }
                prevToken = b->getType();
                m_infixTokenA.Add( b );
            }
        }
    }

    if( pNumber )
    {
        throw ExpError::SyntaxError( _T("Parsing error : parentheses mismatch") );
    }
}


void RPNParser::infixToRpn()
{
    m_rpnTokenA->clear();

    InfixTokenA opStack;

    size_t infA = m_infixTokenA.GetCount();
    for(uint32_t i = 0; i < m_infixTokenA.GetCount(); i++)
    {
        //Logger::logf( LOG_WARNING, _T("OperatorStack : ") + opStackToString( opStack ) );

        ExpressionBase* token = m_infixTokenA.Item( i );

        switch( token->getType() )
        {
            case ExpressionBase::NUMBER :
            case ExpressionBase::VARIABLE :
            case ExpressionBase::STRING :
            {
                m_rpnTokenA->Add( token );
            } break;
            case ExpressionBase::FUNCTION :
            {
                opStack.Add( token );
            } break;
            case ExpressionBase::COMMA :
            {
                bool foundLP = false;

                while( !opStack.IsEmpty() )
                {
                    ExpressionBase* lastToken = opStack.Last();

                    if( lastToken->getType() != ExpressionBase::LP )
                    {
                        opStack.RemoveAt( opStack.GetCount()-1 );
                        m_rpnTokenA->Add( lastToken );
                        continue;
                    }
                    else
                    {
                        foundLP = true;
                        break;
                    }
                }

                if( !foundLP )
                {
                    throw ExpError::SyntaxError( _T("InfixToRPN error :Missing ( or , misplaced") );
                }
            } break;
            case ExpressionBase::COLON :
            {
                bool foundQuery = false;

                while( !opStack.IsEmpty() )
                {
                    ExpressionBase* lastToken = opStack.Last();

                    if( lastToken->getOperator() != _T("?") )
                    {
                        opStack.RemoveAt( opStack.GetCount()-1 );
                        m_rpnTokenA->Add( lastToken );
                        continue;
                    }
                    else
                    {
                        foundQuery = true;
                        break;
                    }
                }

                if( !foundQuery )
                {
                    throw ExpError::SyntaxError( _T("InfixToRPN error :Missing ( or , misplaced") );
                }
            } break;
            case ExpressionBase::LP :
            {
                opStack.Add( token );
            } break;
            case ExpressionBase::RP :
            {
                bool foundLP = false;

                while( !opStack.IsEmpty() )
                {
                    ExpressionBase* lastToken = opStack.Last();

                    if( lastToken->getType() == ExpressionBase::OPERATOR )
                    {
                        opStack.RemoveAt( opStack.GetCount()-1 );
                        m_rpnTokenA->Add( lastToken );
                        continue;
                    }
                    else if( lastToken->getType() == ExpressionBase::LP )
                    {
                        opStack.RemoveAt( opStack.GetCount()-1 );

                        foundLP = true;
                        break;
                    }
                    else
                    {
                        throw ExpError::SyntaxError( _T("InfixToRPN error : bad token") );
                    }
                }

                if( !opStack.IsEmpty() )
                {
                    ExpressionBase* lastToken = opStack.Last();

                    if( lastToken->getType() == ExpressionBase::FUNCTION )
                    {
                        opStack.RemoveAt( opStack.GetCount()-1 );
                        m_rpnTokenA->Add( lastToken );
                    }

                }

                if( !foundLP )
                {
                    throw ExpError::SyntaxError( _T("InfixToRPN error :Missing (") );
                }
            } break;
            case ExpressionBase::OPERATOR:
            {
                while( !opStack.IsEmpty() )
                {
                    ExpressionBase* lastToken = opStack.Last();

                    if( token->getPrec() <= lastToken->getPrec() )
                    {
                        opStack.RemoveAt( opStack.GetCount()-1 );
                        m_rpnTokenA->Add( lastToken );
                        continue;
                    }
                    else
                    {
                        break;
                    }

                }

                opStack.Add( token );
            } break;
            default:
            {
                throw ExpError::SyntaxError( _T("InfixToRPN error : bad token") );
            }
        }
   }

   while( !opStack.IsEmpty() )
   {
        m_rpnTokenA->Add( opStack.Last() );
        opStack.RemoveAt( opStack.GetCount()-1 );
   }
//   m_infixTokenA.clear();
/*
   //TEST
   wxString msg = wxEmptyString;

   for(uint32_t i=0; i<m_rpnTokenA->GetCount(); i++ )
   {
       msg += m_rpnTokenA->Item(i)->toString() + _T(" ");
   }

   Logger::logf( LOG_WARNING, _T("Tokens : ") + msg );
*/
}

double RPNParser::evaluate()
{
    m_stack.clear();
    RPNTokenBase* e;

    for(uint32_t i =0; i< m_rpnTokenA->GetCount(); i++){
        e = m_rpnTokenA->Item( i );
        e->interpret( m_stack );
    }

    RPNStackItem* stackItem = m_stack.pop();
    double result = stackItem->getNumber();
    delete stackItem;
    return result;
}

double RPNParser::calculate( const wxString& expr )
{
    //
    double result = 0.0;



    try{
        if( m_cache.find( expr ) == m_cache.end() )
        {
            //Logger::logf( LOG_WARNING, _T("RPNParser::calculate() new RPNToken %d"), m_cache.size());
            m_rpnTokenA = new RPNTokenA();
            parseInfix( expr );
            infixToRpn();
            m_cache[ expr ] = m_rpnTokenA;
        }
        else
        {
            m_rpnTokenA = m_cache[ expr ];
        }
        result = evaluate();
    }catch( ExpError::SyntaxError& e ){
        if ( m_shExceptions )
        Logger::log( LOG_WARNING, _T("RPNParser::calculate() throw SyntaxError: ") + e.toString() +_T(" Expr: ")+  expr);
        WX_CLEAR_ARRAY( *m_rpnTokenA );

        delete m_rpnTokenA;
        m_rpnTokenA = NULL;
    }
    catch( ExpError::VariableException& e){
        wxString variableName = e.toString();
        if ( m_variableNotFound.find(expr) == m_variableNotFound.end() ){
            m_variableNotFound[expr] = variableName;
             wxString msg = _T("RPNParser::calculate() throw VariableException: ");
            msg += _T("(") +expr+ _T(") ")+ variableName;

            if ( m_shExceptions )
            Logger::logf( LOG_WARNING, msg );

        }
        RPNTokenA* a = m_cache[expr];
        m_cache.erase( expr );
        WX_CLEAR_ARRAY( *a );
        //WX_CLEAR_ARRAY( *m_rpnTokenA );
        delete m_rpnTokenA;
        m_rpnTokenA = NULL;

    }catch( ExpError::ExpError& e ){
        wxString variableName = e.toString();
        if ( m_badExpression.find(expr) == m_badExpression.end() ){
            m_badExpression[expr] = variableName;
            wxString msg = _T("RPNParser::calculate() throw Exception: ");
            msg += _T("(") +expr+ _T(") ")+ variableName;

            if ( m_shExceptions )
            Logger::logf( LOG_WARNING, msg );

        }
        RPNTokenA* a = m_cache[expr];
        m_cache.erase( expr );
        WX_CLEAR_ARRAY( *a );
        WX_CLEAR_ARRAY( *m_rpnTokenA );
        delete m_rpnTokenA;
        m_rpnTokenA = NULL;
        //need handling?

    }
    catch(...)
    {
        RPNTokenA* a = m_cache[expr];
        m_cache.erase( expr );
        WX_CLEAR_ARRAY( *a );
        WX_CLEAR_ARRAY( *m_rpnTokenA );
        delete m_rpnTokenA;
        m_rpnTokenA = NULL;
//        LOGP(_T("--------------------------------"))
    }

    //Logger::logf( LOG_WARNING, _T("RPNParser::calculate() m_cache size: %d"), m_cache.size());
    return result;
}

double RPNParser::calculateNoCache( const wxString& expr )
{
    if( !m_rpnTokenA )
    {
        m_rpnTokenA = new RPNTokenA();
    }

    double result = 0.0;

    try{

        parseInfix( expr );
        infixToRpn();
        //result = evaluate();

    }catch( ExpError::ExpError ){
        //need handling?
    }

    if ( m_rpnTokenA )
    {
        WX_CLEAR_ARRAY(*m_rpnTokenA);
        delete m_rpnTokenA ;
        m_rpnTokenA = NULL;
    }

    return result;
}

void RPNParser::clearCache()
{
    RPNCache::iterator it;

    for( it = m_cache.begin(); it != m_cache.end(); ++it )
    {
        WX_CLEAR_ARRAY( *(it->second) );
    }

    WX_CLEAR_HASH_MAP( RPNCache, m_cache );
}

//get variables from an expression
StringHashSet RPNParser::getVariables( const wxString& expr )
{
    try{
        parseInfix( expr );
    }catch( ExpError::SyntaxError& e ){

        if ( m_shExceptions )
        Logger::log( LOG_WARNING, _T("RPNParser::calculate() throw SyntaxError: ") + e.toString() );
    }

    ExpressionBase* eBase = NULL;
    StringHashSet   variables;

    for( uint32_t i = 0; i < m_infixTokenA.GetCount(); i++ )
    {
        eBase = m_infixTokenA.Item( i );

        if( eBase->getType() == ExpressionBase::VARIABLE )
        {
            variables.insert( eBase->toString() );
        }
    }

    WX_CLEAR_ARRAY( m_infixTokenA );

    return variables;
}

// get variables from expression -> sorts by time
Int2ArrayStringHash RPNParser::getPVariables( const wxString& expr )
{
    try
    {
        parseInfix( expr );
    }
    catch ( ExpError::SyntaxError& e )
    {
        if ( m_shExceptions )
        Logger::logf(LOG_MSG, _T("RPNParser::getPVariables() Syntax Error:") + e.toString() );
    }

    ExpressionBase* eBase = NULL;
    Int2ArrayStringHash   variables;

    for( uint32_t i = 0; i < m_infixTokenA.GetCount(); i++ )
    {
        eBase = m_infixTokenA.Item( i );

        if( eBase->getType() == ExpressionBase::VARIABLE )
        {
            wxString varname = eBase->toString();

            wxRegEx regex( _T("([^\\[]*)\\[(-?[0-9]+)\\]"));
            if ( regex.Matches( varname ) )
            {
                wxString symbol = regex.GetMatch( varname, 1 );
                wxString shiftS = regex.GetMatch( varname, 2 );

                long shift; shiftS.ToLong( &shift );
                variables[shift].Add( symbol );
            }
            else
            {
                variables[0].Add( varname );
            }
        }
    }

    WX_CLEAR_ARRAY( m_infixTokenA );

    return variables;
}

wxArrayString RPNParser::getVariablesArray( const wxString& expr )
{
    try{
        parseInfix( expr );
    }catch( ExpError::SyntaxError& e ){

        if ( m_shExceptions )
        Logger::log( LOG_WARNING, _T("RPNParser::calculate() throw SyntaxError: ") + e.toString() );
    }

    ExpressionBase* eBase = NULL;
    wxArrayString   variables;

    for( uint32_t i = 0; i < m_infixTokenA.GetCount(); i++ )
    {
        eBase = m_infixTokenA.Item( i );

        if( eBase->getType() == ExpressionBase::VARIABLE )
        {
            variables.Add( eBase->toString() );
        }
    }

    WX_CLEAR_ARRAY( m_infixTokenA );

    return variables;
}
