#include "ExtendedHeaderParser.h"

#include <wx/regex.h>
#include <wx/filename.h>

#include "parserutil.h"
#include "MyMakeParser.h"

#include "../logger.h"
#include "../vemsprotocol/SIPRTypes.h"

const wxString SECTION_END      = _T("};");
const wxString UNIT_UINT8_T     = _T("uint8_t");
const wxString UNIT_UINT16_T    = _T("uint16_t");
const wxString UNIT_UINT32_T    = _T("uint32_t");

enum State {
    FINISHED,
    INSIDE_SECTION,
    OUTSIED_SECTION
};

/// sections (eg. "config_t" => { page:5, prefix:false }, "config_ext_t" => { page: 9, prefix:true } etc..)
ExtendedHeaderParser::ExtendedHeaderParser( SymbolMgr* symbolMgr, wxVariantHashHash& sections, bool parseMyMake ) : m_symbolMgr( symbolMgr )
{
    m_hastoParseMyMake = parseMyMake;
    m_mymakeParsed = false;

    m_offset = 0;
    m_state  = -1;

    m_defines.clear();
    m_ifdeflevel = 0;

    for ( size_t i = 0; i < 255; ++i )
        m_validDef[i] = true;

    // create config section struct names
    for ( wxVariantHashHash::iterator it = sections.begin(); it != sections.end(); ++it )
    {
        wxString str = it->first;
        m_sectNames.Add( str );

        str = _T("struct ") + str + _T(" {");
        m_structNames.Add( str );

        long page = (it->second)[_T("page")].GetLong();
        m_sectPages.Add( page );

        long prefix = ( (it->second)[_T("prefix")].GetBool() == true ) ? 1 : 0;
        m_sectPrefix.Add( prefix );

        long address = ( it->second)[_T("address")].GetLong();
        m_sectAddresses.Add( address );
    }
}

void ExtendedHeaderParser::parseLine()
{
    if ( ! m_mymakeParsed )
    {
        parseMyMake();
    }

    int mult = 1;

    if ( m_state >= 0 )
    {
        m_line.Trim( false );   //removes leading space

        // end of section
        if ( m_line.StartsWith( SECTION_END ) )
        {
            m_state = -1;
        }

        // begins with comment
        if ( m_line.StartsWith( ParserUtil::COMMENT ) )
        {
            return;
        }

        // strip comment from the end of the line
        ParserUtil::truncateComment( m_line );


        /// check ifdef/ifndefs
        if ( m_line.StartsWith( _T("#ifdef") ) )
        {
            ++m_ifdeflevel;

            wxRegEx reg( _T("\\s*#ifdef\\s+([a-zA-Z0-9_]+)"), wxRE_ADVANCED );

            if ( reg.Matches( m_line ) )
            {
                //LOGP( m_line )
                wxString defName = reg.GetMatch( m_line, 1 );

                String2IntHash::iterator it = m_defines.find( defName );

                wxVariantHash::iterator it2 = m_globalHdefs.find( defName );

                if ( it == m_defines.end() && it2 == m_globalHdefs.end() )
                {
                    //LOGP( _T("NOT DEFINED->false"))
                    m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] && false;
                }
                else
                {
                    m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] && true;
                    //LOGP( _T("DEFINED->true"))
                }
                return;
            }
        }

        if ( m_line.StartsWith( _T("#ifndef") ) )
        {
            ++m_ifdeflevel;
            wxRegEx reg( _T("\\s*#ifndef\\s+([a-zA-Z0-9_]+)"), wxRE_ADVANCED );
            if ( reg.Matches( m_line ) )
            {
                // LOGP( m_line )

                wxString defName = reg.GetMatch( m_line, 1 );

                String2IntHash::iterator it = m_defines.find( defName );

                wxVariantHash::iterator it2 = m_globalHdefs.find( defName );

                if ( it == m_defines.end() && it2 == m_globalHdefs.end() )
                {
                    // LOGP( _T("NOT DEFINED->true"))
                    m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] &&true;
                }
                else
                {
                    // LOGP( _T("DEFINED->false"))
                    m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] &&false;
                }
                return;
            }
        }
        if ( m_line.StartsWith( _T("#endif") ) )
        {
            --m_ifdeflevel;
            return;
        }


        /// check if this is a valid level
        if ( m_validDef[m_ifdeflevel] == false )
            return;


        wxString unit, symbolName;

        ParserUtil::getUnitAndValue( m_line, unit, symbolName );

        if ( unit.CompareTo( UNIT_UINT8_T ) == 0 )
        {
            mult = 1;
        }
        else if ( unit.CompareTo( UNIT_UINT16_T ) == 0 )
        {
            mult = 2;
        }
        else if ( unit.CompareTo( UNIT_UINT32_T ) == 0 )
        {
            mult = 4;
        }
        else
        {
            return;
        }

        m_page = m_sectPages[m_state];

        wxString arrayName, sizeStr;

        long size = 1;

        if( ParserUtil::getArrayAndSize( symbolName, arrayName, sizeStr ) == 0 )
        {
            //array symbol
            sizeStr.ToLong( &size );
            size *= mult;

            createSymbol( arrayName, size );

        }
        else
        {
            size *= mult;
            createSymbol(symbolName, size);
        }

        return;
    }
    // outside a section
    else
    {
        m_line.Trim( false );   //removes leading space

        // removes trailing spaces
        m_line.Trim();

        // check
        for ( size_t i = 0; i < m_structNames.GetCount(); ++i )
        {
            if ( m_line.StartsWith( m_structNames[i] ) )
            {
                m_state = i;
                m_offset = m_sectAddresses[i];
                return;
            }
        }
        if ( m_line.StartsWith(_T("#error") ) && m_validDef[m_ifdeflevel] )
        {
            int pos = m_line.Find(_T("#error") ) + 7;
            wxString errStr = m_line.Mid( pos );
            LOGP(_T("PARSER ERROR: ") + errStr + wxString::Format(_T(" in global.h (line %d)"), m_lineNumber ) );
        }
        else if ( m_line.StartsWith(_T("#else") ) )
        {
            m_validDef[m_ifdeflevel] = !m_validDef[m_ifdeflevel];
            return;
        }
        else if ( m_line.EndsWith( _T("#endif")))
        {
            --m_ifdeflevel;
            return;
        }
        else if ( m_line.StartsWith( _T("#if") ) )
        {
            if ( m_line.StartsWith(_T("#ifdef") ) )
            {
                ++m_ifdeflevel;
                wxRegEx reg2( _T("\\s*#ifdef\\s+([a-zA-Z0-9_]+)"), wxRE_ADVANCED );
                if ( reg2.Matches( m_line ) )
                {
                    wxString defName = reg2.GetMatch( m_line, 1 );
                    String2IntHash::iterator it = m_defines.find( defName );
                    wxVariantHash::iterator it2 = m_globalHdefs.find( defName );
                    if ( it == m_defines.end() && it2 == m_globalHdefs.end() )
                    {
                        m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] &&false;
                    }
                    else
                    {
                        m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] &&true;
                    }

                    //LOGP( m_line + wxString::Format(_T(" is %d"), m_validDef[m_ifdeflevel] ) )

                    return;
                }
            }
            else if ( m_line.StartsWith(_T("#ifndef") ) )
            {
                ++m_ifdeflevel;
                wxRegEx reg2( _T("\\s*#ifndef\\s+([a-zA-Z0-9_]+)"), wxRE_ADVANCED );
                if ( reg2.Matches( m_line ) )
                {
                    //LOGP( m_line )
                    wxString defName = reg2.GetMatch( m_line, 1 );
                    String2IntHash::iterator it = m_defines.find( defName );
                    wxVariantHash::iterator it2 = m_globalHdefs.find( defName );
                    if ( it == m_defines.end() && it2 == m_globalHdefs.end() )
                    {
                        m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] &&true;
                    }
                    else
                    {
                        m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] &&false;
                    }

                    //LOGP( m_line + wxString::Format(_T(" is %d"), m_validDef[m_ifdeflevel] ) )

                    return;
                }
            }
            else
            {
                // #if(...)
                // only format #if[spc*]([spc*]expression[spc*])
                // expression::=VARNAME [[>|<|== | != | >= | <=] VALUE]
                // where VALUE is a valid double or integer value
                ++m_ifdeflevel;

                wxRegEx reg2(_T("#if\\s*\\(([^\\(]*)\\)"), wxRE_ADVANCED );

                enum operT { EQ, NE, LT, GT, LTE, GTE, NONE } oper;
                wxString operStr;
                if ( reg2.Matches( m_line ) )
                {
                    wxString expression = reg2.GetMatch( m_line, 1 );
                    if ( expression.Find(_T(">=") ) != wxNOT_FOUND ) { oper = GTE; operStr = _T(">="); }
                    else if ( expression.Find(_T("<=") ) != wxNOT_FOUND ) { oper = LTE; operStr = _T("<="); }
                    else if ( expression.Find(_T("!=") ) != wxNOT_FOUND ) { oper = NE; operStr = _T("!="); }
                    else if ( expression.Find(_T("==") ) != wxNOT_FOUND ) { oper = EQ; operStr = _T("=="); }
                    else if ( expression.Find(_T(">") ) != wxNOT_FOUND ) { oper = GT; operStr = _T(">"); }
                    else if ( expression.Find(_T("<") ) != wxNOT_FOUND ) { oper = LT; operStr = _T("<"); }
                    else { oper = NONE; }

                    /// oper == NONE ( simple define )
                    if ( oper == NONE )
                    {
                        String2IntHash::iterator it = m_defines.find( expression );
                        wxVariantHash::iterator it2 = m_globalHdefs.find( expression );
                        if ( it == m_defines.end() && it2 == m_globalHdefs.end() )
                            m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] &&false;
                        else
                            m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] &&true;
                        return;
                    }
                    wxString VarName, Value;
                    /// oper == NE || EQ || LTE || GTE
                    if ( oper == NE || oper == EQ )
                    {
                        int pos = expression.Find( operStr );
                        VarName = expression.Left( pos ); VarName.Trim(); VarName.Trim( false );
                        Value = expression.Mid( pos + 2 ); Value.Trim(); Value.Trim( false );


                    }
                    if ( oper == LT || oper == GT )
                    {
                        int pos = expression.Find( operStr );
                        VarName = expression.Left( pos ); VarName.Trim(); VarName.Trim( false );
                        Value = expression.Mid( pos + 1 ); Value.Trim(); Value.Trim( false );
                    }
                    bool expr = evaluateExpression( VarName, Value, operStr );

                    m_validDef[m_ifdeflevel] = m_validDef[m_ifdeflevel-1] && expr;
                    //LOGP( _T("Value of ") +expression + wxString::Format(_T("%d"), expr ) )
                }
                else LOGP( _T("Passed by if-clause ") + m_line + wxString::Format( _T(" in global.h (line %d)"), m_lineNumber ) )
                }
        }
        else if ( m_line.StartsWith( _T("#define") ) )
        {
            if ( m_validDef[m_ifdeflevel] == false )
                return;

            wxRegEx re(_T("^#define\\s+([a-zA-Z0-9_]+)\\s*\\(?([\\-a-zA-Z0-9_]*)\\)?"), wxRE_ADVANCED );
            if ( re.Matches( m_line ) )
            {
                wxString defname = re.GetMatch( m_line, 1);
                wxString value = re.GetMatch( m_line, 2 );

                wxVariantHash::iterator it = m_globalHdefs.find( defname );
                String2IntHash::iterator it2 = m_defines.find( defname );
                wxVariant v;
                if ( it == m_globalHdefs.end() && it2 == m_defines.end() )
                {
                    // check: number(double/int/string)
                    bool hasDecPoint = false;
                    bool isNumber = false;
                    for ( size_t i = 0; i < value.Length(); ++i )
                    {
                        if ( i == 0 )
                        {
                            if ( value[i] == '-' || value[i] == '+' || (value[i] >= '0' && value[i] <= '9') )
                            {
                                isNumber = true;
                                continue;
                            }
                        }
                        if ( isNumber && value[i] == '.' )
                        {
                            if ( hasDecPoint )
                            {
                                isNumber = false;
                                continue;
                            }
                            hasDecPoint = true;
                            continue;
                        }
                        if ( isNumber && ( value[i] < '0' || value[i] >'9' ) )
                        {
                            LOGP( _T("Passed by define ")+ m_line + wxString::Format( _T(" in global.h (line %d)"), m_lineNumber ) );
                        }
                    }

                    if ( !isNumber ) v = value;
                    if ( isNumber )
                    {
                        if ( hasDecPoint )
                        {
                            double d; value.ToDouble(&d); v = d;
                        }
                        else
                        {
                            long l; value.ToLong(&l); v = l;
                        }
                    }

                    m_globalHdefs[defname] = v;
                    // LOGP( defname )
                }
                else
                {
                    LOGP( _T("WARNING: redefinition of ") + defname + wxString::Format( _T(" in global.h (line %d)"), m_lineNumber ) )
                    return;
                }
            }
        }
    }
}

void ExtendedHeaderParser::addSymbol( const wxString& name )
{
    Symbol* symbol = Symbol::createScalar( name, m_page, m_offset, SIPR_READ_MEMORY_32 );

    // m_symbols[ name ] = symbol;
    m_symbolMgr->addSymbol( symbol );
    ++m_offset;
}

void ExtendedHeaderParser::createSymbol(const wxString& name, size_t size)
{
    wxString prefixed = addPrefix( name );

    if(size == 1)
    {
        addSymbol( prefixed );
        return;
    }

    // check wheter 2 or more-bytes sized data starts on even offset
    if ( size >= 2 )
    {
        if ( m_offset % 2 != 0 )
            LOGP(_T("-- WARNING: symbol ") + name + wxString::Format( _T(" starts on odd offset address = 0x%x (%d), symbol size = %d"), m_offset, m_offset, size ) )
    }

    for ( size_t i = 0; i < size; ++i )
    {

        wxString key;
        key.Printf( _T("[%X]"), i);
        key = prefixed + key;

        addSymbol( key );
    }
}

wxString ExtendedHeaderParser::addPrefix( const wxString& name )
{
    if ( m_state < 0 || m_state >= (int)m_sectNames.GetCount() )
        return name;

    // add prefix if neccessary
    if ( m_sectPrefix[m_state] == 0 )
    {
        return name;
    }
    else
    {
        return m_sectNames[m_state] + _T("_") + name;
    }
}

void ExtendedHeaderParser::parseMyMake()
{
    m_mymakeParsed = true;

    if ( m_hastoParseMyMake == false )
    {
        return;
    }

    wxString fileName = m_fileName;
    wxFileName fn( fileName );

    wxString mymakeName = fn.GetPath() + wxFileName::GetPathSeparator() + _T("my_make");

    if ( ! wxFileExists( mymakeName ) )
    {
        mymakeName += _T(".16x14");

        if ( ! wxFileExists( mymakeName ) )
        {
            LOGP( _T("my_make(.16x14) does not exist, no define included! ") );
            return;
        }
    }

    LOGP( _T("Parsing my_make") )

    MyMakeParser parser;

    parser.parse( mymakeName );

    m_defines = parser.getDefines();

//    for ( String2IntHash::iterator it = m_defines.begin(); it != m_defines.end(); ++it )
//    {
//        LOGP( _T("#defined '") + it->first + _T("'"))
//    }

    LOGP( _T("End of parsing of my_make") )
}

bool ExtendedHeaderParser::evaluateExpression( wxString& operand1, wxString& operand2, wxString& op )
{
    // operand2 is a double number
    if ( operand2.Find(_T(".") ) )
    {
        double val = 0; operand2.ToDouble(&val);

        double defval = -9E9;

        // search for operand1 in mymake
        String2IntHash::iterator it = m_defines.find( operand1 );
        if ( it != m_defines.end() )
        {
             defval = (double)( it->second );
        }
        else
        {
            wxVariantHash::iterator it2 = m_globalHdefs.find( operand1 );
            if ( it2 != m_globalHdefs.end() )
                defval = it2->second.GetDouble();
        }

        if ( op == _T("==") ) return ( val == defval );
        if ( op == _T("!=") ) return ( val != defval );
        if ( op == _T(">=") ) return ( defval >= val );
        if ( op == _T("<=") ) return ( defval <= val );
        if ( op == _T(">") ) return ( defval > val );
        if ( op == _T("<") ) return ( defval < val );
    }
    // else it is an integer
    else
    {
        long val = 0; operand2.ToLong(&val);

        long defval = -9000000;

        // search for operand1 in mymake
        String2IntHash::iterator it = m_defines.find( operand1 );
        if ( it != m_defines.end() )
        {
             defval = (long)( it->second );
        }
        else
        {
            wxVariantHash::iterator it2 = m_globalHdefs.find( operand1 );
            if ( it2 != m_globalHdefs.end() )
                defval = it2->second.GetLong();
        }

        if ( op == _T("==") ) return ( val == defval );
        if ( op == _T("!=") ) return ( val != defval );
        if ( op == _T(">=") ) return ( defval >= val );
        if ( op == _T("<=") ) return ( defval <= val );
        if ( op == _T(">") ) return ( defval > val );
        if ( op == _T("<") ) return ( defval < val );
    }
    //
    return true;
}
