#ifndef FIRMWARECONFIGDATA_H
#define FIRMWARECONFIGDATA_H
#include <wx/wx.h>

#include "../common/types/String2IntHash.h"
#include "../common/types/String2CharRefHash.h"
#include "../common/types/BoolHash.h"
#include "../common/types/StringHash.h"
#include "../common/types/StringSet.h"
#include "../common/VemsMatrix.h"
#include "../expression/VariableData.h"
#include "../Symbol.h"

class FirmwareConfigDifferences;
class SymbolTable;


class FirmwareConfigData
{
    public:
        size_t m_configLineCount;
        FirmwareConfigData( const wxString& name = wxEmptyString ) { m_name = name; }
        virtual ~FirmwareConfigData();

        void addScalar( const wxString& name, int value, int lineNumber = -1 );
        void addTableLine( const wxString& name, size_t rowIndex, const wxArrayString& line, int lineNumber = -1 );
        void addTableLineMSQ( const wxString& name, size_t rowIndex, const wxArrayString& line, int lineNumber = -1 );
        void remove(const FirmwareConfigData& cfgData);
        void removeScalar( const wxString& name);
        void removeTable( const wxString& name );

        void filter(const FirmwareConfigData& cfgData);
        void filterScalar( const wxString& name);

        void setFWVersion( const wxString& version ) { m_strFWVersion = version; }

        bool                        getScalar(const wxString& name, int& value);
        const String2IntHash&       getScalarValuesH() const { return m_scalarValueH; }
        const wxArrayString&        getScalarOrderA() const { return m_scalarOrderA; }
        const wxArrayString&        getTableOrderA() const { return m_tableOrderA; }
        const StringSet&            getFilteredScalarH() const { return m_filteredScalarH; }


        void    addTable(const wxString& name, VemsMatrix<uint8_t>* table );
        void    addTable(const wxString& name, const uint8_t* data, size_t rows, size_t cols);
        wxString getName() const { return m_name; }
        bool is12x12();
        bool isHighBoost() const;
        const BoolHash& getDetectedParams();
        wxString getFWVersion() const { return m_strFWVersion; }

        FirmwareConfigDifferences* compare( const FirmwareConfigData& fcData );
        FirmwareConfigDifferences* compare( const SymbolTable& sTable );

        void clear();

        //debug functions
        void        printScalarsToFile( const wxString& fileName );
        void        printTablesToFile( const wxString& fileName );
        wxString    tableDataToString();
        wxString    scalarDataToString();
        wxString    tableDataToHtml();
        wxString    scalarDataToHtml();

        VemsMatrix<uint8_t>*    getTable( const wxString& name ) const;
        void                    fillVariableData(VariableDataDH& vd);

        void        tableDataToArrayString( wxArrayString& data );

        StringHash          header;

        ///
        void setConvertedTable(const Symbol& symbol, const VemsMatrix<uint8_t>& table);
        void setXaxisName(const wxString& tableName, const wxString& xAxisName);
        void setYaxisName(const wxString& tableName, const wxString& yAxisName);

        const VemsMatrix<uint8_t>* getConvertedTable(const wxString name) const;
        wxString getXaxisName(const wxString& tableName) const;
        wxString getYaxisName(const wxString& tableName) const;
        const VemsMatrixByteHash&  getConvertedTableH(){ return m_ConvertedTableH; };
        const SymbolTable& getConvertedTableSymbols() { return m_tableSymbolH; };

        bool isPartialCfg();

    protected:
    private:
        String2CharRefHash  m_tableDataH;       //[tableName] = (char*)tableData
        String2IntHash      m_tableSizeH;       //[tableName] = (int)tableSize
        String2IntHash      m_tableColCountH;   //[tableName] = (int)tableRowLength
        String2IntHash      m_tableStartLineNo; // start line number of table in the file

        String2IntHash      m_scalarValueH;     //[symbolName] = (int)symbolValue
        String2IntHash      m_scalarLineH;      // line number of scalar in the file
        wxArrayString       m_scalarOrderA;
        wxArrayString       m_tableOrderA;
        wxString            m_name;
        BoolHash            m_boolHash;
        VemsMatrixByteHash  m_tableHash;
        wxString            m_strFWVersion;

        VemsMatrixByteHash  m_ConvertedTableH;
        StringHash        m_XaxisNameH;
        StringHash        m_YaxisNameH;
        SymbolTable       m_tableSymbolH;

        StringSet         m_filteredScalarH;


        void        searchMissingScalars( const SymbolTable& sTable, FirmwareConfigDifferences* fcDiff );
        void        searchMissingTables( const SymbolTable& sTable, FirmwareConfigDifferences* fcDiff );
        void        searchExtraScalars( const SymbolTable& sTable, FirmwareConfigDifferences* fcDiff );
        void        searchExtraTables( const SymbolTable& sTable, FirmwareConfigDifferences* fcDiff );
        wxString    tableDataToString( const wxString& eol, const wxString& tab );
        wxString    scalarDataToString( const wxString& eol, const wxString& tab );

        void compareScalars(FirmwareConfigDifferences* fcDiff, const FirmwareConfigData& fcData);
        void compareTables(FirmwareConfigDifferences* fcDiff, const FirmwareConfigData& fcData);
        void compareTables( const wxString& configName,
                            const wxString& tableName,
                            VemsMatrix<uint8_t>* matrix1,
                            VemsMatrix<uint8_t>* matrix2,
                            wxArrayString& differentTables );
};



#endif // FIRMWARECONFIGDATA_H
