#include "HeaderParser.h"

#include <wx/regex.h>

#include "../logger.h"
#include "../Utils.h"
#include "../fileparser/MyMakeParser.h"

HeaderParser::HeaderParser( const wxString& mymake_filename )
{
    m_insideComment = false;

    for ( size_t i = 0; i < MAX_IFDEF_LEVEL; ++i )
        m_ifdeflevels[i] = true;

    m_currentifdeflevel = 0;

    m_defines = new VariableList;

    // parsing mymake
    if ( mymake_filename != _T("") )
    {
        parseMyMake( mymake_filename );
    }

    m_exprParser.setVariableList( m_defines );

    m_in_function = false;
    m_function_opened_brackets = 0;

    m_parsingStruct = false;

    m_currentStruct = NULL;

}

void HeaderParser::parseMyMake( const wxString& filename )
{
    MyMakeParser parser;
    parser.parse( filename );

    String2IntHash defines = parser.getDefines();

    String2IntHash::const_iterator it = defines.begin();

    for ( ; it != defines.end(); ++it )
    {
        m_defines->add( it->first, (double)it->second );
    }
    // Logger::logf(LOG_MSG, _T("...mymake parsed"));
}

void HeaderParser::parseLine()
{
    // check & strip commented lines
    if ( m_insideComment )
    {
        if ( checkCommentEnd() == false )
            return;

        m_insideComment = false;
    }
    else
    {
        checkCommentBegin();
    }

    stripComments();

    // trim both ends
    trim();

    // if line is empty continue
    if ( m_line.IsEmpty() )
        return;

    if ( m_line.StartsWith( _T("#endif") ) )
    {
        m_currentifdeflevel--;
        return;
    }

    if ( m_line.StartsWith( _T("#else") ) )
    {
        m_ifdeflevels[m_currentifdeflevel] = !m_ifdeflevels[m_currentifdeflevel];
        return;
    }

    // inside a function?
    if ( m_in_function )
    {
        m_function_opened_brackets += countChars( m_line, _T("{"));
        m_function_opened_brackets -= countChars( m_line, _T("}"));

        if ( m_function_opened_brackets == 0 )
        {
            m_in_function = false;
        }
        return;
    }

    if ( m_line.StartsWith(_T("#ifdef") ) )
    {
        doIfdefCheck( true );
        return;
    }
    else if ( m_line.StartsWith(_T("#ifndef") ) )
    {
        doIfdefCheck( false );
        return;
    }

    if ( m_line.StartsWith(_T("#if")))
    {
        checkIfCondition();
        return;
    }

    if ( m_ifdeflevels[m_currentifdeflevel] == false )
        return;

    if ( m_line.StartsWith(_T("#define") ) )
    {
        createDefine();
        return;
    }

    if ( m_line.StartsWith(_T("#undef") ) )
    {
        removeDefine();
        return;
    }

    if ( m_line.StartsWith(_T("#warning") ) )
    {
        showErrWarn( false ); // false->warning
        return;
    }

    if ( m_line.StartsWith(_T("#error") ) )
    {
        showErrWarn( true );
        Abort();
        return;
    }

    /// ok, it should be a variable/type definition or a function definition (block)

    // it is a function
    if ( m_line.StartsWith(_T("static") ) )
    {
        m_in_function = true;
        if ( m_line.Find(_T("{")) != wxNOT_FOUND )
        {
            int no1 = countChars( m_line, _T("{"));
            int no2 = countChars( m_line, _T("}"));
            if ( no1 == no2 )
            {
                m_in_function = false;
                m_function_opened_brackets = 0;
            }
            else
            {
                m_function_opened_brackets = no1-no2;
            }
        }
        // Logger::logf(LOG_MSG, _T("..Function skipped in line %d"), m_lineNumber);

        return;
    }

    m_line.Replace( _T("EXTERN"), _T(""), true );
    m_line.Replace( _T("extern"), _T(""), true );
    m_line.Replace( _T("volatile"), _T(""), true );

    stripMultispaces( m_line );

    if ( m_parsingStruct )
    {
        parseStructLine();
        return;
    }

    wxRegEx re( _T("struct ([0-9a-zA-Z_]*?) {"), wxRE_ADVANCED );
    if ( re.Matches( m_line ) )
    {
        createNewStruct( re.GetMatch( m_line, 1 ) );
        return;
    }
    re.Compile( _T("struct ([0-9a-zA-Z_]*?) ([0-9a-zA-Z_\\[\\]]*).*;"), wxRE_ADVANCED );
    if ( re.Matches( m_line ) )
    {
        StructDescriptor descr;

        wxString structtype = re.GetMatch( m_line, 1 );
        wxString varname = re.GetMatch( m_line, 2 );

        //LOGP( structtype + _T(" ") + varname + wxString::Format( _T(" - line %d"), m_lineNumber ))

        for ( size_t i = 0; i < m_parsedStructTypes.GetCount(); ++i )
        {
            if ( m_parsedStructTypes[i].name == structtype )
            {
                StructureVar svar;
                svar.size = 1;

                if ( varname.Find( _T("[") )  != wxNOT_FOUND )
                {
                    wxString tmp = varname;
                    varname = tmp.BeforeFirst( '[' );
                    wxString len = tmp.AfterFirst('[').BeforeFirst(']');
                    long arrlen = 0;
                    if ( len.ToLong( &arrlen ) == false )
                    {
                        try {
                            arrlen = m_exprParser.parse( len, false, true );
                        }
                        catch (...)
                        {
                            LOGP( wxString::Format( _T("Error in line %d: ") , m_lineNumber ) + len  );
                        }
                    }

                    svar.size = arrlen;
                }

                svar.structType = m_parsedStructTypes[i];
                m_parsedStructVariables[varname] = svar;
                //Logger::logf(LOG_MSG, varname + _T(" is definied as ") + structtype + ((svar.size>1)?wxString::Format( _T("[%d]"),svar.size):_T("")) );
                return;
            }
        }
    }
}

void HeaderParser::stripComments()
{
    int pos = m_line.Find(_T("//"));
    if ( pos != wxNOT_FOUND )
    {
        m_line.Truncate( pos );
    }
}

void HeaderParser::trim( wxString& line )
{
    line.Trim( true );
    line.Trim( false );
}

void HeaderParser::trim()
{
    m_line.Trim( true );
    m_line.Trim( false );
}

bool HeaderParser::checkCommentEnd()
{
    if ( m_insideComment )
    {
        int pos = m_line.Find( _T("*/") );
        if ( pos != wxNOT_FOUND )
        {
            if ( pos + 2 >= (int)m_line.Length() )
                m_line = _T("");
            else
                m_line = m_line.Mid( pos + 2 );

            return true;
        }
        else
            return false;
    }
    return true;
}

void HeaderParser::checkCommentBegin()
{
    int pos = m_line.Find( _T("/*") );

    if ( pos != wxNOT_FOUND )
    {
        int pos1 = m_line.Find(_T("*/"));

        if ( pos1 == wxNOT_FOUND )
        {
            if ( pos == 0 )
                m_line = _T("");
            else
                m_line.Truncate( pos - 1 );

            m_insideComment = true;

            return;
        }
        wxString tmp = m_line.Left( pos );
        if ( pos1 + 2 < (int)m_line.Length() )
            tmp += m_line.Mid( pos1 + 2 );

        m_line = tmp;

        m_insideComment = false;
    }
}

void HeaderParser::stripMultispaces( wxString& line )
{
    line.Replace( _T("\t"), _T(" "), true );
    line.Replace( _T("  "), _T(" "), true );
}

void HeaderParser::doIfdefCheck( bool ifdef )
{
    wxArrayString arr;
    wxString tmpStr = m_line;

    stripMultispaces( tmpStr );

    Utils::tokenizeToArray( tmpStr, _T(" "), arr );

    wxString& define_name = arr[1];

    bool define_exist = m_defines->exist( define_name );

    if ( define_exist )
    {
        if ( ifdef )
        {
//            Logger::logf(LOG_MSG, define_name + _T(" is defined, level is true"));
            m_currentifdeflevel++;
            m_ifdeflevels[m_currentifdeflevel] = m_ifdeflevels[m_currentifdeflevel-1] && true;
        }
        else
        {
//            Logger::logf(LOG_MSG, define_name + _T(" is defined, level is false"));
            m_currentifdeflevel++;
            m_ifdeflevels[m_currentifdeflevel] = m_ifdeflevels[m_currentifdeflevel-1] && false;
        }
    }
    else
    {
        if ( ifdef )
        {
//            Logger::logf(LOG_MSG, define_name + _T(" is not defined, level is false"));
            m_currentifdeflevel++;
            m_ifdeflevels[m_currentifdeflevel]  = m_ifdeflevels[m_currentifdeflevel-1] && false;
        }
        else
        {
//            Logger::logf(LOG_MSG, define_name + _T(" is not defined, level is true"));
            m_currentifdeflevel++;
            m_ifdeflevels[m_currentifdeflevel]  = m_ifdeflevels[m_currentifdeflevel-1] && true;
        }
    }
}

void HeaderParser::createDefine()
{
    wxArrayString arr;
    wxString tmpStr = m_line;

    stripMultispaces( tmpStr );

    Utils::tokenizeToArray( tmpStr, _T(" "), arr );

    wxString& define_name = arr[1];

    wxString define_expr;
    for ( size_t i = 2; i < arr.GetCount(); ++i )
        define_expr += arr[i] + _T(" ");

    trim( define_expr );

    if ( define_expr == _T("") )
        define_expr = _T("1");

    define_expr.Replace( _T("("),_T("( "), define_expr );

    m_exprParser.reset();

    try {
        double result = m_exprParser.parse( define_expr, false, true );

        m_defines->add( define_name, result );
//        LOGP( define_name + wxString::Format( _T(" is defined as %.1f"),result  ))
    }
    catch ( ExpressionException e )
    {
//        e.setRow( m_lineNumber + 1 );
//        if ( e.errcode != ExpressionException::UNKNOWN_VAR && e.errcode != ExpressionException::UNKNOWN_FUNC )
//        {
//            Logger::logf(LOG_MSG, e.err );
//            Logger::logf(LOG_MSG, e.getExprErr() );
//        }
    }
}

void HeaderParser::removeDefine()
{
    wxArrayString arr;
    wxString tmpStr = m_line;

    stripMultispaces( tmpStr );

    Utils::tokenizeToArray( tmpStr, _T(" "), arr );

    wxString& define_name = arr[1];

    m_defines->del( define_name );
}

void HeaderParser::showErrWarn( bool isError )
{
    wxString tmpStr;

    if ( isError )
        tmpStr = m_line.Mid( 7 );
    else
        tmpStr = m_line.Mid( 9 );

    stripMultispaces( tmpStr );

    tmpStr = ( isError ) ? _T("ERROR: '") + tmpStr : _T("Warning: '") + tmpStr;
    tmpStr += wxString::Format( _T("' in line %d"), m_lineNumber );
    Logger::logf(LOG_MSG, tmpStr );

}

void HeaderParser::checkIfCondition()
{
    wxArrayString arr;
    wxString condition = m_line;
    condition = condition.Mid( 3 );

    stripMultispaces( condition );

    trim( condition );

    // testing '#if defined(A) || !defined(B)' cases
    wxRegEx re(_T("(defined\\s*\\(?\\s*)([0-9a-zA-Z_]*)(\\s*\\)?)?"), wxRE_ADVANCED);

    while ( re.Matches(condition) )
    {
        wxString defname = re.GetMatch( condition, 2 );

        if ( m_defines->exist( defname ) )
        {
            re.ReplaceFirst( &condition,_T("1"));
        }
        else
            re.ReplaceFirst( &condition,_T("0") );
    }
    //

    m_exprParser.reset();

    try {
        double result = m_exprParser.parse( condition, false, true );

        m_currentifdeflevel++;

        if ( result == 0 )
        {
            m_ifdeflevels[m_currentifdeflevel]  = m_ifdeflevels[m_currentifdeflevel-1] && false;
        }
        else
        {
            m_ifdeflevels[m_currentifdeflevel]  = m_ifdeflevels[m_currentifdeflevel-1] && true;
        }
    }
    catch ( ExpressionException e )
    {
//        e.setRow( m_lineNumber + 1 );
//        Logger::logf(LOG_MSG, e.err );
//        Logger::logf(LOG_MSG, e.getExprErr() );
    }
}

inline int HeaderParser::countChars( wxString str_in, wxString str_what )
{
    return str_in.Replace( str_what, str_what, true );
}

void HeaderParser::parseStructLine()
{
    if ( m_line.StartsWith(_T("}") ) )
    {
        m_parsedStructTypes.Add( *m_currentStruct );
        m_parsingStruct = false;

        //Logger::logf(LOG_MSG, _T("struct ") + m_currentStruct->name + _T(" parsed, sizeof(")+m_currentStruct->name + wxString::Format( _T(") = %d"), m_currentStruct->size ) );
        return;
    }
    // check for uint8t, uint16t
    wxArrayString tokens;
    Utils::tokenizeToArray( m_line, _T(" "), tokens );
    if ( tokens.GetCount() > 1 )
    {
        VariableDescriptor var;

        wxString v_ = wxEmptyString;
        wxString structtype = _T("");

        if ( tokens[0] == _T("struct") )
        {
            structtype = tokens[1];

            for ( size_t i = 2; i < tokens.GetCount(); ++i )
                v_ += tokens[i];
        }
        else
        {
            for ( size_t i = 1; i < tokens.GetCount(); ++i )
                v_ += tokens[i];
        }

        trim( v_ );
        v_.RemoveLast( 1 );

        bool isArray = false;

        long arraysize = 0;

        if ( v_.Find(_T("[") ) != wxNOT_FOUND )
        {
            var.name = v_.BeforeFirst( '[' );
            v_ = v_.AfterFirst( '[' );
            v_ = v_.BeforeFirst( ']' );

            isArray = true;

            // if it is not an integer number
            if ( v_.ToLong( &arraysize ) == false )
            {
                // maybe it is an expression
                try
                {
                    arraysize = m_exprParser.parse( v_ );
                }
                catch ( ... )
                {
                    LOGP( _T("Something wrong with array: ") + m_line )
                }

            }
        }
        else
        {
            var.name = v_;
        }

        if ( tokens[0] == _T("uint8_t") || tokens[0] == _T("uint16_t") || tokens[0] == _T("uint32_t") )
        {
            int esize = 1;
            if ( tokens[0] == _T("uint8_t") ) esize = 1;
            else if ( tokens[0] == _T("uint16_t") ) esize = 2;
            else if ( tokens[0] == _T("uint32_t") ) esize = 4;
            var.strType = tokens[0];
            var.isElementary = true;
            var.isArray = isArray;
            var.isStruct = false;
            var.elemsize = esize;
            var.arraySize = arraysize;
            var.structDescriptor = NULL;
        }
        else if ( tokens[0] == _T("struct") )
        {
            var.strType = _T("struct");
            var.isElementary = false;
            var.isArray = isArray;
            var.isStruct = true;

            StructDescriptor* descr = findStructDescr( structtype );
            if ( descr == NULL )
            {
                LOGP( _T("ERROR: ") + var.name + _T(" is unknown!!") );
            }
            else
            {
                var.elemsize = descr->size;
                var.arraySize = arraysize;
                var.structDescriptor = descr;
            }
        }
        else
        {
            LOGP( _T("ERROR: unknown type ") + tokens[0] + _T(" (allowed types are: uint8/16/32_t, struct)"))
        }

        m_currentStruct->addVariable( var );
    }
}

void HeaderParser::createNewStruct( const wxString& structName )
{
    m_currentStruct = new StructDescriptor;

    m_currentStruct->name = structName;

    m_parsingStruct = true;
}

StructDescriptor* HeaderParser::findStructDescr( const wxString& structname )
{
    StructDescriptor* result = NULL;
    for ( size_t i = 0; m_parsedStructTypes.GetCount(); ++i )
    {
        if ( m_parsedStructTypes[i].name == structname )
        {
            result = &( m_parsedStructTypes[i] );
            break;
        }
    }

    return result;
}
