#include "AIMTestFrame.h"

#include <wx/statline.h>

#include "../logger.h"
#include "../AIMParser.h"
#include "../serialconfig.h"
#include "../serialporthandler.h"
#include "../SerialException.h"

/*
Signal      Channel Freq
                    [Hz] Transform Units
RPM             1   10  y=x         RPM         OK
Wheel speed     5   10  y=x/10      Km/h        OK
Oil pressure    9   5   y=x/1000    Bar         OK  miliBar
Oil temp        13  2   y=x/10-100  Deg C       OK
Water temp      17  2   y=x/10-100  Deg C       OK
Fuel pressure   21  5   y=x/1000    Bar         OK
Battery voltage 33  5   y=x/100     Volts       ERR y=x/1000 Volts
Throttle angle  45  10  y=x/10      Deg         OK
Manifold press  69  10  y=x         MBar        OK  miliBar
Air charge temp 97  2   y=x/10-100  Deg C       OK
Exhaust temp    101 2   y=x/10-100  Deg C       OK
Lambda Sensor   105 10  y=x/1000    Lambda      OK
Fuel temp       109 2   y=x/10-100  Deg C       OK
Gear            113 5   y=x 0=rev, 1=neutral, 2=first, 3=second...      OK
Errors          125 2   - ECU-specific error flags
*/

enum AimChannels
{
    AIM_RPM             = 1,
    AIM_WHEEL_SPEED     = 5,
    AIM_OIL_PRESSURE    = 9,
    AIM_OIL_TEMP        = 13,
    AIM_WATER_TEMP      = 17,
    AIM_FUEL_PRESSURE   = 21,
    AIM_BATTERY_VOLTAGE = 33,
    AIM_THROTTLE_ANGLE  = 45,
    AIM_MAINFOLD_PRESS  = 69,
    AIM_AIR_CHARGE_TEMP = 97,
    AIM_EXHAUST_TEMP    = 101,
    AIM_LAMBDA_SENSORS  = 105,
    AIM_FUEL_TEMP       = 109,
    AIM_GEAR            = 113,
    AIM_ERRORS          = 125
};


struct AimChDesc
{
    wxString    label;
    int         min;
    int         max;
    int         id;

    AimChDesc( int id, const wxString& label, int min, int max )
    {
        this->id    = id;
        this->label = label;
        this->min   = min;
        this->max   = max;
    }
};

const wxString  DEFAULT_PORT    = _T("COM5");
const int       DEFAULT_SPEED   = 19200;

WX_DEFINE_ARRAY( AimChDesc*, AimChDescArray );

const int TIMER_ID_AIM      = wxNewId();
const int SET_BUTTON_ID     = wxNewId();
const int START_BUTTON_ID   = wxNewId();
const int STOP_BUTTON_ID    = wxNewId();

BEGIN_EVENT_TABLE( AIMTestFrame, wxFrame )
    EVT_BUTTON( wxID_ANY, AIMTestFrame::OnButton )
    EVT_SCROLL( AIMTestFrame::OnChValueChanged )
    EVT_TIMER( TIMER_ID_AIM, AIMTestFrame::OnTimer )
END_EVENT_TABLE()

AIMTestFrame::AIMTestFrame( wxWindow* parent )
    :   wxFrame( parent, wxID_ANY, _T("AIM Test") ),
        m_timer( this, TIMER_ID_AIM )
{
    m_pAIMParser    = new AIMParser();
    m_interval      = 200;

    SerialConfig* sc = new SerialConfig();
    sc->setPortName( DEFAULT_PORT );
    sc->setBaudRate( DEFAULT_SPEED );
    m_pSerialPortHandler = new SerialPortHandler( sc );

    AimChDescArray channels;

    channels.Add( new AimChDesc( AIM_RPM, _T("RPM (1/min)"), 0, 16000) );
    channels.Add( new AimChDesc( AIM_WHEEL_SPEED, _T("Wheel speed (km/h)"), 0, 1000) );

    channels.Add( new AimChDesc( AIM_OIL_PRESSURE, _T("Oil pressure (bar*1000)"), 0, 10000) );
    channels.Add( new AimChDesc( AIM_OIL_TEMP, _T("Oil temp (C)"), -100, 100) );
    channels.Add( new AimChDesc( AIM_WATER_TEMP, _T("Water temp (C)"), -100, 100) );
    channels.Add( new AimChDesc( AIM_FUEL_PRESSURE, _T("Fuel pressure (bar*1000)"), 0, 10000) );
    channels.Add( new AimChDesc( AIM_BATTERY_VOLTAGE, _T("Battery voltage (mV)"), 0, 18000) );
    channels.Add( new AimChDesc( AIM_THROTTLE_ANGLE, _T("Throttle angle (%*10)"), 0, 1000) );
    channels.Add( new AimChDesc( AIM_MAINFOLD_PRESS, _T("Mainfold press (mBar)"), 0, 10000) );
    channels.Add( new AimChDesc( AIM_AIR_CHARGE_TEMP, _T("Air charge temp (C)"), -100, 200) );
    channels.Add( new AimChDesc( AIM_EXHAUST_TEMP, _T("Exhaust temp (C)"), -100, 500) );
    channels.Add( new AimChDesc( AIM_LAMBDA_SENSORS, _T("Lambda sensors (l*1000)"), 0, 10000) );
    channels.Add( new AimChDesc( AIM_FUEL_TEMP, _T("Fuel temp (C)"), -100, 200) );
    channels.Add( new AimChDesc( AIM_GEAR, _T("Gear"), 0, 10) );

    wxPanel* mainPanel      = new wxPanel( this );
    wxBoxSizer* mainSizer   = new wxBoxSizer( wxVERTICAL );

    for ( size_t i = 0; i < channels.GetCount(); ++i )
    {
        AimChDesc* desc = channels[i];

		wxString str = wxString( desc->label + wxString::Format( _T(" (%i)"),desc->id));
        wxStaticText* label = new wxStaticText( mainPanel, wxID_ANY, str );
        wxSlider* slider    = new wxSlider( mainPanel, desc->id, 0, desc->min, desc->max,
                                            wxDefaultPosition, wxDefaultSize, wxSL_LABELS );

        wxBoxSizer* lineSizer = new wxBoxSizer( wxHORIZONTAL );
        lineSizer->Add( label, 1, wxEXPAND | wxALL, 5 );
        lineSizer->Add( slider, 5, wxEXPAND | wxALL, 5 );

        mainSizer->Add( lineSizer, 0, wxEXPAND, 0 );

        m_pAIMParser->getChannelDataHash()[ desc->id ] = 0;

    }

    mainSizer->AddStretchSpacer();
    mainSizer->Add( new wxStaticLine( mainPanel ), 0, wxEXPAND | wxLEFT | wxRIGHT, 5 );

    m_pPortTextCtrl = new wxTextCtrl( mainPanel, wxID_ANY, DEFAULT_PORT );
    m_pBaudTextCtrl = new wxTextCtrl( mainPanel, wxID_ANY, wxString::Format( _T("%d"), DEFAULT_SPEED ) );

    wxButton* setButton = new wxButton( mainPanel, SET_BUTTON_ID, _T("Set") );
    wxButton* startButton = new wxButton( mainPanel, START_BUTTON_ID, _T("Start") );
    wxButton* stopButton = new wxButton( mainPanel, STOP_BUTTON_ID, _T("Stop") );
    wxBoxSizer* lineSizer = new wxBoxSizer( wxHORIZONTAL );

    lineSizer->Add( new wxStaticText( mainPanel, wxID_ANY, _T("Port:") ), 0, wxEXPAND | wxALL, 5 );
    lineSizer->Add( m_pPortTextCtrl, 0, wxEXPAND | wxALL, 5 );
    lineSizer->Add( new wxStaticText( mainPanel, wxID_ANY, _T("Speed:") ), 0, wxEXPAND | wxALL, 5 );
    lineSizer->Add( m_pBaudTextCtrl, 0, wxEXPAND | wxALL, 5 );
    lineSizer->Add( setButton, 0, wxEXPAND | wxALL, 5 );
    lineSizer->Add( startButton, 0, wxEXPAND | wxALL, 5 );
    lineSizer->Add( stopButton, 0, wxEXPAND | wxALL, 5 );
    mainSizer->Add( lineSizer, 0, wxEXPAND, 0 );
    mainPanel->SetSizer( mainSizer );

    mainPanel->SetInitialSize();
    SetInitialSize();
}

AIMTestFrame::~AIMTestFrame()
{
   m_timer.Stop();
   delete m_pAIMParser;
   delete m_pSerialPortHandler;
}

void AIMTestFrame::OnTimer( wxTimerEvent& event )
{
    //Logger::logf( LOG_MSG, _T("AIMTestFrame::OnTimer() : interval = %d"), m_interval );

    Int2IntHash& channelDataH = m_pAIMParser->getChannelDataHash();

    Int2IntHash::iterator it = channelDataH.begin();
    for ( ; it != channelDataH.end(); ++it )
    {
        m_pAIMParser->sendChannel( it->first );

        int txCount = 0;
        m_pAIMParser->tx( m_txBuffer, txCount );
        if( txCount > 0)
        {
            try
            {
                m_pSerialPortHandler->SendString( m_txBuffer, txCount );
                wxThread::Sleep(10);
            }
            catch ( SerialException se )
            {
                Logger::logf( LOG_MSG, _T("AIMTestFrame::OnTimer() - Serial Exception!!!") );
            }
        }
    }
}

void AIMTestFrame::OnChValueChanged( wxScrollEvent& event )
{
    int id      = event.GetId();
    int value   = 0;

    switch( id )
    {
        case AIM_RPM :
        case AIM_MAINFOLD_PRESS:
        case AIM_WHEEL_SPEED :
            value = event.GetPosition();
            break;
        case AIM_THROTTLE_ANGLE :
            value = event.GetPosition();
            break;
        case AIM_BATTERY_VOLTAGE :
            value = event.GetPosition();
            break;
        case AIM_OIL_PRESSURE :
        case AIM_FUEL_PRESSURE :
        case AIM_LAMBDA_SENSORS :
            value = event.GetPosition();
            break;
        case AIM_OIL_TEMP :
        case AIM_WATER_TEMP :
        case AIM_AIR_CHARGE_TEMP :
        case AIM_EXHAUST_TEMP :
        case AIM_FUEL_TEMP :
            //value = event.GetPosition() / 10 - 100;
            value = ( event.GetPosition() + 100 ) * 10;
            break;
         case AIM_GEAR:
            value = ( event.GetPosition() + 1 );
            break;
    }

    m_pAIMParser->getChannelDataHash()[ id ] = value;
}

void AIMTestFrame::OnButton( wxCommandEvent& event )
{
    int id = event.GetId();

    if ( id == START_BUTTON_ID )
    {
        if ( !m_timer.IsRunning() ) { m_timer.Start( m_interval ); }
    }
    else if ( id == STOP_BUTTON_ID ) { m_timer.Stop(); }
    else if ( id == SET_BUTTON_ID )
    {
        m_timer.Stop();

        SerialConfig* sc = new SerialConfig();
        sc->setPortName( m_pPortTextCtrl->GetValue() );

        long speed = DEFAULT_SPEED;
        m_pPortTextCtrl->GetValue().ToLong( &speed );
        sc->setBaudRate( (int)speed );

        delete m_pSerialPortHandler;
        m_pSerialPortHandler = new SerialPortHandler( sc );

        m_timer.Start();
    }
}
