#include "DeviceControllerProgressBarManager.h"
#include "../logger.h"
#include "../events/VemsDeviceControllerEvent.h"
#include "../events/CommunicationEvent.h"
#include "../bcWidgets/bcProgressDialog.h"

DeviceControllerProgressBarManager::DeviceControllerProgressBarManager( wxWindow* parent )

    : m_timer ( this, wxID_OK )
{
    m_parent = parent;
    m_pulseTime = 100;

    m_progressDialog = NULL;

    Connect( wxID_ANY, wxEVT_VEMS_DEVICE_CONTROLLER, VemsDeviceControllerEventHandler( DeviceControllerProgressBarManager::onDeviceControllerEvent ), NULL, this );
    Connect( wxID_ANY, wxEVT_COMMUNICATION, CommunicationEventHandler( DeviceControllerProgressBarManager::onCommunicationEvent ), NULL, this );
    Connect( wxID_OK, wxEVT_TIMER, wxTimerEventHandler( DeviceControllerProgressBarManager::onPulse ), NULL, this );
}

DeviceControllerProgressBarManager::~DeviceControllerProgressBarManager()
{
    Disconnect( wxID_ANY, wxEVT_VEMS_DEVICE_CONTROLLER, VemsDeviceControllerEventHandler( DeviceControllerProgressBarManager::onDeviceControllerEvent ), NULL, this );
    Disconnect( wxID_ANY, wxEVT_COMMUNICATION, CommunicationEventHandler( DeviceControllerProgressBarManager::onCommunicationEvent ), NULL, this );
}

void DeviceControllerProgressBarManager::createProgressDialog( const wxString& title, const wxString& message )
{
    int style = ( wxPD_APP_MODAL | wxPD_AUTO_HIDE  );

    m_progressDialog = new bcProgressDialog(    title,
                                                message,
                                                5,
                                                m_parent,
                                                style );

    m_progressDialog->SetSize( 300, 150 );
    m_progressDialog->SetFocus();
    m_progressDialog->CenterOnParent();

    m_timer.Start( m_pulseTime );
}

void DeviceControllerProgressBarManager::destroyProgressDialog()
{
    m_timer.Stop();

    if ( m_progressDialog != NULL ) {
        m_progressDialog->Destroy();
        m_progressDialog = NULL;
    }
}

void DeviceControllerProgressBarManager::onDeviceControllerEvent( VemsDeviceControllerEvent& event )
{
    LOGP( event.toString() );

    switch ( event.GetId() )
    {
        case VDC_EVENT_STARTED:
            createProgressDialog( _T("VemsTune init"), _T("Searching Device...") );
            break;

        case VDC_EVENT_DEVICE_FOUND:
        {
            if ( m_progressDialog == NULL ) {
                createProgressDialog( _T("VemsTune init"), _T("Device Found...") );
            } else {
                m_progressDialog->Pulse( _T("Device Found...") );
            }

            if ( event.isBootMode() ) {
                destroyProgressDialog();
            }
        } break;

        case VDC_EVENT_DEVICE_CHECKING_PASSWORD:
        {
            if ( m_progressDialog == NULL ) {
                createProgressDialog( _T("VemsTune init"), _T("Searching Device...") );
            }
            m_progressDialog->Pulse( _T("Checking password protection...") );
        } break;

        case VDC_EVENT_DEVICE_NOT_FOUND:
            destroyProgressDialog();
            break;

        case VDC_EVENT_DEVICE_PROTECTED:
        {
            ///TODO
            destroyProgressDialog();
        } break;

//        case VDC_EVENT_DEVICE_NOT_PROTECTED:
//        {
//            if ( m_progressDialog != NULL ) {
//                m_progressDialog->Pulse( _T("Fetching firmware information...") );
//            }
//        } break;

        case VDC_EVENT_DEVICE_FIRMWAREINFO_FETCHING:
        {
            if ( m_progressDialog != NULL ) {
                m_progressDialog->Pulse( _T("Fetching firmware information...") );
            }
        } break;

//        case VDC_EVENT_DEVICE_FIRMWAREINFO_FETCHED:
//        {
//            if ( m_progressDialog != NULL ) {
//                m_progressDialog->Pulse( _T("Syncing pages...") );
//            }
//        } break;

        case VDC_EVENT_DEVICE_FIRMWAREDATA_SYNCING:
        {
            if ( m_progressDialog != NULL ) {
                m_progressDialog->Pulse( _T("Syncing pages...") );
            }
        } break;

        case VDC_EVENT_DEVICE_CONNECTED:
            destroyProgressDialog();
            break;

        case VDC_EVENT_DEVICE_DISCONNECTED:
            destroyProgressDialog();
            break;

        case VDC_EVENT_DEVICE_RECONNECTED:
            destroyProgressDialog();
            break;

        case VDC_EVENT_START_RESUMING:
            if ( !m_progressDialog ) {
                createProgressDialog( _T("VemsTune init"), _T("Resuming...") );
            }
            break;

        case VDC_EVENT_RESUMED:
            destroyProgressDialog();
//            wxMessageBox( _T("Device resumed!") );
            break;

        case VDC_EVENT_LEAVING_BOOT_MODE:
        {
            if ( m_progressDialog == NULL ) {
                createProgressDialog( _T("VemsTune init"), _T("Leaving Boot mode...") );
            } else {
                m_progressDialog->Pulse( _T("Leaving Boot mode...") );
            }
        } break;

        case VDC_EVENT_DEVICE_CANNOT_CONNECT:
            destroyProgressDialog();
            break;

        default:
            break;
    }

}

void DeviceControllerProgressBarManager::onCommunicationEvent( CommunicationEvent& event )
{
    LOGP( wxString::Format( _T("event.GetId(): %d"), event.GetId() ) );

    switch ( event.GetId() )
    {
        case CommunicationEvent::CANNOT_OPEN_PORT:
//        case CommunicationEvent::TIMEOUT:
            destroyProgressDialog();
            break;
//
//        case CommunicationEvent::COMMUNICATION_ERROR:
//            destroyProgressDialog();
////            createProgressDialog( _T("Communication error!"), _T("Please plug off USB-Serial converter!") );
//            createProgressDialog( _T("Communication error!"), _T("TODO") );
//            break;

        case CommunicationEvent::REMOVE_USB:
        {
            wxString message = _T("Please plug off USB-Serial converter!");

            if ( !m_progressDialog ) {
                createProgressDialog( _T("Communication error!"), message );
            }

            m_progressDialog->Pulse(  message );
        } break;

        case CommunicationEvent::PORT_CLOSED:
            destroyProgressDialog();
//            wxMessageBox( _T("Communication error!") );
            break;
    }
}

void DeviceControllerProgressBarManager::onPulse( wxTimerEvent& event )
{
    if ( m_progressDialog ) {
        // crashed on linux
        //m_progressDialog->Pulse();
    }
}
