#include "FileOpenDlg.h"
#include "FileOpenDlgHelper.h"
#include "FileSearchDlg.h"

#include <wx/dir.h>
#include <wx/combo.h>
#include <wx/odcombo.h>
#include <wx/imaglist.h>
#include <wx/arrimpl.cpp>
#include <wx/dnd.h>

#include "../ProgressBar.h"
#include "../RawLogHelper.h"
#include "../MTConfigHelper.h"
#include "../FirmwareInfo.h"

#include "../customwidgets/InfoWindow.h"

#include "../loghistory/LogHistoryMgrHelper.h"
#include "../loghistory/QuickAnalyze.h"
#include "../loghistory/GPSHelper.h"

#include "../logger.h"

#include "../common/FileDropTarget.h"

int idFileOpenDlgCombo      = wxNewId();
int idFileOpenGoParent      = wxNewId();
int idFileOpenAnalyzeFiles  = wxNewId();
int idFileOpenSearchFiles   = wxNewId();
int idFileOpenOpen          = wxNewId();
int idFileOpenOpenSeparate  = wxNewId();
int idFileOpenCancel        = wxNewId();
int idFileOpenList          = wxNewId();
int idFileOpenFilter        = wxNewId();
int idFileOpenExportGPS     = wxNewId();
int idFileOpenKMLMenu       = wxNewId();
int idFileOpenGPXMenu       = wxNewId();
int idFileOpenShowRoute     = wxNewId();

WX_DEFINE_OBJARRAY( wxBitmapRefArray );


/// FILEOPENDLG EVENT TABLE


BEGIN_EVENT_TABLE( FileOpenDlg, wxDialog )

    EVT_SIZE( FileOpenDlg::OnSize )
    EVT_CLOSE( FileOpenDlg::OnClose )
    EVT_COMBOBOX( idFileOpenDlgCombo, FileOpenDlg::OnDirComboBox )
    EVT_COMBOBOX( idFileOpenFilter, FileOpenDlg::OnFilterCombo )

    EVT_TOOL( idFileOpenGoParent, FileOpenDlg::OnGoParentFolder )
    EVT_TOOL( idFileOpenAnalyzeFiles, FileOpenDlg::OnAnalyzeFiles )
    EVT_TOOL( idFileOpenSearchFiles, FileOpenDlg::OnSearchFiles )
    EVT_TOOL( idFileOpenExportGPS, FileOpenDlg::OnExportGPS )
    EVT_TOOL( idFileOpenShowRoute, FileOpenDlg::OnShowGPSRoute )

    EVT_BUTTON( idFileOpenCancel, FileOpenDlg::OnCancelButton )
    EVT_BUTTON( idFileOpenOpen, FileOpenDlg::OnOpenButton )
    EVT_BUTTON( idFileOpenOpenSeparate, FileOpenDlg::OnOpenSepButton )

    EVT_LIST_COL_CLICK( idFileOpenList, FileOpenDlg::OnListHeaderClick )
    EVT_LIST_ITEM_SELECTED( idFileOpenList, FileOpenDlg::OnSelectListItem )
    EVT_LIST_ITEM_ACTIVATED( idFileOpenList, FileOpenDlg::OnDoubleClickList )

    EVT_LIST_COL_RIGHT_CLICK( idFileOpenList, FileOpenDlg::SelectColumns )

    EVT_LIST_BEGIN_DRAG( idFileOpenList, FileOpenDlg::OnDrag )

    EVT_TIMER( FDLG_TIMERID, FileOpenDlg::OnTimer )

END_EVENT_TABLE()


/// PUBLIC FUNCTIONS

FileOpenDlg::FileOpenDlg( wxWindow* parent, const wxString& caption, wxString defaultDir, int defaultFiles )
{

    if ( defaultDir.EndsWith( _T("\\\\") ) )
        defaultDir.RemoveLast();

    m_defaultDir = defaultDir;
    m_defaultFiles = defaultFiles;

    // get visible columns
    m_columnInfoA = FileDialogSettings::getColumns( true );
    ///
    m_numcolumns = m_columnInfoA.GetCount();
    ///

    // get sorting info
    FileDialogSettings::getSorting( m_currColOrder.column, m_currColOrder.ascending );

    m_widthf = new float[m_numcolumns];

    for ( unsigned char j = 0; j < m_numcolumns; ++j )
    {
        FD_ColumnInfo* ci = m_columnInfoA[j];
        m_widthf[j] = ci->column_width;
    }

    wxArtProvider::Push( new FileOpenArtProvider );

    m_progressDlg = NULL;

    m_analyzer = NULL;
    m_popupMenu = NULL;

    m_configLoaded = ( MTConfigHelper::getMTConfigData()->getSymbol( _T("course") ) != NULL );

    buildGUI( parent, caption );

    m_timer.SetOwner( this, FDLG_TIMERID );

    new FileDropTarget<FileOpenDlg>(this);
}

FileOpenDlg::~FileOpenDlg()
{
    WX_CLEAR_ARRAY( m_columnInfoA );
    wxArtProvider::Pop();
    delete [] m_widthf;
}

wxString FileOpenDlg::GetFileName( bool fullPath )
{
    wxFileName fn( m_fileName->GetValue() );

    if ( fn.GetDirCount() > 0 )
    {
        if ( fullPath )
            return fn.GetFullPath();
        else
            return fn.GetFullName();
    }

    wxString f_fname = m_defaultDir.GetFullPath();
    f_fname += m_fileName->GetValue();

    if ( fullPath )
        return f_fname;
    else
        return m_fileName->GetValue();
}

/// CREATE GUI

void FileOpenDlg::buildGUI( wxWindow* parent, const wxString& caption )
{
    wxSize dlgsize = FileDialogSettings::getDlgSize();
    if ( dlgsize.GetWidth() == -1 )
        dlgsize.SetWidth( 650 );
    if ( dlgsize.GetHeight() == -1 )
        dlgsize.SetHeight( 300 );

    SetExtraStyle( wxWS_EX_CONTEXTHELP );
    Create( parent, wxID_ANY, caption, wxDefaultPosition, dlgsize, wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER | wxDIALOG_EX_CONTEXTHELP );

    wxBoxSizer* mainsizer = new wxBoxSizer( wxVERTICAL );

    // upper sizer
    wxBoxSizer* uppersizer = new wxBoxSizer( wxHORIZONTAL );

    wxStaticText* u_txt = new wxStaticText( this, -1, _T("Look in:") );

    m_dirList = new OwnBitmapComboBox( this, idFileOpenDlgCombo, _T(""), wxDefaultPosition,
                    wxSize(350, 24) , 0, NULL, wxCB_READONLY );

    updateDirCombo();

    //
    m_fo_tb = new wxToolBar( this, -1, wxDefaultPosition, wxDefaultSize, wxTB_HORIZONTAL | wxNO_BORDER |
        wxTB_FLAT | wxTB_NODIVIDER );
    wxBitmap upbitmap = wxArtProvider::GetBitmap( FO_GO_UP_ICON, wxART_OTHER , wxSize( 18, 18 ) );
    wxBitmap greyed = (upbitmap.ConvertToImage().ConvertToGreyscale());

    m_fo_tb->AddTool( idFileOpenGoParent, _T("Go to parent directory"), upbitmap, greyed,
        wxITEM_NORMAL, _T("Go to parent directory"), _T("Go to parent directory") );


    wxBitmap analyzeBmp = wxArtProvider::GetBitmap( FO_ICON_ANALYZE, wxART_OTHER, wxSize( 18, 18 ) );
    wxBitmap analyzeGrey = analyzeBmp.ConvertToImage().ConvertToGreyscale();

    m_fo_tb->AddTool( idFileOpenAnalyzeFiles, _T("Analyze all files in current directory"), analyzeBmp, analyzeGrey,
        wxITEM_NORMAL, _T("Analyze all files in current directory"), _T("Analyze all files in current directory") );

    wxBitmap searchBmp = wxArtProvider::GetBitmap( FO_ICON_SEARCH, wxART_OTHER, wxSize( 18, 18 ) );
    wxBitmap searchGr = searchBmp.ConvertToImage().ConvertToGreyscale();

    m_fo_tb->AddTool( idFileOpenSearchFiles, _T("Search files in VEMS home directory"), searchBmp, searchGr,
        wxITEM_NORMAL, _T("Search files in VEMS home directory"), _T("Search files in VEMS home directory") );

    wxBitmap gpsBmp = wxArtProvider::GetBitmap( FO_ICON_GPS, wxART_OTHER, wxSize( 18, 18 ) );
    wxBitmap gpsGr = gpsBmp.ConvertToImage().ConvertToGreyscale();

    m_fo_tb->AddTool( idFileOpenExportGPS, _T("Export GPS data to XML format"), gpsBmp, gpsGr,
        wxITEM_NORMAL, _T("Export GPS data to XML format"), _T("Export GPS data to XML format") );

    wxBitmap satBmp = wxArtProvider::GetBitmap( FO_ICON_SAT, wxART_OTHER, wxSize( 18, 18 ) );
    wxBitmap satGr = satBmp.ConvertToImage().ConvertToGreyscale();

    m_fo_tb->AddCheckTool( idFileOpenShowRoute, _T("Show GPS route"), satBmp, satGr,
        _T("Show GPS route (if calculated)"), _T("Show GPS route (if calculated)") );
    m_fo_tb->ToggleTool( idFileOpenShowRoute, false );

    m_fo_tb->Realize();
    m_fo_tb->EnableTool( idFileOpenExportGPS, false );

    uppersizer->Add( u_txt, 0, wxALL | wxALIGN_CENTER_VERTICAL, 5 );
    uppersizer->Add( m_dirList, 0, wxALL, 5 );
    uppersizer->Add( m_fo_tb, 0, wxALL, 5 );

    // filelist

    m_fileList = new wxListCtrl( this, idFileOpenList, wxDefaultPosition, wxDefaultSize, wxLC_REPORT );

    ///
    wxFlexGridSizer* gsizer = new wxFlexGridSizer( 4 );
    // filename

    wxStaticText* s3_txt = new wxStaticText( this, wxID_ANY, _T("File &name:") );
    m_fileName = new wxTextCtrl( this, -1, _T(""), wxDefaultPosition, wxDefaultSize, wxTE_READONLY );
    wxButton* openButton = new wxButton( this, idFileOpenOpen, _T("&Open") );
    openButton->SetDefault();

    gsizer->Add( s3_txt, 0, wxALL, 5 );
    gsizer->Add( m_fileName, 1, wxEXPAND | wxALL, 5 );
    gsizer->Add( openButton, 0, wxALL, 5 );
    gsizer->Add( new wxButton(this, idFileOpenOpenSeparate, _T("Open to separate frame")) , 0, wxALL, 5 );
    //gsizer->Add( 30, 18 );

    // extension filter
    wxStaticText* s4_txt = new wxStaticText( this, wxID_ANY, _T("Files of &type") );

    wxArrayString ft_arr; ft_arr.Add( _T("Both config and log files (*.vemscfg, *.vemslog, *.triggerlog)") );
    ft_arr.Add( _T("Only log files (*.vemslog)") ); ft_arr.Add( _T("Only config files (*.vemscfg)") );
    ft_arr.Add( _T("Trigger log files (*.triggerlog)") );

    m_fileFilter = new wxComboBox( this, idFileOpenFilter, _T(""), wxDefaultPosition, wxDefaultSize, ft_arr, wxCB_DROPDOWN | wxCB_READONLY );
    m_fileFilter->SetSelection( 0 );

    wxButton* cancelButton = new wxButton( this, idFileOpenCancel, _T("Cancel") );

    gsizer->Add( s4_txt, 0, wxALL, 5 );
    gsizer->Add( m_fileFilter, 1, wxEXPAND | wxALL, 5 );
    gsizer->Add( cancelButton, 0, wxALL, 5 );
    gsizer->AddStretchSpacer();
    //gsizer->Add( 30, 18 );

    // add sizers
    mainsizer->Add( uppersizer, 0, wxALL | wxEXPAND, 5 );

    m_listsizer = new wxBoxSizer( wxHORIZONTAL );

    m_listsizer->Add( m_fileList, 1, wxALL | wxEXPAND, 5 );

    m_gpsroutepanel = new wxPanel( this );
    wxSizer* gpsps = new wxBoxSizer( wxVERTICAL );
    m_gpspreview = new wxStaticBitmap( m_gpsroutepanel, wxID_ANY, wxNullBitmap, wxDefaultPosition, wxSize( 240, -1 ) );
    s3_txt = new wxStaticText( m_gpsroutepanel, wxID_ANY, _T("GPS route preview (only visible when GPS data exists in logfile and the map is generated already or a config/log file is opened)") );
//            wxDefaultPosition, wxSize(220,200) );
    s3_txt->Wrap( 200 );

    gpsps->Add( s3_txt, 0, wxALL | wxEXPAND, 0 );
    gpsps->Add( m_gpspreview, 0, wxTOP, 5 );

    m_gpsroutepanel->SetSizer( gpsps );
    gpsps->Fit( m_gpsroutepanel );
    gpsps->SetSizeHints( m_gpsroutepanel );

    m_gpsroutepanel->Show( false );

    m_listsizer->Add( m_gpsroutepanel, 0, wxALL | wxEXPAND, 5 );

    mainsizer->Add( m_listsizer, 1, wxALL | wxEXPAND, 5 );

    mainsizer->Add( gsizer, 0, wxALL | wxEXPAND, 5 );

    SetSizer( mainsizer );
    mainsizer->Fit( this );
    mainsizer->SetSizeHints( this );

    SetSize( dlgsize );

    // update listbox
    initListCtrl();
    fillListWithFiles( m_currColOrder.column, m_currColOrder.ascending );

    SetEscapeId( idFileOpenCancel );
}

/// EVENT HANDLERS

void FileOpenDlg::OnClose( wxCloseEvent& event )
{
    FileDialogSettings::setDlgSize( GetSize() );
    FileDialogSettings::setSorting( m_currColOrder.column, m_currColOrder.ascending );
    Destroy();
    event.Skip();
}

void FileOpenDlg::SelectColumns( wxListEvent& event )
{
    FileDialogSettingsDialog dlg( this );
    if ( dlg.ShowModal() == 1 )
    {
        rebuildList();
    }
}

void FileOpenDlg::OnDirComboBox( wxCommandEvent& event )
{
    int id = m_dirList->GetSelection();

    if ( id < 0 || id >= (int)(m_comboRawDirs.GetCount()) )
        return;

    m_defaultDir = m_comboRawDirs[id];
    fillListWithFiles( m_currColOrder.column, m_currColOrder.ascending );
    m_fileName->SetValue( wxEmptyString );
}

void FileOpenDlg::OnFilterCombo( wxCommandEvent& event )
{
    int no = m_fileFilter->GetSelection();

    if ( no < 0 || no >= (int)(m_fileFilter->GetCount()) )
    {
        event.Skip();
        return;
    }

    switch ( no )
    {
        case 0: m_defaultFiles = VF_VEMS_ALL; break;
        case 1: m_defaultFiles = VF_VEMS_LOG; break;
        case 2: m_defaultFiles = VF_VEMS_CFG; break;
        case 3: m_defaultFiles = VF_VEMS_TLOG; break;
    }

    fillListWithFiles( m_currColOrder.column, m_currColOrder.ascending );
}

void FileOpenDlg::OnExportGPS( wxCommandEvent& event )
{
    if ( m_popupMenu == NULL )
    {
        m_popupMenu = new wxMenu;
        m_popupMenu->Append( idFileOpenKMLMenu, _T("Export to KML format") );
        m_popupMenu->Append( idFileOpenGPXMenu, _T("Export to GPX format") );

        Connect( idFileOpenKMLMenu, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( FileOpenDlg::OnExportGPSMenu ), NULL, this );
        Connect( idFileOpenGPXMenu, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( FileOpenDlg::OnExportGPSMenu ), NULL, this );
    }

    PopupMenu( m_popupMenu );
}

void FileOpenDlg::OnExportGPSMenu( wxCommandEvent& event )
{
    int id = event.GetId();

    wxString fname = m_defaultDir.GetFullPath();
    fname += m_fileName->GetValue();

    GPSHelper gps;

    if ( id == idFileOpenKMLMenu )
        gps.analyzeGPS( fname, 0 );
    if ( id == idFileOpenGPXMenu )
        gps.analyzeGPS( fname, 1 );
}

void FileOpenDlg::OnSelectListItem( wxListEvent& event )
{
    wxString str = event.GetText();

    int idx = event.GetIndex();

    int col = -1;
    for ( unsigned char j = 0; j < m_columnInfoA.GetCount(); ++j )
    {
        FD_ColumnInfo* gpsCol = m_columnInfoA[j];
        if ( gpsCol->name.CmpNoCase(_T("GPS Data")) == 0 )
        {
            col = j;
            break;
        }
    }

    wxListItem item;
    item.SetColumn( col );
    item.SetId( idx );
    item.SetMask(wxLIST_MASK_TEXT);
    m_fileList->GetItem( item );

    if ( item.GetText() == _T("yes") )
    {
        if ( m_configLoaded )
            m_fo_tb->EnableTool( idFileOpenExportGPS, true );

        wxString f_fname = m_defaultDir.GetFullPath();
        f_fname += str;
        wxFileName fn( f_fname );

        fn.SetExt( _T("png") );
        if ( wxFileExists( fn.GetFullPath() ) )
        {
            wxBitmap bmp( fn.GetFullPath(), wxBITMAP_TYPE_PNG );
            m_gpspreview->SetBitmap( bmp );
        }
        else
        {
            if ( m_configLoaded )
            {
                GPSHelper h;
                h.analyzeGPS( f_fname, -1 );
                wxBitmap bmp( fn.GetFullPath(), wxBITMAP_TYPE_PNG );
                m_gpspreview->SetBitmap( bmp );
            }
            else
                m_gpspreview->SetBitmap( wxNullBitmap );
        }

    }
    else
    {
        m_fo_tb->EnableTool( idFileOpenExportGPS, false );
        m_gpspreview->SetBitmap( wxNullBitmap );
    }
    m_gpspreview->Refresh(true);
//    m_gpspreview->Update();
    Refresh( false );

    m_fileName->SetValue( str );
}

void FileOpenDlg::OnCancelButton( wxCommandEvent& event )
{
    if ( IsModal() )
    {
        FileDialogSettings::setDlgSize( GetSize() );
        FileDialogSettings::setSorting( m_currColOrder.column, m_currColOrder.ascending );
        EndDialog( wxID_CANCEL );
    }
    else
    {
        FileDialogSettings::setDlgSize( GetSize() );
        FileDialogSettings::setSorting( m_currColOrder.column, m_currColOrder.ascending );
        SetReturnCode( wxID_CANCEL );
        Hide();
    }
}

void FileOpenDlg::OnOpenSepButton( wxCommandEvent& event )
{
    if ( m_fileName->GetValue() == wxEmptyString )
    {
        event.Skip();
        return;
    }
    // check whether selection is a dir
    wxString str = m_fileName->GetValue();
    wxFileName f_name( str );

    if ( str == f_name.GetFullPath() )
    {
        f_name = m_defaultDir.GetPathWithSep() + str;
    }
    //LOGP( f_name.GetFullPath() )
    if ( wxDir::Exists( f_name.GetFullPath() ) )
    {
        event.Skip();
        return;
    }

    if ( IsModal() )
    {
        FileDialogSettings::setDlgSize( GetSize() );
        FileDialogSettings::setSorting( m_currColOrder.column, m_currColOrder.ascending );
        EndDialog( wxID_YES );
    }
    else
    {
        FileDialogSettings::setDlgSize( GetSize() );
        FileDialogSettings::setSorting( m_currColOrder.column, m_currColOrder.ascending );
        SetReturnCode( wxID_YES );
        Hide();
    }
}

void FileOpenDlg::OnOpenButton( wxCommandEvent& event )
{
    if ( m_fileName->GetValue() == wxEmptyString )
    {
        event.Skip();
        return;
    }
    // check whether selection is a dir
    wxString str = m_fileName->GetValue();
    wxFileName f_name( str );

    if ( str == f_name.GetFullPath() )
    {
        f_name = m_defaultDir.GetPathWithSep() + str;
    }
    //LOGP( f_name.GetFullPath() )
    if ( wxDir::Exists( f_name.GetFullPath() ) )
    {
        event.Skip();
        return;
    }

    if ( IsModal() )
    {
        FileDialogSettings::setDlgSize( GetSize() );
        FileDialogSettings::setSorting( m_currColOrder.column, m_currColOrder.ascending );
        EndDialog( wxID_OK );
    }
    else
    {
        FileDialogSettings::setDlgSize( GetSize() );
        FileDialogSettings::setSorting( m_currColOrder.column, m_currColOrder.ascending );
        SetReturnCode( wxID_OK );
        Hide();
    }
}

void FileOpenDlg::OnSize( wxSizeEvent& event )
{
    Layout();
    float w = (float)( m_fileList->GetSize().x - 4 );

    int* nw = new int[m_numcolumns];
    calcColumnWidth( w, nw );

    for ( uint8_t i = 0; i < m_numcolumns; ++i )
    {
        m_fileList->SetColumnWidth( i, nw[i] );
    }
}

void FileOpenDlg::OnListHeaderClick( wxListEvent& event )
{
    int column = event.GetColumn();

    if ( column < 0 || column >= m_numcolumns )
    {
        event.Skip();
        return;
    }

    if ( column == m_currColOrder.column )
    {
        bool asc = !m_currColOrder.ascending;
        fillListWithFiles( column, asc );
        wxListItem item;

        m_fileList->GetColumn( column, item );

        if ( asc )
            item.SetImage( ICON_UP );
        else
            item.SetImage( ICON_DOWN );

        m_fileList->SetColumn( column, item );

    }
    else
    {
        wxListItem item;

        // remove image from current column
        m_fileList->GetColumn( m_currColOrder.column, item );
        item.SetImage( ICON_NULLBMP ); // nullbitmap
        m_fileList->SetColumn( m_currColOrder.column, item );

        // update list
        fillListWithFiles( column, true );

        // add up arrow
        m_fileList->GetColumn( column, item );
        item.SetImage( ICON_UP );
        m_fileList->SetColumn( column, item );
    }
}

void FileOpenDlg::OnDoubleClickList( wxListEvent& event )
{
    wxString selected = m_fileName->GetValue();

    // search result?
    wxFileName fn( selected );

    if ( fn.GetDirCount() > 0 )
    {
        //LOGP( fn.GetFullPath() )
        if ( wxFileExists( fn.GetFullPath() ) )
        {
            //LOG
            FileDialogSettings::setDlgSize( GetSize() );
            FileDialogSettings::setSorting( m_currColOrder.column, m_currColOrder.ascending );
            EndModal( wxID_OK );
            return;
        }
    }

    // else
    wxFileName temp( m_defaultDir );
    temp.AppendDir( selected );

    if ( temp.DirExists() )
    {
        m_defaultDir = temp;
        m_fileName->SetValue( wxEmptyString );
        fillListWithFiles( m_currColOrder.column, m_currColOrder.ascending );
    }

    temp = m_defaultDir;
    temp.SetFullName( selected );

    if ( wxFileExists( temp.GetFullPath() ) )
    {
        FileDialogSettings::setDlgSize( GetSize() );
        FileDialogSettings::setSorting( m_currColOrder.column, m_currColOrder.ascending );
        EndModal( wxID_OK );
    }
}

void FileOpenDlg::OnGoParentFolder( wxCommandEvent& event )
{
    wxFileName temp = m_defaultDir;
    temp.RemoveLastDir();

    if ( m_defaultDir == temp )
        return;

    m_defaultDir = temp;
    fillListWithFiles( m_currColOrder.column, m_currColOrder.ascending );
    m_fileName->SetValue( wxEmptyString );
}

void FileOpenDlg::OnAnalyzeFiles( wxCommandEvent& event )
{
    if ( wxMessageBox( _T("Current log file (if any opened) will be closed.\nThe analyzis may take a while"),
        _T("Information"), wxOK | wxCANCEL | wxICON_WARNING ) == wxCANCEL )
    {
        return;
    }

    FileOpenSQL files;

    wxString dir = m_defaultDir.GetFullPath();
    if ( dir.EndsWith( _T("/") ) || dir.EndsWith( _T("\\") ) )
        dir = dir.RemoveLast();

    files.setPath( dir );

    wxString cAS = _T("Currently analyzed: ");

    m_analyzeInfo.Clear();

    m_analyzeInfo = files.getFiles( VF_VEMS_LOG );

    if ( m_analyzer != NULL )
        delete m_analyzer;

    m_analyzer = NULL;
    m_progressDlg2 = new wxProgressDialog( _T("Progress"), cAS, m_analyzeInfo.GetCount(), NULL,
        wxPD_APP_MODAL | wxPD_SMOOTH | wxPD_AUTO_HIDE | wxPD_CAN_ABORT );


    m_progressDlg2->SetSize( 450, 200 );

    size_t i;

    wxString fileName;

    for ( i = 0; i < m_analyzeInfo.GetCount(); ++i )
    {
        FileInfo* info = m_analyzeInfo[i];
        fileName = dir + DirHelper::Separ() + info->fileName;
        m_progressDlg2->Update( i, cAS + info->fileName );
        if ( QuickAnalyze::isAnalyzed( fileName ) == false && wxFileExists( fileName ) )
            break;
    }
    if ( i < m_analyzeInfo.GetCount() )
    {
        LogHistoryMgr* lhmgr = LogHistoryMgrHelper::getInstance();
        lhmgr->emptyHistory();
        lhmgr->loadHistoryFromLog( fileName );
        FirmwareInfo fwInfo;
        RawLogHelper::fillFirmwareInfo(fwInfo, lhmgr->getHeaderHash());
        MTConfigHelper::parseMTConfigData( fwInfo, NULL, false );

        m_analyzer = new QuickAnalyze( lhmgr, fileName );
        m_analyzer->registerObserver( this );
        m_aIndex = i;
        m_analyzer->start();
    }
    else
    {
        WX_CLEAR_ARRAY( m_analyzeInfo );
        delete m_progressDlg2;
    }
}

void FileOpenDlg::OnSearchFiles( wxCommandEvent& event )
{
    FileDialogSearchDialog dlg( this );
    if ( dlg.ShowModal() == - 1 )
        return;

    FileInfoA sresults = dlg.getFiles();
    rebuildList( &sresults );

    event.Skip();
}

void FileOpenDlg::OnShowGPSRoute( wxCommandEvent& event )
{
    int state = event.GetInt();

    if ( state == 1 || state == 0 )
    {
        m_gpsroutepanel->Show( (bool) state );
        m_listsizer->RecalcSizes();
        m_listsizer->Layout();
    }

    event.Skip();
}

void FileOpenDlg::OnDrag( wxListEvent& event )
{
    m_canClose = true;

    wxString fname = m_defaultDir.GetFullPath();
    fname +=  m_fileList->GetItemText( event.GetIndex() );

    if ( wxFileExists( fname ) == false )
    {
        event.Skip();
        return;
    }

    wxFileDataObject fdo;

    fdo.AddFile( fname );

    wxDropSource ds( fdo, m_fileList );
    ds.DoDragDrop( wxDragCopy );

    m_timer.Start( 500, true );
}

bool FileOpenDlg::onDropFiles( const wxArrayString& filenames )
{
    m_canClose = false;

    return true;
}

/// OTHER PRIVATE METHODS

void FileOpenDlg::initListCtrl()
{
    // header

    int w = m_fileList->GetSize().x -4;

    float fw = (float) w;

    int* factw = new int[m_numcolumns];
    calcColumnWidth( fw, factw );

    wxListItem* headers = new wxListItem[m_numcolumns];

    for ( unsigned char i = 0; i < m_numcolumns; ++i )
    {
        headers[i].SetMask( wxLIST_MASK_TEXT | wxLIST_MASK_DATA );
        headers[i].SetWidth( factw[i] );

        FD_ColumnInfo* ci = m_columnInfoA[i];
        // size
        if ( ci->id == 2 )
        {
            headers[i].SetAlign( wxLIST_FORMAT_RIGHT );
        }
        else
            if ( ci->id == 9 ) // gps
                headers[i].SetAlign( wxLIST_FORMAT_CENTER );
            // others
            else headers[i].SetAlign( wxLIST_FORMAT_LEFT );

        headers[i].SetText( ci->name );

        headers[i].SetData( 1 );
    }

    // images
    wxImageList *imglist = new wxImageList(16,16,true);

    imglist->Add( wxArtProvider::GetBitmap( FO_VEMS_FILE, wxART_OTHER , wxSize(16,16) ) );
    imglist->Add( wxArtProvider::GetBitmap( FO_ICON_DOWN, wxART_OTHER, wxSize(16,16) ) );
    imglist->Add( wxArtProvider::GetBitmap( FO_ICON_UP, wxART_OTHER, wxSize(16,16) ) );
    wxBitmap nullbitmap(16,16, 4);
    imglist->Add( nullbitmap );
    imglist->Add( wxArtProvider::GetBitmap( FO_VEMS_CFG, wxART_OTHER, wxSize(16,16) ) );
    imglist->Add( wxArtProvider::GetBitmap( FO_FOLDER_ICON, wxART_OTHER, wxSize( 16,16 ) ) );

    m_fileList->SetImageList( imglist, wxIMAGE_LIST_SMALL );

    for ( unsigned char i = 0; i < m_numcolumns; ++i )
    {
        m_fileList->InsertColumn( i, headers[i] );
    }

    // default sort order: name,asc
    wxListItem item;
    m_fileList->GetColumn( 0, item );
    item.SetImage( ICON_UP );
    m_fileList->SetColumn( 0, item );
}

void FileOpenDlg::fillListWithFiles( uint8_t colOrder, bool ascending )
{
    m_fileList->DeleteAllItems();

    FileOpenSQL files;

    files.registerObserver( this );

    files.setPath( m_defaultDir.GetFullPath() );

    FO_order order;

    int colNo = (m_columnInfoA[colOrder])->id - 1; // id starts with 1

    switch ( colNo )
    {
        case 1: order = FO_OB_SIZE; break;
        case 2: order = FO_OB_TYPE; break;
        case 3: order = FO_OB_DATE; break;
        case 4: order = FO_OB_FIRMWARE; break;
        case 5: order = FO_OB_RPM; break;
        case 6: order = FO_OB_MAP; break;
        case 7: order = FO_OB_TRIGERR; break;
        case 8: order = FO_OB_GPSDATA; break;
        case 0:
        default: order = FO_OB_FNAME; break;
    }

    FileInfoA finfoA = files.getFiles( m_defaultFiles, order, ascending );

    Freeze();

    // get & and subdirs
    wxDir dir( m_defaultDir.GetFullPath() );
    wxString tmp;

    bool res = dir.GetFirst( &tmp ,wxEmptyString, wxDIR_DIRS | wxDIR_HIDDEN );

    while ( res ) {
        AddDir( tmp );
        res = dir.GetNext( &tmp );
    }

    // add files

    for ( size_t i = 0; i < finfoA.GetCount(); ++i )
    {
        FileInfo* info = finfoA[i];
        AddLine( info );
    }

    Thaw();

    files.unRegisterObserver( this );

    WX_CLEAR_ARRAY( finfoA );

    m_currColOrder.ascending = ascending;
    m_currColOrder.column = colOrder;

    updateDirCombo();
}

void FileOpenDlg::AddLine( FileInfo* finfo )
{
    wxListItem* items = new wxListItem[m_numcolumns];

    for ( uint8_t i = 0; i < m_numcolumns; ++i )
    {
        items[i].SetId( m_fileList->GetItemCount() );
        items[i].SetColumn( i );
    }

    for ( unsigned char i = 0; i < m_numcolumns; ++i )
    {
        FD_ColumnInfo* ci = m_columnInfoA[i];
        int id = ci->id;

        switch ( id )
        {
            // filename
            case 1:
            {
                if ( finfo->isLogFile() )
                    items[i].SetImage( ICON_FILE );
                else
                    items[i].SetImage( ICON_CFG );

                items[i].SetText( finfo->fileName );
                break;
            }

            // file size
            case 2:
            {
                wxString fs = wxFileName::GetHumanReadableSize( (wxULongLong) finfo->fileSize, _T("N/A") );
                items[i].SetText( fs );
                break;
            }

            // file type
            case 3:
            {
                items[i].SetText( finfo->fileType );
                break;
            }

            // mod.date
            case 4:
            {
                wxDateTime dt( (time_t) finfo->fileDate );
                wxString dateS = dt.FormatISODate() + _T(" ") + dt.FormatISOTime();
                items[i].SetText( dateS );
                break;
            }

            // firmware version
            case 5:
            {
                wxString fwv = wxString::Format( _T("%d.%d.%d"), finfo->fw_major, finfo->fw_minor, finfo->fw_build );
                items[i].SetText( fwv );
                break;
            }

            // max rpm value
            case 6:
            {
                wxString rpm = ( finfo->rpmMax >= 0 ) ? wxString::Format( _T("%d"), finfo->rpmMax ) : _T("");
                items[i].SetText( rpm );
                break;
            }

            // max map value
            case 7:
            {
                wxString map = ( finfo->mapMax >= 0 ) ? wxString::Format( _T("%d"), finfo->mapMax ) : _T("");
                items[i].SetText( map );
                break;
            }

            // no of trigger errors
            case 8:
            {
                wxString tErr = ( finfo->triggerErr >= 0 ) ? wxString::Format( _T("%d"), finfo->triggerErr ) : _T("");
                items[i].SetText( tErr );
                break;
            }

            // GPS data
            case 9:
            {
                wxString hasGPS = ( finfo->hasGPSData < 0 ) ? _T("N/A") : ( ( finfo->hasGPSData == 1 ) ? _T("yes") : _T("no") );
                items[i].SetText( hasGPS );
                break;
            }


        }
    }

    m_fileList->InsertItem( items[0] );

    for ( uint8_t i = 1; i < m_numcolumns; ++i )
        m_fileList->SetItem( items[i] );
}

void FileOpenDlg::AddDir( const wxString& dirname )
{
    wxListItem* items = new wxListItem[m_numcolumns];

    for ( uint8_t i = 0; i < m_numcolumns; ++i )
    {
        items[i].SetId( m_fileList->GetItemCount() );
        items[i].SetColumn( i );
    }

    for ( unsigned char i = 0; i < m_numcolumns; ++i )
    {
        FD_ColumnInfo* ci = m_columnInfoA[i];
        int id = ci->id;

        switch ( id )
        {
            // name
            case 1:
            {
                items[i].SetImage( ICON_FOLDER );
                items[i].SetText( dirname );
                break;
            }
            // type
            case 3:
            {
                #ifdef WIN32
                items[i].SetText( _T("Folder") );
                #else
                items[i].SetText( _T("Directory") );
                #endif
                break;
            }
            default:
                items[i].SetText( _T("") );
        }

    }

    m_fileList->InsertItem( items[0] );

    for ( uint8_t i = 1; i < m_numcolumns; ++i )
        m_fileList->SetItem( items[i] );
}

void FileOpenDlg::updateDirCombo()
{
    wxBitmapRefArray  bmparray;

    m_dirList->Clear();
    m_dirList->ClearBitmapArray();

    m_comboRawDirs.Clear();

    /// subdirs
    wxArrayString subdirs = m_defaultDir.GetDirs();

    /// drives #1
    wxString volume;
    #ifdef WIN32
    enumDrives();

    volume = m_defaultDir.GetVolume();
    if ( volume != wxEmptyString )
    {
        volume += _T(":\\");
        wxFSVolume vol( volume );

        wxString displayName = vol.GetDisplayName();
        wxBitmap bmp; bmp.CopyFromIcon( vol.GetIcon( wxFS_VOL_ICO_SMALL ) );
        wxImage img = bmp.ConvertToImage(); wxBitmap bmp2( img );

        m_dirList->Append( displayName );
        m_dirList->AddBitmap( bmp2 );
        m_comboRawDirs.Add( volume );
    }

    #endif

    wxBitmap folderbmp( wxArtProvider::GetBitmap( FO_FOLDER_ICON, wxART_OTHER, wxSize( 16,16 ) ) );

    IntArray folderids;

    wxString dir;
    #ifdef WIN32
    dir = volume;
    #else
    dir = DirHelper::Separ();
    #endif

    for ( size_t i = 0; i < subdirs.GetCount(); ++i )
    {
        wxString str = subdirs[i];
        m_dirList->Append( str );
        m_dirList->AddBitmap( folderbmp );

        if ( m_defaultDir.HasVolume() )
            folderids.Add ( i + 1 );
        else
            folderids.Add( i );

        dir += str + DirHelper::Separ();
        m_comboRawDirs.Add( dir );
    }

    /// drives # 2
    // volumes(drives) are not supported under linux

    #ifdef WIN32
    if ( m_drives.GetCount() > 0 )
    {
        VolumeInfoA::const_iterator it = m_drives.begin();

        for ( ; it < m_drives.end(); ++it )
        {
            VolumeInfo drive = **it;
            wxString driveName = drive.Name;

            if ( driveName == volume )
                continue;

            wxImage img = drive.Icon.ConvertToImage();

            wxBitmap bmp2(img);

            m_dirList->AddBitmap( bmp2 );

            m_dirList->Append( drive.DisplayName );

            m_comboRawDirs.Add( driveName );
        }
    }

    #endif

    /// ADD Vems default working dir
    m_comboRawDirs.Add( DirHelper::GetPath(dLOG) );
    wxBitmap vemsbmp( wxArtProvider::GetBitmap( FO_VEMS_FILE, wxART_OTHER, wxSize( 16,16 ) ) );
    m_dirList->AddBitmap( vemsbmp );
    m_dirList->Append( _T("VEMS Default log directory") );

    m_dirList->SetFolderIds( folderids );

    if ( subdirs.GetCount() != 0 )
        m_dirList->SetSelection( subdirs.GetCount() );
    else
        m_dirList->SetSelection( 0 );
}

/// HELPERS

void FileOpenDlg::calcColumnWidth( float w, int *cw )
{
    for ( uint8_t i = 0; i < m_numcolumns; ++i )
    {
        cw[i] = (int)( w * m_widthf[i] );
    }
}

void FileOpenDlg::enumDrives()
{
    #ifdef WIN32
    WX_CLEAR_ARRAY( m_drives );

    wxArrayString arr = wxFSVolume::GetVolumes();

    for ( size_t i = 0; i < arr.GetCount(); ++i )
    {
        wxString drive = arr[i];

        VolumeInfo* vinfo = new VolumeInfo;

        wxFSVolume vol( drive );

        vinfo->Name = vol.GetName();
        vinfo->DisplayName = vol.GetDisplayName();
        vinfo->Kind = vol.GetKind();


        vinfo->Icon.CopyFromIcon( vol.GetIcon( wxFS_VOL_ICO_SMALL ) );

        m_drives.Add( vinfo );
    }
    #endif
}

void FileOpenDlg::notify( const wxString& str )
{
    if ( str != _T("qaFinished") )
        return;

    wxString dir = m_defaultDir.GetFullPath();
    if ( dir.EndsWith( _T("/") ) || dir.EndsWith( _T("\\") ) )
        dir = dir.RemoveLast();
    wxString cAS = _T("Currently analyzed: ");

    wxSafeYield( m_progressDlg2, true );
    m_aIndex++;

    // pressed cancel
    if ( m_progressDlg2->Update( m_aIndex ) == false )
    {
        m_aIndex = m_analyzeInfo.GetCount();

        WX_CLEAR_ARRAY( m_analyzeInfo );

        delete m_progressDlg2;

        return;
    }

    size_t i;
    wxString fileName;

    for ( i = m_aIndex; i < m_analyzeInfo.GetCount(); ++i )
    {
        FileInfo* info = m_analyzeInfo[i];
        fileName = dir + DirHelper::Separ() + info->fileName;

        m_progressDlg2->Update( i, cAS + info->fileName );

        if ( QuickAnalyze::isAnalyzed( fileName ) == false && wxFileExists( fileName ) )
            break;
    }
    if ( i < m_analyzeInfo.GetCount() )
    {
        LogHistoryMgr* lhmgr = LogHistoryMgrHelper::getInstance();
        lhmgr->emptyHistory();
        lhmgr->loadHistoryFromLog( fileName );

        FirmwareInfo fwInfo;

        RawLogHelper::fillFirmwareInfo(fwInfo, lhmgr->getHeaderHash());
        MTConfigHelper::parseMTConfigData( fwInfo, NULL, false );

        m_analyzer->set( lhmgr, fileName );
        m_aIndex = i;
        m_analyzer->start();
    }
    else
    {
        WX_CLEAR_ARRAY( m_analyzeInfo );
        delete m_progressDlg2;
        InfoWindow::Display( 0, _T("Info"), _T("All files have been analyzed") );
        rebuildList();
    }
}

void FileOpenDlg::notify( int i )
{
    static ProgressBar* pb = NULL;

    if ( m_progressDlg == NULL )
    {
        m_progressDlg = new wxDialog( this, -1, _T("Please wait..."), wxDefaultPosition,
            wxSize( 300, 200 ), wxDEFAULT_DIALOG_STYLE | wxSTAY_ON_TOP );
        wxBoxSizer* sizer = new wxBoxSizer( wxVERTICAL );

        wxStaticText* txt = new wxStaticText( m_progressDlg, -1, _T("Parsing and caching of files is in progress...") );
        pb = new ProgressBar( m_progressDlg, -1, wxDefaultPosition, wxSize( 280, 18 ) );
        pb->SetValueFormat( _T("%d") );
        pb->Update( 0 );

        sizer->Add( txt, 0, wxALL, 5 );
        sizer->Add( pb, 0, wxEXPAND | wxALL, 5 );

        m_progressDlg->SetSizer( sizer );
        sizer->Fit( m_progressDlg );
        sizer->SetSizeHints( m_progressDlg );

        m_progressDlg->Show();
    }
    else
    {
        wxSafeYield( m_progressDlg, true );
        pb->Update( i );
    }

    // finished
    if ( i == 100 && m_progressDlg != NULL )
    {
        m_progressDlg->Destroy();
        m_progressDlg = NULL;
        pb = NULL;
    }
}

void FileOpenDlg::rebuildList( FileInfoA* fileList )
{
    // rebuild columns
    m_fileList->ClearAll();

    WX_CLEAR_ARRAY( m_columnInfoA );
    m_currColOrder.ascending = true;
    m_currColOrder.column = 0;

    delete [] m_widthf;

    // get visible columns
    m_columnInfoA = FileDialogSettings::getColumns( true );
    m_numcolumns = m_columnInfoA.GetCount();

    m_widthf = new float[m_numcolumns];

    for ( unsigned char j = 0; j < m_numcolumns; ++j )
    {
        FD_ColumnInfo* ci = m_columnInfoA[j];
        m_widthf[j] = ci->column_width;
    }

    initListCtrl();
    if ( fileList == NULL )
        fillListWithFiles();
    else
        showSearchResult( fileList );
}

void FileOpenDlg::showSearchResult( FileInfoA* filelist )
{
    m_fileList->DeleteAllItems();

    Freeze();

    // add files

    for ( size_t i = 0; i < filelist->GetCount(); ++i )
    {
        FileInfo* info = (*filelist)[i];
        AddLine( info );
    }

    Thaw();

    WX_CLEAR_ARRAY( *filelist );

    updateDirCombo();
    m_dirList->Append( _T("Search result") );
    wxBitmap bmp( wxArtProvider::GetBitmap( FO_ICON_MAGNIFY, wxART_OTHER, wxSize( 14,14 ) ) );
    m_dirList->AddBitmap( bmp );
    m_dirList->SetSelection( m_dirList->GetCount() - 1 );
}

void FileOpenDlg::OnTimer( wxTimerEvent& event )
{
    if ( m_canClose )
    {
        FileDialogSettings::setDlgSize( GetSize() );
        FileDialogSettings::setSorting( m_currColOrder.column, m_currColOrder.ascending );
        Destroy();
    }
}
