#include "GPSConfigDialog.h"

#include <wx/statline.h>

#include "../DirHelper.h"
#include "../SerialPortHelper.h"

#include "../ConfigBase.h"
#include "../configmanager/ConfigHelper2.h"
#include "../configmanager/SerialConfigMgr.h"

#include "../serialporthandler.h"
#include "../SerialException.h"

#include "../logger.h"
#include "../VemsTuneController.h"

GPSDetectProcess::GPSDetectProcess( wxArrayString comports, wxArrayString baudrates )
{
    m_comports = comports;
    long temp = 0;

    for ( size_t i = 0; i < baudrates.GetCount(); ++i )
    {
        if ( baudrates[i].ToLong( &temp ) )
            m_baudrates.Add( temp );
    }
    m_t = m_baudrates.GetCount() * m_comports.GetCount();
    m_running = false;
    m_stopped = false;
}

bool GPSDetectProcess::process()
{
    if ( m_running || m_stopped )
        return false;

    m_running = true;

    int state = 0;

    SerialConfig scfg;

    for ( size_t i = 0; i < m_comports.GetCount(); ++i )
    {
        wxString comport = m_comports[i];

        scfg.setPortName( comport );

        CommunicationBase* combase = NULL;

        try {

            for ( size_t j = 0; j < m_baudrates.GetCount(); ++j )
            {
//                Logger::logf(LOG_MSG, _T("Stopped=%d"), m_stopped);
                if ( m_stopped )
                {
                    if ( combase )
                    {
                        delete combase;
                    }
                    return false;
                }

                int br = m_baudrates[j];

                scfg.setBaudRate( br );

                combase = new SerialPortHandler( &scfg );

                wxString rStr = readPort( combase );

                // valid GPS found
                if ( rStr.Find(_T("$GPGGA")) != wxNOT_FOUND )
                {
                    // parse it
                    wxString posStr = parseGPSString( rStr );

                    combase->clear();
                    combase->close();

                    delete combase;

                    wxVariant vsender = (void*)this;
                    wxVariant vstatus = 0L;
                    wxVariant vcomport = comport;
                    wxVariant vbr = (long)br;
                    wxVariant vpos = posStr;

                    wxVariantHash hash;
                    hash[_T("sender")] = vsender;
                    hash[_T("status")] = vstatus;
                    hash[_T("comport")] = vcomport;
                    hash[_T("baudrate")] = vbr;
                    hash[_T("position")] = vpos;
                    observerNotify( hash );

                    observerNotify( m_t );
                    stop();

                    return false;
                }

                ++state;

                combase->clear();
                combase->close();

                delete combase;
                combase = NULL;

                wxVariant vsender = (void*)this;
                wxVariant vstatus = -2L;
                wxVariant vcomport = comport;
                wxVariant vbr = (long)br;

                wxVariantHash hash;

                hash[_T("sender")] = vsender;
                hash[_T("status")] = vstatus;
                hash[_T("comport")] = vcomport;
                hash[_T("baudrate")] = vbr;
                observerNotify( hash );
                observerNotify( state );
            }
        }
        catch ( SerialException e )
        {
//            Logger::logf(LOG_MSG, _T("GPSDetectProcess: ") + comport + _T(" unavailable") );
            state += m_baudrates.GetCount();

            wxVariant vstatus = -1L;
            wxVariant vcomport = comport;
            wxVariant vsender = (void*)this;

            wxVariantHash hash;
            hash[_T("sender")] = vsender;
            hash[_T("status")] = vstatus;
            hash[_T("comport")] = vcomport;
            observerNotify( hash );
            observerNotify( state );
        }

        if ( combase )
        {
            delete combase;
        }
        if ( m_stopped )
        {
            return false;
        }

    }
    observerNotify( m_t );
    return false;
}

void GPSDetectProcess::stop()
{
    m_running = false;
    m_stopped = true;
    setDisabled( true );
    ProcessBase::stop();
}

void GPSDetectProcess::finished()
{
    setDisabled( true );
    m_running = false;
    m_stopped = true;
}

wxString GPSDetectProcess::readPort( CommunicationBase * cbase )
{
    uint8_t str[257];

    cbase->clear();
    wxMilliSleep( 1000 );

    // cbase->ReadString( str, 256 );
    cbase->Read( str, 256 );
    wxString readS = wxString::From8BitData( (char*)str, 256 );

    return readS;
}

wxString GPSDetectProcess::parseGPSString( const wxString& str )
{
    size_t len = str.Len();
    NMEAParse parser;
    for ( size_t i = 0; i < len;++i )
    {
        parser.feedGPScharacter( str.GetChar(i) );
    }

    GPSDataStruct data = parser.gpsdata;

    wxString posStr;

    if ( data.noofsatellite == 0 )
    {
        posStr = _T("No satellite in use");
    }
    else
    {
        posStr= wxString::Format( _T("Current position:\n%c%.6f, %c%.6f"),
            (char)data.latitudeindicator, ( data.latitude / (1000.0 * 1000.0 ) ),
            (char)data.longnitudeindicator, ( data.longnitude / (1000.0 * 1000.0 ) ) );
    }

    return posStr;
}

///

int idGPSDetect = wxNewId();
int idGPSSave   = wxNewId();
int idGPSTest   = wxNewId();
int idGPSClose  = wxNewId();
int idGPSCancel = wxNewId();

BEGIN_EVENT_TABLE( GPSConfigDialog, wxDialog )
    EVT_BUTTON( idGPSDetect, GPSConfigDialog::OnDetect )
    EVT_BUTTON( idGPSSave, GPSConfigDialog::OnSave )
    EVT_BUTTON( idGPSTest, GPSConfigDialog::OnTest )
    EVT_BUTTON( idGPSClose, GPSConfigDialog::OnClose )
    EVT_BUTTON( idGPSCancel, GPSConfigDialog::OnCancelButton )
END_EVENT_TABLE()

static int CompareStringLen(const wxString& first, const wxString& second)
{
    if ( first.length() != second.length() )
        return first.length() - second.length();
    else return first.CmpNoCase(second);
}

GPSConfigDialog::GPSConfigDialog( wxWindow* parent, VemsTuneController* vtc ) :
    wxDialog( parent, wxID_ANY, _T("GPS Communication settings"), wxDefaultPosition,
    wxSize(450, 220), wxSTAY_ON_TOP | wxDEFAULT_DIALOG_STYLE )
{
    SetIcon( wxIcon( _T("vemsLogoIcon2") ) );

    m_vtc = vtc;

    m_wasECUOnline = m_vtc->isECUOnline();

    m_vtc->setECUOnline( false, true );
    m_vtc->setGPSOnline( false );


    ConfigBase* configBase = ConfigHelper2::getConfig();
    SerialConfigMgr cfgMgr(configBase, _T("/gps"));
    cfgMgr.load(&m_serialConfig);

    m_baudRateS.Add(_T("600")); m_baudRateS.Add(_T("1200"));m_baudRateS.Add(_T("2400"));
    m_baudRateS.Add(_T("4800")); m_baudRateS.Add(_T("9600"));m_baudRateS.Add(_T("14400"));
    m_baudRateS.Add(_T("19200")); m_baudRateS.Add(_T("38400"));m_baudRateS.Add(_T("57600"));
    m_baudRateS.Add(_T("115200"));

    if ( m_baudRateS.Index( m_serialConfig.getBaudRateStr() ) == wxNOT_FOUND )
    {
        m_baudRateS.Add(m_serialConfig.getBaudRateStr() );
    }

    m_process = NULL;
    m_saved = false;

    m_bmp.LoadFile( DirHelper::GetPath(dICON) + _T("icon_sat.png"), wxBITMAP_TYPE_PNG );

    buildGUI();
}

GPSConfigDialog::~GPSConfigDialog()
{
    LOG
    //if ( m_config != NULL )
    //    delete m_config;

    if ( m_wasECUOnline )
         m_vtc->setECUOnline( true, true );

    if ( m_found && m_saved && ( m_enabled->GetValue() == _T("yes") ) )
    {
        m_vtc->setGPSOnline( true, true );
    }
}

void GPSConfigDialog::buildGUI()
{
    wxStaticBoxSizer* sizer = new wxStaticBoxSizer( wxVERTICAL, this, _T("GPS Settings") );

    wxFlexGridSizer* fgs = new wxFlexGridSizer( 2 );

    wxStaticText* dis = new wxStaticText( this, -1, _T("GPS enabled:") );
    wxArrayString enArr; enArr.Add(_T("no")); enArr.Add(_T("yes"));

    wxString msg;
    if( m_serialConfig.isEnabled() ){ msg = _T("yes");}
    else{ msg = _T("no"); }

    m_enabled = new wxComboBox( this, -1, msg, wxDefaultPosition, wxDefaultSize, enArr, wxCB_DROPDOWN | wxCB_READONLY );

    Connect( m_enabled->GetId(), wxEVT_COMMAND_COMBOBOX_SELECTED, wxCommandEventHandler( GPSConfigDialog::OnGPSEnabledCombo ) );

    fgs->Add( dis, 0, wxALL, 3 ); fgs->Add( m_enabled, 1, wxEXPAND | wxALL, 3 );

    wxStaticText* txt = new wxStaticText( this, -1, _T("COM port:") );

    wxArrayString ports;
    getComports( ports );
    ports.Sort(CompareStringLen);

    m_comports = new wxComboBox( this, wxID_ANY, m_serialConfig.getPortName(), wxDefaultPosition, wxDefaultSize, ports,
     wxCB_DROPDOWN);

    fgs->Add( txt, 0, wxALL, 3 ); fgs->Add( m_comports, 1, wxEXPAND | wxALL, 3 );

    txt = new wxStaticText( this, -1, _T("Baud rate:") );
    m_baudRate = new wxComboBox( this, wxID_ANY, m_serialConfig.getBaudRateStr(), wxDefaultPosition, wxDefaultSize, m_baudRateS,
     wxCB_DROPDOWN | wxCB_READONLY );

    fgs->Add( txt, 0, wxALL, 3 ); fgs->Add( m_baudRate, 1, wxEXPAND | wxALL, 3 );
    fgs->Add( 10, 10 );

    m_testButton = new wxButton( this, idGPSTest, _T("Test") );

    fgs->Add( m_testButton, 0, wxALL, 3 );
    sizer->Add( fgs, 5, wxEXPAND | wxALL, 5 );

    // -------

    wxBoxSizer* bs = new wxBoxSizer( wxVERTICAL );

    wxStaticLine* line = new wxStaticLine( this, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxLI_HORIZONTAL );

    m_txtDetect = new wxStaticText( this, wxID_ANY, _T("Click on Detect to search for GPS") );
    m_progressbar = new ProgressBar( this, wxID_ANY, wxDefaultPosition, wxSize(220,20) );
    m_progressbar->SetWindowStyle( wxBORDER_SUNKEN );
    m_progressbar->SetProgressStyle( PB_PERCENTTEXT_INNER );
    m_progressbar->SetValueFormat( _T("%.1f %%") );
    m_progressbar->SetValueLimits( 0, 100.0, 1.0 );
    m_progressbar->Update( 0 );

    m_detectButton = new wxButton( this, idGPSDetect, _T("Detect") );

    bs->Add( m_txtDetect, 0, wxALL, 5 );
    bs->Add( line, 0, wxEXPAND | wxLEFT | wxRIGHT | wxALIGN_CENTER, 10 );
    bs->Add( m_progressbar, 0, wxALL | wxEXPAND, 5 );
    bs->Add( m_detectButton, 0, wxALL | wxALIGN_CENTER, 3 );

    wxBoxSizer* rs = new wxBoxSizer( wxHORIZONTAL );
    m_statBmp = new wxStaticBitmap( this, -1, wxNullBitmap, wxDefaultPosition, wxSize( 32, 32 ) );
    m_result = new wxStaticText( this, -1, _T("") );
    rs->Add( m_statBmp, 0, wxALL, 5 );
    rs->Add( m_result, 1, wxEXPAND | wxALL, 5 );

    bs->Add( rs, 0, wxEXPAND | wxALL, 5 );

    wxBoxSizer* bs2 = new wxBoxSizer( wxHORIZONTAL );
    m_saveButton = new wxButton( this, idGPSSave, _T("Save") );
    m_closeButton = new wxButton( this, idGPSClose, _T("Close") );
    m_cancelButton = new wxButton( this, idGPSCancel, _T("Cancel") );
    m_cancelButton->Enable( false );

    bs2->Add( m_saveButton, 0, wxALL | wxALIGN_CENTER, 3 );
    bs2->Add( m_closeButton, 0, wxALL | wxALIGN_CENTER, 3 );
    bs2->Add( m_cancelButton, 0, wxALL | wxALIGN_CENTER, 3 );

    bs->Add( bs2, 0, wxALL | wxALIGN_CENTER, 5 );

    sizer->Add( bs, 0, wxEXPAND | wxALL, 5 );

    SetSizer( sizer );

    sizer->Fit( this );
    sizer->SetSizeHints( this );

    if ( !m_serialConfig.isEnabled() )
    {
        m_testButton->Enable( false );
        m_baudRate->Enable( false );
        m_comports->Enable( false );
        m_detectButton->Enable( false );
    }
}

void GPSConfigDialog::OnClose( wxCommandEvent& event )
{
    EndModal( 0 );
}

void GPSConfigDialog::OnSave( wxCommandEvent& event )
{

    bool isEnabled = ( m_enabled->GetValue() == _T("yes") ) ? true : false;
    wxString baudrateStr = m_baudRate->GetValue();
    long baudRate;
    baudrateStr.ToLong(&baudRate);

    m_serialConfig.setEnabled(isEnabled);
    m_serialConfig.setPortName(m_comports->GetValue());
    m_serialConfig.setBaudRate(baudRate);

    ConfigBase* configBase = ConfigHelper2::getConfig();
    SerialConfigMgr cfgMgr(configBase, _T("/gps"));
    cfgMgr.save(&m_serialConfig);

/*
    m_config->write( _T("/gps/enabled"), gpsEnabled );
    m_config->write( _T("/gps/port"), m_comports->GetValue() );
    m_config->write( _T("/gps/baudrate"), m_baudRate->GetValue() );

    m_config->save();
*/
    m_saved = true;

    //delete m_config;
    //m_config = NULL;

    EndModal( 0 );
}

void GPSConfigDialog::OnDetect( wxCommandEvent& event )
{
    if ( m_process != NULL )
    {
        delete m_process;
    }

    m_found = false;
    m_saved = false;

    m_process = new GPSDetectProcess( m_comports->GetStrings(), m_baudRateS );

    m_process->registerObserver( this );

    m_process->start();

    enableButtons( false );
}

void GPSConfigDialog::OnTest( wxCommandEvent& event )
{
    long baudrate = 0;
    m_baudRate->GetValue().ToLong( &baudrate );

    SerialConfig scfg;
    wxString comport = m_comports->GetValue();
    // Logger::logf(LOG_MSG, _T("comport =") + comport + _T(", bd= %d"), baudrate );
    scfg.setPortName( comport ); scfg.setBaudRate( baudrate );

    CommunicationBase* combase = NULL;  /// to suppress uninicialized warning
    combase = new SerialPortHandler( &scfg );

    m_found = false;
    m_saved = false;

    try {

        uint8_t str[257];

        combase->clear();
        wxMilliSleep( 1000 );

        combase->ReadString( str, 256 );
        wxString readS = wxString::From8BitData( (char*)str, 256 );

        if ( readS.Find(_T("$GPGGA")) != wxNOT_FOUND )
        {
            size_t len = readS.Len();
            NMEAParse parser;
            for ( size_t i = 0; i < len;++i )
            {
                parser.feedGPScharacter( readS.GetChar(i) );
            }

            wxString posStr;

            GPSDataStruct data = parser.gpsdata;

            if ( data.noofsatellite == 0 )
            {
                posStr = _T("No satellite in use");
            }
            else
            {
                posStr= wxString::Format( _T("Current position:\n%c%.6f, %c%.6f"),
                    (char)data.latitudeindicator, ( data.latitude / (1000.0 * 1000.0 ) ),
                    (char)data.longnitudeindicator, ( data.longnitude / (1000.0 * 1000.0 ) ) );
            }

            wxString sStr = _T("GPS found on ") + comport + wxString::Format( _T(", baudrate = %d"), baudrate );
            m_txtDetect->SetLabel( sStr );
            m_result->SetLabel( _T("GPS found. ") + posStr );
            m_statBmp->SetBitmap( m_bmp );
            m_found = true;
        }
        else
        {
            m_txtDetect->SetLabel( comport + _T(" is unavailable") );
            m_result->SetLabel( _T("No GPS found") );
            m_statBmp->SetBitmap( wxNullBitmap );
        }
        m_progressbar->Update( 100.0 );
        combase->clear();
        combase->close();
    }
    catch ( SerialException e )
    {
        m_txtDetect->SetLabel( comport + _T(" is unavailable") );
        m_result->SetLabel( _T("No GPS found") );
        m_statBmp->SetBitmap( wxNullBitmap );
        m_progressbar->Update( 100.0 );
    }
    delete combase;
}

void GPSConfigDialog::OnCancelButton( wxCommandEvent& event )
{
    if ( m_process )
    {
        m_process->stop();
        delete m_process;
        m_process = NULL;
        m_progressbar->Update( 100.0 );
        m_txtDetect->SetLabel( _T("Cancelled") );
        m_result->SetLabel( _T("GPS not found") );
        m_statBmp->SetBitmap( wxNullBitmap );
        enableButtons( true );
    }
}

void GPSConfigDialog::getComports( wxArrayString& ports )
{
    ports = SerialPortHelper::GetCOMPortList();

    if ( ports.Index( m_serialConfig.getPortName(), false ) == wxNOT_FOUND )
    {
        ports.Add( m_serialConfig.getPortName() );
        Logger::logf(LOG_MSG, _T("GPSConfigDialog::getComports() port name: ")+  m_serialConfig.getPortName());
    }
}

void GPSConfigDialog::enableButtons( bool enable )
{
    m_testButton->Enable( enable );
    m_detectButton->Enable( enable );
    m_saveButton->Enable( enable );
    m_closeButton->Enable( enable );
    m_cancelButton->Enable ( !enable );
    m_enabled->Enable( enable );
}

void GPSConfigDialog::notify( int i )
{
    size_t maxR = m_comports->GetCount() * m_baudRateS.GetCount();
    double percent = (double) i  * 100.0 / (double) maxR;
    m_progressbar->Update( percent );

    if ( i >= (int)maxR )
    {
        m_process->unRegisterObserver( this );
        enableButtons( true );

        if ( m_found )
        {
            m_statBmp->SetBitmap( m_bmp );
        }
        else
        {
            m_statBmp->SetBitmap( wxNullBitmap );
            m_result->SetLabel( _T("GPS not found") );
        }
    }
}

void GPSConfigDialog::notify( wxVariantHash& hash )
{
    wxVariantHash::const_iterator it;

    wxVariant var = hash[_T("sender")];

    if ( var.GetVoidPtr() != m_process )
        return;

    it = hash.find( _T("status") );

    if ( it == hash.end() )
    {
        return;
    }

    int status = hash[_T("status")].GetLong();

    // com port unavailable
    if ( status == -1 )
    {
        wxString comStr = hash[_T("comport")].GetString();
        m_txtDetect->SetLabel( comStr + _T(" is unavailable") );
        return;
    }

    // GPS not found
    if ( status == -2 )
    {
        wxString comStr = hash[_T("comport")].GetString();
        wxString br = wxString::Format( _T("%d"), hash[_T("baudrate")].GetLong() );

        wxString str = _T("GPS not found on ") + comStr + _T(", baudrate = ") + br;
        m_txtDetect->SetLabel( str );
        return;
    }

    // gps found
    if ( status == 0 )
    {
        m_found = true;

        wxString comStr = hash[_T("comport")].GetString();
        wxString br = wxString::Format( _T("%d"), hash[_T("baudrate")].GetLong() );
        wxString posStr = hash[_T("position")].GetString();

        wxString str = _T("GPS found on ") + comStr + _T(", baudrate = ") + br;
        m_txtDetect->SetLabel( str );

        m_result->SetLabel( _T("GPS found. ") + posStr );

        m_comports->SetValue( comStr );
        m_baudRate->SetValue( br );
    }
}

void GPSConfigDialog::OnGPSEnabledCombo( wxCommandEvent& event )
{
    if ( m_enabled->GetValue() == _T("yes") )
    {
        m_testButton->Enable( true );
        m_baudRate->Enable( true );
        m_comports->Enable( true );
        m_detectButton->Enable( true );
    }
    else
    {
        m_testButton->Enable( false );
        m_baudRate->Enable( false );
        m_comports->Enable( false );
        m_detectButton->Enable( false );
    }
}
