#include "FilePlayCommander.h"

#include "../ringbuffer/ThreadLogPlayer.h"
#include "../VemsTuneController.h"
#include "../loghistory/LogHistoryMgr.h"
#include "../logger.h"
#include "../toolbartools/TbTbPlay.h"
#include "../events/GeneralEvent.h"

BEGIN_EVENT_TABLE( FilePlayCommander, wxEvtHandler )
    EVT_SLIDERPOS( wxID_ANY, FilePlayCommander::OnSliderPos )
	EVT_GENERAL ( wxID_ANY, FilePlayCommander::OnGeneralEvent )
END_EVENT_TABLE()

IMPLEMENT_DYNAMIC_CLASS( FilePlayCommander, wxEvtHandler )

FilePlayCommander::FilePlayCommander( VemsTuneController* vtc )
{
    m_vtc = vtc;
    m_logPlayer = NULL;
    m_isPaused = false;
    m_speed = 1;
}

FilePlayCommander::~FilePlayCommander()
{
    if ( m_logPlayer )
    {
        if ( m_logPlayer->isRunning() )
        {
            m_logPlayer->stop();
        }
        delete m_logPlayer;

        m_logPlayer = NULL;
    }
}

void FilePlayCommander::Play()
{
    if ( m_logPlayer )
    {
        if ( m_isPaused && m_logPlayer->isRunning() )
        {
            m_logPlayer->resume();
        }
        else
        {
            m_logPlayer->start();
        }

        m_isPaused = false;
    }

    enableStopTool( true );
}

void FilePlayCommander::Pause()
{
    if ( m_logPlayer && m_isPaused == false )
    {
        if ( m_logPlayer->isRunning() )
        {
            m_logPlayer->pause();
            m_isPaused = true;
        }
    }
}

void FilePlayCommander::Stop()
{
    if ( m_logPlayer && m_logPlayer->isRunning() )
    {
        m_logPlayer->stop();
    }

    setStoppedBitmaps();
}

void FilePlayCommander::setStoppedBitmaps()
{
    ToolHash::iterator it = m_tools.find(_T("Play"));
    if ( it != m_tools.end() )
    {
        if ( it->second != NULL )
        {
            TbTbPlay* tb = dynamic_cast<TbTbPlay*>( it->second );
            if ( tb )
            {
                tb->SetToolBmp( tb->m_bmpHash[ TBI_PLAY_ACTIVE ] );
                tb->m_shortHelpString = _T("Play");
            }
        }
    }

    m_isPaused = false;
}

bool FilePlayCommander::IsPlaying() const
{
    if ( m_logPlayer == NULL )
        return false;

    return ( m_logPlayer->isRunning() && !m_isPaused );
}

void FilePlayCommander::enableTools( bool enabled )
{
    enablePlayTool( enabled );
    enableStopTool( enabled );
    enableSpeedTool( enabled );
    enableRepeatTool( enabled );
}

void FilePlayCommander::Reset()
{
    if ( m_logPlayer )
    {
        if ( m_logPlayer->isRunning() )
        {
            m_logPlayer->stop();
        }
        delete m_logPlayer;

        m_logPlayer = NULL;
    }

    setStoppedBitmaps();

    LogHistoryMgr* lhmgr = m_vtc->getLogHistoryMgr();

    m_logPlayer = new ThreadLogPlayer( this, m_vtc->getLogHistoryMgr(), m_vtc );

    if ( lhmgr->getHistoryCount() < 1 )
    {
        enableTools( false );
    }

    m_isPaused = false;
}

void FilePlayCommander::enablePlayTool( bool enabled )
{
    if ( m_vtc->getLogHistoryMgr()->getHistoryCount() < 1 && enabled )
        return;

    ToolHash::iterator it = m_tools.find(_T("Play"));

    if ( it == m_tools.end() )
        return;

    VemsToolBarToolBase* tb = it->second;

    if ( tb )
    {
        tb->GetToolBar()->EnableTool( tb->GetId(), enabled );
    }

    enableRepeatTool( true );
}

void FilePlayCommander::enableRepeatTool( bool enabled )
{
    if ( m_vtc->getLogHistoryMgr()->getHistoryCount() < 1 && enabled )
        return;

    ToolHash::iterator it = m_tools.find(_T("Repeat"));

    if ( it == m_tools.end() )
        return;

    VemsToolBarToolBase* tb = it->second;

    if ( tb )
    {
        tb->GetToolBar()->EnableTool( tb->GetId(), enabled );
    }
}

void FilePlayCommander::enableSpeedTool( bool enabled )
{
    if ( m_vtc->getLogHistoryMgr()->getHistoryCount() < 1 && enabled )
        return;

    ToolHash::iterator it = m_tools.find(_T("PlaySpeed"));

    if ( it == m_tools.end() )
        return;

    VemsToolBarToolBase* tb = it->second;

    if ( tb )
    {
        tb->GetToolBar()->EnableTool( tb->GetId(), enabled );
    }
}

void FilePlayCommander::enableStopTool( bool enabled )
{
    ToolHash::iterator it = m_tools.find(_T("Stop"));

    if ( it == m_tools.end() )
        return;

    VemsToolBarToolBase* tb = it->second;

    if ( tb )
    {
        tb->GetToolBar()->EnableTool( tb->GetId(), enabled );
    }
}

void FilePlayCommander::registerTool( const wxString& toolname, VemsToolBarToolBase* tool )
{
    m_tools[toolname] = tool;
}

void FilePlayCommander::unregisterTool( VemsToolBarToolBase* tool )
{
    ToolHash::iterator it = m_tools.begin();

    for ( ; it != m_tools.end(); ++it )
    {
        if ( it->second == tool )
        {
            m_tools[it->first] = NULL;
            return;
        }
    }
}

void FilePlayCommander::unregisterTool( const wxString& toolname )
{
    ToolHash::iterator it = m_tools.find( toolname );

    if ( it == m_tools.end() )
        return;

    m_tools[toolname] = NULL;
}

void FilePlayCommander::OnSliderPos( SliderPosEvent& event )
{
    if ( m_logPlayer )
    {
        m_logPlayer->setPosition( event.getCurrentPos() );
    }

    event.Skip();
}

void FilePlayCommander::setReplaySpeed( const unsigned char& speed )
{
    m_speed = speed;

    if ( m_logPlayer )
    {
        m_logPlayer->setSpeed( m_speed );
    }
}

void FilePlayCommander::setRepeat( bool repeat )
{
    m_repeat = repeat;

    if ( m_logPlayer )
    {
        m_logPlayer->setRepeat( m_repeat );
    }

}

void FilePlayCommander::OnGeneralEvent( GeneralEvent& event )
{
	if ( event.GetString() == _T("SET_STOPPED") )
	{
		setStoppedBitmaps();
	}

	event.Skip();
}
