#include "RTDataStat.h"

#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/stats.hpp>
#include <boost/accumulators/statistics/mean.hpp>
#include <boost/accumulators/statistics/max.hpp>
#include <boost/accumulators/statistics/min.hpp>
#include <boost/accumulators/statistics/count.hpp>
#include <boost/accumulators/statistics/variance.hpp>
#include <boost/accumulators/statistics/density.hpp>

#include <wx/tglbtn.h>

#include "../loghistory/LogHistoryProcessFilterThread.h"

#include "../Math.h"
#include "../DirHelper.h"
#include "../logger.h"
#include "../SymbolMgr.h"
#include "../Observable.h"

#include "../gaugedesc/GaugeDescManager.h"

#include "../stdatamanager/STDataMgrBase.h"
#include "../rtdatamanager/RTDataMgrFacade.h"
#include "../rtdatamanager/RTDataMgrBase.h"

#include "../vemsprotocol/SIPRTypes.h"
#include "../vemsprotocol/VemsLogReader.h"
#include "../vemsprotocol/VEMSLogBuffer.h"

#include "../loghistory/LogHistoryBase.h"
#include "../loghistory/LogHistoryMgr.h"

#include "../vemsgauge/VemsTableDesc.h"

#include "../loghistory/ExpressionPanel.h"
#include "../loghistory/LogFilter.h"

#include "../loghistory/ExpressionSQL.h"


using namespace boost::accumulators;

/// constants

const int IDEVALBUTTON  = wxNewId();
const int IDEXPRBUTTON  = wxNewId();
const int IDSPINCTRL    = wxNewId();
const int IDREFRTBL     = wxNewId();
const int IDSELALL      = wxNewId();

/// event handlers

BEGIN_EVENT_TABLE( RTDataStat, wxFrame )

    EVT_RIGHT_DOWN ( RTDataStat::OnRightClick )
    EVT_CLOSE ( RTDataStat::OnClose )
    EVT_BUTTON ( IDEVALBUTTON, RTDataStat::OnEvaluate )
    EVT_BUTTON ( IDREFRTBL, RTDataStat::OnRefreshTables )

    EVT_BUTTON ( IDEXPRBUTTON, RTDataStat::OnExprButton )
    EVT_BUTTON ( IDSELALL, RTDataStat::OnSelectAll )
    EVT_SPINCTRL ( IDSPINCTRL, RTDataStat::OnSpinCtrl )

    // notify
    EVT_SLIDER ( wxID_ANY, RTDataStat::OnNotifyEvtHandler )
//    EVT_SET_FOCUS ( RTDataStat::OnFocus )

END_EVENT_TABLE();

RTDataStat::RTDataStat(wxWindow * parent, LogHistoryBase* loghistory, Observable* loghistoryobs, VemsTableDesc* tableDesc, bool isTuneMode)
{
    m_useWeighting = false;
    m_useOutlierFiltering = true;

    m_tuneMode = isTuneMode;
    m_loghistory = loghistory;
    m_loghistoryobserv = loghistoryobs;
    m_stDataMgr = m_loghistory->getSTDataMgr();
    m_symbolMgr = m_loghistory->getSymbolMgr();

    m_tableDesc = tableDesc;
    m_xBinSymbol = m_symbolMgr->getSymbol( m_tableDesc->xSTSymbol );
    m_yBinSymbol = m_symbolMgr->getSymbol( m_tableDesc->ySTSymbol );

    m_xRTSymbol = m_symbolMgr->getSymbol( m_tableDesc->xRTSymbol );
    m_yRTSymbol = m_symbolMgr->getSymbol( m_tableDesc->yRTSymbol );

    m_cols = (unsigned int) m_xBinSymbol->getShape0();
    m_rows = (unsigned int) m_yBinSymbol->getShape0();

    m_ratioMatrix = NULL;

    m_popupmenu = NULL;
    m_statMatrix = NULL;
    m_throwedDataM = NULL;
    m_outliersDataM = NULL;

    m_index = 0;
    m_logHistoryProcessFilterThread = NULL;
    m_data.Clear();

    m_gdmanager = GaugeDescHelper::GetGDMgr();

    m_xmin = m_ymin = -99999999;
    m_xmax = m_ymax = 99999999;
    m_mincount = 1; m_maxcount = 9000000;

    m_requiredSymbol = _T("lambda");

//    m_stDataMgr->RegisterViewer(this);

    m_guitable = NULL;
    m_progressdlg = NULL;

    m_filter = _T("(pulseWidth > 0) && (lambda < 2) && (lambda > 0.1) && (coolant > 70)");
    // m_expression = _T("lambda * egoCorrection - lambdatarget*100");
    m_expression = _T("(veCurr * egoCorrection * lambda) / (lambdatarget * 100)");

    m_displaycell = invalid_cell;
    m_selBottomRight = invalid_cell;
    m_selTopLeft = invalid_cell;

    m_start = 0;
    m_end = -1;

    m_registered = false;

    m_prec_data = NULL;
    m_prec_data_weight = NULL;
    m_prec_data_max = NULL;
    m_prec_data_min = NULL;
    m_prec_data_mean = NULL;
    m_prec_data_variance = NULL;

    m_logHistoryProcessFilterThread = NULL;

    buildGUI( parent );

}

void RTDataStat::load( unsigned int start, int end)
{
    if ( m_loghistory->getHistoryCount() == 0 )
        return;

    wxString vm = m_varianceMultTxt->GetValue();

    if ( vm.ToDouble( & m_varianceMult ) == false )
    {
        m_useOutlierFiltering = false;
    }
    else
    {
        if ( m_varianceMult == 0 )
            m_useOutlierFiltering = false;
        else
            m_useOutlierFiltering = true;
    }

    m_registered = true;
    m_start = start;
    m_end = end;

    m_outofrange = 0;

    long delay;
    wxString delayStr = m_delay->GetValue();
    delayStr.ToLong(&delay);
    m_loghistory->setPacketDelay( delay );

    // parse expression
    StringHashSet symbols = m_rpnParser.getVariables( m_expression );
    StringHashSet::const_iterator IT = symbols.begin();
    for(; IT != symbols.end(); IT++){
        m_requiredSymbols.insert(*IT);
    }

    // parse filter
    symbols = m_rpnParser.getVariables( m_filter );
    IT = symbols.begin();
    for(; IT != symbols.end(); IT++){
        m_requiredSymbols.insert(*IT);
    }

    // register requires symbols
    require();

    // write out required expression
    m_expressionEdit->SetValue( m_expression );

    // register filter to loghistory
    m_loghistory->registerLogFilter(this);

    if ( end == -1 ) end = m_loghistory->getHistoryCount()-1;
    m_data.Clear();

    if ( m_logHistoryProcessFilterThread != NULL )
    {
        m_logHistoryProcessFilterThread->unRegisterObserver(this);
        delete m_logHistoryProcessFilterThread;
    }

    m_showProgdlg = true;

    m_statLen = ( end + 1 - start );
    // Logger::logf( LOG_MSG, _T("n_statLen=%d, start=%d, stop=%d"), m_statLen, m_start, end );

    m_logHistoryProcessFilterThread = new LogHistoryProcessFilterThread(m_loghistory, start, end);
    m_logHistoryProcessFilterThread->registerObserver(this);
    m_loghistoryobserv->registerObserver(this);
    m_logHistoryProcessFilterThread->start();
}

RTDataStat::~RTDataStat()
{
    // Logger::logf(LOG_MSG, _T("RTDataStat dtor"));
    if ( m_registered )
    {
        m_logHistoryProcessFilterThread->unRegisterObserver(this);
        m_loghistoryobserv->unRegisterObserver(this);
    }

    UnRegisterViewer();
    unregisterSymbols();
    m_guitable->unRegisterObserver(this);

    if ( m_logHistoryProcessFilterThread != NULL )
        delete m_logHistoryProcessFilterThread;

    m_VertLimits.Clear(); m_HorizLimits.Clear();

    delete m_statMatrix;
    delete m_throwedDataM;
    delete m_outliersDataM;

    delete m_prec_data;
    delete m_prec_data_weight;
    delete m_prec_data_max;
    delete m_prec_data_min;
    delete m_prec_data_mean;
    delete m_prec_data_variance;

    if ( m_ratioMatrix )
    {
        delete m_ratioMatrix;
        m_ratioMatrix = NULL;
    }

    m_statMatrix = NULL;
    m_throwedDataM = NULL;
    m_outliersDataM = NULL;

    m_prec_data = NULL;
    m_prec_data_max = NULL;
    m_prec_data_min = NULL;
    m_prec_data_weight = NULL;
    m_prec_data_mean = NULL;
    m_prec_data_variance = NULL;

    m_registered = false;
}

void RTDataStat::refresh(DoubleHash* table)
{
    ///
}

void RTDataStat::require()
{
    wxString xRTName = m_xRTSymbol->getName(); /// instead of rpm
    wxString yRTName = m_yRTSymbol->getName(); /// instead of map

    m_rtDataMgr->registerSymbol( xRTName );
    m_rtDataMgr->registerSymbol( yRTName );

    StringHashSet::const_iterator IT = m_requiredSymbols.begin();
    for(; IT != m_requiredSymbols.end(); IT++){
        wxString symbol = *IT;
        m_rtDataMgr->registerSymbol( symbol );
//        Logger::logf(LOG_MSG, _T("RTDataStat::require ")+symbol +_T(" added"));
    }
//    Logger::logf(LOG_MSG, _T("RTDataStat::require() ended"));
    requireChanged();
}

void RTDataStat::unregisterSymbols()
{
    wxString xRTName = m_xRTSymbol->getName(); /// instead of rpm
    wxString yRTName = m_yRTSymbol->getName(); /// instead of map

    m_rtDataMgr->unregisterSymbol( xRTName );
    m_rtDataMgr->unregisterSymbol( yRTName );

    StringHashSet::const_iterator IT = m_requiredSymbols.begin();
    for(; IT != m_requiredSymbols.end(); IT++){
        wxString symbol = *IT;
        m_rtDataMgr->unregisterSymbol( symbol );
    }
//    Logger::logf(LOG_MSG, _T("RTDataStat::unregisterSymbols() ended"));
}

void RTDataStat::buildGUI(wxWindow* parent)
{
    wxString title = ( m_tuneMode ) ? _T("VE Tune by statistics") : _T("Real-time data statistics");
    Create(parent, wxID_ANY, title, wxDefaultPosition, wxDefaultSize, wxDEFAULT_FRAME_STYLE | wxFRAME_FLOAT_ON_PARENT );

    SetIcon(wxIcon(_T("vemsLogoIcon2")));

    SetBackgroundColour( wxSystemSettings::GetColour( wxSYS_COLOUR_3DFACE ) );
    ClearBackground();

    wxBoxSizer * sizer = new wxBoxSizer(wxVERTICAL);

    /// TOP sizer
    wxPanel* toppanel = new wxPanel(this, -1);
    wxFlexGridSizer * topsizer = new wxFlexGridSizer(2,5,5);

    m_expressionText = new wxStaticText(toppanel, -1, _T("Expression:") );
    m_expressionEdit = new wxComboBox(toppanel, -1, m_expression );

    // expression list elements
    wxArrayString exprlist;
    StringHashA sqllist = ExpressionSQL::getExpressions(false);
    StringHashA::const_iterator it = sqllist.begin();
    for ( ; it!= sqllist.end();++ it)
    {
        exprlist.Add((**it)[_T("expr")]);
    }
    m_expressionEdit->Append(exprlist);
    WX_CLEAR_ARRAY( sqllist );

    // in Tune mode editing expression is not enabled
    if ( m_tuneMode ) m_expressionEdit->Enable( false );

    //

    topsizer->Add(m_expressionText);topsizer->Add(m_expressionEdit, 1, wxEXPAND );

    m_filterText = new wxStaticText(toppanel, -1, _T("Filter:") );
    m_filterEdit = new wxComboBox(toppanel, -1);

    // filter list elements
    wxArrayString filterlist;
    sqllist = ExpressionSQL::getFilters(false);
    it = sqllist.begin();
    for ( ; it!= sqllist.end();++ it)
    {
        filterlist.Add((**it)[_T("expr")]);
    }
    m_filterEdit->Append(filterlist);
    WX_CLEAR_ARRAY( sqllist );
    //

    if ( m_filter != wxEmptyString )
        m_filterEdit->SetValue( m_filter );
    else
        m_filterEdit->SetValue( _T("None") );

    topsizer->Add(m_filterText);topsizer->Add(m_filterEdit, 1, wxEXPAND );
    topsizer->AddGrowableCol(1);
    toppanel->SetSizer(topsizer);
    topsizer->Fit(toppanel);
    topsizer->SetSizeHints(toppanel);

    /// LEFT sizer

    wxPanel* leftpanel = new wxPanel(this, -1);
    leftpanel->SetBackgroundColour( wxSystemSettings::GetColour( wxSYS_COLOUR_3DFACE ) );
    leftpanel->ClearBackground();
    wxFlexGridSizer * leftsizer = new wxFlexGridSizer(2,5,5);


    wxArrayString delayA;
    delayA.Add( _T("0") ); delayA.Add( _T("100") ); delayA.Add( _T("200") );
    delayA.Add( _T("250") ); delayA.Add( _T("300") ); delayA.Add( _T("500") );
    delayA.Add( _T("600") );
    wxStaticText* delayTxt = new wxStaticText( leftpanel, -1, _T("Lambda delay (ms)") );
    wxTextValidator validator(wxFILTER_NUMERIC);
    m_delay = new wxComboBox( leftpanel, wxID_ANY, _T("0"), wxDefaultPosition, wxDefaultSize, delayA, wxCB_DROPDOWN, validator );

    wxStaticText* multTxt =  new wxStaticText( leftpanel, -1, _T("Variance multiplier") );

    wxTextValidator validator2( wxFILTER_NUMERIC | wxFILTER_EXCLUDE_CHAR_LIST );
    wxArrayString val_exc; val_exc.Add(_T("-") ); val_exc.Add(_T("+") );
    validator2.SetExcludes( val_exc );
    m_varianceMultTxt = new wxTextCtrl( leftpanel, -1, _T("2.0"), wxDefaultPosition, wxDefaultSize, 0, validator2 );
    multTxt->SetToolTip(_T("Apply 0 to disable outlier filtering") );

    m_dataInfoText = new wxStaticText(leftpanel, -1, _T("Filtering info: "));
    m_dataInfoEdit = new wxTextCtrl( leftpanel, -1, _T(""), wxDefaultPosition, wxSize(120,50), wxTE_READONLY | wxTE_MULTILINE );

    wxButton* evalButton = new wxButton(leftpanel, IDEVALBUTTON, _T("(Re-)Evaluate"));

    wxButton* exprButton = new wxButton(leftpanel, IDEXPRBUTTON, _T("Edit filters & expressions..."));
    if ( m_tuneMode ) exprButton->Enable( false );

    wxStaticText* mincountTxt = new wxStaticText(leftpanel, -1, _T("Min. # of data in cell"));
    m_minDataSpin = new wxSpinCtrl( leftpanel, IDSPINCTRL, _T(""), wxDefaultPosition, wxDefaultSize,
        wxSP_ARROW_KEYS, 1, m_maxcount, m_mincount);

    leftsizer->Add( delayTxt ); leftsizer->Add( m_delay, 1, wxEXPAND );
    leftsizer->Add( multTxt ); leftsizer->Add( m_varianceMultTxt, 1, wxEXPAND );
    leftsizer->Add(m_dataInfoText);leftsizer->Add(m_dataInfoEdit, 1, wxEXPAND );
    leftsizer->Add(mincountTxt);leftsizer->Add(m_minDataSpin, 1, wxEXPAND );
    leftsizer->Add( evalButton, 0, wxALL, 5 );
    leftsizer->Add( exprButton, 0, wxALL, 5 );

    if ( m_tuneMode )
    {
        wxButton* refrButton = new wxButton( leftpanel, IDREFRTBL, _T("Tune VE Table"));
        leftsizer->Add( refrButton, 0, wxALL, 5 );

        wxButton* selAllButton = new wxButton( leftpanel, IDSELALL, _T("Select/unselect all"));
        leftsizer->Add( selAllButton, 0, wxALL, 5 );
    }
    else
    {
        m_showVEratio = NULL;
    }

    leftsizer->AddGrowableCol(1);

    wxSizer* leftsizer2 = new wxBoxSizer( wxVERTICAL );
    leftsizer2->Add( leftsizer );

    if ( m_tuneMode )
    {
        m_showVEratio = new wxToggleButton( leftpanel, wxID_ANY, _T("Show VEcorr/VE ratio (in %)") );
        Connect( m_showVEratio->GetId(), wxEVT_COMMAND_TOGGLEBUTTON_CLICKED, wxCommandEventHandler( RTDataStat::OnShowVEcorrRation ) );

        leftsizer2->Add( m_showVEratio, 0, wxALL, 5 );
    }

    leftpanel->SetSizer(leftsizer2);
    leftsizer2->Fit(leftpanel);
    leftsizer2->SetSizeHints(leftpanel);

    /// RIGHT sizer

    wxBoxSizer * rightsizer = new wxBoxSizer(wxVERTICAL);
    m_cellinfo = new RTCellInfo(this, wxID_ANY, wxDefaultPosition, wxSize(220,100));
    rightsizer->Add( m_cellinfo, 0, wxALL| wxALIGN_RIGHT, 5);

    wxBoxSizer* uppersizer = new wxBoxSizer(wxHORIZONTAL);
    uppersizer->Add(leftpanel, 1, wxALL | wxEXPAND, 5);
    uppersizer->Add(rightsizer, 1, wxALL | wxALIGN_RIGHT, 5);

    sizer->Add(toppanel, 0, wxALL | wxEXPAND, 5);
    sizer->Add(uppersizer, 0, wxALL | wxEXPAND, 5);

    wxArrayString xbins, ybins;


    // row names
    int i;
    short prev_value = 0;
    for ( i = (int)(m_rows-1); i >=0; --i )
    {
        short value = (short) m_stDataMgr->getDataFromViewerModel(*m_yBinSymbol, 0, i);
        if ( i < (int)(m_rows-1) )
        {
            short limit = (value + prev_value) / 2;
            m_HorizLimits.Add(limit);
        }

        if ( i == (int)(m_rows-2) )
            m_ymax = prev_value + ( prev_value - value );

        if ( i == 0 )
            m_ymin = value - (prev_value - value);

        prev_value = value;

        ybins.Add(wxString::Format( _T("%d"), value ));

    }

    // col names
    for ( i = 0; i < (int)m_cols; ++i )
    {
        short value = (short) m_stDataMgr->getDataFromViewerModel(*m_xBinSymbol,0,i);

        if ( (int)i > 0 )
        {
            short limit = (value + prev_value) / 2;
            m_VertLimits.Add(limit);
        }

        if ( i == (int)(m_cols-1) )
            m_xmax = value + ( value - prev_value );

        if ( i == 1 )
            m_xmin = prev_value - (value - prev_value);

        prev_value = value;

        xbins.Add(wxString::Format( _T("%d"), value ));
    }

    if ( m_statMatrix == NULL )
    {
        m_statMatrix = new AccSetMatrix(m_rows,m_cols);
    }

    if ( m_throwedDataM == NULL )
    {
        m_throwedDataM = new SimpleIntMatrix(m_rows, m_cols);
    }

    if ( m_outliersDataM == NULL )
    {
        m_outliersDataM = new SimpleIntMatrix(m_rows, m_cols);
    }

    // m_prec_xxx -> filter outlier values
    if ( m_prec_data == NULL )
    {
        m_prec_data = new VemsMatrix<DoubleArray>( m_rows, m_cols );
    }

    if ( m_prec_data_weight == NULL )
    {
        m_prec_data_weight = new VemsMatrix<DoubleArray>( m_rows, m_cols );
    }

    if ( m_prec_data_max == NULL )
    {
        m_prec_data_max = new VemsMatrix<double>( m_rows, m_cols );
    }

    if ( m_prec_data_min == NULL )
    {
        m_prec_data_min = new VemsMatrix<double>( m_rows, m_cols );
    }

    if ( m_prec_data_mean == NULL )
    {
        m_prec_data_mean = new VemsMatrix<double>( m_rows, m_cols );
    }

    if ( m_prec_data_variance == NULL )
    {
        m_prec_data_variance = new VemsMatrix<double>( m_rows, m_cols );
    }

    //

    m_guitable = new RTDataStatTable(this,500,300, xbins, ybins, m_requiredSymbol, m_tuneMode );
    m_guitable->setBgColor(wxSystemSettings::GetColour(wxSYS_COLOUR_3DFACE));
    m_guitable->setMinCount( m_mincount );

    // changing statistical feature, selecting cell/group
    m_guitable->registerObserver(this);

    m_stDataMgr->RegisterTableViewer( m_guitable );
    m_guitable->SetDataMgr( m_stDataMgr );

    sizer->Add(m_guitable, 1, wxEXPAND);

    SetSizer( sizer );
    sizer->Fit(this); sizer->SetSizeHints(this);

}


void RTDataStat::insertIntoMatrix( int xval, int yval, double rtDataValue, bool passedFilter)
{
    // search for matrix col by rpm value
    uint8_t i;
    int8_t matcol, matrow; matcol = matrow = -1;

    if ( xval > m_xmax || xval < m_xmin || yval > m_ymax || yval < m_ymin )
    {
        ++m_outofrange;
//        Logger::logf(LOG_MSG, _T("RTDataStat::insertIntoMatrix - cols/row out of range"));
        return;
    }

    for ( i = 0; i < m_cols - 1; ++i )
    {
        if ( xval < m_VertLimits[i])
        {
            matcol = i;
            break;
        }
    }

    for ( i = 0; i < m_rows - 1; ++i )
    {
        if ( yval > m_HorizLimits[i])
        {
            matrow = i;
            break;
        }
    }

    matcol = ( matcol != -1 ) ? matcol : m_cols - 1;
    matrow = ( matrow != -1 ) ? matrow : m_rows - 1;

    // weight?
    double overallWeight = 1.0;

    if ( m_useWeighting )
    {
        double rowmed, colmed;
        rowmed = m_stDataMgr->getDataFromViewerModel(*m_yBinSymbol,0,m_rows-1 - matrow );
        colmed = m_stDataMgr->getDataFromViewerModel(*m_xBinSymbol,0,matcol);

        double temp = ( matcol == 0 ) ? m_xmin : m_VertLimits[matcol-1];
        double xw = (1.0 - (fabs( (double)xval - colmed ) / fabs( colmed - temp )) ) / 2.0 + 0.5;

        temp = ( matrow == (int8_t)m_rows - 1 ) ? m_ymin : m_HorizLimits[matrow];
        double yw = (1.0 - (fabs( (double)yval - rowmed ) / fabs( rowmed - temp )) ) / 2.0 + 0.5;

        overallWeight = xw * yw;

//        Logger::logf(LOG_MSG, _T("%d,%d --> rowmed=%f, yval=%d, temp=%f, w=%f,%f v=%f"), matrow, matcol,
//            rowmed, yval, temp, xw,yw, rtDataValue );
    }

    unsigned matindex = m_prec_data->cols() * matrow + matcol;

    if ( passedFilter )
    {
//        if ( m_useOutlierFiltering )
//        {
            (*m_prec_data)[matindex].Add( rtDataValue );
            (*m_prec_data_weight)[matindex].Add( overallWeight );

//        }
//        else
//        {
//            AccSetT* acccel = m_statMatrix->getData(matindex);
//            (*acccel)(rtDataValue, weight=overallWeight);
//        }
    }
    else{
        (*m_throwedDataM)[matindex]++;
    }
}

void RTDataStat::OnShowVEcorrRation( wxCommandEvent& event )
{
    if ( m_showVEratio == NULL )
    {
        event.Skip();
        return;
    }

    bool on = m_showVEratio->GetValue();

    if ( on )
    {
        m_ratioMatrix = new AccSetMatrix( m_rows, m_cols );

        for ( int i = m_rows - 1; i >= 0; --i )
        {
            for ( size_t j = 0; j < m_cols; ++j )
            {
                AccSetT* vecorr = m_statMatrix->getData( m_rows - 1 - i, j );

                if ( count(*vecorr) > 0 )
                {
                    double vecorr_value = mean(*vecorr);

                    AccSetT* ve_rat_cell = m_ratioMatrix->getData( m_rows - 1 - i, j );
                    Symbol* sym = m_symbolMgr->getSymbol( _T("veTable") );

                    if ( sym )
                    {
                        double ve_currval = m_stDataMgr->getDataFromViewerModel( *sym, i, j );

                        if ( ve_currval != 0 )
                        {
                            double ratio = ( (double)((int)vecorr_value) / (double)((int)ve_currval) - 1.0 ) * 100.0;

                            (*ve_rat_cell)( ratio, weight = 1.0 );
                        }
                    }
                    else
                    {
                        continue;
                    }
                }
                else
                {
                    continue;
                }
            } // for j
        } // for i

        m_guitable->forceNumericFormat( _T("%0.2f") );

        SetColors( 1 );

        m_guitable->setMatrix( m_ratioMatrix );
    }
    else
    {
        if ( m_ratioMatrix )
        {
            delete m_ratioMatrix;
            m_ratioMatrix = NULL;

            m_guitable->forceNumericFormat( _T("%.0f") );

            SetColors( 2 );

            m_guitable->setMatrix( m_statMatrix );
        }
    }
}

void RTDataStat::OnRightClick(wxMouseEvent& event)
{
    if ( m_tuneMode )
    {
        event.Skip();
        return;
    }

    if ( m_popupmenu == NULL )
    {
        m_popupmenu = new wxMenu;
        SymbolGaugeDescArray arr;
        m_gdmanager->GetSymbolDescs(arr);
        unsigned count = arr.GetCount();
        for ( size_t i = 0; i< count; ++i )
        {
            if ( i% (count/4) == 0 && i != 0 ) m_popupmenu->Break();

            int id = wxNewId();

            m_popupmenu->Append(id, arr[i]->Title, arr[i]->Title, wxITEM_RADIO );

            if (arr[i]->Title == m_gdmanager->GetGaugeSymbolTitle(m_requiredSymbol))
            {
                m_popupmenu->Check( id, true);
            }
            m_symbolhash[id] = arr[i]->Name;
            Connect( id, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( RTDataStat::OnSymbolSelect ) );
        }
    }
    PopupMenu(m_popupmenu, event.GetPosition());
}

void RTDataStat::OnSymbolSelect(wxCommandEvent& event)
{
    unsigned int id = event.GetId();

    m_filter = m_filterEdit->GetValue();

    m_guitable->setMatrix(NULL);
    m_guitable->setMinCount( m_mincount );

    delete m_statMatrix;
    delete m_throwedDataM;
    delete m_outliersDataM;

    delete m_prec_data;
    delete m_prec_data_max;
    delete m_prec_data_min;
    delete m_prec_data_weight;
    delete m_prec_data_mean;
    delete m_prec_data_variance;

    if ( m_ratioMatrix )
    {
        delete m_ratioMatrix;
        m_ratioMatrix = NULL;
        if ( m_showVEratio )
            m_showVEratio->SetValue( false );
    }

    m_statMatrix = new AccSetMatrix( m_rows,m_cols );
    m_throwedDataM = new SimpleIntMatrix( m_rows,m_cols );
    m_outliersDataM = new SimpleIntMatrix( m_rows,m_cols );

    m_prec_data = new VemsMatrix<DoubleArray>( m_rows, m_cols );
    m_prec_data_weight = new VemsMatrix<DoubleArray>( m_rows, m_cols );
    m_prec_data_max = new VemsMatrix<double>( m_rows, m_cols );
    m_prec_data_min = new VemsMatrix<double>( m_rows, m_cols );
    m_prec_data_mean = new VemsMatrix<double>( m_rows, m_cols );
    m_prec_data_variance = new VemsMatrix<double>( m_rows, m_cols );

    unregisterSymbols();
    m_expression = m_symbolhash[id];
    load( m_start, m_end );
    m_guitable->setSymbol( m_symbolhash[id] );
    m_expressionEdit->SetValue( m_symbolhash[id] );
    event.Skip();
}

void RTDataStat::OnClose(wxCloseEvent& event)
{
    m_VertLimits.Clear(); m_HorizLimits.Clear();

    if ( m_statMatrix != NULL )
    {
        delete m_statMatrix;
        m_statMatrix = NULL;
    }

    if ( m_throwedDataM != NULL )
    {
        delete m_throwedDataM ;
        m_throwedDataM = NULL;
    }

    if ( m_outliersDataM != NULL )
    {
        delete m_outliersDataM;
        m_outliersDataM = NULL;
    }

    if ( m_prec_data != NULL )
    {
        delete m_prec_data;
        m_prec_data = NULL;
    }

    if ( m_prec_data_max != NULL )
    {
        delete m_prec_data_max;
        m_prec_data_max = NULL;
    }

    if ( m_prec_data_min != NULL )
    {
        delete m_prec_data_min;
        m_prec_data_min = NULL;
    }

    if ( m_prec_data_weight != NULL )
    {
        delete m_prec_data_weight;
        m_prec_data_weight = NULL;
    }

    if ( m_prec_data_mean != NULL )
    {
        delete m_prec_data_mean;
        m_prec_data_mean = NULL;
    }

    if ( m_prec_data_variance != NULL )
    {
        delete m_prec_data_variance;
        m_prec_data_variance = NULL;
    }

    if ( m_ratioMatrix )
    {
        delete m_ratioMatrix;
        m_ratioMatrix = NULL;
    }

    // RTDataAquisition-> frame closed
    observerNotify( 1, this );
    //

    Destroy();
}

void RTDataStat::notify(int i)
{
    // self notify
    wxCommandEvent* evt = new wxCommandEvent( wxEVT_COMMAND_SLIDER_UPDATED, 1 );
    evt->SetInt( i );
    AddPendingEvent(*evt);
}

void RTDataStat::notify(int i, int idx)
{
    if ( i == idFixCell )
    {
        fixSelectedCells();
    }

    if ( i == idMultipleSel )
    {
        // coding tblSize * topleft cell index + bottomright cell index
        size_t tblSize = m_cols * m_rows;
        size_t tlcellI = idx / tblSize;
        size_t brcellI = idx % tblSize;

        cell tlcell, brcell;
        tlcell.row = tlcellI / m_cols;
        tlcell.col = tlcellI % m_cols;

        brcell.row = brcellI / m_cols;
        brcell.col = brcellI % m_cols;

        m_selTopLeft = tlcell;
        m_selBottomRight = brcell;

    }

    if ( i == idSelectCell )
    {
        m_displaycell.row = idx / m_cols;
        m_displaycell.col = idx % m_cols;

        m_selBottomRight = m_selTopLeft = m_displaycell;

        AccSetT* cell;

        if ( m_tuneMode )
        {
            if ( m_showVEratio->GetValue() == false )
            {
                cell = m_statMatrix->getData(idx);
            }
            else
            {
                if ( m_ratioMatrix != NULL )
                {
                    cell = m_ratioMatrix->getData(idx);
                }
                else return;
            }
        }
        else
        {
            cell = m_statMatrix->getData(idx);
        }

        double cmin=0, cmax=0, cmean=0, cvar=0;
        double uf_cmin = 0, uf_cmax = 0, uf_cmean = 0, uf_cvar = 0;

        int dropped = 0;
        int outliers = 0;
        int ccount = count(*cell);
        if ( ccount != 0 )
        {
            cmin = min(*cell);
            cmax = max(*cell);
            cmean = extract_result<tag::weighted_mean>(*cell);
            cvar = sqrtf(extract_result<tag::weighted_variance>(*cell));
            dropped = (*m_throwedDataM)[idx];
            outliers = (*m_outliersDataM)[idx];
        }

        int uf_count = ((*m_prec_data)[idx]).GetCount();
        uf_cmin = (*m_prec_data_min)[idx];
        uf_cmax = (*m_prec_data_max)[idx];
        uf_cmean = (*m_prec_data_mean)[idx];
        uf_cvar = (*m_prec_data_variance)[idx];

        /*
        if ( ccount < (int)m_mincount || ccount > (int)m_maxcount )
            ccount = 0;
        */

        double value;
        switch ( m_guitable->getStatisticalFeature() )
        {
            case STAT_COUNT: value = (double)ccount; break;
            case STAT_MIN: value = cmin; break;
            case STAT_MAX: value = cmax; break;
            case STAT_VARIANCE: value = cvar; break;
            case STAT_MEAN:
            default: value = cmean;
        }

//        histogram_type hist = density(*cell);
//        for ( size_t j = 0; j < hist.size(); ++j )
//        {
//            Logger::logf(LOG_MSG, _T("Bin lower bound =%f, value=%f"), hist[j].first, hist[j].second);
//        }

        if ( m_guitable->isChecked( m_displaycell.row, m_displaycell.col ) )
            m_cellinfo->setChecked(true);
        else
            m_cellinfo->setChecked(false);

        double ve_value = getVE_Value( m_displaycell.row, m_displaycell.col );

        wxColor color = m_guitable->GetCellColor(value);
        m_guitable->SetFocus();

        if ( ccount < (int)m_mincount || ccount > (int)m_maxcount )
        {
            m_cellinfo->setCellBgColor( wxColor(255,255,255) );
            m_cellinfo->setCellTextColor( wxColor(160,160,160) );

            if ( m_tuneMode == false )
            {

                m_cellinfo->setCellInfo(ccount, dropped, outliers, cmin, cmax, cmean, cvar);
            }
            else
            {
                if ( m_showVEratio->GetValue() == false )
                {
                    m_cellinfo->setCellInfo(ccount, dropped, outliers, cmin, cmax, cmean, ve_value, cvar);
                }
                else
                {
                    m_cellinfo->setCellInfo(ccount, 0, 0, cmin, cmax, cmean, cvar);
                }
            }

            m_cellinfo->drawGraphs();
        }
        else
        {
            m_cellinfo->setCellBgColor( color );
            m_cellinfo->setCellTextColor( wxColor(0,0,0) );

            if ( m_tuneMode == false )
            {
                m_cellinfo->setCellInfo(ccount, dropped, outliers, cmin, cmax, cmean, cvar);
            }
            else
            {
                if ( m_showVEratio->GetValue() == false )
                {
                    m_cellinfo->setCellInfo(ccount, dropped, outliers, cmin, cmax, cmean, ve_value, cvar);
                }
                else
                {
                    m_cellinfo->setCellInfo(ccount, 0, 0, cmin, cmax, cmean, cvar);
                }
            }

            m_cellinfo->setUnfilteredInfo( uf_count, uf_cmin, uf_cmax, uf_cmean, uf_cvar );
            // m_cellinfo->drawDistribution();

            DoubleArray& arr = (*m_prec_data)[idx];
            if ( uf_count > 0 )
            {
                m_cellinfo->setHistogram( uf_cmin, uf_cmax, arr );
            }
            else m_cellinfo->setHistogram( 0, 0 );

            m_cellinfo->drawGraphs();

        }
    }

    // select statistical feature
    if ( i == idSelectFeat )
    {
        switch ( idx )
        {
            case 0:
            case 1:
            case 2:
            case 4:
                m_guitable->selectStatisticalFeature( idx );
                break;
            case 3:
            default:
                m_guitable->selectStatisticalFeature( STAT_MEAN );
        }
    }
}

void RTDataStat::notify(const wxString& str)
{
    /// if process is done / all packets were processed -> set matrix

    wxCommandEvent* evt = new wxCommandEvent( wxEVT_COMMAND_SLIDER_UPDATED, 2 );
    AddPendingEvent(*evt);
    return;
}

void RTDataStat::notify( int i, void* classT )
{
    if ( classT == m_guitable )
    {
        if ( i == 1 )
        {
            m_cellinfo->showHistogramWindow();
            m_cellinfo->drawGraphs();
        }
    }
}

void RTDataStat::doOutLierFiltering()
{
    unsigned int cols = m_prec_data->cols(), rows = m_prec_data->rows();

    for ( unsigned i = 0; i < rows; ++i )
    {
        for ( unsigned j = 0; j < cols; ++j )
        {
            unsigned matindex = i * m_prec_data->cols() + j;

            AccSetT* acccel = m_statMatrix->getData(matindex);

            DoubleArray arr = (*m_prec_data)[matindex];
            DoubleArray warr = (*m_prec_data_weight)[matindex];

            unsigned int count = arr.GetCount();

            if ( count < 1 )
                continue;

            double mean = 0.0;
            double sum = 0.0;
            double sum_sqr = 0.0;
            double var = 0.0;

            double minval = 9E9, maxval = -9E9;

            for ( unsigned ctr = 0; ctr < count; ++ctr )
            {
                double val = arr[ctr];
                if ( minval > val ) minval = val;
                if ( maxval < val ) maxval = val;
                sum += val;
                sum_sqr += pow( val, 2 );
            }

            mean = sum / count;

            if ( count == 1 )
                var = 0;
            else
                var = ( 1.0 / (double)( count - 1 ) ) * sum_sqr - ((double) count / (double) ( count - 1 )) * pow( mean, 2 );

            var = sqrtf( var );

            m_prec_data_mean->set( i, j, mean );
            m_prec_data_min->set( i, j, minval );
            m_prec_data_max->set( i, j, maxval );
            m_prec_data_variance->set( i, j, var );

            double fminval;
            double fmaxval;

            if ( m_varianceMult != 0 )
            {
                fminval = mean - m_varianceMult * var;
                fmaxval = mean + m_varianceMult * var;
            }
            else
            {
                fminval = minval;
                fmaxval = maxval;
            }

            for ( unsigned ctr = 0; ctr < count; ++ctr )
            {
                double val = arr[ctr];
                double w = warr[ctr];

                if ( val >= fminval && val <= fmaxval )
                    (*acccel)( val, weight= w );
                else
                    (*m_outliersDataM)[matindex]++;
            }

        }
    }
}

void RTDataStat::OnExprButton(wxCommandEvent& event)
{
    int dlgId = wxNewId();
    const wxString dlgTitle = _T("Edit expressions and filters");

    wxDialog* dlg = new wxDialog(0, dlgId, dlgTitle, wxDefaultPosition,
        wxDefaultSize, wxDEFAULT_DIALOG_STYLE|wxRESIZE_BORDER);

    ExpressionPanel* panel = new ExpressionPanel( dlg );

    wxBoxSizer * sizer = new wxBoxSizer( wxVERTICAL );
    sizer->Add(panel, 1, wxALL | wxEXPAND, 5);
    sizer->Add(dlg->CreateButtonSizer( wxOK | wxCANCEL ), 0, wxEXPAND );

    dlg->SetSizer( sizer );

    sizer->Fit( dlg );
    sizer->SetSizeHints( dlg );

    dlg->ShowModal();

    delete dlg;
}

void RTDataStat::processData(const DoubleHash& data, bool passedFilter)
{
    wxString xRTName = m_xRTSymbol->getName(); /// instead of rpm
    wxString yRTName = m_yRTSymbol->getName(); /// instead of map

    int xvalue = 0;
    int yvalue = 0;
    double value = 0;

    DoubleHash::const_iterator IT = data.find(xRTName);
    if(IT != data.end() ) xvalue = (int)IT->second;

    IT = data.find(yRTName);
    if(IT != data.end() ) yvalue = (int)IT->second;

    IT = data.find(expressionValue);
    if(IT != data.end() ) value = (int)IT->second;

    insertIntoMatrix( xvalue, yvalue, value, passedFilter );
}

void RTDataStat::SetColors()
{
    DoubleHashArray colorA = ExpressionSQL::getExpressionColor(m_expression);

    if ( colorA.GetCount() == 0 )
        m_guitable->resetColorGradients();
    else
    {
        DoubleArray limitsA;
        DoubleArray colorsA;

        for ( size_t i = 0; i < colorA.GetCount(); ++i )
        {
            double limit = (*colorA[i])[_T("value")];
            double color = (*colorA[i])[_T("color")];
            limitsA.Add(limit); colorsA.Add(color);
        }
        m_guitable->setCustomColors(limitsA, colorsA);
    }
    WX_CLEAR_ARRAY( colorA );
}

void RTDataStat::SetColors( const int& coloring_type )
{
    DoubleHashArray colorA = ExpressionSQL::getExpressionColor( coloring_type );

    if ( colorA.GetCount() == 0 )
        m_guitable->resetColorGradients();
    else
    {
        DoubleArray limitsA;
        DoubleArray colorsA;

        for ( size_t i = 0; i < colorA.GetCount(); ++i )
        {
            double limit = (*colorA[i])[_T("value")];
            double color = (*colorA[i])[_T("color")];
            limitsA.Add(limit); colorsA.Add(color);
        }
        m_guitable->setCustomColors(limitsA, colorsA);
    }
    WX_CLEAR_ARRAY( colorA );
}

void RTDataStat::OnEvaluate( wxCommandEvent& event )
{
    m_expression = m_expressionEdit->GetValue();
    m_filter = m_filterEdit->GetValue();
    if ( m_filter == _T("None") )
        m_filter = wxEmptyString;

    m_guitable->setMatrix(NULL);
    m_guitable->setMinCount( m_mincount );

    delete m_statMatrix;
    delete m_throwedDataM;
    delete m_outliersDataM;

    delete m_prec_data;
    delete m_prec_data_max;
    delete m_prec_data_min;
    delete m_prec_data_weight;
    delete m_prec_data_mean;
    delete m_prec_data_variance;

    if ( m_ratioMatrix )
    {
        delete m_ratioMatrix;
        m_ratioMatrix = NULL;
        if ( m_showVEratio )
            m_showVEratio->SetValue( false );
    }


    m_statMatrix = new AccSetMatrix( m_rows,m_cols );
    m_throwedDataM = new SimpleIntMatrix( m_rows,m_cols );
    m_outliersDataM = new SimpleIntMatrix( m_rows, m_cols );
    m_prec_data = new VemsMatrix<DoubleArray>( m_rows, m_cols );
    m_prec_data_weight = new VemsMatrix<DoubleArray>( m_rows, m_cols );
    m_prec_data_max = new VemsMatrix<double>( m_rows, m_cols );
    m_prec_data_min = new VemsMatrix<double>( m_rows, m_cols );
    m_prec_data_mean = new VemsMatrix<double>( m_rows, m_cols );
    m_prec_data_variance = new VemsMatrix<double>( m_rows, m_cols );

    unregisterSymbols();

    load( m_start, m_end );
}

void RTDataStat::OnSpinCtrl( wxSpinEvent& event )
{
    m_mincount = m_minDataSpin->GetValue();
    m_guitable->setMinCount( m_mincount, true );
}

// VE Table tune

void RTDataStat::OnRefreshTables( wxCommandEvent& event )
{
    fixSelectedCells();
}

void RTDataStat::OnSelectAll( wxCommandEvent& event )
{
    if ( m_selTopLeft == m_selBottomRight )
    {
        m_selTopLeft.row = 0;
        m_selTopLeft.col = 0;
        m_selBottomRight.row = m_rows -1;
        m_selBottomRight.col = m_cols - 1;
    }
    else
    {
        m_selTopLeft = m_displaycell;
        m_selBottomRight = m_displaycell;
    }

    m_guitable->setSelection( m_selTopLeft, m_selBottomRight );
}

void RTDataStat::fixSelectedCells()
{
    if ( !m_tuneMode )
        return;

    Symbol* vetable = m_symbolMgr->getSymbol(_T("veTable"));
    int verows = vetable->getShape1();

    size_t verow;

    if ( m_selTopLeft.col == invalid_cell.col && m_selTopLeft.row == invalid_cell.row )
    {
        wxMessageBox( _T("Select cells first!"), _T("Error"), wxOK | wxICON_EXCLAMATION );
        return;
    }

    int i, j;

    unsigned short counter = 0;

    for ( i = m_selTopLeft.row; i <= m_selBottomRight.row; ++i )
    {
        for ( j = m_selTopLeft.col; j <= m_selBottomRight.col; ++j )
        {
            // is it already fixed?
            if  ( m_guitable->isChecked( i, j ) == true )
                continue;

            verow = verows - 1 - i;

            int matindex = ( i * m_statMatrix->getCols() ) + j;
            AccSetT* acc_cell = m_statMatrix->getData(matindex);

            if ( count(*acc_cell) < m_mincount ) continue;

            /// OLD
            /*short value = (short) m_stDataMgr->getDataFromViewerModel(*vetable, verow,j);

            double corr_factor = (mean(*acc_cell) + 100.0) / 100.0;

            /// : -2% - 2%
            /// individual cell can be fixed
            if ( m_displaycell.col != j || m_displaycell.row != i )
                if((corr_factor > 0.98) && (corr_factor < 1.02))
                {
                    continue;
                }

            double corrected = (double) value * corr_factor;*/

            /// NEW: simply overwrite old values
            double corrected = mean(*acc_cell);

            m_stDataMgr->setModelData( *vetable, verow, j, corrected );
            m_stDataMgr->requestWrite( *vetable, verow, j );

            m_guitable->checkCell( i , j );

            if ( m_displaycell.col == j && m_displaycell.row == i )
            {
                m_cellinfo->setChecked(true);
                m_cellinfo->Refresh();
            }

            ++counter;

        }
    }

    m_stDataMgr->updateRegisteredViewers();

    if ( counter > 1 || ( counter == 1 && m_selTopLeft != m_selBottomRight ) )
        wxMessageBox( wxString::Format(_T("%d cells have been fixed"), counter), _T("Information"),
            wxOK | wxICON_INFORMATION );

}

void RTDataStat::OnFocus( wxFocusEvent& event )
{
    //event.GetWindow()->SetFocus();
    event.Skip();
}

void RTDataStat::OnNotifyEvtHandler( wxCommandEvent& event )
{
    int id = event.GetId();

    if ( id == 1 )
    {
        int percent = event.GetInt();

        if ( m_showProgdlg )
        {
            if ( m_progressdlg == NULL )
            {
                wxString title = ( m_tuneMode ) ? _T("Retrieving statistics") : _T("Calculating statistics");
                m_progressdlg = new wxProgressDialog( title, _T("Please wait..."),
                    100, this, wxPD_AUTO_HIDE | wxPD_SMOOTH );
                m_progressdlg->SetSize(260, 120);
            }

            m_progressdlg->Update( percent );
        }
        event.Skip();
        return;
    }

    if ( id == 2 )
    {
        if ( m_showProgdlg )
        {
            m_showProgdlg = false;
            if ( m_progressdlg != NULL )
                delete m_progressdlg;
            m_progressdlg = NULL;
        }

        SetColors();

//        if ( m_useOutlierFiltering )
//        {
            doOutLierFiltering();
//        }

        if ( m_tuneMode )
            m_guitable->forceNumericFormat( _T("%.0f") );

        m_guitable->setMatrix( m_statMatrix );
        m_loghistory->unregisterLogFilter(this);

        m_dataInfoEdit->Clear();

        m_dataInfoEdit->AppendText( wxString::Format( _T("Total: %d\n"), m_statLen ) );

        m_dataInfoEdit->AppendText(wxString::Format(_T("Dropped: %d (Filter+outliers)\nOut of Range: %d"),
            m_throwedDataM->sumOfElements() + m_outliersDataM->sumOfElements(), m_outofrange ));

        m_loghistoryobserv->unRegisterObserver(this);
        event.Skip();
        return;
    }
}

inline double RTDataStat::getVE_Value( int row, int col ) const
{
    const Symbol* sym = m_symbolMgr->getSymbol( _T("veTable") );

    double ve_currval = 0;

    if ( sym )
    {
        ve_currval = m_stDataMgr->getDataFromViewerModel( *sym, m_rows - 1 - row, col );
    }

    return ve_currval;
}
