#include "RTDataStatMath.h"

#include <boost/math/distributions/normal.hpp>
#include <boost/random.hpp>
#include <wx/textfile.h>

#include "../logger.h"

using namespace boost::math;

DblHistogram::DblHistogram()
{
    m_start  = 0;
    m_end    = 0;
    m_bins   = 0;
    m_nb_int = 0;
    m_freq   = NULL;
}

DblHistogram::DblHistogram( const double& start, const double& end, const unsigned int& bins )
{
    m_start = start;
    m_end = end;
    m_bins = bins;
    m_nb_int = (double) m_bins / ( end - start );
    m_freq = new unsigned int[m_bins];

    for ( unsigned i = 0; i < m_bins; ++i )
        m_freq = 0;
}

DblHistogram::DblHistogram( const DblHistogram& hist )
{
    m_start = hist.m_start;
    m_end = hist.m_end;
    m_bins = hist.m_bins;
    m_nb_int = hist.m_nb_int;

    m_freq = new unsigned int[m_bins];

    for ( unsigned i = 0; i < m_bins; ++i )
        m_freq[i] = hist.m_freq[i];
}

DblHistogram& DblHistogram::operator = ( const DblHistogram& hist )
{
    if ( this != &hist )
    {
        m_start = hist.m_start;
        m_end = hist.m_end;
        m_nb_int = hist.m_nb_int;

        if ( m_bins != hist.m_bins )
        {
            m_bins = hist.m_bins;

            if ( m_freq != NULL )
                delete [] m_freq;

            m_freq = new unsigned int[m_bins];
        }

        for ( unsigned i = 0; i < m_bins; ++i )
            m_freq[i] = hist.m_freq[i];
    }

    return *this;
}

DblHistogram::~DblHistogram()
{
    if ( m_freq != NULL )
        delete [] m_freq;
}

void DblHistogram::setProps( const double& start, const double& end, const unsigned int& bins )
{
    m_start = start;
    m_end = end;
    m_bins = bins;
    m_nb_int = (double) m_bins / ( end - start );
    m_freq = new unsigned int[m_bins];

    for ( unsigned i = 0; i < m_bins; ++i )
        m_freq = 0;
}

void DblHistogram::addValue( double value )
{
    if ( m_bins == 0 || m_freq == NULL )
        return;

    unsigned i = (unsigned int)(( value - m_start ) * m_nb_int );

    if ( i < m_bins )
        ++m_freq[i];
}

unsigned int DblHistogram::getCount()
{
    if ( m_bins == 0 || m_freq == NULL )
        return 0;

    unsigned int sum = 0;

    for ( unsigned i = 0; i < m_bins; ++i )
    {
        sum += m_freq[i];
    }

    return sum;
}

const unsigned int* DblHistogram::getFrequencies()
{
    return m_freq;
}

void DblHistogram::getMinMax( double* hmin, double* hmax, double *hbins )
{
    *hmin = m_start;
    *hmax = m_end;
    if ( hbins != NULL )
        *hbins = m_bins;
}

///

double RTDataStatMath::NormalDistribution( double mean, double sigma, double minval, double maxval,
    int bins, double** data )
{
    if ( sigma == 0 || ( minval == maxval ) )
    {
        *data = NULL;
        return 0;
    }

    normal_distribution<> nd( mean, sigma );

    double maximum = 0.0;

    double one_step = ( maxval - minval ) / bins;

    *data = new double[bins];

    double curval = minval;

    for ( int i = 0; i < bins; ++i )
    {
        // calculate the Probability Density (Function)
        double _pd = pdf( nd, curval );
        if ( _pd > maximum )
            maximum = _pd;
        (*data)[i] = _pd;
        curval += one_step;
    }

    return maximum;
}

void RTDataStatMath::RandomNumbersUniform( unsigned count, double minval, double maxval, double** data )
{
    if ( count < 1 || data == NULL )
    {
        *data = NULL;
        return;
    }

    boost::mt19937 bgen(42u);
    boost::uniform_real<> uni_dist( minval, maxval );
    boost::variate_generator<boost::mt19937&, boost::uniform_real<> > uni(bgen, uni_dist);

    *data = new double[count];

    for ( size_t i = 0; i < count; ++i )
    {
        (*data)[i] = uni();
    }
}
void RTDataStatMath::RandomNumbersNormal( unsigned count, double mean, double sigma, double** data )
{
    if ( count < 1 || data == NULL )
    {
        *data = NULL;
        return;
    }

    boost::mt19937 bgen(42u);
    boost::normal_distribution<> normal_dist( mean, sigma );
    boost::variate_generator<boost::mt19937&, boost::normal_distribution<> > norm(bgen, normal_dist);

    *data = new double[count];

    for ( size_t i = 0; i < count; ++i )
    {
        (*data)[i] = norm();
    }
}


void RTDataStatMath::writeToCSV( const wxString& filename, double minval, double maxval, int datasize, double* data )
{
    if ( data == NULL )
        return;

    wxTextFile f( filename );
    if ( f.Exists() )
    {
        f.Open();
        f.Clear();
    }
    else
    {
        f.Create();
    }

    double one_step = ( maxval - minval ) / datasize;
    double curval = minval;

    for ( int i = 0; i < datasize; ++i )
    {
        f.AddLine( wxString::Format( _T("%f,%f"), curval, data[i] ) );
        curval += one_step;
    }
    f.Write();
    f.Close();
}
