#include "ResultHistogram.h"

#include "../logger.h"

ResultHistogram::ResultHistogram(
    uint16_t histogram_elements,
    uint16_t minValue,
    uint16_t maxValue )
{
    //ctor
    m_len = histogram_elements;
    m_histogram = new uint32_t[ histogram_elements + 2];

    m_minValue = minValue;
    m_maxValue = maxValue;
    vectorMult( 0 );

    m_interval = ((maxValue - minValue)<<16) / m_len;
}

ResultHistogram::~ResultHistogram()
{
    //dtor
}

/*
 * returns sum / 0x1000
*/
uint32_t ResultHistogram::getSum()
{
    uint32_t sum = 0;
    int i = m_len + 2; // maxidx is m_len + 1
    do{
        i--;
        uint16_t t16 = m_histogram[ i ] >> 16;
        sum += t16;
    } while( i );

    if( sum == 0 ){
        sum = 1;    // to prevent division by 0 later
    }
    m_sum = sum; // remember sum
    return sum;
}

/* multiplies vector with mult / 0x1000
 * returns sum / 0x1000
*/
uint32_t ResultHistogram::vectorMult( uint16_t mult )
{
    uint32_t sum = 0;
    int i = m_len + 2; // maxidx is m_len + 1
    do{
         i--;
        uint32_t t32 = m_histogram[ i ];
        unsigned long long t64 = t32;
        t64 *= mult; // t64 = t32 * mult does NOT work !!!
        t32 = t64 >> 16;
        m_histogram[ i ] = t32;
        sum += t32 >> 16;
    } while( i );

    if( sum == 0 ){
        sum = 1;    // to prevent division by 0 later
    }
    m_sum = sum; // remember sum
    return sum;
}

void ResultHistogram::vectorPrint(DoubleHash *tableH)
{
    //0-->m_len; 0->2deg 1->3deg
    wxString idxLine;
    wxString binsLine;
    wxString histoLine;

    uint32_t sum = getSum();
    uint16_t i;
    for( i=0; i <= m_len+1; i++){
        idxLine += wxString::Format(_T("%3d"), i);

        uint16_t val = getBin(i);
        binsLine += wxString::Format(_T("%3d"), val);

        // print in percentage. We use m_sum filled by vectorMult() (
        // 0.1 .. 0.9 are printed as .1   .. .9
        // 1..99 are printed normally
        uint16_t histoVal16 = m_histogram[ i ] >> 16;

        wxString key = wxString::Format(_T("%i"), i);
        (*tableH)[key] = 100.0* (double)histoVal16 /sum;

    }
}
/*
 * ResultHistogram::vectorSum()
 * Prints percentage of bins.
 * m_sum must be uptodate before calling, eg. filled by vectorMult()
 */
void ResultHistogram::vectorPrint()
{
    //0-->m_len; 0->2deg 1->3deg
    wxString idxLine;
    wxString binsLine;
    wxString histoLine;

    uint32_t sum = getSum();
    uint16_t i;
    for( i=0; i <= m_len+1; i++){
        idxLine += wxString::Format(_T("%3d"), i);

        uint16_t val = getBin(i);
        binsLine += wxString::Format(_T("%3d"), val);

        // print in percentage. We use m_sum filled by vectorMult() (
        // 0.1 .. 0.9 are printed as .1   .. .9
        // 1..99 are printed normally
        uint16_t histoVal16 = m_histogram[ i ] >> 16;

        // actually not percent, but 1/1000 resolution:
        // unsigned long t64 = 1000 * ... before >> 16 ; would be superfluous
        uint32_t t32 = histoVal16  * 1000;
        uint16_t histoValPercent = t32 / sum ; // range is: 0 .. 1000
        if( histoValPercent <= 9 ){
            histoLine += wxString::Format(_T(" .%d"), histoValPercent);
        } else if( histoValPercent >= 1000 ){ // 100% cannot be displayed, so FF instead:
            histoLine += wxString::Format(_T(" FF"));
        } else { // For 10 .. 999, 1..99 will be printed:
            histoLine += wxString::Format(_T(" %d"), histoValPercent/10);
        }

    }

    // Logger::logf(LOG_MSG,idxLine);
//    Logger::logf(LOG_MSG,binsLine);
  //  Logger::logf(LOG_MSG,histoLine);
    /* DEBUG
    Logger::logf(LOG_MSG, _T("m_minValue=%d, m_maxValue=%d"), m_minValue, m_maxValue );
    uint32_t bin6 = m_histogram[ 6 ];
    uint16_t t16 = bin6 >> 16;
    Logger::logf(LOG_MSG, _T("bin6=%d(%d), from sum=%d (==) m_sum=%d"), t16, bin6, sum, m_sum );
    */
}

/* get minBin treshold
 * 0xFFFF returned for mlen+1 (last collector bin)
*/
uint16_t ResultHistogram::getBin(uint16_t idx)
{
    if( idx == 0 ){
            return 0;
    }
    idx --;
    if (idx >= m_len ){
        return 0xFFFF;
    }
    uint32_t t32 = idx * m_interval;
    uint16_t t16 = t32 >> 16;
    t16 += m_minValue ;
    return t16;
    // there seem to be a gcc bug when doing the above 4 lines in 1 (m_minValue is NOT added !!!)
}

/* degree = 65536*real degree
 * TODO: test this !!!!
*/
uint16_t ResultHistogram::getIdx(uint16_t value)
{
    if( value < m_minValue ){
        return 0;
    }
    uint16_t delta = value - m_minValue;
    uint32_t t32 = delta << 16;
    uint16_t idx = t32 / m_interval;

    if( idx > m_len ){ // idx = m_len + 1 + 1 would be too high
        return m_len +1; // the max index
    }
    return idx+1; // histogram[0] is used for value < m_minValue
}


// looks up idx for value, and adds
void ResultHistogram::increment(uint16_t value)
{
    uint16_t idx = getIdx( value );
    m_histogram[ idx ] += 0x1000;   // decimal dot: 16 bit integer and 16 bit fraction
}

