#include "STDataMgrFacade.h"

#include <wx/listimpl.cpp>

#include "VemsEvent.h"
#include "../Symbol.h"

#include "STDataMgrBase.h"
#include "STDataViewerBase.h"
#include "STDataModelSerial.h"
#include "STDataIOBase.h"

#include "../vemsprotocol/SIPRTypes.h"
#include "../vemsprotocol/SIPRPacket.h"

#include "../expression/VariableData.h"
#include "../expression/RPNParser.h"

#include "../common/types/IntArray.h"

#include "../logger.h"
#include "../firmwareconfig/ConfigValidatorBase.h"

//WX_DEFINE_LIST(STDataViewerBaseList);


STDataMgrFacade::STDataMgrFacade()
{
    m_viewers   = new STDataViewerBaseSet();
    m_tableViewers = new STDataViewerBaseSet();
    m_currInput = NULL;
    Connect(wxEVT_VEMS_CONTROL, VemsControlEventHandler(STDataMgrFacade::OnUpdateViewers) );
}


STDataMgrFacade::~STDataMgrFacade() {}


void STDataMgrFacade::setInput( STDataIOBase* input )
{
    m_currInput = input;
}


void STDataMgrFacade::requestBurn()
{
    if ( m_isConnected )
    {
        m_currInput->burnData();
        clearBurningQueue();
    }
    updateRegisteredViewers();
}


void STDataMgrFacade::requestReadPage()
{
    m_pagesNeedRefresh.clear();

    STDataViewerBaseSet::const_iterator it;
    STDataViewerBase* viewer;

    for( it = m_viewers->begin(); it != m_viewers->end(); ++it )
    {
        viewer = (*it);

        if ( viewer )
        {
            viewer->requestRead();
        }
    }

    IntSet::const_iterator its = m_pagesNeedRefresh.begin();
    for( ; its != m_pagesNeedRefresh.end(); ++its )
    {
        int page = *its;

        if ( m_isConnected )
        {
            m_currInput->read( page );
        }
        else
        {
            Logger::logf( LOG_MSG, _T("STDataMgrFacade::requestReadPage()") );
            m_deviceModel->copyPageTo( page, m_stDataModel );
            m_filter.clearPage( page );
        }
    }

    if ( !m_isConnected ) { updateViewers(); }
}

void STDataMgrFacade::requestReadAllPage()
{
    if ( m_isConnected )
    {
        m_currInput->readAllData();
    }
    else
    {
        m_deviceModel->copyTo( m_stDataModel );
        updateViewers();
    }
}


void STDataMgrFacade::updateViewers()
{
    sendEvent();
}

void STDataMgrFacade::updateRegisteredViewers()
{
    sendEvent();
}

void STDataMgrFacade::sendEvent()
{
    VemsControlEvent event;
    event.SetEventObject( this );
    AddPendingEvent( event );
}

void STDataMgrFacade::OnUpdateViewers( VemsControlEvent& evt )
{
    STDataViewerBaseSet::const_iterator it;
    STDataViewerBase* viewer;

    for( it = m_viewers->begin(); it != m_viewers->end(); it++ )
    {
        viewer = (*it);
        if( viewer ) { viewer->update(); }
    }

    if ( m_configValidator && m_stDataModel )
        m_configValidator->validate( m_stDataModel );
}


void STDataMgrFacade::process( const SIPRPacket& p )
{
    m_deviceModel->process( p );

    if ( isRefreshDeviceModelOnly() ) {
        updateViewers();
        return;
    }

    if ( !m_isConnected )
    {
        m_stDataModel->process( p, &m_filter );
    }
    else
    {
        m_stDataModel->process( p );
        removeRefreshedFromWritingQueue();
    }

    static bool updateEnabled = true;

    if ( p.type() == SIPR_LOG_MARKER ) {
        if ( p.data() && p.size() > 0 ) {
            if ( p.data()[0] == SIPR_REFRESH_START ) {
                //LOGP( _T("SIPR_REFRESH_START") );
                updateEnabled = false;
            } else if ( p.data()[0] == SIPR_REFRESH_STOP ) {
                //LOGP( _T("SIPR_REFRESH_STOP") );
                updateEnabled = true;
            }
        }
    }

    if ( updateEnabled ) {
        //LOGP( _T("UPDATE!") );
    updateViewers();
    }

}

void STDataMgrFacade::requestWriteAllDirty()
{
    m_stDataModel->writeAllDirty( m_currInput );
}

void STDataMgrFacade::requestReadAllDirty()
{
    m_stDataModel->readAllDirty( m_currInput );
}

void STDataMgrFacade::requestWrite( const Symbol& symbol , size_t row, size_t col )
{
    double viewerData = m_stDataModel->getViewerData( symbol, row, col );

    if ( wasDataWritten( symbol, row, col ) )
    {
        removeFromWritingQueue( symbol, row, col );
        return;
    }

    if ( m_isConnected )
    {
        m_deviceModel->setModelData( symbol, row, col, viewerData );
        m_stDataModel->writeViewerData( symbol, m_currInput, row, col );

        removeFromWritingQueue( symbol, row, col );
        addToBurningQueue( symbol, row, col );
    }
}


void STDataMgrFacade::addToBurningQueue( const Symbol& symbol, size_t row, size_t col )
{
    StringToIntSetH::iterator it = m_waitingForBurn.find( symbol.getName() );

    size_t index = symbol.isArray() ? symbol.getArrayColSize() * row + col : 0;

    if ( it == m_waitingForBurn.end() )
    {
        m_waitingForBurn[ symbol.getName() ] = new IntSet();
    }

    (m_waitingForBurn[ symbol.getName() ])->insert( index );
}


void STDataMgrFacade::removeFromBurningQueue( const Symbol& symbol, size_t row, size_t col )
{
    StringToIntSetH::iterator it = m_waitingForBurn.find( symbol.getName() );

    if ( it == m_waitingForBurn.end() ) { return; }

    size_t index = symbol.isArray() ? symbol.getArrayColSize() * row + col : 0;

    IntSet* set = it->second;
    set->erase( index );

    if ( set->empty() )
    {
        m_waitingForBurn.erase( symbol.getName() );
        delete set;
    }
}



void STDataMgrFacade::removeRefreshedFromWritingQueue()
{
    StringToIntSetH::iterator it = m_waitingForWrite.begin();
    wxArrayString needRemove;

    for ( ; it != m_waitingForWrite.end(); ++it )
    {
        IntSet* set = it->second;

        removeRefreshedIndiciesFromWritingQueue( it->first, *(it->second) );

        if ( set->empty() )
        {
            needRemove.Add( it->first );
        }
    }

    for ( size_t i = 0; i < needRemove.GetCount(); ++i )
    {
        IntSet* set = m_waitingForWrite[ needRemove[i] ];
        delete set;

        m_waitingForWrite.erase( needRemove[i] );
    }
}



void STDataMgrFacade::removeRefreshedIndiciesFromWritingQueue( const wxString& name, IntSet& set )
{
    IntSet::iterator it = set.begin();
    IntArray needRemove;

    for( ; it != set.end(); ++it )
    {
        int index = *it;

        if ( wasDataWritten( name, index ) )
        {
            needRemove.Add( index );
        }
    }

    for ( size_t i = 0; i < needRemove.GetCount(); ++i )
    {
        set.erase( needRemove[i] );
    }

}

void STDataMgrFacade::addToWritingQueue( const Symbol& symbol, size_t row, size_t col )
{
    StringToIntSetH::iterator it = m_waitingForWrite.find( symbol.getName() );

    size_t index = symbol.isArray() ? symbol.getArrayColSize() * row + col : 0;

    if ( it == m_waitingForWrite.end() )
    {
        m_waitingForWrite[ symbol.getName() ] = new IntSet();
    }

    (m_waitingForWrite[ symbol.getName() ])->insert( index );

    if ( !m_isConnected ) { m_filter.add( symbol.getPage(), symbol.getOffset() + index ); }
}


void STDataMgrFacade::removeFromWritingQueue( const Symbol& symbol, size_t row, size_t col )
{
    StringToIntSetH::iterator it = m_waitingForWrite.find( symbol.getName() );

    if ( it == m_waitingForWrite.end() ) { return; }

    size_t index = symbol.isArray() ? symbol.getArrayColSize() * row + col : 0;

    IntSet* set = it->second;
    set->erase( index );

    if ( set->empty() )
    {
        m_waitingForWrite.erase( symbol.getName() );
        delete set;
    }
}


void STDataMgrFacade::setModelData( const Symbol& symbol, size_t row, size_t col, double data )
{
    //const double EPSILON = 0.000001;

    //double viewerData = m_stDataModel->getViewerData( symbol, row, col );

    //if ( fabs( viewerData - data ) < EPSILON ) { return; };

    m_stDataModel->setModelData( symbol, row, col, data );

    if ( wasDataWritten( symbol, row, col ) )
    {
        removeFromWritingQueue( symbol, row, col );
    }
    else
    {
        addToWritingQueue( symbol, row, col );
    }
}


void STDataMgrFacade::setRawModelData( const Symbol& symbol, size_t row, size_t col, uint8_t data )
{
    m_stDataModel->setRawModelData( symbol, row, col, data );

    if ( !wasDataWritten( symbol, row, col ) )
    {
        addToWritingQueue( symbol, row, col );
    }
/*

    {
        removeFromWritingQueue( symbol, row, col );
    }
    else
    {
        addToWritingQueue( symbol, row, col );
    }
  */
    //m_stDataModel->setRawModelData( symbol, row, col, data );
}


void STDataMgrFacade::updateFromDeviceModel( const Symbol& symbol )
{
    if ( symbol.isArray() )
    {
        for ( int r = 0; r < symbol.getArrayRowSize(); ++r )
        {
            for ( int c = 0; c < symbol.getArrayColSize(); ++c )
            {
                updateFromDeviceModel( symbol, r, c );
            }
        }
    }
    else if ( symbol.isScalar() )
    {
        updateFromDeviceModel( symbol, 0, 0 );
    }
}


void STDataMgrFacade::updateFromDeviceModel( const Symbol& symbol, size_t row, size_t col )
{
    double data = m_deviceModel->getViewerData( symbol, row, col );

    m_stDataModel->setModelData( symbol, row, col, data );

    removeFromWritingQueue( symbol, row, col );
}


void STDataMgrFacade::updateRawFromDeviceModel( const Symbol& symbol, size_t row, size_t col )
{
    uint8_t data = m_deviceModel->getRawModelData( symbol, row, col );

    m_stDataModel->setRawModelData( symbol, row, col, data );

    //uint8_t viewerData = m_stDataModel->getRawModelData( symbol, row, col );
    m_stDataModel->getRawModelData( symbol, row, col );

    removeFromWritingQueue( symbol, row, col );
}


void STDataMgrFacade::requestWriteAll()
{
    Logger::logf( LOG_MSG, _T("STDataMgrFacade::requestWriteAll()") );

    STDataViewerBaseSet::const_iterator it;
    STDataViewerBase* viewer;

    for( it = m_viewers->begin(); it != m_viewers->end(); ++it )
    {
        viewer = (*it);

        if ( viewer )
        {
            viewer->updateModel();
            viewer->requestWrite();
        }
    }
}

void STDataMgrFacade::clear()
{
    clearBurningQueue();
    clearWritingQueue();
    m_stDataModel->clear();
    m_deviceModel->clear();
    m_filter.clear();
}
