#include "STDataModelSerial.h"
#include <math.h>

#include "STDataIOBase.h"
#include "../mtconfigdata.h"
#include "../Math.h"
#include "../logger.h"
#include "../vemsprotocol/SIPRTypes.h"
#include "../vemsprotocol/SIPRPacket.h"
#include "../vemsprotocol/SIPRDescriptor.h"

#include "../common/types/IntToIntSetHash.h"
#include "../common/types/DoubleHash.h"

#define MAX_PAGE_COUNT  32
#define MAX_BUFFER_SIZE 512


STDataModelSerial::STDataModelSerial( const SymbolTable& symbols, const SIPRDescriptor* siprDesc )
    : m_symbolTable( symbols )
{
    //TEST ONLY!!!
    for( uint16_t i = 0; i < MAX_PAGE_COUNT; i++ )
    {
        m_pageBufferHash[ i ] = new char [ MAX_BUFFER_SIZE ];
        memset ( m_pageBufferHash[ i ], 0, MAX_BUFFER_SIZE );

        m_dirtyOffsets.Add( new IntSet() );
    }

    m_pSIPRDescriptor = siprDesc;

    m_rpnForExpressionBasedSymbols.setVariableData( &m_variablesForExpressionBasedSymbols );
}


STDataModelSerial::~STDataModelSerial()
{
	Int2CharRefHash::iterator it = m_pageBufferHash.begin();

	for ( ; it != m_pageBufferHash.end(); ++it )
	{
		delete[] it->second;
	}

	WX_CLEAR_ARRAY(m_dirtyOffsets);
}

bool STDataModelSerial::getValue( const wxString& name, double* value, IntSet* dirty )
{
    SymbolTable::const_iterator IT = m_symbolTable.find( name );

    if (  IT == m_symbolTable.end() ) { return false; }

    const Symbol* aSymbol = IT->second;

    if( !aSymbol )
    {
        return false; //symbol not found, to do !!!
    }

    int pageIdx     = aSymbol->getPage();
    int pageOffset  = aSymbol->getOffset();

    //msValue   = userValue / scale - translate
	//userValue = (msValue + translate) * scale
    if( aSymbol->isBits() )
    {
        int bitHi = aSymbol->getBitHi();
        int bitLo = aSymbol->getBitLo();
//        int bitOffset = aSymbol->getBitOfs();

        uint8_t ecuValue = m_pageBufferHash[ pageIdx ][ pageOffset ];

        uint8_t mask = ( ( 1 << ( bitHi - bitLo + 1 ) ) - 1 ) << bitLo;

        //*value = (( ecuValue & mask ) >> bitLo) + bitOffset;
        *value = ( ecuValue & mask ) >> bitLo;

        return true;
    }

    if( aSymbol->isScalar() )
    {
        char *buffer        = m_pageBufferHash[pageIdx];
        uint32_t ecuValue   = Math::getNumber( (uint8_t*)&buffer[ pageOffset ], aSymbol->getSize() );

        //uint8_t ecuValue = buffer[pageOffset];

        double viewerValue   = ( ecuValue + aSymbol->getTranslation() ) * aSymbol->getScale() ;

        *value = viewerValue;

        if ( dirty )
        {
            if ( isDataDirty( pageIdx, pageOffset ) ) { dirty->insert( pageOffset ); }
            else { dirty->clear(); }
        }

        return true;
    }

    if ( aSymbol->isExpr() )
    {
        RPNParser rpn;
        rpn.setVariableData( this );

        *value = rpn.calculate( aSymbol->getExpr() );

        return true;
    }

    return false;
}

void STDataModelSerial::setBitSymbol( const Symbol& bitSymbol, double d )
{
    int pageIdx     = bitSymbol.getPage();
    int pageOffset  = bitSymbol.getOffset();
    int bitHi       = bitSymbol.getBitHi();
    int bitLo       = bitSymbol.getBitLo();

    uint8_t oldEcuValue = m_pageBufferHash[ pageIdx ][ pageOffset ];

    uint8_t maxVal = (1 << (bitHi-bitLo+1)) - 1;

    uint8_t val = (uint8_t)(d);

    if ( val > maxVal ) val = maxVal;

    maxVal <<= bitLo;

    uint8_t newEcuValue = ( oldEcuValue & ~maxVal ) | ( val << bitLo );

    setRawData( newEcuValue, pageIdx, pageOffset );
}

void STDataModelSerial::setScalarSymbol( const Symbol& scalarSymbol, double d )
{
    int page            = scalarSymbol.getPage();
    int offset          = scalarSymbol.getOffset();
    int size            = scalarSymbol.getSize();
    double scale        = scalarSymbol.getScale();
    double translation  = scalarSymbol.getTranslation();

    double newEcuValue = d / scale - translation;

    if( newEcuValue < 0.0 ) { newEcuValue = 0.0; }

    double max = pow( 2.0, (double)(size * 8) ) - 1;

    if( floor( newEcuValue + 0.5 ) > max ) { newEcuValue = max; }

    uint32_t newEcuValueBytes = static_cast< uint32_t >( newEcuValue + 0.5 );

    uint8_t tmp[16];
    Math::getBytes( newEcuValueBytes, tmp, size );
    for ( int i = 0; i < size; ++i )
    {
        setRawData( tmp[i], page, offset + i );
    }
}

void STDataModelSerial::setArraySymbol( const Symbol& arraySymbol, size_t row, size_t col, double d )
{
    int page            = arraySymbol.getPage();
    int offset          = arraySymbol.getOffset();
    double scale        = arraySymbol.getScale();
    double translation  = arraySymbol.getTranslation();
    int index           = arraySymbol.getArrayColSize() * row + col;

    double newEcuValue = 0.0;

    if ( arraySymbol.useDefaultTransformation() )
    {
        newEcuValue = d / scale - translation;
    }
    else
    {
        m_variablesForExpressionBasedSymbols.addValue( _T("x"), d );
        newEcuValue = m_rpnForExpressionBasedSymbols.calculate( arraySymbol.getExprToRaw() );
    }
    double max = 255.0 ;

    if( ceil( newEcuValue ) > 255.0 ) { newEcuValue = 255.0; }
    if( ceil( newEcuValue ) < 0.0 ) { newEcuValue = 0.0; }


    newEcuValue += 0.5;
    if( floor( newEcuValue) > max ) { newEcuValue = max; }
    //uint32_t newEcuValueBytes = static_cast< uint32_t >( newEcuValue );

    uint8_t newEcuValueByte = (uint8_t)newEcuValue;

    setRawData( newEcuValueByte, page, offset + index );
}

void STDataModelSerial::setExpressionSymbol( const Symbol& expressionSymbol, double d )
{
    wxString exprReverse    = expressionSymbol.getExprReverse();
    wxString target         = expressionSymbol.getTarget();
    wxString symbol         = expressionSymbol.getName();

    if ( exprReverse.IsEmpty() || target.IsEmpty() ) { return; }

    RPNParser rpn;
    DoubleHash dh;
    VariableDataDH rpnVars( &dh );

    StringHashSet vars = rpn.getVariables( exprReverse );
    vars.erase( symbol );

    StringHashSet::const_iterator it;
    for( it =vars.begin(); it != vars.end(); ++it )
    {
        getValue( *it, &dh[ *it ] );
    }

    dh[ symbol ] = d;
    rpn.setVariableData( &rpnVars );

    double result = rpn.calculate( exprReverse );

    setScalarValue( target, result );
}


// TODO - use symbol manager
void STDataModelSerial::setScalarValue( const wxString& name, double value )
{
    SymbolTable::const_iterator IT = m_symbolTable.find( name );

    if (  IT == m_symbolTable.end() ) { return; }

    const Symbol* scalar = IT->second;

    if( !scalar ) { return; }

    setScalarValue( scalar, value );
}


void STDataModelSerial::setScalarValue( const Symbol* scalar, double value )
{
//    Logger::logf(   LOG_MSG, _T("-----STDataModelSerial::setScalarValue() - symbol: ") + scalar->getName() +
//                    _T(", value: %f"), value );

    int page    = scalar->getPage();
    int offset  = scalar->getOffset();

    double newEcuValue = value / scalar->getScale() - scalar->getTranslation();

    if( newEcuValue < 0.0 ) { newEcuValue = 0.0; }

    double max = pow( 2.0, (double)(scalar->getSize() * 8) ) - 1;

    newEcuValue += 0.5;
    if( floor( newEcuValue) > max ) { newEcuValue = max; }
    uint32_t newEcuValueBytes = static_cast< uint32_t >( newEcuValue );

    uint8_t tmp[16];
    Math::getBytes( newEcuValueBytes, tmp, scalar->getSize() );
    for ( int i = 0; i < scalar->getSize(); ++i )
    {
        setRawData( tmp[i], page, offset + i );
    }
}


void STDataModelSerial::setRawData( const wxString& symbolName, uint8_t value )
{
   //Symbol* aSymbol = m_mtConfigData->getSymbol( symbolName );

    SymbolTable::const_iterator IT = m_symbolTable.find( symbolName );

    if (  IT == m_symbolTable.end() ) { return; }

    const Symbol* aSymbol = IT->second;

    if( aSymbol )
    {
        int pageIdx     = aSymbol->getPage();
        int pageOffset  = aSymbol->getOffset();

        //m_pageBufferHash[ pageIdx ][ pageOffset ] = value; //set ms data
        setRawData( value, pageIdx, pageOffset );
        //Int2DoubleHash tmp;

        //getViewerData( symbolName, &tmp ); //set user data

        //m_symbols[ symbolName ] = tmp[0];
    }
}

void STDataModelSerial::setRawData( uint8_t value, int page, int offset, int source )
{

    if ( page >= MAX_PAGE_COUNT || offset >= MAX_BUFFER_SIZE ) { return; }

    uint8_t mv = m_pageBufferHash[ page ][ offset ];
/*
    if ( page == 2 )
    {
    Logger::logf(   LOG_MSG, _T("STDataModelSerial::setRawData() - ") +
                    getSourceAsString( source ) +
                    _T(" - p: %d, o: %d, v: %u, mv: %u"),
                    page, offset, value, mv );
    }
*/
    switch ( source )
    {
        case USER:
            if ( value == mv ) { return; }
        case USER_FORCE_WRITE:
        {
            //Logger::logf( LOG_MSG, _T("STDataModelSerial::setRawData() - USER_FORCE_WRITE - p: %u, o: %u, data: %u")
            //                , page, offset, value );
            m_pageBufferHash[ page ][ offset ] = value;

            m_dirtyPages.insert( page );
            m_dirtyOffsets[ page ]->insert( offset );

            mv = m_pageBufferHash[ page ][ offset ];
        } break;
        case ECU:
        {
            //if ( value != m_pageBufferHash[ page ][ offset ] )

            if ( value != mv )
            {
                m_pageBufferHash[ page ][ offset ] = value;
            }
            mv = m_pageBufferHash[ page ][ offset ];

            //Logger::logf(   LOG_MSG, _T("STDataModelSerial::setRawData() - ECU - p: %d, o: %d, v: %u, mv: %u"),
            //                page, offset, value, mv );
            if ( !isPageDirty( page ) ) { return; }
            if ( !m_dirtyOffsets[ page ]->erase( offset ) ) { return; }
            if ( m_dirtyOffsets[ page ]->empty() ) { m_dirtyPages.erase( page ); }

        } break;
    }
}

void  STDataModelSerial::setRawData(const Symbol& symbol, VemsMatrix<uint8_t>* table)
{

    int page = symbol.getPage();
    int offset = symbol.getOffset();
    uint8_t value;

    for(unsigned int i = 0; i < table->size(); ++i){
        value = table->get(i);
        setRawData(value, page, offset + i, USER_FORCE_WRITE);
    }
}


uint8_t STDataModelSerial::getRawData( const wxString& symbolName )
{
    //Symbol* aSymbol = m_mtConfigData->getSymbol( symbolName );
    SymbolTable::const_iterator IT = m_symbolTable.find( symbolName );

    if (  IT == m_symbolTable.end() ) { return 0; }

    const Symbol* aSymbol = IT->second;

    if( aSymbol )
    {
        int pageIdx     = aSymbol->getPage();
        int pageOffset  = aSymbol->getOffset();

        return m_pageBufferHash[ pageIdx ][ pageOffset ];
    }
    else
    {
        return 0; ///TODO
    }
}
uint8_t STDataModelSerial::getRawData( const Symbol& alienSymbol ) const
{
    int pageIdx     = alienSymbol.getPage();
    int pageOffset  = alienSymbol.getOffset();

    Int2CharRefHash::const_iterator IT = m_pageBufferHash.find(pageIdx);
    if(IT == m_pageBufferHash.end() ) return 0;
    char* buff = IT->second;
    return buff[ pageOffset ];
}

uint8_t* STDataModelSerial::getRawBuffer(int pageNo)
{
    return (uint8_t* ) m_pageBufferHash[ pageNo ];
}

void STDataModelSerial::process( const SIPRPacket& packet, const ModelFilter* filter )
{
    if ( !m_pSIPRDescriptor ) { return; }

    if ( packet.type() == SIPR_GET_CONFIG_1 || packet.type() == SIPR_GET_CONFIG_2 )      // process config pages
    {
        IntArray pages = m_pSIPRDescriptor->getPages( packet.type(), packet.type() );

        for ( size_t i = 0; i < pages.GetCount(); ++i )
        {
            int pageNo = pages[i];

            //uint8_t* buffer = (uint8_t*)( m_pageBufferHash[ pageNo ] );
            //memcpy( buffer, packet.data(), packet.size() );
            for ( int j = 0; j < packet.size(); ++j )
            {
                if ( filter )
                {
                    if ( filter->isFiltered( pageNo, j ) ) continue;
                }

                setRawData( packet.data()[j], pageNo, j, ECU );
            }

            // Logger::logf( LOG_MSG, _T("STDataModelSerial::process() - pageNo: %d, size: %d, offset:%d"), pageNo, packet.size(), 0 );
        }

        return;
    }
    else if ( packet.type() == SIPR_GET_TABLEDATA ) // process tables
    {
        if ( packet.size() < 2 || !packet.data() ) { return; }

        IntArray pages = m_pSIPRDescriptor->getPages( packet.type(), packet.getSubType() );

        for ( size_t i = 0; i < pages.GetCount(); ++i )
        {
            uint8_t* data       = &packet.data()[1];
            int size        = packet.size() - 1;

            int pageNo = pages[i];
            int offset = m_pSIPRDescriptor->getOffset( packet.type(), packet.getSubType() );
            //int size   = m_pSIPRDescriptor->getSize();

            // verify page
            if ( pageNo < 0 || pageNo > MAX_PAGE_COUNT ) { continue; }

            // verify size and offset
            if ( size == -1 || offset == -1 ) { continue; }

            //uint8_t* buffer = (uint8_t*)( &m_pageBufferHash[ pageNo ][ offset] );
            //memcpy( buffer, data, size );
            for ( int j = 0; j < size; ++j )
            {
                if ( filter )
                {
                    if ( filter->isFiltered( pageNo, offset + j ) ) continue;
                }

                setRawData( data[j], pageNo, offset + j, ECU );
            }
            //Logger::logf( LOG_MSG, _T("STDataModelSerial::process() - pageNo: %d, size: %d, offset:%d"), pageNo, size, offset );
        }
    }
    else if ( packet.type() == SIPR_READ_MEMORY )
    {
        uint8_t* data = &packet.data()[2];
        int size = packet.size() - 2;
        int pageNo = 1;

        for ( int i = 0; i < size; ++i )
        {
            setRawData( data[i], pageNo, i, ECU );
        }
    }
}


uint8_t STDataModelSerial::getRawData( int page, int offset ) const
{
    Int2CharRefHash::const_iterator it = m_pageBufferHash.find( page );

    if ( it == m_pageBufferHash.end() )
    {
        Logger::logf( LOG_ERROR, _T("-----STDataModelSerial::getRawData() - invalid page ( %d )"), page );
        return 0x00;
    }

    if ( offset >= MAX_BUFFER_SIZE || offset < 0 )
    {
        Logger::logf( LOG_ERROR, _T("-----STDataModelSerial::getRawData() - invalid offset ( %d )"), offset );
        return 0x00;
    }

    return it->second[offset];
}


void STDataModelSerial::writeAllDirty( STDataIOBase* iob )
{
    for( IntSet::const_iterator itp = m_dirtyPages.begin() ; itp != m_dirtyPages.end(); ++itp )
    {
        IntSet* aDirtyPage = m_dirtyOffsets[*itp];

        for( IntSet::const_iterator ito = aDirtyPage->begin() ; ito != aDirtyPage->end(); ++ito )
        {
//            Logger::logf(   LOG_MSG, _T(" STDataModelSerial::writeAllDirty() - p: %d, o: %d, v:"),
//                            *itp, *ito, getRawData( *itp, *ito ) );
            iob->write( getRawData( *itp, *ito ), *itp, *ito );
        }
    }
}

void STDataModelSerial::readAllDirty( STDataIOBase* iob )
{
    for( IntSet::const_iterator itp = m_dirtyPages.begin() ; itp != m_dirtyPages.end(); ++itp )
    {
        iob->read( *itp );
    }
}

bool STDataModelSerial::isDataDirty( int page, int offset ) const
{
    //if ( !isPageDirty( page ) ) { return false; }

    //IntSet* dirtyOffsets  = m_dirtyOffsets.find( page )->second;

    //return dirtyOffsets->find( offset ) !=  dirtyOffsets->end();

    return m_dirtyOffsets[ page ]->find( offset ) !=  m_dirtyOffsets[ page ]->end();

}


bool STDataModelSerial::isPageValid( int page ) const
{
    return !( page < 0 || page >= MAX_PAGE_COUNT );
}


int STDataModelSerial::getPageCount() const
{
    return MAX_PAGE_COUNT;
}


int STDataModelSerial::getPageSize( int page ) const
{
    return MAX_BUFFER_SIZE;
}


void STDataModelSerial::clear()
{
    m_dirtyPages.clear();

    for( int i = 0; i < MAX_PAGE_COUNT; ++i )
    {
        memset ( m_pageBufferHash[ i ], 0, MAX_BUFFER_SIZE );
        m_dirtyOffsets[ i ]->clear();
    }
}


wxString STDataModelSerial::getSourceAsString( int src )
{
    switch (  src )
    {
        case ECU:
            return _T("ECU");
        case USER:
            return _T("USER");
        case USER_FORCE_WRITE:
            return _T("USER_FORCE_WRITE");
        default:
            return _T("ERROR - UNKNOWN MODEL SOURCE");
    }
}

/// under development


double STDataModelSerial::getViewerDataByIndex( const wxString& symbolName, size_t index )
{
    double viewerData = 0.0;

    SymbolTable::const_iterator it = m_symbolTable.find( symbolName );

    if (  it != m_symbolTable.end() )
    {
        viewerData = getViewerDataByIndex( *(it->second), index );
    }

    return viewerData;
}


double STDataModelSerial::getViewerDataByIndex( const Symbol& symbol, size_t index )
{
    //Logger::logf( LOG_MSG, _T("STDataModelSerial::getViewerData() - BEGIN") );

    double viewerData = 0.0;

    size_t page         = symbol.getPage();

    Int2CharRefHash::const_iterator it = m_pageBufferHash.find( page );

    if ( it == m_pageBufferHash.end() ) { return viewerData; }

    size_t offset       = symbol.getOffset();
    size_t size         = symbol.getSize();
    double translation  = symbol.getTranslation();
    double scale        = symbol.getScale();

    if ( symbol.isArray() )
    {
        /// use row order, and two dimensional arrays, one byte size
        //size_t index        = offset + symbol.getArrayColSize() * row + col;

        if ( symbol.useDefaultTransformation() )
        {
            viewerData = ( (double)getRawData( page, offset + index ) + translation ) * scale;
        }
        else
        {
            m_variablesForExpressionBasedSymbols.addValue( _T("x"), (double)getRawData( page, offset + index ) );
            viewerData = m_rpnForExpressionBasedSymbols.calculate( symbol.getExprFromRaw() );
            //Logger::logf( LOG_MSG, _T("STDataModelSerial::getViewerDataByIndex() - symbol: ") + symbol.getName() +
            //    _T(" = %f"), viewerData );
        }
    }
    else if ( symbol.isScalar() )
    {
        char *buffer        = it->second;
        uint32_t ecuValue   = Math::getNumber( (uint8_t*)&buffer[ offset ], size );

        viewerData   = ( ecuValue + translation ) * scale;
//        Logger::logf( LOG_MSG, _T("STDataModelSerial::getViewerData() - ") + symbol.getName() +
//                        _T(" - data: %f"), viewerData );
    }
    else if ( symbol.isBits() )
    {
        int bitHi = symbol.getBitHi();
        int bitLo = symbol.getBitLo();

        char *buffer        = it->second;

        uint8_t ecuValue = buffer[ offset ];

        uint8_t mask = ( ( 1 << ( bitHi - bitLo + 1 ) ) - 1 ) << bitLo;

        viewerData = ( ecuValue & mask ) >> bitLo;
    }
    else if ( symbol.isExpr() )
    {
        RPNParser rpn;
        rpn.setVariableData( this );

        viewerData = rpn.calculate( symbol.getExpr() );
    }

    return viewerData;
}


double STDataModelSerial::getViewerData( const wxString& symbolName, size_t row, size_t col )
{
    SymbolTable::const_iterator IT = m_symbolTable.find( symbolName );

    if (  IT == m_symbolTable.end() ) { return -100000000; }

    const Symbol* aSymbol = IT->second;

    //Symbol* aSymbol = m_mtConfigData->getSymbol( symbolName );
    if( !aSymbol )
    {
        return -100000000; //symbol not found, to do !!!
    }
    return getViewerData(*aSymbol, row, col );
}

double STDataModelSerial::getViewerData( const Symbol& symbol, size_t row, size_t col )
{
    size_t index =  symbol.isArray() ? symbol.getArrayColSize() * row + col : 0;

    return getViewerDataByIndex( symbol, index );
}


void STDataModelSerial::setModelData( const Symbol& symbol, size_t row, size_t col, double data )
{
    if ( symbol.isArray() )
    {
        setArraySymbol( symbol, row, col, data );
    }
    else if ( symbol.isScalar() )
    {
        setScalarSymbol( symbol, data );
    }
    else if ( symbol.isBits() )
    {
        setBitSymbol( symbol, data );
    }
    else  if ( symbol.isExpr() )
    {
        setExpressionSymbol( symbol, data );
    }
}

void STDataModelSerial::writeViewerData( const Symbol& symbol, STDataIOBase* iob, size_t row , size_t col )
{
    size_t page     = symbol.getPage();

    Int2CharRefHash::const_iterator it = m_pageBufferHash.find( page );

    if ( it == m_pageBufferHash.end() ) { return; }

    size_t offset   = symbol.getOffset();
    size_t size     = symbol.getSize();

    if ( symbol.isArray() )
    {


        size_t index = symbol.getArrayColSize() * row + col;
        uint8_t data = getRawData( page, offset + index );

        //Logger::logf( LOG_MSG, _T("STDataModelSerial::writeViewerData() - ")  + symbol.getName()
        //               + _T(" - p: %u, o: %u, i: %u, d: %u"), page, offset, index, data );

        iob->write( page, offset + index, data );
    }
    else if ( symbol.isScalar() || symbol.isBits() )
    {
        for( size_t i = 0; i < size; ++i )
        {
            uint8_t data = getRawData( page, offset + i );
            iob->write( page, offset + i, data );
        }
    }
}


void STDataModelSerial::setRawModelData( const Symbol& symbol, size_t row, size_t col, uint8_t data )
{
    size_t page     = symbol.getPage();

    Int2CharRefHash::const_iterator it = m_pageBufferHash.find( page );

    if ( it == m_pageBufferHash.end() ) { return; }

    if ( symbol.getSize() != 1 )
    {
        Logger::logf( LOG_MSG, _T("STDataModelSerial::setRawModelData() - ") + symbol.getName() +
            _(" size != 1 (%u)"), symbol.getSize() );
        return;
    }

    size_t offset   = symbol.getOffset();
    //size_t size     = symbol.getSize();

    if ( symbol.isArray() )
    {
        size_t index = symbol.getArrayColSize() * row + col;
        setRawData( data, page, offset + index );
    }
    else if ( symbol.isScalar() )
    {
        setRawData( data, page, offset );
    }
}


uint8_t STDataModelSerial::getRawModelData( const Symbol& symbol, size_t row, size_t col )
{
    size_t page = symbol.getPage();

    Int2CharRefHash::const_iterator it = m_pageBufferHash.find( page );

    if ( it == m_pageBufferHash.end() ) { return 0x00; }

    if ( symbol.getSize() != 1 )
    {
        Logger::logf( LOG_MSG, _T("STDataModelSerial::getRawModelData() - ") + symbol.getName() +
            _(" size != 1 (%u)"), symbol.getSize() );
        return 0x00;
    }

    size_t offset   = symbol.getOffset();
    //size_t size     = symbol.getSize();

    if ( symbol.isArray() )
    {
        size_t index = symbol.getArrayColSize() * row + col;
        return getRawData( page, offset + index );
    }
    else if ( symbol.isScalar() )
    {
        return getRawData( page, offset );
    }

    return 0x00;
}


bool STDataModelSerial::getValue( const wxString& name, VemsMatrix<double>& data )
{
    //Logger::logf( LOG_MSG, _T("STDataModelSerial::getValue() - name: ") + name );

    SymbolTable::const_iterator IT = m_symbolTable.find( name );

    if (  IT == m_symbolTable.end() ) { return false; }

    Symbol* symbol = IT->second;

    if ( symbol->isArray() )
    {
        int cols = symbol->getArrayColSize();
        int rows = symbol->getArrayRowSize();

        data.setSize( rows, cols );

        for ( int r = 0; r < rows; ++r )
        {
            for ( int c = 0; c < cols; ++c )
            {
                data.set( r, c, getViewerData( *symbol, r, c ) );
            }
        }
    }

    return true;
}

void  STDataModelSerial::copyPageTo( size_t page, STDataModelBase* target )
{
    uint8_t* sourceBuffer = getRawBuffer( page );
    uint8_t* targetBuffer = target->getRawBuffer( page );

    if ( sourceBuffer && targetBuffer )
    {
        memcpy( targetBuffer, sourceBuffer, MAX_BUFFER_SIZE * sizeof(uint8_t) );
    }
}

void  STDataModelSerial::copyTo( STDataModelBase* target )
{
    for ( size_t page = 0; page< MAX_PAGE_COUNT; ++page )
    {
        copyPageTo( page, target );
    }
}

size_t STDataModelSerial::getDirtyCount()
{
    size_t dirtyCount = 0;

    for ( int i = 1; i < getPageCount(); ++i )
    {
        if ( !isPageValid( i ) || !isPageDirty( i ) ) { continue; }

        for ( int j = 0; j < getPageSize( i ); ++j )
        {
            if ( !isDataDirty( i, j ) ) { continue; }

            ++dirtyCount;
        }
    }

    return dirtyCount;
}
