#include "TriggerFrame.h"

#include "../Math.h"
#include "../logger.h"


TriggerFrame::TriggerFrame()
{
    m_Trigger = 0;
    m_data = 0;
    m_crc = 0;
}

/**
frame FORMAT
only_CRC:  N* prim trigg(2 byte),type(1byte),CRC(2byte),7e
SEC_TRIG:  N* prim trigg(2 byte),sec trig(2byte),type, CRC, 7e
event_3:   N* prim trigg(2 byte), event 3 (2byte),type, CRC, 7e
event_4:   N* prim trigg(2 byte), event 4 (2byte),type, CRC, 7e
EVENT_DATA_TIMESTAMP_1:
            N* prim trigg(2 byte), data(2byte,) timestamp (2byte),type, CRC, 7e
N can be 0
*/

/// fills m_primaryA16 array (0..N primary timestamps), m_Trigger and m_crc from source IntArray& frame
/// the only reasonable solution is to apply +clockbase here !
/// the +clockbase is totally missing from here ! It is very error-prone to find out clock-base later than this.
int TriggerFrame::addFrame( const IntArray& frame )
{
    size_t size = frame.GetCount();

    if ( size == 0 )
    {
        return TRFRAME_EMPTY_FRAME;
    }

    if (size < 5 )
    {
        return TRFRAME_SMALL_SIZE;
    }

    // LITTLE ENDIAN
    m_crc = frame.Item( size - 2 ) + 256 * frame.Item( size - 1 );

    // calculate crc;
    m_calc_crc = 0xFFFF;

    for( size_t i = 0; i < size-2; ++i )
    {
        m_calc_crc = Math::crc16_sipr( m_calc_crc, frame.Item( i ) );
    }

    // CRC error
    if( m_crc != m_calc_crc )
    {
        Logger::logf( LOG_ERROR,_T("TriggerFrame::addFrame() CRC error 0x%04X, 0x%04X"), m_crc, m_calc_crc );
        return TRFRAME_CRC_ERROR;
    }

    m_type = frame.Item( size - 3 );

    m_Trigger = 0;

    // byte count
    size_t primCount = 0;

    switch ( m_type )
    {
        case CRC_ONLY:
            primCount = size - 3;
            break;

        case SEC_TRIG:
        case CAMSHAFT_EX:
        case EVENT_4:
        case EVENT_5:
        case ROAD_SPEED1:
        case ROAD_SPEED2:
        case EVENT_8:
        case EVENT_9:
        case EVENT_10:
        case EVENT_11:
        case EVENT_12:
        case EVENT_13:
        case EVENT_14:
        case EVENT_15:
            primCount = size - 5;
            // BIG ENDIAN
            m_Trigger = 256* frame.Item( size - 5 ) + frame.Item( size - 4 );
            break;
        case EVENT_DATA_TIMESTAMP_1:
            primCount = size - 7;
            m_Trigger = 256* frame.Item( size - 5 ) + frame.Item( size - 4 );
            m_data    = 256* frame.Item( size - 7 ) + frame.Item( size - 6 );
            break;
        case EVENT_REALTIME:
            return m_type;
        default:
            // Unsupported frame format
            return UNSUPPORTED_FORMAT;
    }

    // if CRC is good but odd number of primary tr bytes we crash
    primCount &= ~1;

    for( size_t i = 0; i < primCount; i += 2 )
    {
        int tmp = 256 * frame.Item( i ) + frame.Item( i + 1 );

        m_primaryA16.Add( tmp );
    }

    // Logger::logf(LOG_MSG, _T("TRFRame Format = %d"), m_type);
    return m_type;
};

// CRC the bytestream (after unescape, but BEFORE crunching)
bool TriggerFrame::checkCRC()
{
    return ( m_calc_crc == m_crc );
};

uint16_t TriggerFrame::getLastPRTrigger()
{
	int len = m_primaryA16.GetCount();

    if( len < 1 )
    {
        return 0;
    }

    return m_primaryA16.Item( len-1 );
}

void TriggerFrame::convertFrame2Time( Int2DoubleArrayHash& triggerTimes, int& clockbase, int& prevTime )
{
    IntArray& primary = m_primaryA16;

    // Primary convert
    DoubleArray& primaryTime = triggerTimes[CRC_ONLY];

    for ( size_t prIdx = 0; prIdx < primary.GetCount(); ++prIdx )
    {
        int now = clockbase + primary[prIdx];

        // time OVERFLOW !
        if ( now < prevTime )
        {
            // step 65535 ahead
            clockbase += 65536;
            now = clockbase + primary[prIdx];
        }

        primaryTime.Add( (double) now * 4 / 1000 );  // 1ms resolution

        if ( now > prevTime)
        {
            prevTime = now;
        }
    }

    // primary trigger already added, others (id>last_ok) are not supported
    if ( m_type == CRC_ONLY || m_type >= LAST_OK_ID )
    {
        return;
    }

    DoubleArray& trArr = triggerTimes[m_type];

    int now = clockbase + m_Trigger;

    // time OVERFLOW !
    if ( now < prevTime )
    {
        // step 65535 ahead
        clockbase += 65536;
        now = clockbase + m_Trigger;
    }

    // 1ms resolution
    trArr.Add( (double) now * 4 / 1000);

    if ( now > prevTime )
    {
        prevTime = now;
    }

    if(m_type == EVENT_DATA_TIMESTAMP_1){
        // m_dataA.Add(m_data);
    }
}

/// DEPRECATED: only prim&sectrig ( use it if something goes wrong )
/*
void TriggerFrame::convertFrame2Time(DoubleArray& primaryTime, DoubleArray& secondaryTime,
int& clockbase, int& prevTime)
{
        IntArray& primary = m_primaryA16;
        //size_t size  = primary.GetCount();

        //if( size < 2){
            //Logger::logf(LOG_MSG, _T("TriggerLog::fillTRTime() small size:%i!"), size);
            //continue;
        //}
        /// Primary convert
        for(size_t prIdx =0; prIdx < primary.GetCount(); ++prIdx){
            int now = clockbase + primary[prIdx];
            if( now < prevTime ){ // time OVERFLOW !
                clockbase += 65536; // step 65535 ahead
                now = clockbase + primary[prIdx];
            }
            primaryTime.Add((double)now *4 / 1000);  /// 1ms resolution
            if(now > prevTime) prevTime = now;
        }

        if(m_type != SEC_TRIG) return;

        int now = clockbase + m_secTrigger;
        if( now < prevTime ){ // time OVERFLOW !
                clockbase += 65536; // step 65535 ahead
                now = clockbase + m_secTrigger;
        }
        secondaryTime.Add((double)now *4 / 1000); /// 1ms resolution
        if(now > prevTime) prevTime = now;
}
*/
