#include "TriggerLogSaveAudio.h"
#include "../logger.h"

TriggerLogSaveAudio::TriggerLogSaveAudio( DoubleArray& primary, DoubleArray& secondary )
    : m_primary( primary ), m_secondary( secondary )
{
    m_samplerate = 44100; // 44.1kHz
    m_on = 32767; // 16bit, signed
    m_off = -32768;

    m_step = 1.0 / (double)m_samplerate * 1000;

    getSampleNo();
}

bool TriggerLogSaveAudio::saveToRaw( const wxString& filename )
{
    bool ret;

    ret = createFile( filename );

    if ( ret == false )
        return false;

    ret = writeData();
    if ( ret == false )
    {
        m_file.Close();
        return false;
    }

    return true;
}

bool TriggerLogSaveAudio::saveToWaw( const wxString& filename )
{
    bool ret;

    ret = createFile( filename );

    if ( ret == false )
        return false;

    ret = writeHeader();
    if ( ret == false )
    {
        m_file.Close();
        return false;
    }

    ret = writeData();
    if ( ret == false )
    {
        m_file.Close();
        return false;
    }

    return true;
}

void TriggerLogSaveAudio::getSampleNo()
{
    size_t prim_count = m_primary.GetCount();

    double prim_max = 0, sec_max = 0;

    if ( prim_count )
        prim_max = m_primary[prim_count-1];

    size_t sec_count = m_secondary.GetCount();

    if ( sec_count )
        sec_max = m_secondary[sec_count-1];

    double max_time = ( prim_max > sec_max ) ? prim_max : sec_max;

    m_sampleNo = (int)( max_time / m_step ) + 1;
}

bool TriggerLogSaveAudio::createFile( const wxString& name )
{
    m_file.Create( name, true );
    return true;
}

bool TriggerLogSaveAudio::writeHeader()
{
    struct _wavedescr {
        unsigned char riff[4];
        unsigned long size;
        unsigned char wave[4];
    } wavedescr;

    wavedescr.riff[0] = 'R'; wavedescr.riff[1] = 'I'; wavedescr.riff[2] = 'F'; wavedescr.riff[3] = 'F';
    wavedescr.size = 36 + m_sampleNo * 2 * 16 / 8; // see size below
    wavedescr.wave[0] = 'W'; wavedescr.wave[1] = 'A'; wavedescr.wave[2] = 'V'; wavedescr.wave[3] = 'E';

    m_file.Write( &wavedescr, sizeof(wavedescr) );

    ///-------------------
    struct _wavefmt {
        unsigned char id[4];
        unsigned long size;
        short format;
        short channels;
        unsigned long samplerate;
        unsigned long byterate;
        short blockalign;
        short bitspersample;
    } wavefmt;

    wavefmt.id[0] = 'f'; wavefmt.id[1] = 'm'; wavefmt.id[2] = 't'; wavefmt.id[3] = ' ';
    wavefmt.size = 16; // 16 for pcm
    wavefmt.format = 1; // pcm
    wavefmt.channels = 2; // stereo
    wavefmt.samplerate = m_samplerate; // 44.1kHz
    wavefmt.byterate = (m_samplerate * 2 * 16 / 8); // == SampleRate * NumChannels * BitsPerSample/8
    wavefmt.blockalign = 2 * 16 / 8; // == NumChannels * BitsPerSample/8
    wavefmt.bitspersample = 16; // 16 bits/sample

    m_file.Write( &wavefmt, sizeof( wavefmt ) );

    unsigned char data[4] = { 'd', 'a','t','a' };

    m_file.Write( &data, 4 );

    unsigned long size = m_sampleNo * 2 * 16 / 8; // == NumSamples * NumChannels * BitsPerSample/8

    m_file.Write( &size, 4 );

    return true;
}

bool TriggerLogSaveAudio::writeData()
{
    double timecounter = 0;

    double prim_max = 0, sec_max = 0;
    double prim_next = 0, sec_next = 0;

    size_t prim_count = m_primary.GetCount();

    if ( prim_count )
    {
        prim_next = m_primary[0];
        prim_max = m_primary[prim_count-1];
    }


    size_t sec_count = m_secondary.GetCount();

    if ( sec_count )
    {
        sec_next = m_secondary[0];
        sec_max = m_secondary[sec_count-1];
    }


    double max_time = ( prim_max > sec_max ) ? prim_max : sec_max;

    #define SIGN_ON_TIME 40

    uint8_t prim_on = 0;
    uint8_t sec_on = 0;

    int pcnt = 0, scnt = 0;

    static int counter = 0;

    while ( timecounter <= max_time )
    {
        signed short left; // primary
        signed short right; // secondary

        if ( timecounter < prim_next )
        {
            if ( prim_on > 0 )
            {
                if ( prim_on < SIGN_ON_TIME )
                {
                    left = m_on;
                    ++prim_on;
                }
                else
                {
                    left = m_on;
                    prim_on = 0;
                }
            }
            else
            {
                left = m_off;
            }
        }
        if ( timecounter < sec_next )
        {
            if ( sec_on > 0 )
            {
                if ( sec_on < SIGN_ON_TIME )
                {
                    right = m_on;
                    ++sec_on;
                }
                else
                {
                    right = m_on;
                    sec_on = 0;
                }
            }
            else
            {
                right = m_off;
            }
        }

        if ( timecounter >= prim_next )
        {
            left = m_on;
            prim_on = 1;
            if ( ++pcnt < m_primary.GetCount() )
                prim_next = m_primary[pcnt];
        }
        if ( timecounter >= sec_next )
        {
            right = m_on;
            sec_on = 1;
            if ( ++scnt < m_secondary.GetCount() )
                sec_next = m_secondary[scnt];
        }

        m_file.Write( &left, sizeof( left ) );
        m_file.Write( &right, sizeof( right ) );

        if ( counter++ == 1000 )
        {
            observerNotify( (int)( timecounter * 100.0 / max_time ) );
            //Logger::logf(LOG_MSG, _T("current time=%f, max=%f"), timecounter, max_time);
            counter = 0;
        }

        timecounter += m_step;
    }

    m_file.Close();

    observerNotify( 100 );

    return true;

}
