#include <wx/wx.h>
#include "CurveEditor.h"
#include "ConfigletPanel.h"
#include "../mtconfigdata.h"
#include "../SymbolMgr.h"
#include "../stdatamanager/STDataMgrBase.h"
#include "../stdatamanager/STDataModelSerial.h"
#include "../vemsgauge/AggGraphControl.h"
#include "../events/UpdateViewerEvent.h"
#include "../DirHelper.h"
#include "../firmwareconfig/FirmwareConfigLoader.h"
#include "../common/VemsContext.h"
#include "../common/FileDropTarget.h"
#include "../hotkeys/wxHotKeyManager.h"

#include <wx/file.h>
#include <wx/filename.h>


/**********************************************************
*   STArrayViewerTextCtrl methods
**********************************************************/

BEGIN_EVENT_TABLE( STArrayViewerTextCtrl, wxTextCtrl )
    EVT_KILL_FOCUS( STArrayViewerTextCtrl::OnKillFocus )
    EVT_SET_FOCUS( STArrayViewerTextCtrl::OnSetFocus )
    EVT_KEY_DOWN( STArrayViewerTextCtrl::OnKeyDown )
END_EVENT_TABLE()

STArrayViewerTextCtrl::STArrayViewerTextCtrl( wxWindow* parent, wxWindowID id, const wxString& value, const wxPoint& pos, const wxSize& size,
    long style, const wxValidator& validator, const wxString& name ) : wxTextCtrl( parent, id, value, pos, size, style, validator, name )
{
    m_rowIdx = 0;
    m_colIdx = 0;
    m_curveEditor = NULL;
    m_isFirstKillFocus = true;
}

void STArrayViewerTextCtrl::setIndex( uint32_t rowIdx, uint32_t colIdx )
{
    m_rowIdx = rowIdx;
    m_colIdx = colIdx;
}


uint32_t STArrayViewerTextCtrl::getValueIndex()
{
    int dimension = m_symbol.getOrder();

    if( dimension == 1 )
    {
        return m_rowIdx;
    }
    else if( dimension == 2 )
    {
        int rows    = m_symbol.getShape0();
        int cols    = m_symbol.getShape1();
        int offset  = ( rows - 1 - m_rowIdx ) * cols + m_colIdx;
        return offset;
    }

    return 0;
}


void STArrayViewerTextCtrl::update()
{
    size_t row;
    size_t col;

    getItemPosition( row, col );

    if ( m_stDataMgr->wasDataWritten( m_symbol, row, col ) )
    {
        SetForegroundColour( *wxBLACK );
    }
    else
    {
        SetForegroundColour( *wxRED );
    }

    double d = m_stDataMgr->getDataFromViewerModel( m_symbol, row, col );

    ChangeValue( wxString::Format( m_format, d ) );

    GetValue().ToDouble( &d );


    if( m_curveEditor )
    {
        m_curveEditor->update();
    }

    if ( !m_expr.IsEmpty() )
    {
        Enable( m_stDataMgr->calculate( m_expr )!=0?true:false );
    }

}

void STArrayViewerTextCtrl::requestRead()
{
    m_stDataMgr->addPageNeedRefresh( m_symbol.getPage() );
}

void STArrayViewerTextCtrl::updateModel()
{
    double d = 0;

    GetValue().ToDouble( &d );

    size_t row;
    size_t col;

    getItemPosition( row, col );

    m_stDataMgr->setModelData( m_symbol, row, col, d );


    if( m_curveEditor )
    {
        m_curveEditor->update();
    }
}

void STArrayViewerTextCtrl::setFormat( int digits_in )
{
    m_format = _T("%.") + wxString::Format( _T("%i"), digits_in ) + _T("f");
}

void STArrayViewerTextCtrl::getItemPosition( size_t& row, size_t& col )
{
    int dimension = m_symbol.getOrder();

    if( dimension == 1 )
    {
        row = 0;
        col = m_rowIdx;
    }
    else if( dimension == 2 )
    {
        int rows    = m_symbol.getShape0();

        row = rows - 1 - m_rowIdx;
        col = m_colIdx;
    }
}

void STArrayViewerTextCtrl::requestWrite()
{
    size_t row;
    size_t col;

    getItemPosition( row, col );

    m_stDataMgr->requestWrite( m_symbol, row, col );
}

void STArrayViewerTextCtrl::OnKillFocus( wxFocusEvent& event )
{
    double guiValue = 0;

    if( GetValue().ToDouble( &guiValue ) )
    {
        size_t row;
        size_t col;

        getItemPosition( row, col );

        double viewerData = m_stDataMgr->getDataFromViewerModel( m_symbol, row, col );


        if( guiValue != viewerData )
        {
            updateModel();

            requestWrite();
            m_stDataMgr->updateRegisteredViewers();
        }
    }
    else
    {
        Logger::logf( LOG_WARNING, _T("STArrayViewerTextCtrl::OnKillFocus() - invalid value") );
        update();
    }

    SetBackgroundColour( *wxWHITE );
}

void STArrayViewerTextCtrl::OnSetFocus( wxFocusEvent& event )
{
    SetBackgroundColour( wxColor(255,255,0x99) );
}

void STArrayViewerTextCtrl::OnKeyDown( wxKeyEvent& event )
{
    wxString evtstr = m_hotkeyMgr->GetEvent(event);

    if ( evtstr.IsSameAs( _T("CURVE_EDITOR_KEY_INC_SMALL" ) ) ) {
        changeValueWith( m_smallStep );
    } else if ( evtstr.IsSameAs( _T("CURVE_EDITOR_KEY_INC_LARGE") ) ) {
        changeValueWith( m_largeStep );
    } else if ( evtstr.IsSameAs( _T("CURVE_EDITOR_KEY_DEC_SMALL") ) ) {
        changeValueWith( -m_smallStep );
    } else if ( evtstr.IsSameAs( _T("CURVE_EDITOR_KEY_DEC_LARGE") ) ) {
        changeValueWith( -m_largeStep );
    } else {
        event.Skip();
    }
}

void STArrayViewerTextCtrl::changeValueWith( double change )
{
    double currentValue = 0.0;

    if ( !GetValue().ToDouble( &currentValue ) ) {
        return;
    }

    currentValue += change;

    if ( currentValue < m_symbol.getLo() ) {
        currentValue = m_symbol.getLo();
    } else if ( currentValue > m_symbol.getHi() ) {
        currentValue = m_symbol.getHi();
    }

    wxString currentValueStr = wxString::Format( m_format, currentValue );

    size_t row;
    size_t col;

    getItemPosition( row, col );

    double viewerData = m_stDataMgr->getDataFromViewerModel( m_symbol, row, col );

    wxString viewerDataStr = wxString::Format( m_format, viewerData );

    SetValue( currentValueStr );

    if ( currentValueStr.IsSameAs( viewerDataStr ) ) {
        SetForegroundColour( *wxBLACK );
    } else {
        SetForegroundColour( *wxRED );
    }
}

wxString STArrayViewerTextCtrl::toString()
{
    return _T("STArrayViewerTextCtrl - ") + m_symbol.getName();
}

/**********************************************************
*   CurveEditor
**********************************************************/
IMPLEMENT_DYNAMIC_CLASS(CurveEditor, wxPanel);

int REALTIME_X_TEXT_ID = wxNewId();
int REALTIME_Y_TEXT_ID = wxNewId();

BEGIN_EVENT_TABLE( CurveEditor, wxPanel )
    EVT_MOUSEWHEEL( CurveEditor::OnMouseWheel )
    EVT_UPDATE_VIEWER_EVENT( wxID_ANY, CurveEditor::OnUpdateValue )
END_EVENT_TABLE()


CurveEditor::CurveEditor( wxWindow* parent, VemsCurveDialogDesc* desc, SymbolMgr* symbolMgr)
    : wxPanel( parent )
    , StaticViewersController( parent, desc->topicHelp )
{
    m_curve     = NULL;
    m_curveDesc = desc;

    m_symbolMgr = symbolMgr;

    m_seqXAxisName = _T("SEQUENCE");
    m_xBinSymbol = NULL;

    m_xRealTimeName = wxEmptyString;
    m_yRealTimeName = wxEmptyString;
    m_isSymbolRegistered = false;

#ifndef __WINCE_STANDARDSDK__
    new FileDropTarget<CurveEditor>( this );
#endif

    m_hotkeyMgr = new wxHotKeyManager(this);
}

CurveEditor::~CurveEditor()
{

    if(m_curve) delete m_curve;
    Logger::logf( LOG_MSG, _T("CurveEditor::~CurveEditor() ") );
    if ( m_isSymbolRegistered ) {
         Logger::logf( LOG_MSG, _T("CurveEditor::~CurveEditor() Unregister") );
        UnRegisterViewer();
    }
    unregisterSymbols();
}

bool CurveEditor::init( STDataMgrBase* stController, RTDataMgrBase* rtController )
{


    m_xRealTimeName = m_curveDesc->xRealTime;
    m_yRealTimeName = m_curveDesc->yRealTime;

    Symbol* xRealTimeSymbol = NULL;

    Logger::logf( LOG_MSG, _T(" CurveEditor::init() - m_xRealTimeName >") + m_xRealTimeName + _T("<") );

    if ( !m_xRealTimeName.IsEmpty() )
    {
        xRealTimeSymbol = m_symbolMgr->getSymbol( m_xRealTimeName );

        if( !xRealTimeSymbol )
        {
            Logger::logf( LOG_MSG, _T(" CurveEditor::init() - xRealTimeSymbol symbol not found : ") + m_xRealTimeName );
            //return false;
        }
    }
    else if ( !m_yRealTimeName.IsEmpty() )
    {
        Symbol* yRealTimeSymbol = m_symbolMgr->getSymbol( m_yRealTimeName );

        if( !yRealTimeSymbol )
        {
            Logger::logf( LOG_MSG, _T(" CurveEditor::init() - m_yRealTimeName symbol not found : ") + m_yRealTimeName );
            return false;
        }
    }

    wxString xBinName = m_curveDesc->xBins;
    m_xBinSymbol = m_symbolMgr->getSymbol(xBinName );

    if( !m_xBinSymbol )
    {
        Logger::logf( LOG_MSG, _T(" CurveEditor::init() - m_xBinSymbol symbol not found : ") + xBinName );
        Logger::logf( LOG_MSG, _T(" CurveEditor::init() - addr of xRealTimeSymbol: %p "), xRealTimeSymbol );

        if(xBinName == m_seqXAxisName ) { //&& xRealTimeSymbol ){
            m_xBinSymbol = new Symbol();
            //m_xBinSymbol->setBasic(m_seqXAxisName, xRealTimeSymbol->getUnit(), true, -1, -1);
            m_xBinSymbol->setBasic(m_seqXAxisName, _T(""), true, -1, -1);
            m_xBinSymbol->setArray(true, 0, m_curveDesc->divisionsX ,1);
        }else{
            return false;
        }
    }

    m_yBinName = m_curveDesc->yBins;
    Symbol* yBinSymbol = m_symbolMgr->getSymbol( m_yBinName );

    if( !yBinSymbol )
    {
        Logger::logf( LOG_MSG, _T(" CurveEditor::init() - dialog symbol not found : ") + m_yBinName );
        return false;
    }


    m_controller = stController;
    buildGUI( m_controller, rtController );

    return true;
}


void CurveEditor::buildGUI( STDataMgrBase* stController, RTDataMgrBase* rtController )
{
    wxBoxSizer* mainSizer = new wxBoxSizer( wxVERTICAL  );
    SetSizer( mainSizer );

    //add xAxis text controls
    wxBoxSizer* xAxisColumnSizer = new wxBoxSizer( wxVERTICAL );
    wxStaticText* textLabel = new wxStaticText( this, wxID_ANY, m_curveDesc->columnLabelX );
    wxFont font  = textLabel->GetFont();
    font.SetWeight( wxFONTWEIGHT_BOLD );
    textLabel->SetFont( font );
    //box->SetForegroundColour( *wxBLACK );

    xAxisColumnSizer->Add( textLabel, 0, wxLEFT, 5 );


    textLabel = new wxStaticText( this, wxID_ANY, _T("[") + m_xBinSymbol->getUnit() + _T("]") );
    xAxisColumnSizer->Add( textLabel, 0, wxTOP | wxBOTTOM  | wxLEFT, 5 );

    int xAxisColumnSize = m_xBinSymbol->getShape0();

    int i = 0;

    STArrayViewerTextCtrl* textField = NULL;
    wxSize textFieldSize = wxDefaultSize;
    textFieldSize.SetWidth( 60 );


    wxTextValidator validator( wxFILTER_INCLUDE_CHAR_LIST );
    wxArrayString includes;

    includes.Add( _T("0") );
    includes.Add( _T("1") );
    includes.Add( _T("2") );
    includes.Add( _T("3") );
    includes.Add( _T("4") );
    includes.Add( _T("5") );
    includes.Add( _T("6") );
    includes.Add( _T("7") );
    includes.Add( _T("8") );
    includes.Add( _T("9") );
    includes.Add( _T(".") );
    includes.Add( _T("-") );

    includes.Add( _T("q") );
    includes.Add( _T("w") );
    includes.Add( _T("e") );
    includes.Add( _T("r") );
    includes.Add( _T("Q") );
    includes.Add( _T("W") );
    includes.Add( _T("E") );
    includes.Add( _T("R") );

    validator.SetIncludes( includes );
   //SetValidator( validator );

    for( i = 0; i < xAxisColumnSize; ++i )
    {

        long style = 0;
        if( m_curveDesc->isXBinsReadOnly ) { style = wxTE_READONLY; }

        textField = new STArrayViewerTextCtrl(  this,
                                                wxID_ANY,
                                                _T(""),
                                                wxDefaultPosition,
                                                textFieldSize,
                                                style,
                                                validator );
        textField->setHotKeyManager( m_hotkeyMgr );
        textField->setIndex( i, 0 );
        textField->setFormat( m_xBinSymbol->getDigits() );
        textField->SetName( m_xBinSymbol->getName() );
        textField->setSymbol( *m_xBinSymbol );
        textField->setSmallStep( m_curveDesc->xBinsSmallStep );
        textField->setLargeStep( m_curveDesc->xBinsLargeStep );
        xAxisColumnSizer->Add( textField, 0, wxALL, 1 );

        if(m_xBinSymbol->getName() != m_seqXAxisName){
            m_ViewerBaseA.Add( textField );
            stController->RegisterViewer( textField );
        }else{
            double dX = double(m_curveDesc->maxX - m_curveDesc->minX) / (double)(m_curveDesc->divisionsX-1);
            double xValue = m_curveDesc->minX + i * dX;
            textField->ChangeValue( wxString::Format( m_curveDesc->labelFormatX, xValue) );
            textField->Enable(false);
        }
        textField->setCurveEditor( this );
        //textField->SetInitialSize();

        m_xAxisViewers.Add( textField );


    }

    //add yAxis text controls

    wxBoxSizer* yAxisColumnSizer = new wxBoxSizer( wxVERTICAL );
    textLabel = new wxStaticText( this, wxID_ANY, m_curveDesc->columnLabelY );
    font  = textLabel->GetFont();
    font.SetWeight( wxFONTWEIGHT_BOLD );
    textLabel->SetFont( font );
    yAxisColumnSizer->Add( textLabel, 0, wxLEFT, 5 );

    Symbol* yBinSymbol = m_symbolMgr->getSymbol(m_yBinName);

    textLabel = new wxStaticText( this, wxID_ANY, _T("[") + yBinSymbol->getUnit() + _T("]") );
    yAxisColumnSizer->Add( textLabel, 0, wxTOP | wxBOTTOM | wxLEFT, 5 );


    int yAxisColumnSize = yBinSymbol->getShape0();

    textField = NULL;

    for( i = 0; i < yAxisColumnSize; ++i )
    {
        textField = new  STArrayViewerTextCtrl( this, wxID_ANY, _T(""), wxDefaultPosition, textFieldSize );
        textField->setHotKeyManager( m_hotkeyMgr );
        textField->SetValidator( validator );
        textField->setIndex( i, 0 );
        textField->setFormat( yBinSymbol->getDigits() );
        textField->SetName( yBinSymbol->getName() );
 //       textField->setSymbolName( yBinSymbol->getName() );
        textField->setSymbol( *yBinSymbol );
        //textField->setExpression( expr );
        yAxisColumnSizer->Add( textField, 0, wxALL, 1 );
        m_ViewerBaseA.Add( textField );
        stController->RegisterViewer( textField );
        textField->setCurveEditor( this );
        textField->setSmallStep( m_curveDesc->yBinsSmallStep );
        textField->setLargeStep( m_curveDesc->yBinsLargeStep );
        //textField->update();
        m_yAxisViewers.Add( textField );
    }

    //add xAxis graph control
    m_curve = new AggGraphControl( this );
    m_curve->SetInitialSize( wxSize( 400, 400 ) );
    m_curve->setMinX( m_curveDesc->minX );
    m_curve->setMaxX( m_curveDesc->maxX );
    m_curve->setDivisionsX( m_curveDesc->divisionsX );
    m_curve->setMinY( m_curveDesc->minY );
    m_curve->setMaxY( m_curveDesc->maxY );
    m_curve->setDivisionsY( m_curveDesc->divisionsY );
    m_curve->showHorizontalHelpLine( m_curveDesc->showHorizontalHelpLine );
    m_curve->showVerticalHelpLine( m_curveDesc->showVerticalHelpLine );

    m_curve->setRowLabelFormat( m_curveDesc->labelFormatY );
    m_curve->setColLabelFormat( m_curveDesc->labelFormatX );

    if ( !m_xRealTimeName.IsEmpty() )
    {
        m_curve->showCrosshair( true );
        m_curve->setCrosshairs( m_curveDesc->minX, m_curveDesc->minY );

        rtController->RegisterViewer( this );
        this->SetDataMgr( rtController );
    }
    else if ( !m_yRealTimeName.IsEmpty() )
    {
        m_curve->showLine( true );
        m_curve->setHorizontalLine( m_curveDesc->minY );

        rtController->RegisterViewer( this );
        this->SetDataMgr( rtController );
    }


    //update();

    //
    //rtController->RegisterViewer( this );
    //this->SetDataMgr( rtController );

    //set layout

    wxBoxSizer* dataSizer = new wxBoxSizer( wxHORIZONTAL );

    wxBoxSizer* columnSizer = new wxBoxSizer( wxHORIZONTAL );

    columnSizer->Add( xAxisColumnSizer, 0, wxALL, 5 );
    columnSizer->Add( yAxisColumnSizer, 0, wxALL, 5 );

    wxBoxSizer* leftSizer =  new wxBoxSizer( wxVERTICAL );
    leftSizer->Add( columnSizer, 0, 0, 0 );

//    wxStaticText* realtimeXText = new wxStaticText( this, REALTIME_X_TEXT_ID, _T("") );
//    wxFont realtimeXTextFont = realtimeXText->GetFont();
//    realtimeXTextFont.SetWeight( wxFONTWEIGHT_BOLD );
//    realtimeXText->SetFont( realtimeXTextFont );
//    leftSizer->Add( realtimeXText, 0, wxALL | wxALIGN_CENTER_HORIZONTAL | wxEXPAND, 5 );

    dataSizer->Add( leftSizer, 0, 0, 0 );
    dataSizer->Add( m_curve, 1, wxEXPAND, 0 );

    mainSizer->Add( dataSizer, 1, wxEXPAND );

    addConfigletPanel();

    SetLabel( m_curveDesc->title );

    require();
}


void CurveEditor::addConfigletPanel()
{
    wxString configletDir = DirHelper::GetPath( dCONFIGLETS ) + m_curveDesc->name + DirHelper::Separ();

    if ( !::wxDirExists( configletDir ) ) return;

    ConfigletPanel* configletPanel = new ConfigletPanel( this, configletDir, m_symbolMgr );
    m_ViewerBaseA.Add( configletPanel );
    m_controller->RegisterViewer( configletPanel );

    GetSizer()->Add( configletPanel, 0, wxEXPAND );
}


void CurveEditor::update()
{
    m_curve->clearPoints();

    double x = 0;
    double y = 0;

   // size_t xAxisColumnSize = m_xAxisViewers.GetCount();
    size_t size = std::min( m_xAxisViewers.GetCount(), m_yAxisViewers.GetCount() );

    for( uint32_t i = 0; i < size; ++i )
    {
        m_xAxisViewers.Item( i )->GetValue().ToDouble( &x );
        m_yAxisViewers.Item( i )->GetValue().ToDouble( &y );
        m_curve->addPoint( x, y );
    }

    m_curve->Refresh();
}

void CurveEditor::refresh( DoubleHash* table )
{
    double d = 0.0;

    if ( !m_xRealTimeName.IsEmpty() )
    {
        Symbol* xRealTimeSymbol = m_symbolMgr->getSymbol(m_xRealTimeName);

        if( !xRealTimeSymbol ) { return; }

        DoubleHash::const_iterator it = table->find( xRealTimeSymbol->getName() );

        if( it == table->end() ) { return; }

        d = it->second;

        if( d > m_curveDesc->maxX ) { d = m_curveDesc->maxX; }
        if( d < m_curveDesc->minX ) { d = m_curveDesc->minX; }
    }
    else if ( !m_yRealTimeName.IsEmpty() )
    {
        Symbol* yRealTimeSymbol = m_symbolMgr->getSymbol(m_yRealTimeName);

        if( !yRealTimeSymbol ) { return; }

        DoubleHash::const_iterator it = table->find( yRealTimeSymbol->getName() );

        if( it == table->end() ) { return; }

        d = it->second;

        if( d > m_curveDesc->maxY ) { d = m_curveDesc->maxY; }
        if( d < m_curveDesc->minY ) { d = m_curveDesc->minY; }
    }

    UpdateViewerEvent event;
    event.m_userData[_T("data")] = d;
    this->AddPendingEvent( event );
}

/*
void CurveEditor::require( DoubleHash* table)
{
    if( m_xRealTime )
    {
        (*table)[ m_xRealTime->getName() ];
    }
}

*/

void CurveEditor::require()
{
    if ( m_isSymbolRegistered ) return;

    if ( !m_xRealTimeName.IsEmpty() )
    {
        m_rtDataMgr->registerSymbol( m_xRealTimeName );
        m_isSymbolRegistered = true;
    }
    else if ( !m_yRealTimeName.IsEmpty() )
    {
        m_rtDataMgr->registerSymbol( m_yRealTimeName );
        m_isSymbolRegistered = true;
    }
}

void CurveEditor::unregisterSymbols()
{
    if ( !m_isSymbolRegistered ) return;

    if ( !m_xRealTimeName.IsEmpty() )
    {
        m_rtDataMgr->unregisterSymbol( m_xRealTimeName );
        m_isSymbolRegistered = false;
    }
    else if ( !m_yRealTimeName.IsEmpty() )
    {
        m_rtDataMgr->unregisterSymbol( m_yRealTimeName );
        m_isSymbolRegistered = false;
    }
}

void CurveEditor::OnUpdateValue( UpdateViewerEvent& event )
{
    //wxStaticText* realtimeText = (wxStaticText*)wxWindow::FindWindowById( REALTIME_X_TEXT_ID );

    if ( !m_xRealTimeName.IsEmpty() )
    {
        double x = event.m_userData[_T("data")].GetDouble();
        double y = m_curve->interpolate( x );
        m_curve->setCrosshairs( x, y );
        m_curve->Refresh();

        //wxString realtimeTextLabel = wxString::Format( m_curveDesc->columnLabelX + _T(": ") + m_curveDesc->labelFormatX
        //    + _T(" | ") + m_curveDesc->columnLabelY + _T(": ") + m_curveDesc->labelFormatY, x, y );
        //realtimeText->SetLabel( realtimeTextLabel );
    }
    else if ( !m_yRealTimeName.IsEmpty() )
    {
        double y = event.m_userData[_T("data")].GetDouble();
        m_curve->setHorizontalLine( y );
        m_curve->Refresh();

        //wxString realtimeTextLabel = wxString::Format( m_curveDesc->columnLabelY + _T(": ") + m_curveDesc->labelFormatY, y );
        //realtimeText->SetLabel( realtimeTextLabel );
    }
}

bool CurveEditor::onDropFiles( const wxArrayString& filenames )
{
    size_t nFiles = filenames.GetCount();

    if ( nFiles != 1 )
    {
        return false;
    }

    wxString filename = filenames[0];

    if ( !wxFile::Exists( filename ) )
    {
        return false;
    }

    wxFileName fn( filename );

    if( (fn.GetExt() ==  _T("vemscfg") ) || fn.GetExt()  ==  _T("txt") ){
       importFromFile( filename );
    }

    return true;
}

void CurveEditor::importFromFile( const wxString& file )
{
    MTConfigData mtConfigData;
    FirmwareConfigLoader configLoader;
    wxArrayString fileA;
    fileA.Add(file);

    wxObject* obj = VemsContext::getObject(_T("FirmwareInfo"));
    if(obj){
        Logger::logf( LOG_MSG, _T("TableEditor::importFromFile() setDestinationFwInfo") );
        configLoader.setDestinationFwInfo((FirmwareInfo*) obj);
    }

    configLoader.setMTConfigData(&mtConfigData);
    configLoader.parse(fileA);

    loadStaticViewers( configLoader );

    m_controller->updateRegisteredViewers();
}

wxString CurveEditor::getDialogName() const
{
    return m_curveDesc->name;
}

void CurveEditor::loadStaticViewers( FirmwareConfigLoader& configLoader )
{
    STDataModelSerial model( m_symbolMgr->getSymbolTable(), NULL );
    loadModelFromDialog( model );
    configLoader.load( &model );

    Symbol* yBinSymbol = m_symbolMgr->getSymbol( m_yBinName );

    if ( yBinSymbol )
    {
        for ( size_t i = 0; i  < m_yAxisViewers.GetCount(); ++i )
        {
            double value = model.getViewerData( m_yBinName, 0, i );
            m_controller->setModelData( *yBinSymbol, 0, i, value );
        }
    }

    if ( m_xBinSymbol && ( m_xBinSymbol->getName() != m_seqXAxisName ) &&
            ( m_xBinSymbol->getName() != _T("rpmBinsSpark") ) )
    {
        for ( size_t i = 0; i  < m_xAxisViewers.GetCount(); ++i )
        {
            double value = model.getViewerData( *m_xBinSymbol, 0, i );
           m_controller->setModelData( *m_xBinSymbol, 0, i, value );
        }
    }
}

void CurveEditor::loadModelFromDialog( STDataModelBase& model )
{
    Symbol* yBinSymbol = m_symbolMgr->getSymbol( m_yBinName );

    if ( yBinSymbol )
    {
        for ( size_t i = 0; i  < m_yAxisViewers.GetCount(); ++i )
        {
            double value = m_controller->getDataFromViewerModel( *yBinSymbol, 0, i );
            model.setModelData( *yBinSymbol, 0, i, value );
        }
    }

    if ( m_xBinSymbol && ( m_xBinSymbol->getName() != m_seqXAxisName ) &&
            ( m_xBinSymbol->getName() != _T("rpmBinsSpark") ) )
    {
        for ( size_t i = 0; i  < m_xAxisViewers.GetCount(); ++i )
        {
            double value = m_controller->getDataFromViewerModel( *m_xBinSymbol, 0, i );
            model.setModelData( *m_xBinSymbol, 0, i, value );
        }
    }
}

void CurveEditor::OnMouseWheel( wxMouseEvent& event )
{
	bool forwards   = ( event.GetWheelRotation() > 0 );
    double inc      = ( m_curveDesc->maxX  - m_curveDesc->minX ) / 100;

	if( forwards )
	{
        m_value += inc;
	}
	else
	{
		m_value -= inc;
	}

    if( !m_curve ) return;

    double y = m_curve->interpolate( m_value );
    m_curve->setCrosshairs( m_value, y );
    m_curve->Refresh();
}
