#include "InterpolationWidget.h"
#include "InterpolationWidgetDescriptor.h"
#include "../vemsgauge/AggGraphControl.h"
#include "../stdatamanager/STDataMgrBase.h"

#include "../MTConfigHelper.h"

BEGIN_EVENT_TABLE( InterpolationWidgetTextCtrl, wxTextCtrl )
    EVT_KILL_FOCUS( InterpolationWidgetTextCtrl::OnKillFocus )
END_EVENT_TABLE()

InterpolationWidgetTextCtrl::InterpolationWidgetTextCtrl(   wxWindow* parent,
                                                            wxWindowID id,
                                                            InterpolationWidgetInputField* desc,
                                                            STDataMgrBase* stController,
                                                            SymbolMgr * symbolMgr )
    : wxTextCtrl( parent, id )
{
    m_isActive = false;
    Disable();

    if ( desc->type == InterpolationWidgetInputField::NUMBER ) {
        ChangeValue( desc->value );
    } else if ( desc->type == InterpolationWidgetInputField::INVALID ) {
        ChangeValue( _T("INVALID") );
    } else if ( desc->type == InterpolationWidgetInputField::VARIABLE ) {
        Symbol* symbol = symbolMgr->getSymbol( desc->value );
        if ( symbol ) {
            m_symbol = *symbol;
            m_isActive = true;
            Logger::logf( LOG_MSG, _T("InterpolationWidgetTextCtrl::Contructor() - Symbol not found! - ") + desc->value );
        }
        else {
            ChangeValue( _T("INVALID") );
        }
    } else if ( desc->type == InterpolationWidgetInputField::EXPRESSION ) {
        m_expression = desc->value;
        m_isActive = true;
    }

    if ( m_isActive ) {
        m_enableExpr = desc->expression;
        stController->RegisterViewer( this );
    }
}

InterpolationWidgetTextCtrl::~InterpolationWidgetTextCtrl()
{
    if ( m_isActive ) unregisterViewer();
}

void InterpolationWidgetTextCtrl::update()
{
    wxString message = _T("InterpolationWidgetTextCtrl::update()\n");
    message += wxString::Format( _T("m_isActive: %d\n"), m_isActive );
    message += _T("symbol name: ") + m_symbol.getName() + _T("\n");
    message += _T("m_expression: ") + m_expression + _T("\n");
    Logger::logf( LOG_MSG, message );

    if ( !m_isActive ) return;

    if ( !m_symbol.getName().IsEmpty() ) {
        double value = m_stDataMgr->getDataFromViewerModel( m_symbol, 0, 0 );
        wxString message = wxString::Format( _T("m_symbol: %f"), value );
        Logger::logf( LOG_MSG, message );
        wxString text = wxString::Format( _T("%.0f"), value );
        ChangeValue( text );
        bool isEnabled = m_stDataMgr->calculate( m_enableExpr )!=0?true:false;
        Enable( isEnabled );
    } else if ( !m_expression.IsEmpty() ) {
        double value = (int)m_stDataMgr->calculate( m_expression );
        wxString message = wxString::Format( _T("m_expression: %f"), value );
        Logger::logf( LOG_MSG, message );
        wxString text = wxString::Format( _T("%.0f"), value );
        ChangeValue( text );
    }

    if ( GetParent() ) {
        if ( GetParent()->IsKindOf(CLASSINFO(InterpolationWidget)) ) {
          ( (InterpolationWidget*)GetParent() )->update();
        }
    }
}

void InterpolationWidgetTextCtrl::requestRead()
{
    if ( m_isActive == !m_symbol.getName().IsEmpty() ) {
        m_stDataMgr->addPageNeedRefresh( m_symbol.getPage() );
    }
}

void InterpolationWidgetTextCtrl::updateModel()
{
    if ( m_isActive == !m_symbol.getName().IsEmpty() )
    {
        double data = 0.0;

        if ( GetValue().ToDouble( &data ) )
        {
            m_stDataMgr->setModelData( m_symbol, 0, 0, data );
        }
    }
}


void InterpolationWidgetTextCtrl::requestWrite()
{
    if ( m_isActive == !m_symbol.getName().IsEmpty() ) {
        m_stDataMgr->requestWrite( m_symbol, 0, 0 );
    }
}

void InterpolationWidgetTextCtrl::OnKillFocus( wxFocusEvent& event )
{
    if ( m_isActive == !m_symbol.getName().IsEmpty() )
    {
        updateModel();
        requestWrite();
        m_stDataMgr->updateRegisteredViewers();
    }
}

SymbolArray InterpolationWidgetTextCtrl::getSymbols()
{
    SymbolArray symbols;

    if ( m_isActive == !m_symbol.getName().IsEmpty() ) {
        symbols.Add( new Symbol( m_symbol ) );
    }

    return symbols;
}

IMPLEMENT_DYNAMIC_CLASS(InterpolationWidget, wxPanel)

enum
{
    X0_ID = 0,
    X1_ID,
    Y0_ID,
    Y1_ID,
};

InterpolationWidget::InterpolationWidget()
{
    m_desc  = NULL;
    m_curve = NULL;
}

InterpolationWidget::InterpolationWidget(   wxWindow* parent,
                                            InterpolationWidgetDescriptor* desc,
                                            STDataMgrBase* stController,
                                            RTDataMgrBase* rtController )
    : wxPanel( parent )
{
    m_desc = desc;
    //m_stDataMgr = stController;
    m_controller = stController;
    m_rtDataMgr = rtController;

    m_isSymbolRegistered = false;
    m_realtimeSymbol = desc->realtime;

    //stController->RegisterViewer( this );

    buildGUI();

    m_controller->updateRegisteredViewers();
}

InterpolationWidget::~InterpolationWidget()
{
    //unregisterViewer();

    delete m_curve;
}

void InterpolationWidget::update()
{
    if ( !m_desc || !m_curve ) return;

    m_curve->clearPoints();

    double x = 0;
    double y = 0;

    InterpolationWidgetTextCtrl* x0Ctrl = (InterpolationWidgetTextCtrl*) wxWindow::FindWindowById(X0_ID, this );
    InterpolationWidgetTextCtrl* y0Ctrl = (InterpolationWidgetTextCtrl*) wxWindow::FindWindowById(Y0_ID, this );

    if ( x0Ctrl->GetValue().ToDouble(&x) && y0Ctrl->GetValue().ToDouble(&y) ) {
        m_curve->addPoint( x, y );
    }

    InterpolationWidgetTextCtrl* x1Ctrl = (InterpolationWidgetTextCtrl*) wxWindow::FindWindowById(X1_ID, this );
    InterpolationWidgetTextCtrl* y1Ctrl = (InterpolationWidgetTextCtrl*) wxWindow::FindWindowById(Y1_ID, this );

    if ( x1Ctrl->GetValue().ToDouble(&x) && y1Ctrl->GetValue().ToDouble(&y) ) {
        m_curve->addPoint( x, y );
    }

    m_curve->Refresh();
}

void InterpolationWidget::refresh( DoubleHash* table )
{
    DoubleHash::const_iterator it = table->find( m_realtimeSymbol );

    if ( it == table->end() ) return;

    double x = it->second;
    double y = m_curve->interpolate( x );

    m_curve->setCrosshairs( x, y );
    m_curve->Refresh();
}
//void InterpolationWidget::init( STDataMgrBase* stController, RTDataMgrBase* rtController )
//{
//    /*
//
//        m_desc->yMax.type == InterpolationWidgetInputField::NUMBER
//
//
//        m_desc->yMax
//    */
////    MTConfigData* configData = MTConfigHelper::getMTConfigData();
////    m_symbolMgr = &configData->getSymbolMgr();
////
////    m_xRealTimeName = m_curveDesc->xRealTime;
////    m_yRealTimeName = m_curveDesc->yRealTime;
////
////    Symbol* xRealTimeSymbol = NULL;
////
////    Logger::logf( LOG_MSG, _T(" CurveEditor::init() - m_xRealTimeName >") + m_xRealTimeName + _T("<") );
////
////    if ( !m_xRealTimeName.IsEmpty() )
////    {
////        xRealTimeSymbol = m_symbolMgr->getSymbol( m_xRealTimeName );
////
////        if( !xRealTimeSymbol )
////        {
////            Logger::logf( LOG_MSG, _T(" CurveEditor::init() - xRealTimeSymbol symbol not found : ") + m_xRealTimeName );
////            return false;
////        }
////    }
////    else if ( !m_yRealTimeName.IsEmpty() )
////    {
////        Symbol* yRealTimeSymbol = m_symbolMgr->getSymbol( m_yRealTimeName );
////
////        if( !yRealTimeSymbol )
////        {
////            Logger::logf( LOG_MSG, _T(" CurveEditor::init() - m_yRealTimeName symbol not found : ") + m_yRealTimeName );
////            return false;
////        }
////    }
////
////    wxString xBinName = m_curveDesc->xBins;
////    m_xBinSymbol = m_symbolMgr->getSymbol(xBinName );
////
////    if( !m_xBinSymbol )
////    {
////        Logger::logf( LOG_MSG, _T(" CurveEditor::init() - m_xBinSymbol symbol not found : ") + xBinName );
////        Logger::logf( LOG_MSG, _T(" CurveEditor::init() - addr of xRealTimeSymbol: %p "), xRealTimeSymbol );
////
////        if(xBinName == m_seqXAxisName && xRealTimeSymbol ){
////            m_xBinSymbol = new Symbol();
////            m_xBinSymbol->setBasic(m_seqXAxisName, xRealTimeSymbol->getUnit(), true, -1, -1);
////            m_xBinSymbol->setArray(true, 0, m_curveDesc->divisionsX ,1);
////        }else{
////            return false;
////        }
////    }
////
////    m_yBinName = m_curveDesc->yBins;
////    Symbol* yBinSymbol = m_symbolMgr->getSymbol( m_yBinName );
////
////    if( !yBinSymbol )
////    {
////        Logger::logf( LOG_MSG, _T(" CurveEditor::init() - dialog symbol not found : ") + m_yBinName );
////        return false;
////    }
////
////
////    m_controller = stController;
////    buildGUI( m_controller, rtController );
//
//    return true;
//}

void InterpolationWidget::require()
{
    if ( !m_isSymbolRegistered && !m_realtimeSymbol.IsEmpty() ) {
        m_rtDataMgr->registerSymbol( m_realtimeSymbol );
        m_isSymbolRegistered = true;
    }
}

void InterpolationWidget::unregisterSymbols()
{
    if ( m_isSymbolRegistered && !m_realtimeSymbol.IsEmpty() ) {
        m_rtDataMgr->unregisterSymbol( m_realtimeSymbol );
        m_isSymbolRegistered = false;
    }
}

void InterpolationWidget::buildGUI()
{
    MTConfigData* configData = MTConfigHelper::getMTConfigData();
    SymbolMgr* symbolMgr = &configData->getSymbolMgr();

    m_curve = new AggGraphControl( this );
    m_curve->SetInitialSize( wxSize( 400, 400 ) );
    m_curve->setMinX( m_desc->xAxis.min );
    m_curve->setMaxX( m_desc->xAxis.max );
    m_curve->setDivisionsX( m_desc->xAxis.divisions );
    m_curve->setMinY( m_desc->yAxis.min );
    m_curve->setMaxY( m_desc->yAxis.max );
    m_curve->setDivisionsY( m_desc->yAxis.divisions );

    wxStaticText* xTitle = new wxStaticText( this, wxID_ANY, m_desc->xAxis.title );
    wxFont titleFont = xTitle->GetFont();
    titleFont.SetWeight( wxFONTWEIGHT_BOLD );
    xTitle->SetFont( titleFont );

    wxString xUnitString = m_desc->xAxis.unit;
    if ( !xUnitString.IsEmpty() ) {
        xUnitString = _T("[") + xUnitString + _T("]");
    }

    wxStaticText* xUnit = new wxStaticText( this, wxID_ANY, xUnitString );

    wxStaticText* yTitle = new wxStaticText( this, wxID_ANY, m_desc->yAxis.title );
    yTitle->SetFont( titleFont );

    wxString yUnitString = m_desc->yAxis.unit;
    if ( !yUnitString.IsEmpty() ) {
        yUnitString = _T("[") + yUnitString + _T("]");
    }

    wxStaticText* yUnit = new wxStaticText( this, wxID_ANY, yUnitString );

    InterpolationWidgetTextCtrl* y1TextCtrl = new InterpolationWidgetTextCtrl( this, Y1_ID, &m_desc->y1, m_controller, symbolMgr );
    InterpolationWidgetTextCtrl* y0TextCtrl = new InterpolationWidgetTextCtrl( this, Y0_ID, &m_desc->y0, m_controller, symbolMgr );
    InterpolationWidgetTextCtrl* x0TextCtrl = new InterpolationWidgetTextCtrl( this, X0_ID, &m_desc->x0, m_controller, symbolMgr );
    InterpolationWidgetTextCtrl* x1TextCtrl = new InterpolationWidgetTextCtrl( this, X1_ID, &m_desc->x1, m_controller, symbolMgr );

    wxSize size = wxDefaultSize;
    size.SetWidth( 60 );

    y1TextCtrl->SetInitialSize( size );
    y0TextCtrl->SetInitialSize( size );
    x0TextCtrl->SetInitialSize( size );
    x1TextCtrl->SetInitialSize( size );

    wxStaticText* x0Label = new wxStaticText( this, wxID_ANY, m_desc->x0.label );
    wxStaticText* x1Label = new wxStaticText( this, wxID_ANY, m_desc->x1.label );
    wxStaticText* y0Label = new wxStaticText( this, wxID_ANY, m_desc->y0.label );
    wxStaticText* y1Label = new wxStaticText( this, wxID_ANY, m_desc->y1.label );

    wxFont labelFont = x0Label->GetFont();
    labelFont.SetWeight( wxFONTWEIGHT_BOLD );
    x0Label->SetFont( labelFont );
    x1Label->SetFont( labelFont );
    y0Label->SetFont( labelFont );
    y1Label->SetFont( labelFont );

    wxGridSizer* axesSizer = new wxFlexGridSizer( 3, 4, 1, 1 );

    axesSizer->Add( xTitle, 0, wxEXPAND | wxLEFT, 5 );
    axesSizer->Add( xUnit, 0, wxEXPAND , 0 );
    axesSizer->Add( yTitle, 0, wxEXPAND | wxLEFT, 5 );
    axesSizer->Add( yUnit, 0, wxEXPAND, 0 );

    axesSizer->Add( x0Label, 0, wxEXPAND | wxLEFT | wxALIGN_CENTER_VERTICAL, 10 );
    axesSizer->Add( x0TextCtrl, 0, wxEXPAND, 0 );
    axesSizer->Add( y0Label, 0, wxEXPAND | wxLEFT | wxALIGN_CENTER_VERTICAL, 10 );
    axesSizer->Add( y0TextCtrl, 0, wxEXPAND, 0 );

    axesSizer->Add( x1Label, 0, wxEXPAND | wxLEFT | wxALIGN_CENTER_VERTICAL, 10 );
    axesSizer->Add( x1TextCtrl, 0, wxEXPAND, 0 );
    axesSizer->Add( y1Label, 0, wxEXPAND | wxLEFT | wxALIGN_CENTER_VERTICAL, 10 );
    axesSizer->Add( y1TextCtrl, 0, wxEXPAND, 0 );

    wxBoxSizer* mainSizer = new wxBoxSizer( wxVERTICAL );
    mainSizer->Add( m_curve, 1, wxEXPAND, 0 );

    mainSizer->Add( axesSizer, 0, wxTOP | wxBOTTOM , 5 );

    SetSizer( mainSizer );
}
