#include "AggGraphControl.h"

#include <wx/dcbuffer.h>

#include "../Math.h"
#include "../logger.h"
#include "../agg/wxAgg.h"
#include "../DirHelper.h"



BEGIN_EVENT_TABLE( AggGraphControl, wxControl )
    EVT_SIZE ( AggGraphControl::OnSize )
    EVT_PAINT( AggGraphControl::OnPaint )
END_EVENT_TABLE()


AggGraphControl::AggGraphControl( wxWindow* parent )
{
    m_minX = 0;
    m_maxX = 100;
    m_divisionsX = 11;

    m_minY = 0;
    m_maxY = 100;
    m_divisionsY = 11;

    m_crosshairsX = m_minX;
    m_crosshairsY = m_minY;

    m_crosshairsSize    = 5;
    m_borderSize        = 5;
    m_pointSize         = 8;

    m_colLabelFormat    = _T("%.0f");
    m_rowLabelFormat    = _T("%.0f");

    m_isResized = true;
    m_bmp       = NULL;

    createCrosshairs();

    m_showCrosshair = false;
    m_showLine  = false;
    m_showHorizontalHelpLine = false;
    m_showVerticalHelpLine = false;
    m_fontPointSize = 10;

#if defined(__WXMSW__)
    SetBackgroundStyle( wxBG_STYLE_CUSTOM );
#elif defined(__UNIX__)
    SetBackgroundStyle( wxBG_STYLE_SYSTEM );
#endif
    Create( parent, wxID_ANY );
}


AggGraphControl::~AggGraphControl()
{
    delete m_bmp;
    delete m_bmpCrosshairs;
}


void AggGraphControl::createCrosshairs()
{
    int width   = (int)(m_crosshairsSize * 2.0);
    int height  = (int)(m_crosshairsSize * 2.0);

    wxAgg aggCanvas( width, height );
    //aggCanvas.setPen( *wxRED );
    aggCanvas.setPen( wxColor( 255, 255, 128, 255 ) );

    aggCanvas.drawLine( 0, 0, width, height, 2 );
    aggCanvas.drawLine( width, 0, 0, height, 2 );

    m_bmpCrosshairs = aggCanvas.renderToBitmap();
}

void AggGraphControl::updateSize()
{
    const int maxFontPointSize = 10;

    m_width         = GetSize().GetWidth();
    m_height        = GetSize().GetHeight();

    double  divisionWidth =  floor( ( m_width - m_borderSize * 3 )  / ( m_divisionsX + 0.5 ) );


    wxClientDC clientDC( this );
    clientDC.SetFont( wxFont( 10, wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL ) );
    wxSize m_labelsMaxSize = clientDC.GetTextExtent( _T("00000") );

    double maxValueWidth = m_labelsMaxSize.GetWidth();

    if ( maxValueWidth != divisionWidth )
    {
        wxFont font = clientDC.GetFont();
        int fontPointSize = font.GetPointSize();

        fontPointSize = (int)((double)fontPointSize * divisionWidth / maxValueWidth);
        if ( fontPointSize > maxFontPointSize ) fontPointSize = maxFontPointSize;

        font.SetPointSize( fontPointSize );
        clientDC.SetFont( font );
        m_labelsMaxSize = clientDC.GetTextExtent( _T("00000") );
        m_fontPointSize = fontPointSize;
    }

    m_textWidth = m_labelsMaxSize.GetWidth();

    //m_graphWidth    = divisionWidth * ( m_divisionsX - 1 );
    m_graphWidth    = m_width - 1.5 * m_textWidth - m_borderSize * 3; //divisionWidth * ( m_divisionsX - 1 );
    m_offsetX       = m_borderSize * 2 + m_textWidth;
    m_rangeX        = m_maxX - m_minX;
    m_scaleX        = m_graphWidth / m_rangeX;
    m_stepX         = m_rangeX / ( m_divisionsX - 1 );

    m_textHeight = m_labelsMaxSize.GetHeight();

    m_graphHeight  = m_height - m_textHeight - m_borderSize * 3;
    m_rangeY       = m_maxY - m_minY;
    m_scaleY       = m_graphHeight / m_rangeY;
    m_offsetY      = m_height - m_borderSize * 2 - m_textHeight;
    m_stepY        = m_rangeY / ( m_divisionsY - 1 );
}

//void AggGraphControl::updateSize()
//{
//    m_width        = GetSize().GetWidth();
//    m_height       = GetSize().GetHeight();
//
//    double  divisionWidth =  ( m_width - m_borderSize * 3 )  / ( m_divisionsX + 0.5 );
//    Logger::logf( LOG_MSG, _T("divisionWidth: %f"), divisionWidth );
//
//    wxClientDC cDC( this );
//    wxSize textSize = cDC.GetTextExtent( _T("8888.8") );
//
//    m_textWidth    = textSize.GetWidth();
//    m_textHeight   = textSize.GetHeight();
//
//    m_graphWidth   = m_width - m_textWidth * 1.5  - m_borderSize * 3;
//    m_graphHeight  = m_height - m_textHeight * 1.5 - m_borderSize * 3;
//    m_rangeX       = m_maxX - m_minX;
//    m_rangeY       = m_maxY - m_minY;
//    m_scaleX       = m_graphWidth / m_rangeX;
//    m_scaleY       = m_graphHeight / m_rangeY;
//    m_offsetX      = m_borderSize * 2 + m_textWidth;
//    m_offsetY      = m_height - m_borderSize * 2 - m_textHeight;
//    m_stepX        = m_rangeX / ( m_divisionsX - 1 );
//    m_stepY        = m_rangeY / ( m_divisionsY - 1 );
//
//    Logger::logf( LOG_MSG, _T("m_stepX * m_scaleX: %f"), m_stepX * m_scaleX );
//}


wxRealPoint AggGraphControl::translate( double x, double y, bool round )
{
    double tx =  x * m_scaleX + m_offsetX;
    double ty = -y * m_scaleY + m_offsetY;    //mirror y

    if ( round ) {
        tx =  Math::round( tx ) + 0.5;
        ty =  Math::round( ty ) + 0.5;
    }

    return wxRealPoint( tx, ty );
}


double AggGraphControl::interpolate( double x )
{
    if ( !m_points.IsEmpty() )
    {
        if ( x < m_points[0]->x ) {
            return m_points[0]->y;
        } else if ( x > m_points.Last()->x ) {
            return m_points.Last()->y;
        }
    }

    double x0   = 0;
    double y0   = 0;
    double x1   = 0;
    double y1   = 0;
    double y    = m_minY;
    bool found  = false;

    uint32_t i  = 0;
    for( i = 0; i < m_points.GetCount(); ++i )
    {
        if( m_points[ i ]->x < x ) { continue; }

        found = true;

        x1 = m_points[ i ]->x;
        y1 = m_points[ i ]->y;

        if( i == 0 ) { break; }

        x0 = m_points[ i - 1 ]->x;
        y0 = m_points[ i - 1 ]->y;

        break;
    }

    if( found ) {
        y = y0 + ( x - x0 ) * ( y1 - y0 ) / ( x1 - x0 );
    }

    return y;
}


void AggGraphControl::addPoint( double x, double y )
{
    m_points.Add( new wxRealPoint( x, y ) );
    m_isResized = true;
    Refresh();
}


void AggGraphControl::drawGrids()
{
    wxAgg aggCanvas( m_width, m_height );

    aggCanvas.clear( *wxBLACK );
    aggCanvas.setPen( *wxGREEN );

    double x1 = 0.0;
    double y1 = 0.0;
    double x2 = 0.0;
    double y2 = 0.0;

    uint32_t i = 0;

    //draw vertical grid lines
    for( i = 0; i < m_divisionsX; ++i )
    {
        x1 = Math::round( translate( i * m_stepX, 0.0 ).x ) + 0.5;
        //y1 = m_borderSize + m_textHeight / 2;
        y1 = m_borderSize;
        x2 = x1;
        //y2 = m_borderSize * 2 + m_textHeight / 2 + m_graphHeight;
        y2 = m_graphHeight + 2 * m_borderSize;

        aggCanvas.drawLine( x1, y1, x2, y2, 0.5 );
    }

    //draw horizontal grid lines
    for( i = 0; i < m_divisionsY; ++i )
    {
        x1 = m_offsetX - m_borderSize;
        y1 = Math::round( translate( 0.0, i * m_stepY  ).y ) + 0.5;
        //x2 = m_width - m_borderSize - m_textWidth / 2;
        x2 = m_offsetX + m_graphWidth;// + 1.5 * m_textWidth + m_borderSize * 2;
        y2 = y1;

        aggCanvas.drawLine( x1, y1, x2, y2, 0.5 );
    }

    delete m_bmp;
    m_bmp = aggCanvas.renderToBitmap();
}

#include "../agg/wxAgg2.h"

void AggGraphControl::drawLabels()
{
    wxAgg2 canvas( m_bmp );

    canvas.setPen( wxColor( 255, 255, 255, 255 ) );

    wxString value      = wxEmptyString;
    double valueWidth   = 0.0;
    double valueHeight  = 0.0;
    double x = 0;
    double y = m_offsetY + m_borderSize;;

    uint32_t i = 0;

    wxString fontfile = DirHelper::GetPath( dFONTS ) + _T("arial.ttf");
    canvas.loadTTFont( fontfile, m_fontPointSize, 1.0 );

    //draw horizontal labels
    for( i = 0; i < m_divisionsX; ++i )
    {
        //value       = wxString::Format( _T("%4.1f"), i * m_stepX + m_minX );
        value       = wxString::Format( m_colLabelFormat, i * m_stepX + m_minX );
        valueWidth = canvas.getTextWidth( value, m_fontPointSize );

        x = i * m_stepX * m_scaleX + m_offsetX - valueWidth / 4;
        canvas.drawTTFText( value, x, y + m_fontPointSize );
    }

    //draw vertical labels
    for( i = 0; i < m_divisionsY; ++i )
    {
        //value       = wxString::Format( _T("%4.1f"), i * m_stepY + m_minY );
        value       = wxString::Format( m_rowLabelFormat, i * m_stepY + m_minY );

        wxRealRect rect = canvas.getTTFTextBound( value );
        valueHeight = rect.h;
        valueWidth = rect.w;

        x = m_textWidth - valueWidth;
        y = m_offsetY - i * m_stepY * m_scaleY - valueHeight / 2;

        canvas.drawTTFText( value, x, y + m_fontPointSize );
    }

    wxMemoryDC mDC( *m_bmp );
    canvas.render( mDC );

    /// TRANSPARENCY PROBLEMS ON WIN7
    /*wxMemoryDC mDC( *m_bmp );
    mDC.SetBackgroundMode(wxSOLID);
    mDC.SetTextForeground( wxColor(255,255,255,255) );

    wxString value      = wxEmptyString;
    double valueWidth   = 0.0;
    double valueHeight  = 0.0;
    double x = 0;
    double y = m_offsetY + m_borderSize;;

    uint32_t i = 0;

    wxFont font = mDC.GetFont();
    font.SetPointSize( m_fontPointSize );
    mDC.SetFont( font );

    //draw horizontal labels
    for( i = 0; i < m_divisionsX; ++i )
    {
        //value       = wxString::Format( _T("%4.1f"), i * m_stepX + m_minX );
        value       = wxString::Format( m_colLabelFormat, i * m_stepX + m_minX );
        valueWidth  = mDC.GetTextExtent( value ).GetWidth();

        x = i * m_stepX * m_scaleX + m_offsetX - valueWidth / 2;

        mDC.DrawText( value, (int)x , (int)y );

    }

    //draw vertical labels
    for( i = 0; i < m_divisionsY; ++i )
    {
        //value       = wxString::Format( _T("%4.1f"), i * m_stepY + m_minY );
        value       = wxString::Format( m_rowLabelFormat, i * m_stepY + m_minY );
        valueHeight = mDC.GetTextExtent( value ).GetHeight();
        valueWidth  = mDC.GetTextExtent( value ).GetWidth();

        x = m_textWidth - valueWidth;
        y = m_offsetY - i * m_stepY * m_scaleY - valueHeight / 2;

        mDC.DrawText( value, (int)x , (int)y );
    }*/
    }


void AggGraphControl::drawLines()
{
    wxAgg aggCanvas( m_width, m_height );

    aggCanvas.setPen( *wxWHITE );

    wxRealPoint p0;
    wxRealPoint p1;

    uint32_t i = 0;

    for( i = 1; i < m_points.GetCount(); ++i )
    {
        p0 = translate( m_points.Item( i - 1 )->x - m_minX, m_points.Item( i - 1 )->y - m_minY , true );
        p1 =  translate( m_points.Item( i )->x - m_minX, m_points.Item( i )->y - m_minY, true );
        aggCanvas.drawLine( p0.x, p0.y, p1.x, p1.y );
    }

    if ( !m_points.IsEmpty() )
    {
        if ( m_points[0]->x >= m_minX )
        {
            p0 = translate( 0, m_points[0]->y - m_minY, true );
            p1 =  translate( m_points[0]->x - m_minX, m_points[0]->y - m_minY, true );
            aggCanvas.drawLine( p0.x, p0.y, p1.x, p1.y );
        }

        int maxIdx = m_points.size() - 1;

        if (  m_points[maxIdx]->x <= m_maxX )
        {
            p0 = translate( m_points[maxIdx]->x - m_minX, m_points[maxIdx]->y - m_minY, true );
            p1 = translate( m_maxX - m_minX, m_points[maxIdx]->y - m_minY, true );

            aggCanvas.drawLine( p0.x, p0.y, p1.x, p1.y );
        }
    }

    wxMemoryDC mDC( *m_bmp );
    aggCanvas.render( mDC );
}


void AggGraphControl::drawPoints()
{
    wxAgg aggCanvas( m_width, m_height );
    aggCanvas.setBrush( *wxRED );

    wxRealPoint p( 0, 0 );

    double x1 = 0.0;
    double y1 = 0.0;
//    double x2 = 0.0;
//    double y2 = 0.0;

    uint32_t i = 0;

    for( i = 0; i < m_points.GetCount(); ++i )
    {
        p = translate( m_points.Item( i )->x - m_minX, m_points.Item( i )->y - m_minY );

        x1 = Math::round( p.x - m_pointSize / 2 ) + 0.5;
        y1 = Math::round( p.y - m_pointSize / 2 ) + 0.5;

        aggCanvas.drawSolidRectangle( x1, y1, m_pointSize, m_pointSize );
    }

    wxMemoryDC mDC( *m_bmp );
    aggCanvas.render( mDC );
}


void AggGraphControl::drawCrosshairs( wxDC& dc )
{
    wxRealPoint p = translate( m_crosshairsX - m_minX, m_crosshairsY - m_minY );

    int x = (int)(Math::round( p.x - m_crosshairsSize ));
    int y = (int)(Math::round( p.y - m_crosshairsSize ));

    if( m_bmpCrosshairs )
    {
        dc.DrawBitmap( *m_bmpCrosshairs, x, y, true );
    }

    if ( !m_showHorizontalHelpLine && !m_showVerticalHelpLine ) return;

    wxAgg aggCanvas( m_width, m_height );
    aggCanvas.setPen( wxColor( 255, 255, 0 ) );

    wxRealPoint min = translate( m_minX, m_minY );
    p.x = Math::round( p.x ) + 0.5;
    p.y = Math::round( p.y ) + 0.5;

    if ( m_showHorizontalHelpLine )
    {
        aggCanvas.drawLine( min.x, p.y, p.x, p.y, 1.0 );
    }

    if ( m_showVerticalHelpLine )
    {
        aggCanvas.drawLine( p.x, p.y, p.x, min.y, 1.0 );
    }

    aggCanvas.render( dc );
}

void AggGraphControl::showLine( wxDC& dc )
{
    wxAgg aggCanvas( m_width, m_height );
    aggCanvas.setPen( *wxRED );

    double x1 = m_offsetX - m_borderSize;
    //double y1 = Math::round( translate( 0.0, i * m_stepY  ).y ) + 0.5;
    double y1 = Math::round( translate( 0.0, m_lineY - m_minY ).y ) + 0.5;
    double x2 = m_width - m_borderSize - m_textWidth / 2;
    double y2 = y1;

    aggCanvas.drawLine( x1, y1, x2, y2, 2.0 );

    //wxMemoryDC mDC( *m_bmp );
    //aggCanvas.render( mDC );
    //wxMemoryDC mDC( *m_bmp );
    aggCanvas.render( dc );
}

void AggGraphControl::OnSize( wxSizeEvent& event )
{
    m_isResized = true;
    Refresh();
}


void AggGraphControl::OnPaint( wxPaintEvent& event )
{

    wxBufferedPaintDC dc( this );
    dc.SetBackground( *wxBLACK );
    dc.Clear();

    if( m_isResized )
    {
        updateSize();
        drawGrids();
        drawLabels();
        drawLines();
        drawPoints();
        m_isResized = false;
    }

    if( m_bmp )
    {
        dc.DrawBitmap( *m_bmp, 0, 0, true );
    }

    if ( m_showCrosshair ) drawCrosshairs( dc );
    if ( m_showLine ) showLine( dc );


    //DrawExtraLayer(dc);
}


