#include "CommunicationController.h"

#include <climits>
#include "../serialhelper.h"


#include "../CommunicationBase.h"
#include "ProtocolBase.h"
#include "ECUStateMgrBase.h"

#include "ECUInfo.h"
#include "VemsCommand.h"
#include "SIPRTypes.h"

#include "../events/InputEvent.h"
#include "../events/EcuCommEvent.h"
#include "../events/DeviceEvent.h"
#include "../logger.h"
#include "../SerialException.h"
#include "../Math.h"

#include "../ringbuffer/CommunicationThread.h"
#include "../FirmwareInfo.h"
#include "../events/ECUEvent.h"

#include <wx/sstream.h>
#include <wx/datstrm.h>

#include "ComCtrlStates.h"
#include "TerminalPanel.h"

#include <wx/textfile.h>
#include "../DirHelper.h"
#include <wx/regex.h>
#include "../UnicodeHelper.h"
#include "PasswordHelper.h"


const wxEventType ECU_IS_PASSWORD_PROTECTED = wxNewEventType();

CommunicationController::CommunicationController(   ECUStateMgrBase* stateMgr,
                                                    FirmwareInfo* fwInfo,
                                                    wxEvtHandler* pEventHandler )
{
    m_stateMgr = stateMgr;
    m_commBase = stateMgr->getCommunicationBase();
    m_pEventHandler = pEventHandler;

    //if ( m_pEventHandler )
    //{
//    Connect( wxEVT_THREAD, wxCommandEventHandler( CommunicationController::OnThread ) );
    //}

    //m_bWaitForBaudrateChanging = false;
    //m_bWaitForPing             = false;
    m_iSavedState               = ECUInfo::UNKNOWN;

    m_fwInfo = fwInfo;
    if ( !fwInfo )
    {
        m_fwInfo = new FirmwareInfo();
    }

    m_pCurrentProtocol = NULL;

    reset();

    m_device    = ECUInfo::DEVICE_UNKNOWN;

    m_state     = ECUInfo::UNKNOWN;

    m_bReadSensors = true;

    m_direct = false;

    m_detectingProtocolIdx = (size_t)-1; //std::numeric_limits<size_t>::max();
    m_interWriteDelay = 0;
    m_preferredBaudrate = 19200;
    Connect( wxEVT_INPUT, InputEventHandler( CommunicationController::OnSIPRInput ) );

    m_deviceIsProtected = false;
    m_leaveBootModeFlag = false;
}


CommunicationController::~CommunicationController()
{

}


void CommunicationController::reset()
{
    Logger::logf( LOG_MSG, _T("CommunicationController::reset() START") );
    /*
    m_pCtrlState = CCS_Normal::instance();
    setProtocolForState( ECUInfo::SIPR );

    m_pPreviousCtrlState    = NULL;

    if ( m_pCurrentProtocol ) m_pCurrentProtocol->reset();

    m_iPreviousState    = ECUInfo::UNKNOWN;
    m_pCurrentCommand   = NULL;
    m_VemsCommandQueue.clear();

    m_device =  ECUInfo::DEVICE_ARM;
*/

    m_pCtrlState = CCS_Sync::instance();

    //Logger::logf( LOG_MSG, _T("CommunicationController::reset() %p"), m_pCtrlState);

    m_pPreviousCtrlState    = NULL;

    if ( m_pCurrentProtocol ) m_pCurrentProtocol->reset();
   // m_pCurrentProtocol  = NULL;
    m_iPreviousState    = ECUInfo::UNKNOWN;
    //m_pCurrentCommand   = NULL;
    //m_iControllerState  = NORMAL;
    m_VemsCommandQueue.clear();

    m_stateMgr->reset();

    Logger::logf( LOG_MSG, _T("CommunicationController::reset() END") );
}


void CommunicationController::start( int priority )
{
    m_deviceIsProtected = false;

    this->ProcessBase::start();

    ECUCommEvent event;
    event.state = ECUCommEvent::ONLINE;
    m_pEventHandler->AddPendingEvent( event );

}


void CommunicationController::stop()
{
    this->ProcessBase::stop();
    m_stateMgr->reset();
    m_commBase->close();
}

void CommunicationController::finished()
{
    sendOfflineEvents();
}

int CommunicationController::getDevice()
{
    return m_stateMgr->getDevice();
}


int CommunicationController::getState()
{
    //return m_stateMgr->getStoredState();
    if ( m_pCurrentProtocol ) return m_pCurrentProtocol->type();
    return ECUInfo::UNKNOWN;
}


/// Must be non blocking (invoked from thread)
int CommunicationController::readBytes()
{
    //Logger::logf( LOG_MSG, _T("CommunicationController::readBytes()") );
    int rxCount = 1;
    int receivedBytes = 0;
    receivedBytes = m_commBase->Read(m_rxBuff, rxCount);

//    if ( receivedBytes ) TerminalLog::rxLog( m_rxBuff, receivedBytes );

    m_pCurrentProtocol->rx(m_rxBuff, receivedBytes);

    return receivedBytes;
}

/// Must be non blocking (invoked from thread)
int CommunicationController::writeBytes()
{
    //Logger::logf( LOG_MSG, _T("CommunicationController::writeBytes()") );
    int txCount = 0;
    //for(){

    m_pCurrentProtocol->tx(m_txBuff, txCount);

   //Logger::logf( LOG_MSG, _T("CommunicationController::writeBytes() - %d"), txCount );
    if ( txCount > 0 )
    {
        m_commBase->SendString(m_txBuff, txCount);
        //Logger::logf( LOG_MSG, _T("CommunicationController::writeBytes() - size: %d"), txCount );
/*        try{
            m_commBase->SendString(m_txBuff, txCount);
        }catch( SerialException& se ){
            return 0;
        }catch (...){
            return 0;
        }
        */
        //Logger::logf( LOG_MSG, _T("CommunicationController::writeBytes() - NO EXCEPTION") );
        //InputEvent e;
        //e.setPacket( m_rxPacket );
        //m_eventHandler->AddPendingEvent( e );
    }

    if ( txCount ) TerminalLog::txLog( m_txBuff, txCount );

    //}
    return txCount;
}

void CommunicationController::registerProtocol( ProtocolBase* pb )
{
    /// First registered is the thefault;
    if(m_detectingOrder.size() == 0 ){ m_pCurrentProtocol = pb; }

    Logger::logf( LOG_MSG, _T("CommunicationController::registerProtocol() - addr: %p"), pb );
    /// JUST for DEBUG
    //if( pb->type()  != ECUInfo::MEGATUNE) return;

    if ( pb )
    {
        Logger::logf( LOG_MSG, _T("CommunicationController::registerProtocol() type: %i"), pb->type() );
        m_ProtocolHash[ pb->type() ] = pb;
        m_detectingOrder.Add( pb->type() );
    }
    //m_detectingOrder.Add( ECUInfo::MEGATUNE );
    //m_pCurrentProtocol = pb;
    m_detectingProtocolIdx = m_detectingOrder.size() -1;
}


bool CommunicationController::needClear()
{
    if ( m_pCurrentProtocol )
    {
        return m_pCurrentProtocol->needClear();
    }

    return false;
}

int CommunicationController::setState( int state )
{
    return m_stateMgr->setState( m_state, state );
}

void CommunicationController::setBaudRate( int baudRate )
{
    m_commBase->setBaudRate( baudRate );
}


void CommunicationController::request( VemsCommand* command )
{
    //Logger::logf(LOG_MSG, _T("CommunicationController::request() %p"), this);
    //m_pCtrlState->request( this, command );

    m_VemsCommandQueue.push_back( command );

}


void CommunicationController::OnSIPRInput( InputEvent& event )
{
    //m_pCtrlState->AddPendingEvent( event );
    //Logger::logf( LOG_MSG, _T("CommunicationController::OnSIPRInput(%p) %p"), this, m_pCtrlState);
    m_pCtrlState->OnSipr( this, event );

}


void CommunicationController::execute()
{
    if ( m_deviceIsProtected ) {
        return;
    }
    //Logger::logf( LOG_MSG, _T("CommunicationController::execute(%p) START %p"), this, m_pCtrlState);
    processNextCommand();
    //Logger::logf( LOG_MSG, _T("CommunicationController::execute() START 100") );
    communicate();
    //Logger::logf( LOG_MSG, _T("CommunicationController::execute() START 200") );
    m_pCtrlState->execute( this );


    //Logger::logf( LOG_MSG, _T("CommunicationController::execute(%p) STOP %p"), this, m_pCtrlState);
}


bool CommunicationController::isProtocolRegistered( int protocol )
{
    if ( m_ProtocolHash.find( protocol ) != m_ProtocolHash.end() ) { return true; }

    return false;
}

bool CommunicationController::setProtocolForState( int state )
{
    if ( isProtocolRegistered( state ) )
    {
        Logger::logf( LOG_MSG, _T("CommunicationController::setProtocolForState() - ") +
            ECUInfo::getStateAsString( state ) );
        m_pCurrentProtocol = m_ProtocolHash[ state ];
        TerminalLog::protocolLog( ECUInfo::getStateAsString( getState() ) );
        return true;
    }
    else
    {
        m_pCurrentProtocol = NULL;
        TerminalLog::protocolLog( ECUInfo::getStateAsString( getState() ) );
        Logger::logf( LOG_MSG, _T("CommunicationController::setProtocolForState() - no registered protocol for state ") +
                        ECUInfo::getStateAsString( state ) + _T("!") );
        return false;
    }
}


void CommunicationController::processNextCommand()
{
    if( m_VemsCommandQueue.empty() ) return;
    VemsCommand* pCurrentCommand = m_VemsCommandQueue.front();
    if ( !pCurrentCommand )
    {
        /// Empty
        //Logger::logf( LOG_MSG, _T("---CommunicationController::processNextCommand() - m_pCurrentCommand is NULL!!!") );
        return;
    }
    m_VemsCommandQueue.pop_front();
    //Logger::logf( LOG_MSG, _T("---CommunicationController::processNextCommand() - ") + pCurrentCommand->toString() );


    switch ( pCurrentCommand->getType() )
    {
        case VemsCommand::SET_STATE:

            setCtrlState( CCS_SetState::instance() );
            CCS_SetState::instance()->m_expectedState =  pCurrentCommand->getState();
            break;
        //case VemsCommand::SET_STATE_END:
        //    restorePreviousCtrlState();
        //    break;
        case VemsCommand::SET_BAUDRATE:
            break;
        default:
            m_pCurrentProtocol->request( pCurrentCommand );
    }


    ///
/*
    switch ( m_pCurrentCommand->getType() )
    {
        case VemsCommand::SET_STATE:
        case VemsCommand::SET_BAUDRATE:
            delete m_pCurrentCommand;
            m_pCurrentCommand = NULL;
            break;
        default:
        {
            m_pCurrentProtocol->request( m_pCurrentCommand );
            m_pCurrentCommand = NULL;
        }
    }
*/
}

/*
// the callback for our custom event
void CommunicationController::OnThread( wxCommandEvent& event )
{
    Logger::logf( LOG_MSG, _T("-----CommunicationController::OnThread()") );
    // handle thread events here, i. e. a premature end due to internal error
    m_pThread = NULL; // thread terminated itself (detached,  was freed automatically)
    m_stateMgr->reset();

    sendOfflineEvents();
}
*/

void CommunicationController::save()
{
#ifdef COM_TEST
    m_iSavedState = getState();
#else
    m_iSavedState = m_stateMgr->getStoredState();
#endif

}


void CommunicationController::restore()
{
    if ( m_iSavedState != ECUInfo::UNKNOWN )
    {
        request( new SetStateCommand( m_iSavedState ) );
    }
}


bool CommunicationController::process()
{
    try{
        if(m_direct){
            communicate();
        }else{
            execute();
        }
    }catch(SerialException& se){
        Logger::logf( LOG_MSG, _T("-----CommunicationController::process() ") + se.toString() );
        return false;
    }catch (...){
        return false;
    }
    return true;
}
/*
bool CommunicationController::ProcessEvent(wxEvent& event )
{
    //Logger::logf( LOG_MSG, _T("CommunicationController::ProcessEvent()"));
    return this->wxEvtHandler::ProcessEvent(event);
    //return ((wxEvtHandler*)this)->ProcessEvent(event);
}
*/
void CommunicationController::setDirect(ProtocolBase* pb)
{
    if ( pb==NULL ) m_direct = false;
    else {
        m_direct = true;
        m_pCurrentProtocol = pb;

    }
}

FirmwareVersion CommunicationController::getFirmwareVersion() const
{
    return m_fwInfo->getFirmwareVersion();
}


void CommunicationController::fetchRoundFirmwareInfo()
{
    uint8_t* command_serNum = (uint8_t*)"mci.";
    uint8_t* command_Signature = (uint8_t*)"S.";

    wxString response;
    CommunicationBase* comBase = m_stateMgr->getCommunicationBase();

    comBase->getAnswer(command_serNum, 12, 12, 5, &response);

    const wxCharBuffer buffer =  response.To8BitData();
    const char* data = buffer.data();

    wxString serialString = response.BeforeFirst( _T('\0') ) + _T("_");
    if ( response.length() >= 12 ) {
        uint32_t serial =  Math::getNumber((uint8_t*)&data[8], 4 );
        serial -= 0x01010100;
        serialString += wxString::Format( _T("%u"), serial );
    }

    m_fwInfo->setSerialNumber(serialString);

    m_stateMgr->setState(ECUInfo::MEGATUNE);
    comBase->getAnswer(command_Signature, 26, 20, 5, &response);
    m_fwInfo->setSignature(response);

    response = response.AfterLast( _T('v') );
    m_fwInfo->setFirmwareVersion(response);
}

/*
void CommunicationController::fetchV3FirmwareVersion()
{

}
*/


void CommunicationController::fetchV3FirmwareInfo()
{
    uint8_t* command_ctime = (uint8_t*)"Manmdv";
    uint8_t* command_fwversion = (uint8_t*)"ManmdV"; //31 2E 31 2E 34 39 0A 00// 8
    uint8_t* command_serNum = (uint8_t*)"Manmci";    //76 33 2E 33 5F 6B 30 30 30 30 37 39// 12
    uint8_t* command_Signature = (uint8_t*)"Q";
    uint8_t* command_S = (uint8_t*)"S";

    wxString response;
    CommunicationBase* comBase = m_stateMgr->getCommunicationBase();

    comBase->getAnswer(command_ctime, 32, 16, 5, &response);
    m_fwInfo->setCompilationTime(response);
    //Logger::logf( LOG_MSG, _T("ECUStateMgrFirmwareInfoDtor setCompilationTime() ") +response);
    comBase->getAnswer(command_fwversion, 10, 6, 5, &response);
    m_fwInfo->setFirmwareVersion(response);
    //Logger::logf( LOG_MSG, _T("ECUStateMgrFirmwareInfoDtor setFirmwareVersion() ") +response);
    comBase->getAnswer(command_serNum, 12, 12, 5, &response);
    m_fwInfo->setSerialNumber(response);
    //Logger::logf( LOG_MSG, _T("ECUStateMgrFirmwareInfoDtor setSerialNumber() ") +response);

    m_stateMgr->setState(ECUInfo::MEGATUNE);
    comBase->getAnswer(command_Signature, 26, 20, 5, &response);
    //Logger::logf( LOG_MSG, _T("ECUStateMgrFirmwareInfoDtor setSignature() ") +response);
    m_fwInfo->setSignature(response);

    comBase->getAnswer(command_S, 28, 20, 5, &response);
    //Logger::logf( LOG_MSG, _T("ECUStateMgrFirmwareInfoDtor command_S() ") +response);

    m_fwInfo->setFirmwareOptions(response);

    //m_fwInfo.setTableSize(response);
    //m_fwInfo.setKpa(response);

}


bool CommunicationController::fetchV3FirmwareInfoSipr()
{
    CommunicationBase* cb = m_stateMgr->getCommunicationBase();

    SIPRPacket* packet = NULL;
    wxString s;
    uint8_t res = 0x00;

    ///COMPILE_TIME
    packet  = new SIPRPacket( 0, SIPR_GET_COMPILE_TIME );

    cb->SendString( packet->packet_data(), packet->packet_size() );

    res = readSiprRecord( cb );

    if ( res != SIPRParser::SIPR_STATE_CRC_OK ) { return false; }

    s = wxString::From8BitData( (const char*)m_parser.data(), m_parser.size() );

    Logger::logf(   LOG_MSG, _T("CommunicationController::fetchV3FirmwareInfoSipr() COMPILE_TIME- ") + s );

    m_fwInfo->setCompilationTime( s );

    ///VERSION
    packet  = new SIPRPacket( 0, SIPR_GET_VERSION );

    cb->SendString( packet->packet_data(), packet->packet_size() );

    res = readSiprRecord( cb );

    if ( res != SIPRParser::SIPR_STATE_CRC_OK ) { return false; }

    Logger::logf(   LOG_MSG, _T("CommunicationController::fetchV3FirmwareInfoSipr() - version size: %d"), m_parser.size() );

    Logger::logf(   LOG_MSG, _T("CommunicationController::fetchV3FirmwareInfoSipr() - %u.%u.%u"),
                    m_parser.data()[0], m_parser.data()[1], m_parser.data()[2] );

    wxString version = wxString::Format( _T("%d.%d.%d"), m_parser.data()[0], m_parser.data()[1], m_parser.data()[2] );
    m_fwInfo->setFirmwareVersion( version );
    uint8_t op[2] = { m_parser.data()[5], m_parser.data()[6] };
    //uint8_t op1 = m_parser.data()[5];
    //uint8_t op2 = m_parser.data()[6];
    ///HWSIG
    packet  = new SIPRPacket( 0, SIPR_GET_HWSIG );

    cb->SendString( packet->packet_data(), packet->packet_size() );

    res = readSiprRecord( cb );

    if ( res != SIPRParser::SIPR_STATE_CRC_OK ) { return false; }

    s = wxString::From8BitData( (const char*)m_parser.data(), m_parser.size() );
    Logger::logf(   LOG_MSG, _T("CommunicationController::fetchV3FirmwareInfoSipr() - HWSIG - size: %d - ") + s, m_parser.size() );

    m_fwInfo->setSerialNumber( s );


    ///MTSIG
    packet  = new SIPRPacket( 0, SIPR_GET_MT_SIGNATURE );

    cb->SendString( packet->packet_data(), packet->packet_size() );

    res = readSiprRecord( cb );

    if ( res != SIPRParser::SIPR_STATE_CRC_OK ) { return false; }

    s = wxString::From8BitData( (const char*)m_parser.data(), m_parser.size() );

    Logger::logf(   LOG_MSG, _T("CommunicationController::fetchV3FirmwareInfoSipr() - MTSIG - size: %d - ") + s, m_parser.size() );

    m_fwInfo->setSignature( s );

    ///FW OPTIONS
    s = _T("VEMS Genboard v3, ") + version;
    //int options = 256 * op[0] + op[1];

    wxChar c1( op[0] );
    wxChar c2( op[1] );
    //wxStringOutputStream sos( &s );
    //wxDataOutputStream os( sos );
    //os.Write8( op, 2 );
    s += c1;
    s += c2;
    Logger::logf(   LOG_MSG, _T("CommunicationController::fetchV3FirmwareInfoSipr() - options - size: %d - "), s.Length() );
    m_fwInfo->setFirmwareOptions( s );

    return true;

}


uint8_t CommunicationController::readSiprRecord( CommunicationBase* cb )
{
    m_parser.reset();

    uint8_t c = 0x00;
    int result;
    int recCount = 0;
    int tryCount = 0;
    do {
        result = cb->ReadString(&c, 1);
        uint8_t state = m_parser.feed( c );
        recCount += result;
        if( state == SIPRParser::SIPR_STATE_CRC_OK ) return SIPRParser::SIPR_STATE_CRC_OK;
        if( state == SIPRParser::SIPR_STATE_CRC_ERROR ) return SIPRParser::SIPR_STATE_CRC_ERROR;

        if(result == 0){
            tryCount++;
            wxThread::Sleep(1);
        }
    } while(tryCount < 1000);

    return SIPRParser::SIPR_STATE_NO_DATA;
}


void CommunicationController::clear()
{
    const int CLEARED_COUNT = 10;

    CommunicationBase* cb = m_stateMgr->getCommunicationBase();

    uint8_t c           = 0x00;
    int     result      = 0;
    int     clearCount  = 0;

    do {
        result = cb->ReadString( &c, 1 );

        result ? clearCount = 0 : ++clearCount;

    } while ( clearCount < CLEARED_COUNT );
}


void CommunicationController::sendDeviceEvent( int device )
{
    DeviceEvent deviceEvent;
    deviceEvent.device = (ECUInfo::Device) device;

    if ( m_pEventHandler )
    {
        m_pEventHandler->AddPendingEvent(deviceEvent);
    }
}


void CommunicationController::sendECUStateEvent( int state )
{
    ECUEvent ecuEvent;
    ecuEvent.state = (ECUEvent::State)state;

    if ( m_pEventHandler )
    {
        m_pEventHandler->AddPendingEvent(ecuEvent);
    }
}

void CommunicationController::sendDeviceEvent()
{
    sendDeviceEvent( m_device );
}

void CommunicationController::sendConnectedEvent()
{
    sendECUStateEvent( ECUEvent::CONNECTED );
}

void CommunicationController::sendNotConnectedEvent()
{
    Logger::logf( LOG_MSG, _T("CommunicationController::sendNotConnectedEvent()") );
    sendECUStateEvent( ECUEvent::NOT_CONNECTED );
}

void CommunicationController::sendBootModeEvent()
{
    sendECUStateEvent( ECUEvent::BOOT_MODE );
}

void CommunicationController::sendPasswordProtectedEvent()
{
    if ( m_pEventHandler ) {
        LOG
        wxCommandEvent event( ECU_IS_PASSWORD_PROTECTED, wxID_ANY );
        m_pEventHandler->AddPendingEvent( event );
    }
}

int CommunicationController::communicate()
{
    //Logger::logf( LOG_MSG, _T("CommunicationController::communicate()") );

    int sentBytes       = writeBytes();
    //Logger::logf( LOG_MSG, _T("CommunicationController::communicate()2") );

    if(sentBytes && m_interWriteDelay){
        wxThread::Sleep( m_interWriteDelay );
    }

    int receivedBytes   = readBytes();
    //Logger::logf( LOG_MSG, _T("CommunicationController::communicate()1") );

    if ( !receivedBytes && !needClear() && !sentBytes )
    //if ( ( !receivedBytes && !sentBytes ) || needClear() )
    {
        //Logger::logf( LOG_MSG, _T("CommunicationController::execute() - Sleep(1)") );
        wxThread::Sleep( 1 );
        return 0;
    }

    return 1;
}


void CommunicationController::setCtrlState( ComCtrlState* state )
{
    //Logger::logf( LOG_MSG, _T("CommunicationController::setCtrlState()1: %p"), m_pCtrlState );
    m_pPreviousCtrlState = m_pCtrlState;
    m_pCtrlState = state;
     //Logger::logf( LOG_MSG, _T("CommunicationController::setCtrlState()2: %p"), m_pCtrlState );
}

void CommunicationController::restorePreviousCtrlState()
{

    if ( !m_pPreviousCtrlState )
    {
        Logger::logf( LOG_MSG, _T("-----CommunicationController::restorePreviousCtrlState() - m_pPreviousCtrlState is NULL!") );
        return;
    }

    m_pCtrlState = m_pPreviousCtrlState;
    m_pPreviousCtrlState = NULL;
    //debugPrint(_T("restorePreviousCtrlState"));
    Logger::logf( LOG_MSG, _T("-----CommunicationController::restorePreviousCtrlState() - set m_pCtrlState to !") + m_pCtrlState->toString()  );
}

void CommunicationController::sendOfflineEvents()
{
    Logger::logf( LOG_MSG, _T("CommunicationController::sendOfflineEvents()") );

    ECUCommEvent event;
    event.state = ECUCommEvent::OFFLINE;
    m_pEventHandler->AddPendingEvent( event );

    DeviceEvent deviceEvent;
    deviceEvent.device = ECUInfo::DEVICE_UNKNOWN;
    m_pEventHandler->AddPendingEvent( deviceEvent );
}

void CommunicationController::debugPrint(const wxString& msg)
{
    Logger::logf( LOG_MSG, _T("CommunicationController::debugPrint( ")+msg+_T(" )m_pCtrlState: %p"), m_pCtrlState );
}

void CommunicationController::setPreferredBaudrate()
{
    const int detectedBaudrate = m_stateMgr->getBaudRate();

    if ( detectedBaudrate == m_preferredBaudrate ) return;

    m_stateMgr->setBaudRate( m_preferredBaudrate );
}

bool CommunicationController::checkPassword()
{
    uint8_t* command = (uint8_t*)".";

    wxString response;
    wxString serialNumber;

    CommunicationBase* comBase = m_stateMgr->getCommunicationBase();

    comBase->getAnswer( command, 32, 18, 5, &response );

    //LOGP( response );

    if ( !response.StartsWith( _T("BAD pw\r\n"), &serialNumber ) ) {
        LOGP( _T("NO PASSWORD PROTECTION") );
        m_deviceIsProtected = false;
        return true;
    }

    serialNumber = serialNumber.AfterFirst( _T('n') );
    wxString crcHex;

    int checkResult = PasswordHelper::checkPassword( serialNumber, &crcHex );

    switch ( checkResult )
    {
        case PasswordHelper::PASSWORD_ACCEPTED:
        {
            crcHex += _T(".");
            command = (uint8_t*)UnicodeHelper::U2C( crcHex );
            comBase->getAnswer( command, 32, 4, 5, &response );

            if ( response.StartsWith( _T("pwOK" ) ) ) {
                LOGP( _T("PASSWORD_ACCEPTED") );
                m_deviceIsProtected = false;
            } else {
                LOGP( _T("INVALID_PASSWORD") );
                m_deviceIsProtected = true;
            }
        } break;

        case PasswordHelper::PASSWORD_FILE_NOT_FOUND:
            LOGP( _T("PASSWORD_FILE_NOT_FOUND") );
            m_deviceIsProtected = true;
            break;

        case PasswordHelper::PASSWORD_NOT_FOUND:
            LOGP( _T("PASSWORD_NOT_FOUND") );
            m_deviceIsProtected = true;
            break;

        case PasswordHelper::INVALID_PASSWORD:
            LOGP( _T("INVALID_PASSWORD") );
            m_deviceIsProtected = true;
            break;
    }

    return !m_deviceIsProtected;
            }

void CommunicationController::leaveBootMode()
{
    if ( !m_leaveBootModeFlag ) {
        return;
    }

    uint8_t* command = (uint8_t*)"E";
    wxString response;

    CommunicationBase* comBase = m_stateMgr->getCommunicationBase();
    comBase->getAnswer( command, 8, 6, 5, &response);

    m_leaveBootModeFlag = false;
}
