    #ifndef MANPROTOCOL_H
#define MANPROTOCOL_H

#include "ProtocolBase.h"
#include "VemsCommand.h"

class VemsCommand;
class ManProtocol;
class ManProtocolState;
class SIPRPacket;


///--------------------------------------------------------------------------------------ManCommand
class ManCommand
{
    public:
        ManCommand( int             commandType,        // VemsCommand type
                    const uint8_t*  data,               // data must exist, data copied
                    uint8_t         size,               // must greater than > 0
                    int             expectedRxSize,
                    int             timeoutCount,
                    int             clearCount,
                    int             sleepAfterSend = 0 );

        ManCommand( int             type,
                    const wxString& command,
                    int             expectedRxSize,
                    int             timeoutCount,
                    int             clearCount,
                    int             sleepAfterSend = 0 );

        ManCommand( int             type,
                    const wxString& command );

        virtual ~ManCommand() { delete[] m_pData; }

        // getter methods
        int             getExpectedRxSize() const { return m_iExpectedRxSize; }
        int             getTimeoutCount() const { return m_iTimeoutCount; }
        int             getClearCount() const { return m_iClearCount; }
        const uint8_t*  getData() const { return m_pData; }
        uint8_t         getSize() const { return m_size; }
        int             getType() const { return m_iType; }
        bool            isWaitForResponse() const { return m_iExpectedRxSize > 0; }
        int             isNeedClear() const { return m_iClearCount > 0; }
        int             getSleepAfterWrite() const { return m_iSleepAfterSend; }

    protected:
        uint8_t*    m_pData;
        uint8_t     m_size;
        int         m_iExpectedRxSize;
        int         m_iTimeoutCount;
        int         m_iClearCount;
        int         m_iType;
        int         m_iSleepAfterSend;
    private:
        void        init(   int             type,
                            const uint8_t*  data,
                            uint8_t         size,
                            int             expectedRxSize,
                            int             timeoutCount,
                            int             clearCount,
                            int             sleepAfterSend );

        void        init(   int             type,
                            const wxString& command,
                            int             expectedRxSize,
                            int             timeoutCount,
                            int             clearCount,
                            int             sleepAfterSend );
};

WX_DECLARE_LIST( ManCommand, ManCommandList ); //store pointers


///-------------------------------------------------------------------------------------------ManProtocolMode
class ManProtocolMode
{
    public:
        enum ModeType { NORMAL, DOWNLOAD, SAVE_SD };
    public:
        ManProtocolMode() {}
        virtual ~ManProtocolMode() {}

        int             getType() const { return m_iType; }
        virtual void    execute( ManProtocol* p ) = 0;

        // StateMachine functions
        virtual bool    isInTxState() = 0;
        virtual bool    isInRxState() = 0;
        virtual bool    isIdle() = 0;
        bool            isInState( int stateType );
        void            changeState( ManProtocolState* state );

    protected:
        ManProtocolState*   m_pCurrentState;

    protected:
        void    setType( int type ) { m_iType = type; }

    private:
        int                 m_iType;
};


///--------------------------------------------------------------------------------------ManProtocolModeNormal
class ManProtocolModeNormal : public ManProtocolMode
{
    public:
        ManProtocolModeNormal();
        virtual ~ManProtocolModeNormal() {}

        bool    isInTxState();
        bool    isInRxState();
        bool    isIdle();
        void    execute( ManProtocol* p );
};


///--------------------------------------------------------------------------------------ManProtocolModeDownload
class ManProtocolModeDownload : public ManProtocolMode
{
    public:
        ManProtocolModeDownload();
        virtual ~ManProtocolModeDownload() {}

        bool    isInTxState();
        bool    isInRxState();
        bool    isIdle();
        void    execute( ManProtocol* p );
};


///--------------------------------------------------------------------------------------ManProtocolModeSaveSD
class ManProtocolModeSaveSD : public ManProtocolMode
{
    public:
        ManProtocolModeSaveSD();
        virtual ~ManProtocolModeSaveSD() {}

        bool    isInTxState();
        bool    isInRxState();
        bool    isIdle();
        void    execute( ManProtocol* p );
};


///---------------------------------------------------------------------------------------------ManProtocolState
class ManProtocolState
{
    public:
        enum StateType
        {
            IDLE,
            WAIT,
            CLEAR,
            //TIMEOUT,

            DOWNLOAD_IDLE,
            DOWNLOAD_WAIT,
            DOWNLOAD_CLEAR,

            SAVE_SD_NEED_RT_SIZE,
            SAVE_SD_WAIT_RT_SIZE,
            SAVE_SD_CLEAR_AFTER_RT_SIZE,
            SAVE_SD_NEED_DATA,
            SAVE_SD_WAIT_DATA,
            SAVE_SD_CLEAR_AFTER_DATA,
        };
    public:
        ManProtocolState() {}
        virtual ~ManProtocolState() {}

        int             getType() const { return m_iType; }
        bool            isSameTypeAs( int type )const  { return m_iType == type; }
        virtual void    execute( ManProtocol* p ) = 0;
    protected:
        void    setType( int type ) { m_iType = type; }
    private:
        int     m_iType;
};


///----------------------------------------------------------------------------------ManProtocolModeNormal States
///----------------------------------------------------------------------------------- IDLE, WAIT, CLEAR, TIMEOUT

///------------------------------------------------------------------------------------------ManProtocolStateIdle
class ManProtocolStateIdle : public ManProtocolState
{
    public:
        static ManProtocolStateIdle*    instance();
        void                            execute( ManProtocol* p );
    private:
        ManProtocolStateIdle() { setType( ManProtocolState::IDLE ); }
        virtual ~ManProtocolStateIdle() {}
    private:
        static ManProtocolStateIdle* s_pInstance;
};


///------------------------------------------------------------------------------------------ManProtocolStateWait
class ManProtocolStateWait : public ManProtocolState
{
    public:
        static ManProtocolStateWait*    instance();
        void                            execute( ManProtocol* p );
    private:
        ManProtocolStateWait() { setType( ManProtocolState::WAIT ); }
        virtual ~ManProtocolStateWait() {}
    private:
        static ManProtocolStateWait* s_pInstance;
};


///------------------------------------------------------------------------------------------ManProtocolStateClear
class ManProtocolStateClear : public ManProtocolState
{
    public:
        static ManProtocolStateClear*   instance();
        void                            execute( ManProtocol* p );
    private:
        ManProtocolStateClear() { setType( ManProtocolState::CLEAR ); }
        virtual ~ManProtocolStateClear() {}
    private:
        static ManProtocolStateClear* s_pInstance;
};

/*
///------------------------------------------------------------------------------------------ManProtocolStateTimeout
class ManProtocolStateTimeout : public ManProtocolState
{
    public:
        static ManProtocolStateTimeout*     instance();
        void                                execute( ManProtocol* p );
    private:
        ManProtocolStateTimeout() { setType( ManProtocolState::TIMEOUT ); }
        virtual ~ManProtocolStateTimeout() {}
    private:
        static ManProtocolStateTimeout* s_pInstance;
};
*/

///------------------------------------------------------------------------------------ManProtocolModeDownload States

///--------------------------------------------------------------------------------------ManProtocolStateDownloadIdle
class ManProtocolStateDownloadIdle : public ManProtocolState
{
    public:
        static ManProtocolStateDownloadIdle*    instance();
        void                                    execute( ManProtocol* p );
    private:
        ManProtocolStateDownloadIdle() { setType( ManProtocolState::DOWNLOAD_IDLE ); }
        virtual ~ManProtocolStateDownloadIdle() {}
    private:
        static ManProtocolStateDownloadIdle* s_pInstance;
};


///----------------------------------------------------------------------------------ManProtocolStateDownloadWait
class ManProtocolStateDownloadWait : public ManProtocolState
{
    public:
        static ManProtocolStateDownloadWait*    instance();
        void                                    execute( ManProtocol* p );
    private:
        ManProtocolStateDownloadWait() { setType( ManProtocolState::DOWNLOAD_WAIT ); }
        virtual ~ManProtocolStateDownloadWait() {}
    private:
        static ManProtocolStateDownloadWait* s_pInstance;
};


///--------------------------------------------------------------------------------------ManProtocolModeSaveSD States
///-----------------------------SAVE_SD_NEED_RT_SIZE, SAVE_SD_WAIT_RT_SIZE, SAVE_SD_NEED_DATA, SAVE_SD_SAVE_WAIT_DATA
///-----------------------------SD_SEARCH_LAST_WAIT, MPSSearchLastWaitData

///----------------------------------------------------------------------------------ManProtocolStateSaveSDNeedRTSize
class ManProtocolStateSaveSDNeedRTSize : public ManProtocolState
{
    public:
        static ManProtocolStateSaveSDNeedRTSize*    instance();
        void                                        execute( ManProtocol* p );
    private:
        ManProtocolStateSaveSDNeedRTSize() { setType( ManProtocolState::SAVE_SD_NEED_RT_SIZE ); }
        virtual ~ManProtocolStateSaveSDNeedRTSize() {}
    private:
        static ManProtocolStateSaveSDNeedRTSize* s_pInstance;
};


///----------------------------------------------------------------------------------ManProtocolStateSaveSDWaitRTSize
class ManProtocolStateSaveSDWaitRTSize : public ManProtocolState
{
    public:
        static ManProtocolStateSaveSDWaitRTSize*    instance();
        void                                        execute( ManProtocol* p );
    private:
        ManProtocolStateSaveSDWaitRTSize() { setType( ManProtocolState::SAVE_SD_WAIT_RT_SIZE ); }
        virtual ~ManProtocolStateSaveSDWaitRTSize() {}
    private:
        static ManProtocolStateSaveSDWaitRTSize* s_pInstance;
};

///----------------------------------------------------------------------------------ManProtocolStateSaveSDClearAfterRTSize
class ManProtocolStateSaveSDClearAfterRTSize : public ManProtocolState
{
    public:
        static ManProtocolStateSaveSDClearAfterRTSize*      instance();
        void                                                execute( ManProtocol* p );
    private:
        ManProtocolStateSaveSDClearAfterRTSize() { setType( ManProtocolState::SAVE_SD_CLEAR_AFTER_RT_SIZE ); }
        virtual ~ManProtocolStateSaveSDClearAfterRTSize() {}
    private:
        static ManProtocolStateSaveSDClearAfterRTSize* s_pInstance;
};

///----------------------------------------------------------------------------------ManProtocolStateSaveSDNeedData
class ManProtocolStateSaveSDNeedData : public ManProtocolState
{
    public:
        static ManProtocolStateSaveSDNeedData*    instance();
        void                                        execute( ManProtocol* p );
    private:
        ManProtocolStateSaveSDNeedData() { setType( ManProtocolState::SAVE_SD_NEED_DATA ); }
        virtual ~ManProtocolStateSaveSDNeedData() {}
    private:
        static ManProtocolStateSaveSDNeedData* s_pInstance;
};

///----------------------------------------------------------------------------------ManProtocolStateSaveSDWaitData
class ManProtocolStateSaveSDWaitData : public ManProtocolState
{
    public:
        static ManProtocolStateSaveSDWaitData*    instance();
        void                                        execute( ManProtocol* p );
    private:
        ManProtocolStateSaveSDWaitData() { setType( ManProtocolState::SAVE_SD_WAIT_DATA ); }
        virtual ~ManProtocolStateSaveSDWaitData() {}
    private:
        static ManProtocolStateSaveSDWaitData* s_pInstance;
};

///----------------------------------------------------------------------------------MPSSearchLastWaitData
class MPSSearchLastWaitData : public ManProtocolState
{
    public:
        static MPSSearchLastWaitData*   instance();
        void                            execute( ManProtocol* p );
    private:
        MPSSearchLastWaitData() { setType( ManProtocolState::SAVE_SD_WAIT_DATA ); }
        virtual ~MPSSearchLastWaitData() {}
    private:
        static MPSSearchLastWaitData* s_pInstance;
};

///----------------------------------------------------------------------------------ManProtocolStateSaveSDClearAfterData
class ManProtocolStateSaveSDClearAfterData : public ManProtocolState
{
    public:
        static ManProtocolStateSaveSDClearAfterData*    instance();
        void                                            execute( ManProtocol* p );
    private:
        ManProtocolStateSaveSDClearAfterData() { setType( ManProtocolState::SAVE_SD_CLEAR_AFTER_DATA ); }
        virtual ~ManProtocolStateSaveSDClearAfterData() {}
    private:
        static ManProtocolStateSaveSDClearAfterData* s_pInstance;
};

///-----------------------------------------------------------------------------------------------ManCommandFactory
class ManCommandFactory
{
    public:
        static int s_timeout;
        static ManCommand* create( VemsCommand* command );
    private:
        ~ManCommandFactory() {}

};


///------------------------------------------------------------------------------------------------------ManProtocol
class ManProtocol : public ProtocolBase
{
    public:
        //static const int    s_iRX_BUFFER_SIZE               = 1024;
        static const int    s_iDOWNLOAD_MIN_LINE_COUNT      = 2;
        static const int    s_iDOWNLOAD_SKIP_EMPTY_LINES    = 2;
        static const int    s_iSD_BLOCK_SIZE                = 514;
        static const int    s_iSD_BLOCK_COUNT               = 60;
        static const int    s_iMAX_CRC_ERROR_COUNT          = 4;
        static const int    s_iSD_RT_SIZE_BLOCK_INDEX       = 510;
        static const int    s_iRX_BUFFER_SIZE               = s_iSD_BLOCK_SIZE * s_iSD_BLOCK_COUNT;

    public:
        ManProtocol( wxEvtHandler* eventHandler = NULL, int timeout = 0 );
        virtual ~ManProtocol();

        /// ProtocolBase virtual functions ( communication )
        int     rx( uint8_t* rxBuffer, int size );
        int     tx( uint8_t* txBuffer, int& size );
        void    request( VemsCommand* command );
        bool    hasError() { return false; }
        int     type() const { return ECUInfo::MAN; }
        void    disable();
        void    enable();
        bool    isDisabled();



        /// StateMachine functions
        bool    isInState( int stateType );
        void    changeState( ManProtocolState* state );
        void    changeMode( ManProtocolMode* mode );

        bool    isCleared();
        void    increaseClearCount() { ++m_iClearCount; }

        bool    isDataReceived();
        void    clearTimeoutCount();
        void    addReceivedData();
        bool    isExpectedDataReceived();
        void    notify( int blockIndex );
        void    notifyTimeout();

        void    increaseTimeoutCount();
        bool    isTimedOut();

        bool    hasNextCommand() const { return !m_txQueue.IsEmpty(); }
        void    sendCommand();
        void    resetRxProperties();
        bool    isWaitForResponse();
        bool    isNeedClear();
        bool    needDisable() const { return m_bNeedDisable; }

        void    setDisabled();
        void    setEnabled()        { m_bIsDisabled = false; }


        void    sleep();

        /// download functions
        bool    isEOL() const;
        bool    isEmptyLine();
        void    resetDownloadProperties();
        void    increaseEmptyLineCount() { ++m_iEmptyLineCount; }
        void    increaseLineCount() { ++m_iLineCount; }
        bool    hasTooMuchEmptyLines() const
        {
            return m_iEmptyLineCount >= s_iDOWNLOAD_SKIP_EMPTY_LINES;
        }
        bool    isNotEnoughtLines() const { return m_iLineCount < s_iDOWNLOAD_MIN_LINE_COUNT; }
        int     getLineCount() const { return m_iLineCount; }
        void    sendConfigLine();

        /// sd functions
        bool    isCRCValid();
        void    getNextBlocks();
        bool    isValid();
        bool    isNextLog();
        bool    isSDCardEmpty();
        bool    isCheckCardEmpty();
        void    increaseBlockIndex();
        void    resetSDVariables();
        bool    isClearedAfterRTSize();
        void    sendValidBlocks();
        void    resetCheckCardEmpty();

        /// test functions
        void        printConfigLine();
        bool        isCurrentCommandSameAs( int vemsCommandType );
        void        notify();
        wxString    toString();
        void        sendEvent( int type );

        bool        isSearchStartIndex()        { return m_bSearchStartIndex; }
        void        resetSearchStartIndex()     { m_bSearchStartIndex = false; }
        void        setSearchIndex();
        void        verifyBlocks();
        bool        isIdle() { return m_pCurrentMode->isIdle(); }

        void        reset();

    private:
        ManProtocolMode*    m_pCurrentMode;
        VemsCommandList     m_RequestQueue;
        ManCommandList      m_txQueue;
        ManCommand*         m_pCurrentCommand;
        wxEvtHandler*       m_pEventHandler;


        // rx variables
        int                 m_iClearCount;
        int                 m_iTimeoutCount;
        int                 m_iRxDataLength;
        uint8_t*            m_pRxBuffer;
        uint8_t*            m_pTxBuffer;
        int*                m_pTxBufferSize;
        int                 m_iRxBufferSize;
        uint8_t             m_rxBuffer[ s_iRX_BUFFER_SIZE ];

        // download variables
        int                 m_iLineCount;
        int                 m_iEmptyLineCount;

        // sd variables
        bool                m_bIsNextLog;
        bool                m_bCheckCardEmpty;
        int                 m_iBlockIndex;
        int                 m_iBlockCount;
        int                 m_iCRCErrorCount;
        int                 m_iLastSecCountInBlock;
        int                 m_iInvalidBlockIndex;
        int                 m_iRTDataSize;
        int                 m_iNextLogIndex;

        // sd search variables for SD Card save log last T minute
        bool                m_bSearchStartIndex;                    // protocol need search
        bool                m_bSearchStartIndexUp;                  // serach count up/down
        int                 m_iSearchCount;
        int                 m_iSearchMask;
        int                 m_iBackTime;

        // under test
        bool                m_bNeedDisable;
        bool                m_bIsDisabled;

    private:
        void                pushCommand( VemsCommand* command );
        //int                 getModeOfCommand( int command );
        void                changeMode( VemsCommand* command );
        uint16_t            getCRC( uint8_t* data, int len );
        bool                isCRCValid( uint8_t* data, int len );
        bool                isDataValid( uint8_t* data, int len );
        bool                isNextLog( uint8_t * data );
        // under test
        void                sendEvent( const SIPRPacket& packet );
        void                initSearchMode( int iBackTime );
        void                processNextCommand();
};

#endif // MANPROTOCOL_H
