#include "MegatuneProtocol.h"
#include "MegatuneHelper.h"
#include "SIPRTypes.h"
#include "SIPRPacket.h"
#include "SIPRDescriptor.h"
#include "VemsCommand.h"
#include "../logger.h"
#include "../SymbolMgr.h"


#include "../events/InputEvent.h"
#include "SIPRPacketHelper.h"
#include "../UserLog.h"
#include "../Math.h"
#include <wx/listimpl.cpp>
#include "ECUInfo.h"

WX_DEFINE_LIST(PageRequestList);

WX_DEFINE_LIST( MegatuneCommandList );

///-----------------------------------------------------------------------------------------------MegatuneCommand
MegatuneCommand::MegatuneCommand(   int             commandType,
                                    const uint8_t*  data,
                                    uint8_t         size,
                                    int             expectedRxSize,
                                    int             timeoutCount,
                                    int             clearCount,
                                    int             pageNo )

{
    init( commandType, data, size, expectedRxSize, timeoutCount, clearCount, pageNo );
}

MegatuneCommand::MegatuneCommand(   int             commandType,
                                    const wxString& command,
                                    int             expectedRxSize,
                                    int             timeoutCount,
                                    int             clearCount,
                                    int             pageNo )

{
    init( commandType, command, expectedRxSize, timeoutCount, clearCount, pageNo );
}

void MegatuneCommand::init( int             type,
                            const uint8_t*  data,
                            uint8_t         size,
                            int             expectedRxSize,
                            int             timeoutCount,
                            int             clearCount,
                            int             pageNo )
{
    m_pData = new uint8_t[ size ];
    memcpy( m_pData, data, size );
    m_size  = size;
    m_iExpectedRxSize   = expectedRxSize;
    m_iTimeoutCount     = timeoutCount;
    m_iClearCount       = clearCount;
    m_iPageNumber       = pageNo;
    m_iType             = type;
}

void MegatuneCommand::init( int             type,
                            const wxString& command,
                            int             expectedRxSize,
                            int             timeoutCount,
                            int             clearCount,
                            int             pageNo )
{
#if defined (WIN32) || (__WINCE_STANDARDSDK__)
    char* c = _strdup( command.mb_str() );
#else
    char* c = strdup( command.mb_str() );
#endif

    init( type, (uint8_t*)c, strlen( c ), expectedRxSize, timeoutCount, clearCount, pageNo );

    delete c;
}


///-------------------------------------------------------------------------------------------MegatuneProtocol
MegatuneProtocol::MegatuneProtocol( SIPRDescriptor* siprDesc,
                                    int sensorType,
                                    wxEvtHandler* eventHandler,
                                    int timeout,
                                    int rtErrorCount )
    : m_rtCRCCheckMinimumVersion ( 1, 1, 75 )
{
    m_pSIPRDescriptor       = siprDesc;
    m_pEventHandler         = eventHandler;
    m_iRealtimeLength       = 0;
    m_iRxDataLength         = 0;
    m_isReadSensors         = false;
    m_realtimeErrorCount    = 0;
    m_realtimeCRCErrorCount    = 0;
    m_realtimeCount         = 0;

    m_state                 = TIME_OUT;
    m_mode                  = SYNC;

    m_syncState             = NEED_RT_SIZE;
    m_pCurrentCommand       = NULL;
    m_iNumberOfPages        = m_pSIPRDescriptor->getPageCount();
    m_missingPages          = 0;
    m_iCurrentPageIndex     = 0;
    m_bNeedDisable          = false;
    m_bIsDisabled           = false;
    m_bSyncStarted          = true;
    m_strReadRealtimeCommand = wxEmptyString;
    m_iSensorType            = sensorType;

    m_iTimeout               = timeout;

    m_maxRealtimeErrorCount = rtErrorCount;
    m_realtimeErrorCountPerMinute = 0;
    m_realtimeErrorCountTime = wxDateTime::Now();
    m_maxRealtimeErrorCountPerMinute = 3;
    m_isRTErrorMessageSent = false;

    m_startTime = wxDateTime::Now();
    m_device = ECUInfo::DEVICE_V3;
    m_syncPageTryCount = 0;
    m_readPageTryCount = 0;
}

MegatuneProtocol::~MegatuneProtocol()
{
}

int MegatuneProtocol::rx( uint8_t* rxBuffer, int size )
{
    switch ( m_state )
    {

        case WAIT :
            rxWaitState( rxBuffer, size );
            break;

        case CLEAR :
            rxClearState( rxBuffer, size );
            break;

    }

    return 0;
}

void MegatuneProtocol::rxWaitState( uint8_t* rxBuffer, int size )
        {
    if ( size ) {
        rxWaitReadState( rxBuffer, size );
    } else {
        rxWaitEmptyState( rxBuffer, size );
    }
}

void MegatuneProtocol::rxWaitReadState( uint8_t* rxBuffer, int size )
            {
                // clear timeout count
                m_iTimeoutCount = 0;

                // add received data
                m_rxBuffer[ m_iRxDataLength++ ] = rxBuffer[0];

    // expected data not received
    if ( m_pCurrentCommand->getExpectedRxSize() != m_iRxDataLength ) return;

    // expected data received
    switch ( m_mode )
                {
        case NORMAL:
            rxDataReceivedNormalMode( rxBuffer, size );
            break;

        case SYNC:
            rxDataReceivedSyncMode( rxBuffer, size );
            break;
                }

}

void MegatuneProtocol::rxDataReceivedNormalMode( uint8_t* rxBuffer, int size )
                {
    if ( !checkPageCrc() ) return;

                            notify();

                            if ( m_maxRealtimeErrorCount > m_realtimeErrorCount ) {
                                m_state = CLEAR;
                            } else {
                                m_realtimeErrorCount = 0;
                                m_state = TIME_OUT;
                                sendEvent( SIPR_TIMEOUT );
                            }
}

bool MegatuneProtocol::checkPageCrc()
{
    if ( m_pCurrentCommand->getType() != VemsCommand::READ_PAGE ) return true;

    if ( !isPageCrcEnabled() ) return true;

    if ( MegatuneHelper::checkCrc( m_rxBuffer, m_iRxDataLength ) ) return true;

    if ( m_readPageTryCount < READ_PAGE_TRY_COUNT_MAX ) {
        ++m_readPageTryCount;
        m_txQueue.push_front( m_pCurrentCommand );
        m_pCurrentCommand = NULL;
        m_state = IDLE;
        return false;
    } else {
        m_state = TIME_OUT;
        sendEvent( SIPR_TIMEOUT );
        return false;
    }
}

void MegatuneProtocol::rxDataReceivedSyncMode( uint8_t* rxBuffer, int size )
                        {
                            switch ( m_syncState)
                            {
                                case GET_RT_SIZE:
                                {
                                    m_syncState = NEED_TEST_RT_SIZE;
                                    m_state = IDLE;
                                    Logger::logf( LOG_MSG, _T("MegatuneProtocol::rx() - GET_RT_SIZE - %d"), m_iRxDataLength );
                                }break;

                                case TEST_RT_SIZE:
                                {
                                    m_syncState = HAVE_RT_SIZE;
                                    m_state = IDLE;
                                    Logger::logf( LOG_MSG, _T("MegatuneProtocol::rx() - TEST_RT_SIZE - %d"), m_iRxDataLength );
                                }break;

                                case GET_PAGE_SIZE:
                                {
                                    m_syncState = HAVE_PAGE_SIZE;
                                    m_state = IDLE;
                                    notify();
                                }break;

                                case TEST_PAGE_SIZE:
                                {
                                    m_syncState = HAVE_PAGE_SIZE;
                                    m_state = IDLE;
                                    notify();
                                }break;
                            }
                    }

void MegatuneProtocol::rxWaitEmptyState( uint8_t* rxBuffer, int size )
            {
    if ( ++m_iTimeoutCount < m_pCurrentCommand->getTimeoutCount() ) return;

    //TIMEOUT
                    switch ( m_mode )
                    {
                        case NORMAL:
            rxTimeoutNormalMode( rxBuffer, size );
            break;

        case SYNC:
            rxTimeoutSyncMode( rxBuffer, size );
            break;

    }
}

void MegatuneProtocol::rxTimeoutNormalMode( uint8_t* rxBuffer, int size )
                        {
                            if ( VemsCommand::READ_REALTIME == m_pCurrentCommand->getType() )
                            {
                                if ( m_maxRealtimeErrorCount > m_realtimeErrorCount++ )
                                {
                                    wxDateTime now = wxDateTime::Now();

                                    if ( now.GetMinute() == m_realtimeErrorCountTime.GetMinute() )
                                    {
                                        if ( ++m_realtimeErrorCountPerMinute > m_maxRealtimeErrorCountPerMinute )
                                        {
                                            if ( !m_isRTErrorMessageSent )
                                            {
                                                 wxString message = _T("Too many realtime read error!");
                                                 //UserLog::uLog( ULOG_WARNING, message );
                                                 m_isRTErrorMessageSent = true;
                                            }
                                            Logger::logf( LOG_MSG, _T("m_realtimeErrorCountPerMinute: %d"), m_realtimeErrorCountPerMinute );
                                        }
                                    }
                                    else
                                    {
                                        m_realtimeErrorCountTime = now;
                                        m_realtimeErrorCountPerMinute = 1;
                                    }

                                    Logger::logf( LOG_MSG, _T("m_realtimeErrorCount: %d"), m_realtimeErrorCount );
                                    m_state = CLEAR;
            return;

                                    //m_realtimeErrorCountTime
                                }
                            }
                            m_realtimeErrorCount = 0;
                            m_state = TIME_OUT;
                            Logger::logf( LOG_MSG, _T("MegatuneProtocol::rx() - TIME_OUT")
                                            _T(" - command: %d, expected size: %d, received size : %d"),
                                            m_pCurrentCommand->getType(),
                                            m_pCurrentCommand->getExpectedRxSize(),
                                            m_iRxDataLength );


                            sendEvent( SIPR_TIMEOUT );
}

void MegatuneProtocol::rxTimeoutSyncMode( uint8_t* rxBuffer, int size )
                        {
                            switch ( m_syncState )
                            {
                                case GET_RT_SIZE:
            rxTimeoutSyncModeGetRtSize( rxBuffer, size );
            break;

        case TEST_RT_SIZE:
            rxTimeoutSyncModeTestRtSize( rxBuffer, size );
            break;

        case GET_PAGE_SIZE:
            rxTimeoutSyncModeGetPageSize( rxBuffer, size );
            break;

        case TEST_PAGE_SIZE:
            rxTimeoutSyncModeTestPageSize( rxBuffer, size );
            break;
    }
}

void MegatuneProtocol::rxTimeoutSyncModeGetRtSize( uint8_t* rxBuffer, int size )
                                {
    if ( m_iRxDataLength == 0 && m_strReadRealtimeCommand.IsSameAs( _T("D") ) ) {
                                        m_syncState = NEED_RT_SIZE;
    } else {
                                        m_syncState = NEED_TEST_RT_SIZE;
                                    }

                                    m_state = IDLE;

                                    Logger::logf( LOG_MSG, _T("MegatuneProtocol::rx() - GET_PAGE_SIZE - %d"), m_iRxDataLength );
}

void MegatuneProtocol::rxTimeoutSyncModeTestRtSize( uint8_t* rxBuffer, int size )
                                {
                                    Logger::logf( LOG_MSG, _T("MegatuneProtocol::rx() - SYNC- TEST_RT_SIZE - TIMEOUT") );
                                    m_mode  = NORMAL;
                                    m_state = TIME_OUT;
                                    sendEvent( SIPR_TIMEOUT );
}

void MegatuneProtocol::rxTimeoutSyncModeGetPageSize( uint8_t* rxBuffer, int size )
                                {
                                    if ( m_iRxDataLength == 0 )
                                    {
                                        ++m_missingPages;
                                        m_syncState = HAVE_PAGE_SIZE;
                                        m_state     = IDLE;
                                        int pageNumber = m_pSIPRDescriptor->getPageNumber( m_iCurrentPageIndex );
                                        wxString message = wxString::Format( _T("Page: %d not found!"), pageNumber );
                                        Logger::logf( LOG_MSG, _T("MegatuneProtocol::rx() - ") + message );
                                    }
                                    else
                                    {
        if ( isPageCrcEnabled() ) {
            if ( MegatuneHelper::checkCrc( m_rxBuffer, m_iRxDataLength ) ) {
                m_syncPageTryCount = 0;
                                        m_syncState = HAVE_PAGE_SIZE;
                                        m_state     = IDLE;
                                        notify();
            } else if ( m_syncPageTryCount < SYNC_PAGE_TRY_COUNT_MAX ) {
                m_syncPageTryCount++;
                m_syncState = NEED_PAGE_SIZE;
                m_state     = IDLE;
            } else {
                m_mode  = NORMAL;
                m_state = TIME_OUT;
                sendEvent( SIPR_TIMEOUT );
                                    }
        } else {
            m_syncState = HAVE_PAGE_SIZE;
            m_state     = IDLE;
            notify();
}
    }
}

void MegatuneProtocol::rxTimeoutSyncModeTestPageSize( uint8_t* rxBuffer, int size )
                                {
                                    Logger::logf( LOG_MSG, _T("MegatuneProtocol::rx() - SYNC- TEST_PAGE_SIZE - TIMEOUT") );
                                    m_mode  = NORMAL;
                                    m_state = TIME_OUT;
                                    sendEvent( SIPR_TIMEOUT );
                            }

void MegatuneProtocol::rxClearState( uint8_t* rxBuffer, int size )
{
    if ( size ) {
        if ( ++m_iClearCount >= m_pCurrentCommand->getClearCount() ) {
            m_state = IDLE;
                    }
    } else {
        m_state = IDLE;
                }
            }

int MegatuneProtocol::tx( uint8_t* txBuffer, int& size )
            {
    if ( m_mode == SYNC ) {
        if ( txSyncMode( txBuffer, size ) == 0 ) return 0;
                }

    switch( m_state )
            {
        case WAIT:
            txWait( txBuffer, size );
            break;

        case IDLE:
            txIdle( txBuffer, size );
            break;
            }

    return 0;
}

int MegatuneProtocol::txSyncMode( uint8_t* txBuffer, int& size )
{
		if ( m_bSyncStarted )
		{
            sendEvent( SIPR_SYNC_STARTED );
		    m_bSyncStarted = false;
		}

		switch ( m_syncState )
		{
		    case NEED_RT_SIZE:
            txSyncModeNeedRtSize( txBuffer, size );
            return -1;

        case NEED_TEST_RT_SIZE:
            txSyncModeNeedTestRtSize( txBuffer, size );
            return -1;

        case HAVE_RT_SIZE:
            txSyncModeHaveRtSize( txBuffer, size );
            return 0;

        case NEED_PAGE_SIZE:
            return txSyncModeNeedPageSize( txBuffer, size );

        case NEED_TEST_PAGE_SIZE:
            txSyncModeNeedTestPageSize( txBuffer, size );
            return -1;

        case HAVE_PAGE_SIZE:
            return txSyncModeHavePageSize( txBuffer, size );
    }

    return -1;
}

void MegatuneProtocol::txSyncModeNeedRtSize( uint8_t* txBuffer, int& size )
		    {
		        setReadRealTimeCommand();
		        m_iRealtimeLength = 256;
		        request( new ReadRealTimeCommand() );
		        m_syncState = GET_RT_SIZE;
}

void MegatuneProtocol::txSyncModeNeedTestRtSize( uint8_t* txBuffer, int& size )
		    {
                m_iRealtimeLength = m_iRxDataLength;
                request( new ReadRealTimeCommand() );
                m_syncState = TEST_RT_SIZE;
}

void MegatuneProtocol::txSyncModeHaveRtSize( uint8_t* txBuffer, int& size )
            {
                Logger::logf( LOG_MSG, _T("MegatuneProtocol::tx() - SYNC - HAVE_RT_SIZE - size: %d "), m_iRxDataLength );
                m_syncState = NEED_PAGE_SIZE;
}

int MegatuneProtocol::txSyncModeNeedPageSize( uint8_t* txBuffer, int& size )
		    {
                int pageNumber = m_pSIPRDescriptor->getPageNumber( m_iCurrentPageIndex );

                if ( pageNumber == -1 )
                {
                    m_mode = NORMAL;
                    sendEvent( SIPR_SYNC_FINISHED );
                    return 0;
                }

    int pageSize = isPageCrcEnabled() ? 258: 256;

    request( new ReadPageCommand( pageNumber, pageSize ) );

                m_syncState = GET_PAGE_SIZE;

                uint8_t packetData = SIPR_REFRESH_START;
                SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &packetData );
                sendEvent( packet );

    return -1;
}

void MegatuneProtocol::txSyncModeNeedTestPageSize( uint8_t* txBuffer, int& size )
		    {
                int pageNumber = m_pSIPRDescriptor->getPageNumber( m_iCurrentPageIndex );
                request( new ReadPageCommand( pageNumber, m_iRxDataLength ) );
                m_syncState = TEST_PAGE_SIZE;
}

int MegatuneProtocol::txSyncModeHavePageSize( uint8_t* txBuffer, int& size )
            {
                 int pageNumber = m_pSIPRDescriptor->getPageNumber( m_iCurrentPageIndex );

    int configPageSize = m_pSIPRDescriptor->getPageSize( pageNumber );

                Logger::logf( LOG_MSG,
        _T("MegatuneProtocol::tx() - SYNC - HAVE_PAGE_SIZE - idx: %d, page:%d, cmd page: %d, size: %d, config size: %d "),
        m_iCurrentPageIndex, pageNumber, m_pCurrentCommand->getPageNumber(), m_iRxDataLength, configPageSize );

                m_syncState = NEED_PAGE_SIZE;
                m_SizeOfPages[ m_pCurrentCommand->getPageNumber() ] = m_iRxDataLength;

                if ( (size_t)(++m_iCurrentPageIndex)  >= m_pSIPRDescriptor->getPageCount() )
                {
                    if ( m_missingPages == (int)(m_pSIPRDescriptor->getPageCount()) )
                    {
                        Logger::logf( LOG_MSG, _T("MegatuneProtocol::tx() - all pages missing!") );

                        m_mode  = NORMAL;
                        m_state = TIME_OUT;
                        sendEvent( SIPR_TIMEOUT );
                        return 0;
                    }
                    m_mode = NORMAL;
                    sendEvent( SIPR_SYNC_FINISHED );

                    uint8_t packetData = SIPR_REFRESH_STOP;
                    SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &packetData );
                    sendEvent( packet );

                    return 0;
                }

                Logger::logf( LOG_MSG, _T("MegatuneProtocol::tx() - m_iCurrentPageIndex: %d, page count: %d"),
                    m_iCurrentPageIndex, m_pSIPRDescriptor->getPageCount() );

                sendEvent( SIPR_SYNC_PAGE );

    return 0;
		}

void MegatuneProtocol::txWait( uint8_t* txBuffer, int& size )
    {
    if ( m_txQueue.IsEmpty() ) { return; }

        MegatuneCommand* command = m_txQueue.front();

    if ( command->getType() != VemsCommand::WRITE_PAGE ) { return; }

        m_txQueue.pop_front();

        memcpy( txBuffer, command->getData(), command->getSize() );
        size = command->getSize();

        delete command;

        SIPRPacket p( 0, SIPR_ACK );
        sendEvent( p );
}

void MegatuneProtocol::txIdle( uint8_t* txBuffer, int& size )
    {
        if ( m_bNeedDisable )
        {
            if ( ( m_mode == NORMAL ) &&  ( m_txQueue.IsEmpty() ) )
            {
                m_bNeedDisable = false;
                m_bIsDisabled = true;
            }
        }

    if ( m_bIsDisabled ) { return; }

        if ( m_txQueue.IsEmpty() )
        {
            if ( m_isReadSensors ) { request( new ReadRealTimeCommand() ); }
        }
        else
        {
            MegatuneCommand* command = m_txQueue.front();

            if ( command->getType() == VemsCommand::WRITE_PAGE && m_isReadSensors )
            {
                //m_txQueue.Insert( new MegatuneCommand( VemsCommand::READ_REALTIME, m_strReadRealtimeCommand, m_iRealtimeLength, m_iTimeout, 2 ) );
            }
        }

    if ( m_txQueue.IsEmpty() ) { return; }

        if ( m_pCurrentCommand ) { delete m_pCurrentCommand; } // ???

        // get command from txQueue
        m_pCurrentCommand = m_txQueue.front();
        m_txQueue.pop_front();

    if ( !m_pCurrentCommand ) { return; }

        memcpy( txBuffer, m_pCurrentCommand->getData(), m_pCurrentCommand->getSize() );
        size = m_pCurrentCommand->getSize();

        resetRxProperties();

        switch ( m_pCurrentCommand->getType() )
        {
            case VemsCommand::READ_REALTIME :
            case VemsCommand::PING:
            case VemsCommand::GET_MT_SIGNATURE:
            case VemsCommand::SET_STATE:
            case VemsCommand::GENERIC_STRING:
                m_state = WAIT;
                break;
            case VemsCommand::WRITE_PAGE :
            {
                SIPRPacket p( 0, SIPR_ACK );
                sendEvent( p );
            } break;

            case VemsCommand::READ_PAGE :
            {
                if ( m_mode == NORMAL ) {
                    PageRequest* request = m_readPageRequests.front();

                    if ( request ) {
                        if ( !request->active ) {
                            request->active = true;

                            uint8_t packetData = SIPR_REFRESH_START;
                            SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &packetData );
                            sendEvent( packet );
                        }
                    }
                }

                m_state = WAIT;
            } break;
        }
    }

bool MegatuneProtocol::checkRtCrc()
{
    if (    m_firmwareVersion < m_rtCRCCheckMinimumVersion ||
            !m_strReadRealtimeCommand.IsSameAs( _T("D") ) ||
            m_device != ECUInfo::DEVICE_V3 ) {
        return true;
    }

    uint16_t crc=0xffff;
    uint16_t rec_crc=0;

    uint8_t rt_len = m_iRealtimeLength-2;

    for (int i=0;i<m_iRealtimeLength;++i){
        if (i<rt_len) crc = Math::crc16_sipr( crc, m_rxBuffer[i] );
        else {
            if ( i == rt_len ){
                rec_crc=  m_rxBuffer[i];
            }
            if ( i == rt_len+1 ){
                rec_crc=  rec_crc + ((uint16_t)m_rxBuffer[i]<<8);
            }
        }
    }

    return ( crc == rec_crc );
}

void MegatuneProtocol::sendRtTimePacket()
{
    wxDateTime now = wxDateTime::UNow();
    wxTimeSpan timeSpan = now.Subtract( m_startTime );
    long elapsedTime = timeSpan.GetMilliseconds().ToLong();

    Math::getBytes( elapsedTime, m_buffer, 4 );
    SIPRPacket timePacket( 0, SIPR_RT_TIME, 4, m_buffer );
    sendEvent( timePacket );
}

void MegatuneProtocol::sendRtPacket()
{
    m_buffer[0] = 0;
    m_buffer[1] = 0;

    switch ( m_device )
    {
        case ECUInfo::DEVICE_ROUND:
            m_buffer[2] = SiprSensorType::SIPR_RT_ROUND;
            break;
        case ECUInfo::DEVICE_V3:
            m_buffer[2] = SiprSensorType::SIPR_RT_MT_NORMAL;
            break;
    }

    memcpy( &m_buffer[3], m_rxBuffer, m_iRealtimeLength );

    SIPRPacket p( 0, SIPR_READ_SENSORS, m_iRealtimeLength + 3, m_buffer );
    sendEvent( p );
}

void MegatuneProtocol::notify()
{
    if ( !m_pEventHandler ) { return; }

    switch ( m_pCurrentCommand->getType() )
    {
        case VemsCommand::READ_REALTIME :
        {
            m_realtimeCount ++;
            if ( checkRtCrc() ) {
                sendRtTimePacket();
                sendRtPacket();
            } else {
                ++m_realtimeCRCErrorCount;
                int rate = 0;
                if (m_realtimeCount) rate = m_realtimeCRCErrorCount / m_realtimeCount * 100;
                Logger::logf( LOG_MSG, _T("MegatuneProtocol::notify() READ_REALTIME crc error, Error rate=%d (%d / %d packet) "),rate , m_realtimeErrorCount , m_realtimeCount );
            }
        } break;

        case VemsCommand::READ_PAGE :
        {
            sendSIPRGetTableEvent( m_pCurrentCommand->getPageNumber() );

            if ( m_mode == SYNC ) {
                break;
            }

            if ( m_readPageRequests.empty() ) {
                LOGP( _T("TODO!") );
            } else {
                PageRequest* request = m_readPageRequests.front();

                if ( request ) {
                    if ( request->requestedPages.erase( m_pCurrentCommand->getPageNumber() ) ) {
                        LOGP( wxString::Format( _T("page id: %d removed!"), m_pCurrentCommand->getPageNumber() ) );

                        if ( request->requestedPages.empty() ) {
                            LOGP( wxString::Format( _T("request removed!"), m_pCurrentCommand->getPageNumber() ) );

                            m_readPageRequests.pop_front();
                            delete request;

                            uint8_t packetData = SIPR_REFRESH_STOP;
                            SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &packetData );
                            sendEvent( packet );
                        }
                    } else {
                        LOGP( _T("TODO") );
                    }
                } else {
                    m_readPageRequests.pop_front();
                    LOGP( _T("TODO!") );
                }
            }

            LOGP( wxString::Format( _T("remained requests: %d"), m_readPageRequests.size() ) );


        } break;

        case VemsCommand::PING:
        {
            wxString s = wxString::From8BitData( (const char *)m_rxBuffer, m_iRxDataLength );
            SIPRPacket p( 0, SIPR_ACK );
            sendEvent( p );

        } break;

        case VemsCommand::GET_MT_SIGNATURE:
        {
            SIPRPacket p( 0, SIPR_GET_MT_SIGNATURE, m_iRxDataLength, m_rxBuffer );
            sendEvent( p );
        } break;
    }
}

//void MegatuneProtocol::reqestWritePage( int pageID,uint8_t pageOffset, uint8_t value );
void MegatuneProtocol::sendEvent( const SIPRPacket& packet )
{
    InputEvent e;
    e.setPacket( packet );
    m_pEventHandler->AddPendingEvent( e );
}


void MegatuneProtocol::sendEvent( int type )
{
    SIPRPacket p( 0, type );
    sendEvent( p );
}


void MegatuneProtocol::resetRxProperties()
{
    //Logger::logf( LOG_MSG, _T("MegatuneProtocol::resetRxProperties()") );
    m_iTimeoutCount = 0;
    m_iClearCount   = 0;
    m_iRxDataLength = 0;
    //m_realtimeErrorCount = 0;
    //m_rxBuffer[256];
}

void MegatuneProtocol::reset()
{
    resetRxProperties();
    m_txQueue.Clear();
    m_state             = IDLE;
    m_mode              = SYNC;
    m_strReadRealtimeCommand = wxEmptyString;

    m_bSyncStarted      = true;
    m_syncState         = NEED_RT_SIZE;
    m_iCurrentPageIndex = 0;
    m_missingPages      = 0;

    m_SizeOfPages.clear();

    WX_CLEAR_LIST( PageRequestList, m_readPageRequests );

    m_syncPageTryCount = 0;
    m_readPageTryCount = 0;

}

void MegatuneProtocol::sendSIPRGetTableEvent( int megatunePageNo )
{
    SIPRPacketList packets = SIPRPacketHelper::createSIPRPackets(*m_pSIPRDescriptor, m_rxBuffer, megatunePageNo);
    for ( SIPRPacketList::iterator it = packets.begin(); it != packets.end(); ++it ){
        SIPRPacket* p = *it;
        sendEvent( *p );
        delete p;
    }
}

bool MegatuneProtocol::needClear()
{
    return m_state == CLEAR ? true : false;
}

MegatuneCommand* MegatuneProtocol::createReadPageCommand( int pageNo )
{
    wxString strCommand = m_pSIPRDescriptor->getPageReadCommand( pageNo );

    if  ( strCommand.IsEmpty() ) return NULL;

    int pageSize        = 255;

    if ( m_SizeOfPages.find( pageNo ) != m_SizeOfPages.end() )
    {
        pageSize = m_SizeOfPages[ pageNo ];
    }

    if ( pageSize == 0 ) { return NULL; }

    if ( strCommand.StartsWith( _T("0x") ) || strCommand.StartsWith( _T("0X") ) ) {
        unsigned long val = 0 ;

        if ( strCommand.ToULong(&val,0) ) {
            uint8_t buff[2];
            buff[0] = val;


            size_t cmdSize = 1;

    if ( m_device == ECUInfo::DEVICE_ROUND ) {
                buff[1] = '.';
                cmdSize = 2;
            }

            return new MegatuneCommand( VemsCommand::READ_PAGE, buff, cmdSize, pageSize, m_iTimeout, 10, pageNo );

        } else {
            LOGP( _T("INVALID COMMAND - ") + strCommand );
            return NULL;
        }
    }

    if ( m_device == ECUInfo::DEVICE_ROUND ) {
        strCommand += _T(".");
    }

    return new MegatuneCommand( VemsCommand::READ_PAGE, strCommand, pageSize, m_iTimeout, 10, pageNo );
}

void MegatuneProtocol::request( VemsCommand* command )
{
//    if ( command ) { Logger::logf( LOG_MSG, _T("MegatuneProtocol::request() - %d"), command->getType() ); }
//    else { Logger::logf( LOG_MSG, _T("MegatuneProtocol::request() - NULL!!!") ); }

    MegatuneCommand*    mtCommand = NULL;
    wxString            strCommand;

//    int pages;
    switch ( command->getType() )
    {
        case VemsCommand::PING:

            mtCommand = new MegatuneCommand( VemsCommand::PING, _T("S"), 26, m_iTimeout, 2, 0 );
            break;
        case VemsCommand::SET_STATE:
            switch ( command->getState() )
            {
                case ECUInfo::MAN:
                    Logger::logf( LOG_MSG, _T("MegatuneProtocol::request() - SET_STATE - ECUInfo::MAN") );
                    mtCommand = new MegatuneCommand( VemsCommand::SET_STATE, _T("Man"), 8, m_iTimeout, 2, 0 );
                    break;
                case ECUInfo::SIPR:
                    mtCommand = new MegatuneCommand( VemsCommand::SET_STATE, _T("Mas"), 2, m_iTimeout, 2, 0 );
                    break;
            }
            break;
        case VemsCommand::GET_MT_SIGNATURE:
            //Logger::logf( LOG_MSG, _T("MegatuneProtocol::request() - GET_MT_SIGNATURE") );
            //mtCommand = new MegatuneCommand( VemsCommand::GET_MT_SIGNATURE, _T("Q"), 30, 0, 0, 0 );
            mtCommand = new MegatuneCommand( VemsCommand::GET_MT_SIGNATURE, _T("Q"), 26, m_iTimeout, 0, 0 );
            //Logger::logf( LOG_MSG, _T("MegatuneProtocol::request() - GET_MT_SIGNATURE - command: %p"), mtCommand );
            break;

        case VemsCommand::READ_PAGE: {
            //Logger::logf( LOG_MSG, _T("MegatuneProtocol::request() - READ_PAGE: %d"), command->getPageID() );
            mtCommand =  createReadPageCommand( command->getPageID() );
            if ( !mtCommand ) break;
            if ( command->getPageSize() > 0 ) { mtCommand->setExpectedRxSize( command->getPageSize() ); }
            if ( m_mode == SYNC ) break;

            //IntSet* requestedPages = new IntSet();
            PageRequest* pageReq = new PageRequest();
            pageReq->requestedPages.insert( command->getPageID() );
            m_readPageRequests.push_back( pageReq );
            LOGP( wxString::Format( _T("add requests! requests count: %d"), m_readPageRequests.size() ) );

            uint8_t packetData = SIPR_REFRESH_START;
            SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &packetData );
            sendEvent( packet );
        } break;

        case VemsCommand::READ_ALL_PAGE:
        {
            const IntArray& pagesA = m_pSIPRDescriptor->getPagesOrder();

            //Logger::logf( LOG_MSG, _T("MegatuneProtocol::request() - READ_ALL_PAGE: %d"), pagesA.GetCount() );

            PageRequest* pageReq = new PageRequest();

            for( size_t i = 0; i < pagesA.GetCount(); ++i )
            {
                MegatuneCommand* readPageCommand = createReadPageCommand( pagesA[i] );
                if ( readPageCommand ) {
                    m_txQueue.push_back( readPageCommand );
                    pageReq->requestedPages.insert( pagesA[i] );
            }
            }


            m_readPageRequests.push_back( pageReq );
            LOGP( wxString::Format( _T("add requests! requests count: %d"), m_readPageRequests.size() ) );

            uint8_t packetData = SIPR_REFRESH_START;
            SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &packetData );
            sendEvent( packet );

        } break;

        case VemsCommand::WRITE_PAGE:
        {
            strCommand = m_pSIPRDescriptor->getPageWriteCommand( command->getPageID() );

            if ( strCommand.IsEmpty() ) break;

            uint8_t cmdSize = 3;

            if ( m_device == ECUInfo::DEVICE_ROUND ) {
                cmdSize = 4;
            }

            uint8_t* cmd = new uint8_t[ cmdSize ];

             if ( strCommand.StartsWith( _T("0x") ) || strCommand.StartsWith( _T("0X") ) ) {
                unsigned long val = 0;
                if ( strCommand.ToULong(&val,0) ) {
                    cmd[0] = val;
                } else {
                    LOGP( _T("INVALID COMMAND - ") + strCommand );
                    delete[] cmd;
                    break;
                }

            } else {
            cmd[0]  = strCommand.GetChar( 0 );
            }

            cmd[1]  = (uint8_t)command->getPageOffset();
            cmd[2]  = command->getValue();

            if ( m_device == ECUInfo::DEVICE_ROUND ) {
                cmd[3] = '.';
            }

            mtCommand = new MegatuneCommand( VemsCommand::WRITE_PAGE, cmd, cmdSize, 0, 0, 0 );

			delete[] cmd;

        } break;

        case VemsCommand::BURN:
        {
            strCommand = m_pSIPRDescriptor->getBurnCommand();

            if ( m_device == ECUInfo::DEVICE_ROUND ) {
                strCommand += _T(".");
            }

            mtCommand = new MegatuneCommand( VemsCommand::BURN, strCommand, 1, 0, 0, 0 );
        } break;
        case VemsCommand::READ_REALTIME:
            //Logger::logf( LOG_MSG, _T("MegatuneProtocol::request( VemsCommand::READ_REALTIME - cmd: " + strCommand ) );
            //mtCommand   = new MegatuneCommand(  VemsCommand::READ_REALTIME, strCommand, m_iRealtimeLength, 10, 2 );
            //Logger::logf( LOG_MSG, _T("MegatuneProtocol():request() - m_iTimeout = %d"), m_iTimeout );
            mtCommand   = new MegatuneCommand(  VemsCommand::READ_REALTIME, m_strReadRealtimeCommand, m_iRealtimeLength, m_iTimeout, 2 );
//            mtCommand   = new MegatuneCommand(  VemsCommand::READ_REALTIME, _T("D"), m_iRealtimeLength, m_iTimeout, 0 );
            break;
        case VemsCommand::ENABLE_READ_REALTIME:
            //Logger::logf( LOG_MSG, _T("MegatuneProtocol::request( VemsCommand::ENABLE_READ_REALTIME : %d )"), command->isReadRealTimeEnabled() );
            m_isReadSensors = command->isReadRealTimeEnabled();
            break;
        case VemsCommand::SYNC:
            m_mode = SYNC;
            break;
        case VemsCommand::GENERIC_STRING:
        {
            wxString  aCommand = ((StringCommand*)command)->m_str;
            mtCommand   = new MegatuneCommand( VemsCommand::GENERIC_STRING, aCommand, 13, m_iTimeout, 32 );
            //Logger::logf( LOG_MSG, _T("MegatuneProtocol::request() - VemsCommand::GENERIC_STRING - ") + aCommand );
        } break;
        case VemsCommand::RAW_DATA:
        {
            TxData rawData = ((RawCommand*)command)->rawData;
            mtCommand   = new MegatuneCommand( VemsCommand::RAW_DATA, (uint8_t*)rawData.data, rawData.size, 0, m_iTimeout, 32 );
        } break;
        default:
            if ( m_pEventHandler ){
                Logger::logf( LOG_MSG, _T("-----MegatuneProtocol::request() - SEND TIME_OUT") );
                SIPRPacket p( 0, SIPR_TIMEOUT );
                InputEvent e;
                e.setPacket( p );
                m_pEventHandler->AddPendingEvent( e );
            }
    }

    if ( mtCommand ) { m_txQueue.push_back( mtCommand ); }

    //Logger::logf( LOG_MSG, _T("MegatuneProtocol::request( ) - queue size: %d"), m_txQueue.size() );
}

void MegatuneProtocol::disable()
{
    if ( !m_bIsDisabled ) { m_bNeedDisable = true; }
}

void MegatuneProtocol::enable()
{
    m_bIsDisabled = false;
}

bool MegatuneProtocol::isDisabled()
{
    return m_bIsDisabled;
}


void MegatuneProtocol::setReadRealTimeCommand()
{
    switch ( m_device )
    {
        case ECUInfo::DEVICE_V3:
        {
    if ( m_strReadRealtimeCommand.IsSameAs( _T("D") ) )
    {
        m_strReadRealtimeCommand = _T("A");
        return;
    }

    m_strReadRealtimeCommand = m_pSIPRDescriptor->getMTReadRealTimeCommand( m_iSensorType );

    if ( m_strReadRealtimeCommand.IsEmpty() )
    {
        m_strReadRealtimeCommand    = _T("D");
            }

        } break;

        case ECUInfo::DEVICE_ROUND:
        {
            m_strReadRealtimeCommand = _T("A.");
        } break;

        default:
            break;
    }

    Logger::logf( LOG_MSG, _T("MegatuneProtocol::setReadRealTimeCommand() ->") + m_strReadRealtimeCommand + _T("<-") );
}

bool MegatuneProtocol::isPageCrcEnabled()
{
    return ( m_firmwareVersion >= FirmwareVersion( 1,1,86 ) );
}
