#include "VemsDeviceController.h"
#include "SIPRTypes.h"
#include "VemsDeviceControllerFSM.h"
#include "../serialconfig.h"
#include "../logger.h"
#include "../events/InputEvent.h"
#include "../events/CommunicationEvent.h"
#include "../MTConfigHelper.h"
#include "../communication/SerialPortHandlerFactory.h"
#include "../CommunicationBase.h"
#include "../communication/CommunicationLogger.h"
#include "../communication/CommunicationController2.h"
#include "../communication/CommunicationControllerEvent.h"
#include "../communication/clients/PasswordChecker.h"
#include "../communication/clients/V3FirmwareInfoFetcher.h"
#include "../communication/clients/RoundFirmwareInfoFetcher.h"
#include "../communication/clients/MegatuneSyncer.h"
#include "../communication/clients/Megatune.h"
#include "../communication/clients/BaudRateSelector.h"
#include "../communication/clients/DeviceModeSelector.h"
#include "../communication/clients/PortScanner.h"
#include "../communication/clients/OffsetReaderClient.h"
#include "../Interruptable.h"
#include "../SerialPortHelper.h"
#include "../DirHelper.h"


const wxEventType START_TIMER    = wxNewEventType();


VemsDeviceController::VemsDeviceController( const CommunicationConfig& config )
    : m_timer ( this, wxID_OK )
    , m_detectWaitTime (1000)
{
    m_config = config;

    m_supportedBaudrates = SerialPortHelper::getSerialBaudratesInt();

    m_fsm = new VemsDeviceControllerFSM( this );

    setInterWriteDelay( m_config.interWriteDelay );
    setTimeout( m_config.timeout );

    m_commController = new CommunicationController2();

    m_isWorking = false;
    m_isResuming = false;
    m_isSkipCannotOpenPort = true;

    Connect( COMMUNICATION_CONTROLLER_START, wxCommandEventHandler( VemsDeviceController::OnStartEvent ), NULL, this );
    Connect( wxID_OK, wxEVT_TIMER, wxTimerEventHandler( VemsDeviceController::OnStartDetection ), NULL, this );
    Connect(  START_TIMER, wxCommandEventHandler( VemsDeviceController::OnStartTimer ), NULL, this );
    Connect( wxID_ANY, wxEVT_COMMUNICATION, CommunicationEventHandler( VemsDeviceController::OnCommunicationEvent ), NULL, this );

    m_eventHandler = NULL;
    m_flags =  VDC_FLAG_DEFAULT;
    m_isSendCannotOpenPortEventEnabled = true;
    m_isUSBCrashed = false;
    m_connectTryCount = 0;
}

VemsDeviceController::~VemsDeviceController()
{
    Disconnect( COMMUNICATION_CONTROLLER_START, wxCommandEventHandler( VemsDeviceController::OnStartEvent ), NULL, this );
    Disconnect( wxID_OK, wxEVT_TIMER, wxTimerEventHandler( VemsDeviceController::OnStartDetection ), NULL, this );
    Disconnect(  START_TIMER, wxCommandEventHandler( VemsDeviceController::OnStartTimer ), NULL, this );
    Disconnect( wxID_ANY, wxEVT_COMMUNICATION, CommunicationEventHandler( VemsDeviceController::OnCommunicationEvent ), NULL, this );

	m_commController->stop();
	delete m_commController;

	clearCommBase();

    delete m_fsm ;
    delete m_commClient;
    delete m_prevCommClient;
}

void VemsDeviceController::setDetectionSleep( int sleep )
{
    m_detectionSleep = sleep;
}

void VemsDeviceController::start( int flags )
{
    m_connectTryCount = 0;

    if ( m_commController ) {
        if ( m_commController->isRunning() ) {
            m_commController->stop();
        }
    }

    m_flags = flags;

    m_isSendCannotOpenPortEventEnabled = true;

    //m_isSkipCannotOpenPort  = false;
    m_isFirstDetection      = true;
    m_isWorking             = true;

    m_deviceInfo.device     = ECUInfo::DEVICE_UNKNOWN;
    m_deviceInfo.deviceMode = ECUInfo::UNKNOWN;

    m_fsm->reset();

    initCommunication();
    initUserMegatuneConfig();

    registerAndStartCommunicationClient( new PortScanner( m_fsm, m_commBase, getSupportedBaudrates(), m_commBase->getBaudRate() ) );

    sendDeviceControllerEvent( VDC_EVENT_STARTED );


}

void VemsDeviceController::start( const CommunicationConfig& config, int flags )
{
    m_config = config;

    LOGP( wxString::Format( _T("m_config.timeout: %d"), m_config.timeout ) );

    start( flags );
}

void VemsDeviceController::resume( int flag )
{
    LOGP( m_prevDeviceInfo.toString() );

    if ( m_prevDeviceInfo.isUnknown() ) {
        start();
        return;
    }

    sendDeviceControllerEvent( VDC_EVENT_START_RESUMING );

    m_deviceInfo.device     = ECUInfo::DEVICE_UNKNOWN;
    m_deviceInfo.deviceMode = ECUInfo::UNKNOWN;

    m_isResuming = true;
    m_isWorking = true;
     m_isSendCannotOpenPortEventEnabled = true;



    m_fsm->resume( flag & VDC_FLAG_GET_PAGES );

    initCommunication();

    registerAndStartCommunicationClient( new PortScanner( m_fsm, m_commBase, getSupportedBaudrates(), m_commBase->getBaudRate() ) );
}

void VemsDeviceController::stop( int flag )
{
    m_timer.Stop();

    if(m_commController) m_commController->stop();
    if(m_commBase) m_commBase->close();
    ///TODO
//    clearCommBase();

    m_isWorking = false;

    save();

    if ( !( flag & VDC_FLAG_NOEVENT ) )  {
        sendDeviceControllerEvent( VDC_EVENT_STOPPED );
    }
}

bool VemsDeviceController::isRunning()
{
    return m_commController->isRunning();
}

void VemsDeviceController::OnStartEvent(wxCommandEvent& event)
{
    LOG

    delete m_prevCommClient;
    m_prevCommClient = NULL;

    LOGP( wxString::Format( _T("m_isWorking: %d"), m_isWorking ) );
    if ( m_isWorking ) {
        m_commController->start();

///TODO
//        if ( m_fsm->isReconnecting() ) {
//                WX_CLEAR_LIST( SIPRPacketList, m_megatuneConfig.pages );
//                m_commClient->request( new ReadAllPageCommand() );
//        }
    } else {
        delete m_commClient;
        m_commClient = NULL;
    }
}

void VemsDeviceController::OnStartDetection( wxTimerEvent& event )
{
    LOG
    startDetecting();
}

void VemsDeviceController::OnCommunicationEvent( CommunicationEvent& event )
{
    LOGP( event.toString() );

    switch ( event.GetId() )
    {
//        case CommunicationEvent::REMOVE_USB:
//        {
//            m_isUSBCrashed = true;
//        } break;

        case CommunicationEvent::CANNOT_OPEN_PORT:
        {
            m_fsm->reset();

            if ( m_isSkipCannotOpenPort ) {
                m_timer.Start( m_detectWaitTime, true );
            }

            if (  m_isSendCannotOpenPortEventEnabled ) {
                m_isSendCannotOpenPortEventEnabled = false;
            } else {
                return;
            }
        } break;

        case CommunicationEvent::PORT_CLOSED:
        {
            m_fsm->reset();
            resetDevice();

//            if ( !m_isUSBCrashed ) {
                sendDeviceControllerEvent( VDC_EVENT_DEVICE_DISCONNECTED );
//            }

            m_isWorking = true;
            startDetectingTimer();
        } break;
    }

    m_eventHandler->AddPendingEvent( event );
}

void VemsDeviceController::readOffsetSIPR( wxArrayString addresses, wxEvtHandler* evtHandler )
{
    FirmwareInfo& info = m_deviceInfo.firmwareInfo;
    FirmwareVersion ver = info.getFirmwareVersion();

    /// above 1.1.80 responds 32 byte + 2byte CRC
    bool crc = ( (ver.major >= 1) && (ver.minor >= 1) && (ver.build>=80) );

    CommunicationControllerClient* client = new OffsetReaderClient( getDeviceMode(), addresses, evtHandler, crc );
    registerAndStartCommunicationClient( client, 512, 100 );
}

void VemsDeviceController::readOffset( bool start, wxArrayString offsets )
{
    FirmwareInfo& info = m_deviceInfo.firmwareInfo;
    FirmwareVersion ver = info.getFirmwareVersion();

    /// above 1.1.80 responds 32 byte + 2byte CRC
    bool crc = ( (ver.major >= 1) && (ver.minor >= 1) && (ver.build>=80) );

    if ( start && ( offsets.GetCount() > 0 ) )
    {
        IntArray baudrates = SerialPortHelper::getSerialBaudratesInt();
        CommunicationControllerClient* client = new OffsetReaderClient( m_commBase, baudrates, getDeviceMode(), offsets, crc );
        registerAndStartCommunicationClient( client, 512, 100 );
    }
    else
    {
        /// try to cast to Interruptable
        Interruptable* client = dynamic_cast<Interruptable*>( m_commClient );
        if ( client )
        {
            client->Abort();
        }
        else
        {
            LOGP( _T("stop readoffset - cannot cast client to Interruptable!") )
        }
    }
}

void VemsDeviceController::readOffset( bool start, StringHash vars, wxEvtHandler* evthandler )
{
    FirmwareInfo& info = m_deviceInfo.firmwareInfo;
    FirmwareVersion ver = info.getFirmwareVersion();

    // above 1.1.80 responds 32 byte + 2byte CRC
    bool crc = ( (ver.major >= 1) && (ver.minor >= 1) && (ver.build>=80) );

    if ( start && ( vars.size() > 0 ) )
    {
        CommunicationControllerClient* client = new OffsetReaderClient( getDeviceMode(), vars, evthandler, crc );
        registerAndStartCommunicationClient( client, 512, 100 );
    }
    else
    {
        // try to cast to Interruptable
        Interruptable* client = dynamic_cast<Interruptable*>( m_commClient );
        if ( client )
        {
            client->Abort();
        }
        else
        {
            LOGP( _T("stop readoffset - cannot cast client to Interruptable!") )
        }
    }
}


/*void VemsDeviceController::readOffsetSIPR( wxArrayString addresses )
{
    CommunicationControllerClient* client = new OffsetReaderClient( getDeviceMode(), addresses, this );
    registerAndStartCommunicationClient( client, 512, 100 );
}

void VemsDeviceController::readOffset( bool start, wxArrayString offsets )
{
    FirmwareInfo& info = m_deviceInfo.firmwareInfo;
    FirmwareVersion ver = info.getFirmwareVersion();

    // above 1.1.80 responds 32 byte + 2byte CRC
    bool crc = ( (ver.major >= 1) && (ver.minor >= 1) && (ver.build>=80) );

    if ( start && ( offsets.GetCount() > 0 ) )
    {
        CommunicationControllerClient* client = new OffsetReaderClient( getDeviceMode(), offsets, crc );
        registerAndStartCommunicationClient( client, 512, 100 );
    }
    else
    {
        // try to cast to Interruptable
        Interruptable* client = dynamic_cast<Interruptable*>( m_commClient );
        if ( client )
        {
            client->Abort();
        }
        else
        {
            LOGP( _T("stop readoffset - cannot cast client to Interruptable!") )
        }
    }
}

void VemsDeviceController::readOffset( bool start, StringHash vars, wxEvtHandler* evthandler )
{
    FirmwareInfo& info = m_deviceInfo.firmwareInfo;
    FirmwareVersion ver = info.getFirmwareVersion();

    // above 1.1.80 responds 32 byte + 2byte CRC
    bool crc = ( (ver.major >= 1) && (ver.minor >= 1) && (ver.build>=80) );

    if ( start && ( vars.size() > 0 ) )
    {
        CommunicationControllerClient* client = new OffsetReaderClient( getDeviceMode(), vars, evthandler, crc );
        registerAndStartCommunicationClient( client, 512, 100 );
    }
    else
    {
        // try to cast to Interruptable
        Interruptable* client = dynamic_cast<Interruptable*>( m_commClient );
        if ( client )
        {
            client->Abort();
        }
        else
        {
            LOGP( _T("stop readoffset - cannot cast client to Interruptable!") )
        }
    }
}*/

void VemsDeviceController::changeBaudRate( size_t baudrate )
{
    LOGP( wxString::Format( _T(" m_commBase->getBaudRate(): %d, baudrate: %d"),  m_commBase->getBaudRate(), baudrate ) );

    if ( m_commBase->getBaudRate() == (int)baudrate ) {
            sendDeviceControllerEvent( VDC_EVENT_DEVICE_BAUDRATE_CHANGED );
        return;
    }

    m_commController->stop();

    IntArray baudrates = SerialPortHelper::getSerialBaudratesInt();
    CommunicationControllerClient* client = new BaudRateSelector( m_fsm, m_fsm, m_commBase, baudrates, getDeviceMode(), m_commBase->getBaudRate(), baudrate );
    registerAndStartCommunicationClient( client, 512, 100 );

    return;
}

void VemsDeviceController::leaveBootMode()
{
    sendDeviceControllerEvent( VDC_EVENT_LEAVING_BOOT_MODE );

    CommunicationControllerClient* client = new DeviceModeSelector( m_fsm, m_fsm,
                                                                    m_commBase,
                                                                    getSupportedBaudrates(),
                                                                    ECUInfo::BOOT,
                                                                    ECUInfo::MEGATUNE,
                                                                    &m_deviceInfo.deviceMode );
    registerAndStartCommunicationClient( client, 512, 100 );
}

void VemsDeviceController::stayBootMode()
{
//    stop( VDC_FLAG_NOEVENT );
   // connect();
}

bool VemsDeviceController::isWorking() const
{
    return m_isWorking;
}

bool VemsDeviceController::isConnected() const
{
    ///TODO
    if ( !m_isWorking || !m_fsm ) {
        return false;
    }

    return m_fsm->isConnected();
}

bool VemsDeviceController::isDetecting() const
{
    ///TODO
    return m_fsm->isDetecting();
}

void VemsDeviceController::receive( const SIPRPacket& packet )
{
//    InputEvent e;
//    e.setPacket( packet );
//    m_eventHandler->AddPendingEvent( e );
}

void VemsDeviceController::request( VemsCommand* command )
{
    m_commClient->request( command );
}

void VemsDeviceController::sendPages()
{

}

void VemsDeviceController::checkPassword()
{
//    if ( !isV3() ) {
//        m_fsm->handleEvent( CommunicationClientEvent::DEVICE_NOT_PROTECTED );
//        return;
//    }

    sendDeviceControllerEvent( VDC_EVENT_DEVICE_CHECKING_PASSWORD );

    CommunicationControllerClient* client = new PasswordChecker( m_fsm );
    registerAndStartCommunicationClient( client, 512, 100 );
}

void VemsDeviceController::fetchFirmwareInfo()
{
    sendDeviceControllerEvent( VDC_EVENT_DEVICE_FIRMWAREINFO_FETCHING );

    CommunicationControllerClient* client = NULL;

    if ( isV3() ) {
        client = new V3FirmwareInfoFetcher( m_fsm, m_fsm );
    } else {
        client = new RoundFirmwareInfoFetcher( m_fsm );
    }

    registerAndStartCommunicationClient( client, 512, 100 );
}

void VemsDeviceController::sync()
{
    sendDeviceControllerEvent( VDC_EVENT_DEVICE_FIRMWAREDATA_SYNCING );

    initDetectedMegatuneConfig();

#ifdef COMM_MODEL_WITH_GUI
    MTConfigHelper::parseMTConfigDataV2( m_deviceInfo.firmwareInfo );
#else
    MTConfigHelper::parseMTConfigData( m_deviceInfo.firmwareInfo );
#endif

    ///TODO
    //    if ( !isFileParsed ) {
//        /// TODO - handle error
//        return;
//    }


    MTConfigData* mtConfig = MTConfigHelper::getMTConfigData();
    SIPRDescriptor* siprDesc = mtConfig->getSIPRDescriptor();

    CommunicationControllerClient* client = new MegatuneSyncer( m_fsm, m_fsm, siprDesc, &m_megatuneConfig );
    registerAndStartCommunicationClient( client, 512, 100 );
}



void VemsDeviceController::connect()
{
    ///TODO - Man, SIPR ?
//    m_isSkipCannotOpenPort = true;
    m_isFirstDetection = false;
    m_connectTryCount = 0;

    save();

//    if ( isBootMode() ) {
//        return;
//    }

    ///TODO handle reconnect
   // MTConfigHelper::parseMTConfigData( m_fwInfo );

    logConnectionParams();

    CommunicationControllerClient* client =  new Megatune( m_fsm, this, m_megatuneConfig, MTConfigHelper::getMTConfigData()->getSIPRDescriptor() );
    registerAndStartCommunicationClient( client, 512, 100 );

    sendDeviceControllerEvent( VDC_EVENT_DEVICE_CONNECTED );
}

bool VemsDeviceController::processPage( InputEvent& event )
{
    static bool addPageEnabled = false;

    SIPRPacket packet = event.getPacket();

    switch ( packet.type() )
    {
        case SIPR_GET_CONFIG_1:
        case SIPR_GET_CONFIG_2:
        case SIPR_GET_TABLEDATA:
        {   if ( addPageEnabled ) {
                m_megatuneConfig.pages.Append( new SIPRPacket( packet ) );
            }
        } break;

        case SIPR_LOG_MARKER:
        {
            if ( packet.data() && packet.size() > 0 ) {
                if ( packet.data()[0] == SIPR_REFRESH_START ) {
                    addPageEnabled = true;
                } else if ( packet.data()[0] == SIPR_REFRESH_STOP ) {
                    addPageEnabled = false;
                    Disconnect( wxEVT_INPUT, InputEventHandler( VemsDeviceControllerFSM::OnSIPRInput ), NULL, m_fsm );
                    return true;
                }
            }
        } break;

        default:
            return false;
    }

    return false;
}

void VemsDeviceController::getAllPages()
{
//    ///TODO - Man, SIPR ?
//
//    if ( isBootMode() ) {
//        return;
//    }
    WX_CLEAR_LIST( SIPRPacketList, m_megatuneConfig.pages );

    Connect( wxEVT_INPUT, InputEventHandler( VemsDeviceControllerFSM::OnSIPRInput ), NULL, m_fsm );

    CommunicationControllerClient* client =  new Megatune( m_fsm, this, m_megatuneConfig, MTConfigHelper::getMTConfigData()->getSIPRDescriptor() );
    client->request( new ReadAllPageCommand() );
    registerAndStartCommunicationClient( client, 512, 100 );
}

void VemsDeviceController::deviceResumed()
{
    logConnectionParams();

    CommunicationControllerClient* client =  new Megatune( m_fsm, this, m_megatuneConfig, MTConfigHelper::getMTConfigData()->getSIPRDescriptor() );
    registerAndStartCommunicationClient( client, 512, 100 );

    sendDeviceControllerEvent( VDC_EVENT_RESUMED );
}

void VemsDeviceController::timeout()
{
    if ( m_connectTryCount >= MAX_CONNECT_TRY_COUNT ) {
        m_connectTryCount = 0;
        sendDeviceControllerEvent( VDC_EVENT_DEVICE_CANNOT_CONNECT );
        resetDevice();
        stop();
        return;
    }

    m_fsm->reset();
    m_isWorking = true;
    m_isFirstDetection = false;
    Disconnect( wxEVT_INPUT, InputEventHandler( VemsDeviceControllerFSM::OnSIPRInput ), NULL, m_fsm );
//    save();
    resetDevice();

    sendDeviceControllerEvent( VDC_EVENT_DEVICE_DISCONNECTED );

    startDetectingTimer();
}

void VemsDeviceController::resetDevice()
{
    m_deviceInfo.device        = ECUInfo::DEVICE_UNKNOWN;
    m_deviceInfo.deviceMode    = ECUInfo::UNKNOWN;
}

void VemsDeviceController::startDetectingTimer()
{
    LOG
    //m_timer.Start( m_detectWaitTime, true );
    wxCommandEvent event(START_TIMER);
    AddPendingEvent( event );
}

void VemsDeviceController::OnStartTimer( wxCommandEvent& event)
{
    LOG
    m_timer.Start( m_detectWaitTime, true );
}

void VemsDeviceController::startDetecting()
{
    //clearCommBase();
    LOG

    registerAndStartCommunicationClient( new PortScanner( m_fsm, m_commBase, getSupportedBaudrates() ) );

//    wxArrayString ports;
//    ports.Add( m_deviceInfo.serialConfig.getPortName() );

//    delete m_serialPortScanner;
//    m_serialPortScanner = new SerialPortScanner( m_fsm, ports, baudrates, m_config.timeout, m_config.interWriteDelay );
}

void VemsDeviceController::clearCommBase()
{
    LOGP( wxString::Format( _T("m_commBase: %p"), m_commBase ) );
    if ( m_commBase ) {
        m_commBase->close();
        delete m_commBase;
        m_commBase = NULL;
    }

    LOGP( wxString::Format( _T("m_commBase: %p"), m_commBase ) );
}

void VemsDeviceController::selectMegatuneMode()
{
    IntArray baudrates = SerialPortHelper::getSerialBaudratesInt();
    CommunicationControllerClient* client = new DeviceModeSelector( m_fsm, m_fsm, m_commBase, baudrates, getDeviceMode(), ECUInfo::MEGATUNE, &m_deviceInfo.deviceMode );
    registerAndStartCommunicationClient( client, 512, 100 );
}

void VemsDeviceController::disconnectFSMSiprInput()
{
    Disconnect( wxEVT_INPUT, InputEventHandler( VemsDeviceControllerFSM::OnSIPRInput ), NULL, m_fsm );
}

void VemsDeviceController::sendDeviceControllerEvent( VemsDeviceControllerEventType eventType )
{
    VemsDeviceControllerEvent event( eventType );
    event.setDeviceInfo( m_deviceInfo );
    m_eventHandler->AddPendingEvent( event );
}

void VemsDeviceController::save()
{
    m_prevDeviceInfo = m_deviceInfo;

    wxString prevVerStr = m_prevDeviceInfo.firmwareInfo.getFirmwareVersionAsString();
    wxString verStr = m_deviceInfo.firmwareInfo.getFirmwareVersionAsString();

    LOGP( _T("prevVer: ") + prevVerStr + _T(", ver: ") + verStr );

}

void VemsDeviceController::stopCommunicationController()
{
    m_commController->stop();
}

void VemsDeviceController::baudrateChanged()
{
    sendDeviceControllerEvent( VDC_EVENT_DEVICE_BAUDRATE_CHANGED );
}

bool VemsDeviceController::isSyncingNeeded() const
{
    bool isSyncingNeeded = !m_deviceInfo.firmwareInfo.isVersionAndCompileOptionsIdentical( m_prevDeviceInfo.firmwareInfo );
    return isSyncingNeeded;
}

bool VemsDeviceController::hasPreferredBaudrate() const
{
    return ( m_deviceInfo.getBaudrate() != m_config.preferredBaudrate );
}


void VemsDeviceController::initCommunication()
{
    setInterWriteDelay( m_config.interWriteDelay );


    clearCommBase();

    if ( m_config.mode.IsSameAs( _T("SERIAL") ) )
    {

		setBaseTimeout(0);
		setTimeout( m_config.timeout );


		m_deviceInfo.serialConfig.setPortName( m_config.port );
		m_deviceInfo.serialConfig.setBaudRate( m_config.baudrate );
//		m_deviceInfo.serialConfig.setBaudRate( m_config.preferredBaudrate );

//        wxString logFileName = DirHelper::GetPath( dDEBUGLOG ) + _T("CommLog.txt");
//        m_commBase = new CommunicationLogger( SerialPortHandlerFactory::create(&m_deviceInfo.serialConfig), logFileName );
        m_commBase = SerialPortHandlerFactory::create(&m_deviceInfo.serialConfig);
//        LOGP( wxString::Format( _T("Create SerialPort! - port: ") + m_config.port + _T(", baudrate: %d"), m_config.baudrate ) );
    }
    else
    {
        //setBaseTimeout(m_config.tcpTimeout);
        setTimeout( m_config.tcpTimeout );
        wxString host = m_config.ip.BeforeFirst( _T(':') );
        long port = 0;
        m_config.ip.AfterLast( _T(':') ).ToLong( &port );

        m_commBase =  SerialPortHandlerFactory::create( host, port );

        LOGP( wxString::Format( _T("Create TCPClientHandler! - ip: ") + host + _T(", port: %d"), port ) );
    }

    m_isFirstDetection  = true;
    m_detectionSleep    = 1000;


//    m_commController        = new CommunicationController2( this, m_commClient, m_commBase );

    m_commController->setEventHandler(this);
    m_commController->setCommunicationBase(m_commBase);
    m_commController->registerClient(m_commClient);
}

void VemsDeviceController::initMegatuneConfig()
{
    initDetectedMegatuneConfig();
    initUserMegatuneConfig();
}

void VemsDeviceController::initDetectedMegatuneConfig()
{
    if ( m_deviceInfo.isV3() ) {

        m_megatuneConfig.rtType = MegatuneConfig::EXTENDED;

        if ( m_megatuneConfig.fwVersion >= FirmwareVersion(1,1,86 ) ) {
            m_megatuneConfig.rtCheckType = MegatuneConfig::TRIGGER;
            m_megatuneConfig.pageCheckType = MegatuneConfig::CRC;
            char rtCmd = 0x7E;
            m_megatuneConfig.readRealtimeCommand = wxString::From8BitData(&rtCmd, 1);
        } else if ( m_megatuneConfig.fwVersion >= FirmwareVersion(1,1,75 ) ) {
            m_megatuneConfig.rtCheckType    = MegatuneConfig::CRC;
            m_megatuneConfig.pageCheckType  = MegatuneConfig::LENGTH;
            m_megatuneConfig.readRealtimeCommand = wxString::From8BitData("D");
        } else {
            m_megatuneConfig.rtCheckType    = MegatuneConfig::LENGTH;
            m_megatuneConfig.pageCheckType  = MegatuneConfig::LENGTH;
            m_megatuneConfig.readRealtimeCommand = wxString::From8BitData("D");
        }

    } else {

        m_megatuneConfig.rtType         = MegatuneConfig::NORMAL;
        m_megatuneConfig.rtCheckType    = MegatuneConfig::LENGTH;
        m_megatuneConfig.pageCheckType  = MegatuneConfig::LENGTH;
        m_megatuneConfig.readRealtimeCommand = wxString::From8BitData("A.");

    }

    m_megatuneConfig.device                 = m_deviceInfo.device;
//    m_megatuneConfig.readRealtimeCommand    = _T("D");
//    m_megatuneConfig.realtimeCRCEnabled     = true;
    m_megatuneConfig.realtimeLength         = 100;
    m_megatuneConfig.minRealtimeLength      = isV3() ? 50 : 30; ///TODO - TEST
    m_megatuneConfig.maxRealtimeLength      = 128;

    switch ( m_deviceInfo.device )
    {
        case ECUInfo::DEVICE_ROUND:
            m_megatuneConfig.siprSensorType = SiprSensorType::SIPR_RT_ROUND;
            break;
        case ECUInfo::DEVICE_V3:
            m_megatuneConfig.siprSensorType = SiprSensorType::SIPR_RT_MT_NORMAL;
            break;

        default:
            break;
    }
//    m_megatuneConfig.siprSensorType         = SiprSensorType::SIPR_RT_MT_NORMAL;

    WX_CLEAR_LIST( SIPRPacketList, m_megatuneConfig.pages );

    LOGP( wxString::Format( _T("m_megatuneConfig.pages.GetCount(): %d"), m_megatuneConfig.pages.GetCount() ) );
}

void VemsDeviceController::initUserMegatuneConfig()
{
    if ( m_config.mode.IsSameAs( _T("SERIAL") ) ) {
        m_megatuneConfig.timeout = m_config.timeout;
    } else {
        m_megatuneConfig.timeout = m_config.tcpTimeout;
    }

    m_megatuneConfig.interWriteDelay    = m_config.interWriteDelay;
    m_megatuneConfig.maxRealtimeError   = m_config.rtErrorCount;
}

void VemsDeviceController::logConnectionParams()
{
    MTConfigData* mtConfig = MTConfigHelper::getMTConfigData();
    SIPRDescriptor* siprDesc = mtConfig->getSIPRDescriptor();

    wxString rtCheckTypeStr;

    switch ( m_megatuneConfig.rtCheckType )
    {
        case MegatuneConfig::TRIGGER:
            rtCheckTypeStr = _T("TRIGGER");
            break;

        case MegatuneConfig::CRC:
            rtCheckTypeStr = _T("CRC");
            break;

        case MegatuneConfig::LENGTH:
            rtCheckTypeStr = _T("LENGTH");
            break;
    }

    wxString pageCheckTypeStr;

    if (m_megatuneConfig.pageCheckType == MegatuneConfig::CRC ) {
        pageCheckTypeStr = _T("CRC");
    } else {
        pageCheckTypeStr = _T("LENGTH");
    }

    wxString message = _T("\n>>>DEVICE CONNECTED SUCCESSFULLY!\n");
    message += _T("\tPort: ") + m_deviceInfo.getPortName() + _T(", Baudrate: ") + m_deviceInfo.getBaudrateAsString() + _T("\n");
    message += _T("\tdevice: ") + ECUInfo::getDeviceAsString( m_deviceInfo.device );
    message += _T(", mode: ") + ECUInfo::getStateAsString( m_deviceInfo.deviceMode ) + _T("\n");
    message += _T("\tFirmware version: ") + m_deviceInfo.firmwareInfo.getFirmwareVersionAsString() + _T("\n");
    message += _T("\tRealtime Check Type: ") + rtCheckTypeStr + _T("\n");

    //message += _T("\tRealtime Read Command: ") + m_megatuneConfig.readRealtimeCommand + _T("\n");
    message += wxString::Format( _T("\tRealtime length: %d\n"), m_megatuneConfig.realtimeLength );
    message += _T("\tPage Check Type: ") + pageCheckTypeStr + _T("\n");
    message += wxString::Format( _T("\tPages detected(Ini/ECU): %d / %d \n"), siprDesc->getPageCount(), m_megatuneConfig.pageSize.size() );

    for ( size_t i = 0; i < siprDesc->getPageCount(); ++i ) {
        int pageNumber =  siprDesc->getPageNumber( i );
        int cfgPageSize = siprDesc->getPageSize( pageNumber );
        int ecuPageSize = 0;

        Int2IntHash::const_iterator it = m_megatuneConfig.pageSize.find( pageNumber );

        if ( it != m_megatuneConfig.pageSize.end() ) {
            ecuPageSize = it->second;
        }

        message += wxString::Format( _T("\t\tpage: %d, Config size: %d, ECU size: %d, \n"),
            pageNumber, cfgPageSize, ecuPageSize );
    }

    message += _T(">>>\n");

    LOGP( message );
}

void VemsDeviceController::registerAndStartCommunicationClient( CommunicationControllerClient* client,
                                                                size_t clearingReadCount, size_t sleepBeforeClearing )
{
//    m_commController->clear( clearingReadCount, sleepBeforeClearing );
    m_commController->clear( 64, 64 );

    m_prevCommClient = m_commClient;
    m_commClient = client;

    ///TODO handle TCP
    m_commClient->setTimeout( m_timeout );
    m_commClient->setInterWriteDelay( m_interWriteDelay );
    m_commController->registerClient( m_commClient );

    LOGP( wxString::Format( _T("m_timeout: %d, m_interWriteDelay: %d"), m_timeout,m_interWriteDelay ) );

    wxCommandEvent startEvent( COMMUNICATION_CONTROLLER_START, wxID_ANY );
    AddPendingEvent( startEvent );
}

void VemsDeviceController::deviceFound()
{
    ++m_connectTryCount;
}

void VemsDeviceController::deviceNotFound()
{
    LOGP( wxString::Format( _T("m_isFirstDetection: %d"), m_isFirstDetection ) );

    if ( m_isFirstDetection ) {
        m_isFirstDetection = false;
        sendDeviceControllerEvent( VDC_EVENT_DEVICE_NOT_FOUND );
    }

    //m_timer.Start( m_detectWaitTime, true );
    startDetectingTimer();
}
/**
void VemsDeviceController::deviceFoundDetecting( const DeviceInfo& deviceInfo )
{
    m_isSkipCannotOpenPort = true;

//    delete m_serialPortScanner;
//    m_serialPortScanner = NULL;

//    m_deviceInfo = deviceInfo;

//    delete m_commBase;
//    m_commBase = SerialPortHandlerFactory::create(&m_deviceInfo.serialConfig);
//    m_commController->setCommunicationBase( m_commBase );

    sendDeviceControllerEvent( VDC_EVENT_DEVICE_FOUND );
}


void VemsDeviceController::deviceNotFoundDetecting( const DeviceInfo& deviceInfo )
{
//    delete m_serialPortScanner;
//    m_serialPortScanner = NULL;

    m_deviceInfo = deviceInfo;

    if ( m_isFirstDetection ) {
        m_isFirstDetection = false;

        sendDeviceControllerEvent( VDC_EVENT_DEVICE_NOT_FOUND );
    }

    m_timer.Start( m_detectWaitTime, true );
}

void VemsDeviceController::cannotOpenPortDetecting( const DeviceInfo& deviceInfo )
{
//    delete m_serialPortScanner;
//    m_serialPortScanner = NULL;

    m_deviceInfo = deviceInfo;

    if ( m_isSkipCannotOpenPort ) {
        LOG
        m_timer.Start( m_detectWaitTime, true );
    } else {
        LOG
        m_isSkipCannotOpenPort = true;
        m_isWorking = false;
        CommunicationEvent commEvent( CommunicationEvent::CANNOT_OPEN_PORT );
        AddPendingEvent( commEvent );
    }
}

void VemsDeviceController::deviceFoundResuming( const DeviceInfo& deviceInfo )
{
    LOGP( deviceInfo.toString() );

//    delete m_serialPortScanner;
//    m_serialPortScanner = NULL;

    m_deviceInfo.device = deviceInfo.device;
    m_deviceInfo.deviceMode = deviceInfo.deviceMode;
    m_deviceInfo.serialConfig = deviceInfo.serialConfig;

//    m_deviceInfo = deviceInfo;


    m_commBase = SerialPortHandlerFactory::create(&m_deviceInfo.serialConfig);
    m_commController->setCommunicationBase( m_commBase );

    LOGP( m_deviceInfo.toString() );
}

void VemsDeviceController::deviceNotFoundResuming( const DeviceInfo& deviceInfo )
{

}

void VemsDeviceController::cannotOpenPortResuming( const DeviceInfo& deviceInfo )
{
    m_isWorking = false;
}
*/

bool VemsDeviceController::isDeviceResumed() const
{
    bool isResumed = (  ( m_deviceInfo.device == m_prevDeviceInfo.device ) &&
                        ( m_deviceInfo.deviceMode == m_prevDeviceInfo.deviceMode ) );

    return isResumed;
}

bool VemsDeviceController::isBaudrateResumed()
{
    bool isResumed = ( m_deviceInfo.serialConfig.getBaudRate() == m_prevDeviceInfo.serialConfig.getBaudRate() );
    return isResumed;
}

bool VemsDeviceController::isDeviceReconnected() const
{
    LOGP( wxString::Format( _T("m_prevDeviceInfo.isUnknown(): %d"), m_prevDeviceInfo.isUnknown() ) );
    LOGP( wxString::Format( _T("m_prevDeviceInfo.isUnknownMode(): %d"), m_prevDeviceInfo.isUnknownMode() ) );
    LOGP( wxString::Format( _T("m_flags: %d"), m_flags ) );

    if ( m_prevDeviceInfo.isUnknown() || m_prevDeviceInfo.isUnknownMode() || ( m_flags & VDC_FLAG_NO_RECONNECT ) ) {
        return false;
    }

    return m_deviceInfo.firmwareInfo.isVersionAndCompileOptionsIdentical( m_prevDeviceInfo.firmwareInfo );
}

void VemsDeviceController::resumeBaudrate()
{
    changeBaudRate( m_prevDeviceInfo.serialConfig.getBaudRate() );
}

void VemsDeviceController::changeToPreferredBaudrate()
{
    changeBaudRate( m_config.preferredBaudrate );
}

bool VemsDeviceController::isOnline() const
{
    return m_isWorking;
}

void VemsDeviceController::enableRealtime()
{
    Megatune* client = dynamic_cast<Megatune*>( m_commClient );

    if ( client ) {
        client->enableRealtime();
    }
}

void VemsDeviceController::refreshDeviceInfoBaudrate()
{
    m_deviceInfo.serialConfig.setBaudRate( m_commBase->getBaudRate() );
}

bool VemsDeviceController::isUseSerialComm() const
{
    return m_config.mode.IsSameAs( _T("SERIAL") );
}

bool VemsDeviceController::isLeaveBootModeEnabled()
{
    return ( m_flags & VDC_FLAG_AUTO_LEAVE_BOOT_MODE );
}

