#include "VemsLogReader.h"

#include "../Utils.h"
#include "../logger.h"
#include "../RawLogHelper.h"
#include "SIPRTypes.h"
#include "SIPRParser.h"
#include "VemsLogHeaderParser.h"

VemsLogReader::VemsLogReader()
{
    init();
}

VemsLogReader::~VemsLogReader()
{
	m_rawFile.Close();
	if ( m_blockBuffer ) { delete m_blockBuffer; }
}

void VemsLogReader::init()
{
    m_blockBuffer = NULL;
	m_gpsDataLength = 0;
    m_gpsBlockIdx = 0;
    m_datalen = 0;
    m_blocksize = 0;
    m_blockIdx = 0;

    m_seq = 0;
    m_isGPSCaptured = false;
    m_isFirstAfterGPS = false;
    //m_tmpPacket;

    m_isValid = false;
    m_seekIdx = 0;
    m_isSeekSet = false;
}

void VemsLogReader::reset()
{
    //Logger::logf(LOG_MSG,_T("RTDataInputRawLog::reset()"));
    m_rawFile.Seek(HEADERBLOCKSIZE);
    m_blockIdx = 0;
}

bool VemsLogReader::SetInputFile(const wxString &filename)
{
    init();
    m_fileName = filename;
    return true;
}

/** @brief GetFileSize
  *
  * @todo: document this function
  */
unsigned int VemsLogReader::GetFileSize()
{
    return m_rawFile.Length();
}

/** @brief GetFilePos
  *
  * @todo: document this function
  */
unsigned int VemsLogReader::GetFilePos()
{
    return m_rawFile.Tell();
}

int VemsLogReader::getNext(SIPRPacket& siprPacket)
{
    if(m_vemsLogVersion == VEMSLOG_VERSION_1){
        return getNextVersion1(siprPacket);
    }
    if ( m_vemsLogVersion == VEMSLOG_VERSION_2 )
        return getNextVersion2(siprPacket);
    if ( m_vemsLogVersion == VEMSLOG_VERSION_3 )
        return getNextVersion3(siprPacket);

    return -1;
}

int VemsLogReader::loadNextBlock()
{
    // Logger::logf( LOG_MSG, _T("VemsLogReader::loadNextBlock()") );

    if(m_isSeekSet){
        m_rawFile.Seek( m_seekIdx * m_blocksize, wxFromStart);
         //observerNotify(0);
        m_isSeekSet = false;
    }

    int res = m_rawFile.Read(m_blockBuffer,m_blocksize);
    observerNotify(m_rawFile.Tell(), m_rawFile.Length());


    if(res != m_blocksize){
        return -2;
    }
    m_blockIdx = 0;
    return 0;
}

int VemsLogReader::getNextVersion3(SIPRPacket& siprPacket)
{
    return getNextVersion2( siprPacket );
}

int VemsLogReader::getNextVersion2(SIPRPacket& siprPacket)
{
     if( !m_isValid ){
        return -1;
    }

    SIPRParser siprParser;
    bool found = false;
    while(!found){
        for(int i = m_blockIdx; i < m_blocksize; i++){

            int res = siprParser.feed(m_blockBuffer[i]);

            if ( res == SIPRParser::SIPR_STATE_SEQ )
            {
                //Logger::logf( LOG_MSG, _T("VemsLogReader::getNextVersion2() - SIPRPacket found at BlockIdx: %d "), i );
            }

            if( res == SIPRParser::SIPR_STATE_CRC_OK){
                    //uint8_t crc = m_blockBuffer[i];
                    //Logger::logf(LOG_MSG,_T("VemsLogReader::getNextVersion2() CRC_OK m_blockIdx:%i, crc: %02X"), i, crc);
                    SIPRPacket packet(  siprParser.sequence(), siprParser.type(), siprParser.size(),
                        siprParser.data(),siprParser.crc() );

                    if( (packet.type() == SIPR_READ_SENSORS) &&
                        (packet.getSubType() == SiprSensorType::GPS) ){
                        m_isGPSCaptured = true;
                        m_isFirstAfterGPS = true;
                    }else if(m_isFirstAfterGPS){
                        m_isFirstAfterGPS = false;
                    }else{
                        m_isGPSCaptured = false;
                    }
                    siprPacket = packet;
                    m_blockIdx = i + 1;

                    return 0;
            }else if(res == SIPRParser::SIPR_STATE_CRC_ERROR) {
                    if ( m_vemsLogVersion == VEMSLOG_VERSION_2 )
                        Logger::logf(LOG_MSG,_T("VemsLogReader::getNextVersion2() CRC_ER m_blockIdx:%i"), i );
                    else if ( m_vemsLogVersion == VEMSLOG_VERSION_3 )
                        Logger::logf(LOG_MSG,_T("VemsLogReader::getNextVersion3() CRC_ER m_blockIdx:%i"), i );
                    if( loadNextBlock() < 0) return -2;
            }
        }
        if( loadNextBlock() < 0) return -2;
        siprParser.reset();
    }
    return -2;
}

int VemsLogReader::getNextVersion1(SIPRPacket& siprPacket)
{
    //Logger::logf(LOG_MSG,_T("VemsLogReader::getNextVersion1() - BEGIN") );

    if( !m_isValid ){
        return -1;
    }
    //Logger::logf(LOG_MSG,_T("VemsLogReader::getNext()") );

    if(m_isFirstAfterGPS){
        siprPacket = m_tmpPacket; /// copy send RT
        m_isFirstAfterGPS = false;
        return 0;
    }

    if( m_blocksize - m_blockIdx <= ( m_datalen + m_gpsDataLength ) ){
         if( loadNextBlock() < 0) return -2;
    }

    /// SAVE RT
    //Logger::logf(LOG_MSG,_T("VemsLogReader::getNextVersion1() - copy tmp packet") );
    m_dataBuffer[2] =SiprSensorType::SIPR_RT_MT_NORMAL;
    memcpy( m_dataBuffer+3, &(m_blockBuffer[ m_blockIdx ]), m_datalen );
    m_tmpPacket = SIPRPacket(m_seq, SIPR_READ_SENSORS, m_datalen+3,  m_dataBuffer); /// copy to tmp

    int gpsCaptureIdx = m_gpsBlockIdx + m_gpsDataLength-1;
    //Logger::logf(LOG_MSG,_T("VemsLogReader::getNext() m_blockIdx: %i, gpsCapture:%i"), m_blockIdx, m_blockBuffer[ gpsCaptureIdx ] );

    if( (m_gpsDataLength == 14) && ((m_blockIdx/m_datalen) == m_blockBuffer[ gpsCaptureIdx ]) ){
        m_dataBuffer[2] =SiprSensorType::GPS;
        memcpy( m_dataBuffer+3, &(m_blockBuffer[ m_gpsBlockIdx ]), m_gpsDataLength );
        SIPRPacket packet(m_seq, SIPR_READ_SENSORS, m_datalen+3,  m_dataBuffer);
        siprPacket = packet; /// copy GPS
        m_isGPSCaptured = true;
        m_isFirstAfterGPS = true;
    }else{
        m_isGPSCaptured = false;
        siprPacket = m_tmpPacket; /// copy RT
    }
    //Logger::logf(LOG_MSG,_T("VemsLogReader::getNext() - assign tmp packet") );
    siprPacket = m_tmpPacket; /// copy RT
    m_blockIdx += m_datalen;

    //Logger::logf(LOG_MSG,_T("VemsLogReader::getNext() END") );
    return 0;
}

wxString VemsLogReader::getHeaderValue(const wxString& key) const
{
    StringHash::const_iterator IT = m_headerHash.find(key);
    if(IT == m_headerHash.end() ) return wxEmptyString;
    return IT->second;
}

int VemsLogReader::ParseHeader()
{
    int res;
    if( res = parseHeaderPriv() != 0) return res;

    long logVersion = 0;
    /// Blocksize is mandatory
    if( m_headerHash.find(RawLogHelper::blockSize) == m_headerHash.end() ) return -10;
    m_headerHash[RawLogHelper::blockSize].ToLong(&m_blocksize);

    /// check log version
    if( m_headerHash.find(RawLogHelper::logVersion) == m_headerHash.end() ){
            m_vemsLogVersion  = VEMSLOG_VERSION_1;
    }else{
        m_headerHash[RawLogHelper::logVersion].ToLong(&logVersion);
        if(logVersion != 2 && logVersion != 3 ){
            m_vemsLogVersion = VEMSLOG_VERSION_1;
        }
        else
        {
            if ( logVersion == 3 ) m_vemsLogVersion = VEMSLOG_VERSION_3;
            else m_vemsLogVersion = VEMSLOG_VERSION_2;
        }
    }
    Logger::logf(LOG_MSG,_T("VemsLogReader::ParseHeader() version:%i"), m_vemsLogVersion);

    if(m_vemsLogVersion == VEMSLOG_VERSION_1){
        if( m_headerHash.find(RawLogHelper::dataLen) == m_headerHash.end() ) return -11;
        m_headerHash[RawLogHelper::dataLen].ToLong(&m_datalen);
        m_headerHash[ RawLogHelper::gpsDataLen ].ToLong( &m_gpsDataLength );
        if(m_datalen < 1.0) return -12;

        //m_RTDataPerBlock =  (m_blocksize - m_gpsDataLength)/m_datalen;
        //Logger::logf(LOG_MSG,_T("VemsLogReader::ParseHeader() RTDataPerBlock: %i"), m_RTDataPerBlock);
        int remains = m_blocksize % m_datalen;
        if( remains < m_gpsDataLength ){
            m_gpsBlockIdx = m_blocksize - m_datalen - remains;
        }else{
            m_gpsBlockIdx = m_blocksize - remains;
        }
    }
    m_isValid = true;
    m_blockBuffer = new char[ m_blocksize ];
    m_rawFile.Read(m_blockBuffer,m_blocksize); /// Read first data block

    //m_symbolMgr->setGroupOffset( _T("GPSData"), m_datalen );
    //observerNotify(1);
    return 0;

}

int VemsLogReader::parseHeaderPriv()
{
    m_headerHash.clear();

    if( !m_rawFile.Open(m_fileName,wxFile::read) )
        return -1;

    VemsLogHeaderParser headerparser( m_rawFile );
    m_headerHash = headerparser.getHeaderHash();
    if ( m_headerHash.size() == 0 )
        return -1;

    /*
    if( m_rawFile.Read(m_headerBuffer,HEADERBLOCKSIZE) != HEADERBLOCKSIZE )
        return -1;

    wxString tmp = wxString::From8BitData(m_headerBuffer, HEADERBLOCKSIZE);
    Logger::logf(LOG_MSG,_T("VemsLogReader::parseHeaderPriv(): header size:%d"), tmp.size() );

    wxArrayString tmparray;

    Utils::tokenizeToArray( tmp, _T("\n"), tmparray);

    for(uint16_t a=0; a<tmparray.GetCount();a++) {
        Logger::logf(LOG_MSG,tmparray.Item(a));
        //if( tmparray.Item(a).StartsWith(0) continue;
        if( tmparray.Item(a).StartsWith(_T("#"))) continue;
        wxArrayString tmpitemarray;
        Utils::tokenizeToArray( tmparray.Item(a), _T("="), tmpitemarray);
        if(tmpitemarray.GetCount() < 2) continue;

        wxString key = tmpitemarray.Item(0);
        wxString value = tmpitemarray.Item(1);

        for(size_t i = 2; i < tmpitemarray.GetCount(); i++){
            value += _T("=")+tmpitemarray.Item(i);
        }
        m_headerHash[key] = value;
    }*/
    return 0;
}


void VemsLogReader::SetFilePos(double percent)
{
    if( !m_rawFile.IsOpened() ){ return; }

    uint16_t fullblock = m_rawFile.Length() / m_blocksize;
    m_seekIdx = (int)(fullblock * percent);
    m_isSeekSet = true;
}

void VemsLogReader::start()
{
    m_seekIdx = m_blocksize;
    m_isSeekSet = true;
}



