#include "VTFilesUpdater.h"

#include <wx/tokenzr.h>
#include <wx/event.h>

#include "../logger.h"
#include "../appHelper.h"

#include "../ConfigFile.h"
#include "../configmanager/ConfigHelper2.h"

#include "../web/WebHelper.h"
#include "../web/WIPAccessories.h"
#include "../DirHelper.h"
#include "../ProgressBar.h"
#include "../customwidgets/InfoWindow.h"
#include "../customwidgets/ConfirmationDialog.h"

#include "VTFilesMD5CheckerThread.h"

class PrgDialog : public wxDialog
{
    public:
        PrgDialog( wxWindow* parent, const wxString& caption );
        void    setMessage( const wxString& msg );
        void    updateValue( const double& val, const double& val2 );
    private:
        ProgressBar*    m_pb1, *m_pb2;
        wxStaticText*   m_message;
        void    OnStop( wxCommandEvent& event );
};

PrgDialog::PrgDialog( wxWindow* parent, const wxString& caption ) :
    wxDialog( parent, -1, caption, wxDefaultPosition, wxDefaultSize, wxDEFAULT_DIALOG_STYLE | wxSTAY_ON_TOP )
{
    wxBoxSizer* mainsizer = new wxBoxSizer( wxVERTICAL );
    m_message = new wxStaticText( this, -1, _T("") );
    m_pb1 = new ProgressBar( this, -1, wxDefaultPosition, wxSize( 300, 18 ) );
    m_pb1->SetValueFormat( _T("%.1f") ); m_pb1->SetValueLimits( 0.0, 100.0 );
    m_pb2 = new ProgressBar( this, -1, wxDefaultPosition, wxSize( 300, 18 ) );
    m_pb2->SetValueFormat( _T("%.1f") ); m_pb2->SetValueLimits( 0.0, 100.0 );
    m_pb1->Update( 0 ); m_pb2->Update( 0 );
    int id = wxNewId();
    wxButton* stop = new wxButton( this, id, _T("Stop") );
    Connect( id, wxEVT_COMMAND_BUTTON_CLICKED, wxCommandEventHandler( PrgDialog::OnStop ), NULL, this );
    mainsizer->Add( m_message, 0, wxEXPAND | wxALL, 5 );
    mainsizer->Add( m_pb1, 0, wxEXPAND | wxALL, 5 );
    mainsizer->Add( m_pb2, 0, wxEXPAND | wxALL, 5 );
    mainsizer->Add( stop, 0, wxALIGN_CENTER | wxALL, 5 );
    SetSizer( mainsizer ); mainsizer->Fit( this ); mainsizer->SetSizeHints( this );
    SetReturnCode( 0 );
}

void PrgDialog::setMessage( const wxString& msg )
{
    m_message->SetLabel( msg );
    Refresh(false);
}

void PrgDialog::updateValue( const double& val, const double& val2 )
{
    if ( val >= 0 ) m_pb1->Update( val );
    if ( val2 >= 0 ) m_pb2->Update( val2 );
}

void PrgDialog::OnStop( wxCommandEvent& event )
{
    EndModal( -1 );
}

/// VTFilesUpdater
/// Public methods

BEGIN_EVENT_TABLE( VTFilesUpdater, wxEvtHandler )
    EVT_GENERAL( wxID_ANY, VTFilesUpdater::hEvtHandler )
END_EVENT_TABLE()

VTFilesUpdater::VTFilesUpdater( const wxString& WXUNUSED(baseDir) )
{
    m_md5checker = NULL;
    m_baseDir = DirHelper::GetPath( dPROGRAMROOT );

    m_Update = 0;

    m_prgDialog = NULL;

    m_isBusy = false;

    getSettings();
}

VTFilesUpdater::~VTFilesUpdater()
{
    if ( m_prgDialog )
        m_prgDialog->Destroy();

    if ( m_md5checker )
        delete m_md5checker;
}

void VTFilesUpdater::Check( bool forceUpdate )
{
    if ( m_isBusy )
    {
        wxMessageBox(_T("Checking in files in progress."), _T("Information"), wxOK | wxICON_INFORMATION );
        return;
    }

    // manually update/restore
    if ( forceUpdate )
    {
        if ( m_Update < 1 )
        {
            m_Update = 1;
        }
    }

    // never check updates
    if ( m_Update == 0 )
    {
        return;
    }

    // gather file info
    checkServer();

    if ( m_serverVersion == m_localVersion )
    {
        if ( !forceUpdate )
            return;
    }

    m_forceUpdate = forceUpdate;
    // check files & their integrity
    checkMD5();
}

void VTFilesUpdater::doDownload()
{
    bool download_them = false;

    m_toDownload = m_md5checker->getFilesToUpdate();

    int no_files_Found = m_toDownload.GetCount();

    Logger::logf(LOG_MSG, _T("no of=%d"),no_files_Found);

    if ( m_forceUpdate && ( no_files_Found == 0 ) )
    {
        InfoWindow::Display( NULL, _T("Information"), _T("No new/modified files found") );
        return;
    }

    /// missing files
    if ( no_files_Found > 0 )
    {
        if ( m_Update == 1 )
        {
            wxString msg = wxString::Format( _T("%d new files are available on the server\nDownload them now?"), no_files_Found );

            ConfirmationDialog dlg( NULL, msg, _T("Information"), wxDefaultPosition, wxDefaultSize, wxYES | wxNO,
                ConfirmationDialog::CONFIRMDLG_EXTENDED, _T("Do not ask me again until newer updates are available") );

            if ( dlg.ShowModal() == wxID_YES )
            {
                download_them = true;
                saveLastUpdateTime();
            }
            else
            {
                download_them = false;
                bool do_not_askmeagain;
                if ( dlg.getCheckBoxState( do_not_askmeagain ) == true )
                {
                    if ( do_not_askmeagain == true )
                    {
                        saveLastUpdateTime();
                        return;
                    }
                }
            }
        }
        else if ( m_Update == 2 ) {
            download_them = true;
            saveLastUpdateTime();
        }
    }else{
        saveLastUpdateTime();
    }


    /// download files
    m_fileCnt = 0;

    if ( download_them )
    {
        m_prgDialog = new PrgDialog( NULL, _T("Downloading files") );

        // start download
        wxString link = m_baseurl + _T("/") + m_toDownload[0];
        wxString dest = m_baseDir + m_toDownload[0];

        // apply appropriate dir separator
        dest.Replace( _T("/"), DirHelper::Separ(), true );
        dest.Replace( _T("\\"), DirHelper::Separ(), true );

        wxFileName fn( dest );
        DirHelper::CheckDir( fn.GetPath() );

        m_fileCnt = m_toDownload.GetCount();

        wxSafeYield( m_prgDialog );

        m_prgDialog->setMessage( _T("Downloading: ") + m_toDownload[0] );

        if ( wxFileExists(dest) )
        {
            wxString backup_name = dest+_T(".bak");
           wxRenameFile(dest,backup_name,true); 
        }

        DownloadProcess* dt = new DownloadProcess( link, dest ,true);

        dt->registerObserver( this );
        dt->start();

        m_prgDialog->ShowModal();
    }
}

void VTFilesUpdater::notify( int i, int j )
{
    if ( j == WEB_ENDSESSION )
    {
        GeneralEvent evt;

        evt.SetVoidPtr( this );
        evt.SetLong( FILEDL_FILEDOWNLOADED );

        AddPendingEvent( evt );

        return;
    }
    if ( j == WEB_FILESIZE )
    {
        // Logger::logf(LOG_MSG, _T("filesize=%d"),i);
        m_filesize = i; m_read = 0;

        GeneralEvent evt;

        evt.SetVoidPtr( this );
        evt.SetLong( FILEDL_FILESIZE );

        AddPendingEvent( evt );

        return;

    }
    if ( j == WEB_TRANSMITTEDDATASIZE )
    {
        m_read += i;

        GeneralEvent evt;

        evt.SetVoidPtr( this );
        evt.SetLong( FILEDL_DLINPROGRESS );

        AddPendingEvent( evt );

        return;
    }
}

/// Private methods

void VTFilesUpdater::checkStopPressed()
{
    if ( m_prgDialog == NULL )
        return;

    if ( m_prgDialog->GetReturnCode() == -1 )
    {
        m_toDownload.Clear();

        GeneralEvent evt;

        evt.SetVoidPtr( this );
        evt.SetLong( FILEDL_STOPPRESSED );

        AddPendingEvent( evt );
    }
}

void VTFilesUpdater::checkMD5()
{
    m_isBusy = true;

    if ( m_md5checker )
    {
        if ( m_md5checker->isRunning() == false )
            delete m_md5checker;
        else
            return;
    }

    m_md5checker = new VTFilesMD5CheckerThread( this, m_fileHash, DirHelper::GetPath( dPROGRAMROOT ) );
    m_md5checker->start();
/*    m_toDownload.Clear();

    StringHash::const_iterator it = m_fileHash.begin();

    for ( ; it != m_fileHash.end(); ++it )
    {
        wxString fname = m_baseDir + it->first;
        if ( ! wxFileExists( fname ) )
        {
            m_toDownload.Add( it->first );
            continue;
        }
        hashwrapper* wrapper = new md5wrapper();

        std::string temp = wxString2stdString( fname );
        wxString tmp;

        std::string hash2 = wrapper->getHashFromFile( temp );
        wxString md5hash = stdString2wxString (hash2 );

        delete wrapper;

        if ( md5hash != it->second )
        {
            m_toDownload.Add( it->first );
        }
    }*/
}

void VTFilesUpdater::checkServer()
{
    wxString result = WebHelper::DownloadPage( m_phpUrl );

    wxString temp = result.BeforeFirst( '\n' );
    m_baseurl = temp.BeforeFirst( '|' );
    m_serverVersion = temp.AfterFirst( '|' );
    m_serverVersion.Trim( true );
    m_serverVersion.Trim( false );

    if ( m_baseurl.EndsWith(_T("/") ) )
        m_baseurl.RemoveLast();

    result = result.AfterFirst( '\n' );

    wxStringTokenizer tkz( result, _T("\n") );

    m_fileHash.clear();

    while ( tkz.HasMoreTokens() )
    {
        wxString token = tkz.GetNextToken();
        wxString md5 = token.Left( 32 );    // first 32 char = md5 length
        wxString fname = token.Mid( 33 );   // rest is file name
        m_fileHash[fname] = md5;
    }
}

void VTFilesUpdater::getSettings()
{
    // get last update time
    wxTextFile lastupdatefile( DirHelper::GetPath( dCONFIG ) + _T("filesupdate.cfg") );
    if ( lastupdatefile.Exists() == false )
    {
        m_localVersion = _T("");
    }
    else
    {
        lastupdatefile.Open();
        m_localVersion = lastupdatefile.GetFirstLine();
        m_localVersion.Trim( true );
        m_localVersion.Trim( false );
        lastupdatefile.Close();
    }

    ConfigBase* config = ConfigHelper2::getConfig( );

    long int autoUpdate = 0;

    wxString phpName = _T("");
    wxString base = _T("");

    if ( config->read( _T("/updates/updateserver"), &base ) == false )
    {
        config->read( _T("/webservice/baseserver"), &base );
    }

//    ///TEST ->
//    base = _T("http://localhost/vems");
//    /// <-- TEST

    if ( config->read( _T("/webservice/inihandler"), &phpName ) == false )
    {
        phpName = _T("getfiles.php");
    }

    m_phpUrl = base + _T("/") + phpName;

    if ( config->read( _T("/updates/update"), &autoUpdate ) == false )
    {
        autoUpdate = 0;
    }
    if ( autoUpdate < 0 || autoUpdate > 2 )
    {
        autoUpdate = 0;
    }

    m_Update = (unsigned char) autoUpdate;
    //delete config;
}

/// helpers
void VTFilesUpdater::hEvtHandler( GeneralEvent& event )
{
    // only accept events from self or md5checker
    if ( event.GetVoidPtr() != this && event.GetVoidPtr() != m_md5checker )
    {
        event.Skip();
        return;
    }

    int id = event.GetLong();

    if ( id == 300 )
    {
        m_isBusy = false;
        doDownload();
        return;
    }

    if ( id == FILEDL_ALLDOWNLOADED )
    {
        // finished
        InfoWindow::Display( NULL, _T("Information"), _T("All files have been downloaded") );
        return;
    }
    if ( id == FILEDL_STOPPRESSED )
    {
        InfoWindow::Display( NULL, _T("Info"), _T("Download aborted"),
            IW_FADE_IN | IW_MOVE | IW_WARNING );
        return;
    }

    // WEB_ENDSESSION
    if ( id == FILEDL_FILEDOWNLOADED )
    {
        m_filesize = 0; m_read = 0;

        if ( m_prgDialog )
            m_prgDialog->updateValue( 0, -1 );

        m_toDownload.RemoveAt(0);

//        Logger::logf(LOG_MSG,_T("IniUpdater::hEvtHandler() m_toDownload.GetCount=%d"),m_toDownload.GetCount());

        if ( m_toDownload.GetCount() == 0 )
        {
            m_prgDialog->Destroy();
            m_prgDialog = NULL;

            GeneralEvent evt;

            evt.SetVoidPtr( this );
            evt.SetLong( FILEDL_ALLDOWNLOADED );

            AddPendingEvent( evt );
            return;
        }
        // start download
        wxString link = m_baseurl + _T("/") + m_toDownload[0];
        wxString dest = m_baseDir + m_toDownload[0];

        // apply appropriate dir separator
        dest.Replace( _T("/"), DirHelper::Separ(), true );
        dest.Replace( _T("\\"), DirHelper::Separ(), true );

        //Logger::logf(LOG_MSG, dest);

        wxFileName fn( dest );
        DirHelper::CheckDir( fn.GetPath() );

        if ( m_prgDialog )
            m_prgDialog->setMessage( _T("Downloading: ") + m_toDownload[0] );

        if ( wxFileExists(dest) )
        {
            wxString backup_name = dest+_T(".bak");
           wxRenameFile(dest,backup_name,true); 
        }

        DownloadProcess* dt = new DownloadProcess( link, dest ,true);

        unsigned done = m_fileCnt - m_toDownload.GetCount();

        double percent = (double) done * 100.0 / (double) m_fileCnt;

        if ( m_prgDialog )
            m_prgDialog->updateValue( -1, percent );

        dt->registerObserver( this );
        dt->start();

        checkStopPressed();
    }

    if( id == FILEDL_FILESIZE ) {

        if ( m_prgDialog )
            m_prgDialog->updateValue( 0, -1 );

        checkStopPressed();

    }

    if( id == FILEDL_DLINPROGRESS ) {

        double percent = ( m_read * 100.0 / (double) m_filesize );

        if ( m_prgDialog )
            m_prgDialog->updateValue( percent, -1 );

        checkStopPressed();
    }

    event.Skip();
}

void VTFilesUpdater::saveLastUpdateTime()
{
    wxTextFile lastupdatefile( DirHelper::GetPath( dCONFIG ) + _T("filesupdate.cfg") );
    if ( lastupdatefile.Exists() )
    {
        lastupdatefile.Open();
        lastupdatefile.Clear();
    }
    else
    {
        lastupdatefile.Create();
    }
    lastupdatefile.AddLine( m_serverVersion );
    lastupdatefile.Write();
    lastupdatefile.Close();
}
