#ifndef WXREALRECT_H
#define WXREALRECT_H

#include "wxRealSize.h"

class wxRealRectArray;

class wxRealRect: public wxObject
{
    DECLARE_DYNAMIC_CLASS(wxRealRect);

    public:
        wxRealRect() { Init(-1.0,-1.0,0,0); };
        wxRealRect(const wxRealRect& rect);
        wxRealRect(const wxRect& rect);
        wxRealRect(double xx, double yy, double ww, double hh, double precision = -1.0) { Init(xx,yy,ww,hh,precision); };
        wxRealRect(const wxRealPoint& topLeft, const wxRealPoint& bottomRight, double precision = -1.0) {
            Init(topLeft.x, topLeft.y, bottomRight.x - topLeft.x, bottomRight.y - topLeft.y, precision);
        };
        wxRealRect(const wxRealPoint& pos, const wxRealSize& size, double precision = -1.0) {
            Init(pos.x, pos.y, size.w, size.h, precision);
        };

        virtual ~wxRealRect();

        void Init(double xx, double yy, double ww, double hh, double precision = -1.0);

        void operator=(const wxRealRect& rect);
        void operator=(const wxRect& rect);
        bool operator==(const wxRealRect& rect) const;
        bool operator!=(const wxRealRect& rect) const;

        wxRealRect& operator+=(const wxRealRect& rect);
        wxRealRect operator+(const wxRealRect& rect) const;

        bool Contains(double cx, double cy) const;
        bool Contains(const wxRealPoint& p) const { return Contains(p.x,p.y); };
        bool Contains(const wxRealRect& rect) const { return Contains(rect.GetTopLeft()) && Contains(rect.GetBottomRight()); };

        wxRealRect& Inflate(const wxRealSize& ds) { return Inflate(ds.w,ds.h); };
        wxRealRect& Inflate(double diff) { return Inflate(diff,diff); };
        wxRealRect& Inflate(double dx, double dy);

        wxRealRect& Deflate(double dx, double dy) { return Inflate(-dx,-dy); };
        wxRealRect& Deflate(const wxRealSize& ds) { return Inflate(-ds.w,-ds.h); };
        wxRealRect& Deflate(double diff) { return Inflate(-diff,-diff); };

        void Offset(double dx, double dy) { x += dx; y += dy; };
        void Offset(const wxRealPoint& dp) { Offset(dp.x,dp.y); };

        void OffsetRef(const wxRealSize& offs);


        double GetLeft() const { return x; };
        double GetRight() const { return (x + w); };
        double GetTop() const { return y; };
        double GetBottom() const { return (y + h); };

        void SetLeft(double left) { x = left; };
        void SetRight(double right) { w = right - x; };
        void SetTop(double top) { y = top; };
        void SetBottom(double bottom) { h = bottom - y; };

        wxRealPoint GetPosition() const { return wxRealPoint(x,y); };
        wxRealSize  GetSize() const { return wxRealSize(w,h); };
        wxRealPoint GetTopLeft() const { return wxRealPoint(GetLeft(),GetTop()); };
        wxRealPoint GetTopRight() const { return wxRealPoint(GetRight(),GetTop()); };
        wxRealPoint GetBottomLeft() const { return wxRealPoint(GetLeft(),GetBottom()); };
        wxRealPoint GetBottomRight() const { return wxRealPoint(GetRight(),GetBottom()); };

        wxRealPoint GetCenter() const { return wxRealPoint( x + w/2, y + h/2 ); };

        double GetHeight() const { return h; };
        double GetWidth() const { return w; };
        double GetX() const { return x; };
        double GetY() const { return y; };

        void SetX(double px) { x = px; }
        void SetY(double py) { y = py; }
        void SetWidth(double pw) { w = pw; }
        void SetHeight(double ph) { h = ph; }

        wxString GetAsString(int dec = 2);
        wxString GetAsStringBorders(int dec = 2);

        void SetTopLeft(double top, double left)  { x=left; y=top; };
        void SetTopLeft(const wxRealPoint& p)  { x=p.x; y=p.y; };

        void SetBottomRight(double bottom, double right) { w = right - x; h = bottom - y; }
        void SetBottomRight(const wxRealPoint& p) { SetBottomRight(p.y,p.x); }

        void SetSize(const wxRealSize& size) { w = size.w, h = size.h; }

        wxRealRect& Intersect(const wxRealRect& rect);
        wxRealRect Intersect(const wxRealRect& rect) const;
        bool Intersects(const wxRealRect& rect) const;

        void Normalize();
        void SubPixelNormalize(double linewidth = 1.0);

        wxRealRect& Union(const wxRealRect& rect);
        wxRealRect Union(const wxRealRect rect) const
        {
            wxRealRect r = *this;
            r.Union(rect);
            return r;
        }

        wxRect GetOuterBox() const;
        wxRect GetInnerBox() const;

        wxRealRectArray Subtract(const wxRealRect& rect);
        wxRealRectArray Subtract(const wxRealRectArray& rects, bool removeNullRects = true);

    public:

        double x;
        double y;
        double w;
        double h;


    private:

        double m_precision;
};


extern wxRealRect wxNullRealRect;
extern wxRealRect wxDefaultRealRect;
extern wxRect wxDefaultRect;

DECLARE_VARIANT_OBJECT(wxRealRect);

#endif // WXREALRECT_H
